/* Automatically @generated by tree-sitter v0.25.10 */

#include "tree_sitter/parser.h"

#if defined(__GNUC__) || defined(__clang__)
#pragma GCC diagnostic ignored "-Wmissing-field-initializers"
#endif

#ifdef _MSC_VER
#pragma optimize("", off)
#elif defined(__clang__)
#pragma clang optimize off
#elif defined(__GNUC__)
#pragma GCC optimize ("O0")
#endif

#define LANGUAGE_VERSION 15
#define STATE_COUNT 442
#define LARGE_STATE_COUNT 2
#define SYMBOL_COUNT 142
#define ALIAS_COUNT 9
#define TOKEN_COUNT 75
#define EXTERNAL_TOKEN_COUNT 3
#define FIELD_COUNT 0
#define MAX_ALIAS_SEQUENCE_LENGTH 9
#define MAX_RESERVED_WORD_SET_SIZE 0
#define PRODUCTION_ID_COUNT 17
#define SUPERTYPE_COUNT 0

enum ts_symbol_identifiers {
  anon_sym_ATimport = 1,
  anon_sym_COMMA = 2,
  anon_sym_SEMI = 3,
  anon_sym_ATmedia = 4,
  anon_sym_ATcharset = 5,
  anon_sym_ATnamespace = 6,
  anon_sym_ATkeyframes = 7,
  aux_sym_keyframes_statement_token1 = 8,
  anon_sym_LBRACE = 9,
  anon_sym_RBRACE = 10,
  sym_from = 11,
  anon_sym_to = 12,
  anon_sym_ATsupports = 13,
  anon_sym_ATscope = 14,
  anon_sym_LPAREN = 15,
  anon_sym_RPAREN = 16,
  sym_nesting_selector = 17,
  anon_sym_STAR = 18,
  anon_sym_DOT = 19,
  anon_sym_has = 20,
  anon_sym_not = 21,
  anon_sym_is = 22,
  anon_sym_where = 23,
  anon_sym_host = 24,
  anon_sym_host_DASHcontext = 25,
  anon_sym_nth_DASHchild = 26,
  anon_sym_nth_DASHlast_DASHchild = 27,
  anon_sym_COLON_COLON = 28,
  anon_sym_POUND = 29,
  anon_sym_LBRACK = 30,
  anon_sym_EQ = 31,
  anon_sym_TILDE_EQ = 32,
  anon_sym_CARET_EQ = 33,
  anon_sym_PIPE_EQ = 34,
  anon_sym_STAR_EQ = 35,
  anon_sym_DOLLAR_EQ = 36,
  anon_sym_RBRACK = 37,
  anon_sym_GT = 38,
  anon_sym_TILDE = 39,
  anon_sym_PLUS = 40,
  anon_sym_PIPE = 41,
  anon_sym_LPAREN2 = 42,
  anon_sym_even = 43,
  anon_sym_odd = 44,
  anon_sym_of = 45,
  sym__nth_functional_notation = 46,
  anon_sym_COLON = 47,
  sym_important = 48,
  anon_sym_and = 49,
  anon_sym_or = 50,
  anon_sym_only = 51,
  anon_sym_selector = 52,
  aux_sym_color_value_token1 = 53,
  anon_sym_SQUOTE = 54,
  aux_sym_string_value_token1 = 55,
  anon_sym_DQUOTE = 56,
  aux_sym_string_value_token2 = 57,
  sym_escape_sequence = 58,
  aux_sym_integer_value_token1 = 59,
  aux_sym_float_value_token1 = 60,
  sym_unit = 61,
  anon_sym_LBRACK2 = 62,
  anon_sym_DASH = 63,
  anon_sym_SLASH = 64,
  aux_sym_class_name_token1 = 65,
  sym_identifier = 66,
  sym_at_keyword = 67,
  sym_js_comment = 68,
  sym_comment = 69,
  sym_plain_value = 70,
  sym_important_value = 71,
  sym__descendant_operator = 72,
  sym__pseudo_class_selector_colon = 73,
  sym___error_recovery = 74,
  sym_stylesheet = 75,
  sym_import_statement = 76,
  sym_media_statement = 77,
  sym_charset_statement = 78,
  sym_namespace_statement = 79,
  sym_keyframes_statement = 80,
  sym_keyframe_block_list = 81,
  sym_keyframe_block = 82,
  sym_to = 83,
  sym_supports_statement = 84,
  sym_scope_statement = 85,
  sym_postcss_statement = 86,
  sym_at_rule = 87,
  sym_rule_set = 88,
  sym_selectors = 89,
  sym_block = 90,
  sym__selector = 91,
  sym_universal_selector = 92,
  sym_class_selector = 93,
  sym_pseudo_class_selector = 94,
  sym__nth_child_pseudo_class_selector = 95,
  sym_pseudo_element_selector = 96,
  sym_id_selector = 97,
  sym_attribute_selector = 98,
  sym_child_selector = 99,
  sym_descendant_selector = 100,
  sym_sibling_selector = 101,
  sym_adjacent_sibling_selector = 102,
  sym_namespace_selector = 103,
  sym_pseudo_class_arguments = 104,
  sym_pseudo_class_with_selector_arguments = 105,
  sym_pseudo_class_nth_child_arguments = 106,
  sym_pseudo_element_arguments = 107,
  sym_declaration = 108,
  sym_last_declaration = 109,
  sym__query = 110,
  sym_feature_query = 111,
  sym_parenthesized_query = 112,
  sym_binary_query = 113,
  sym_unary_query = 114,
  sym_selector_query = 115,
  sym__value = 116,
  sym_parenthesized_value = 117,
  sym_color_value = 118,
  sym_string_value = 119,
  sym_integer_value = 120,
  sym_float_value = 121,
  sym_grid_value = 122,
  sym_call_expression = 123,
  sym_binary_expression = 124,
  sym_arguments = 125,
  sym_class_name = 126,
  aux_sym_stylesheet_repeat1 = 127,
  aux_sym_import_statement_repeat1 = 128,
  aux_sym_keyframe_block_list_repeat1 = 129,
  aux_sym_keyframe_block_repeat1 = 130,
  aux_sym_postcss_statement_repeat1 = 131,
  aux_sym_selectors_repeat1 = 132,
  aux_sym_block_repeat1 = 133,
  aux_sym_pseudo_class_arguments_repeat1 = 134,
  aux_sym_pseudo_class_arguments_repeat2 = 135,
  aux_sym_declaration_repeat1 = 136,
  aux_sym_string_value_repeat1 = 137,
  aux_sym_string_value_repeat2 = 138,
  aux_sym_grid_value_repeat1 = 139,
  aux_sym_arguments_repeat1 = 140,
  aux_sym_class_name_repeat1 = 141,
  alias_sym_attribute_name = 142,
  alias_sym_feature_name = 143,
  alias_sym_function_name = 144,
  alias_sym_id_name = 145,
  alias_sym_keyframes_name = 146,
  alias_sym_keyword_query = 147,
  alias_sym_namespace_name = 148,
  alias_sym_property_name = 149,
  alias_sym_tag_name = 150,
};

static const char * const ts_symbol_names[] = {
  [ts_builtin_sym_end] = "end",
  [anon_sym_ATimport] = "@import",
  [anon_sym_COMMA] = ",",
  [anon_sym_SEMI] = ";",
  [anon_sym_ATmedia] = "@media",
  [anon_sym_ATcharset] = "@charset",
  [anon_sym_ATnamespace] = "@namespace",
  [anon_sym_ATkeyframes] = "@keyframes",
  [aux_sym_keyframes_statement_token1] = "at_keyword",
  [anon_sym_LBRACE] = "{",
  [anon_sym_RBRACE] = "}",
  [sym_from] = "from",
  [anon_sym_to] = "to",
  [anon_sym_ATsupports] = "@supports",
  [anon_sym_ATscope] = "@scope",
  [anon_sym_LPAREN] = "(",
  [anon_sym_RPAREN] = ")",
  [sym_nesting_selector] = "nesting_selector",
  [anon_sym_STAR] = "*",
  [anon_sym_DOT] = ".",
  [anon_sym_has] = "class_name",
  [anon_sym_not] = "not",
  [anon_sym_is] = "class_name",
  [anon_sym_where] = "class_name",
  [anon_sym_host] = "class_name",
  [anon_sym_host_DASHcontext] = "class_name",
  [anon_sym_nth_DASHchild] = "class_name",
  [anon_sym_nth_DASHlast_DASHchild] = "class_name",
  [anon_sym_COLON_COLON] = "::",
  [anon_sym_POUND] = "#",
  [anon_sym_LBRACK] = "[",
  [anon_sym_EQ] = "=",
  [anon_sym_TILDE_EQ] = "~=",
  [anon_sym_CARET_EQ] = "^=",
  [anon_sym_PIPE_EQ] = "|=",
  [anon_sym_STAR_EQ] = "*=",
  [anon_sym_DOLLAR_EQ] = "$=",
  [anon_sym_RBRACK] = "]",
  [anon_sym_GT] = ">",
  [anon_sym_TILDE] = "~",
  [anon_sym_PLUS] = "+",
  [anon_sym_PIPE] = "|",
  [anon_sym_LPAREN2] = "(",
  [anon_sym_even] = "plain_value",
  [anon_sym_odd] = "plain_value",
  [anon_sym_of] = "of",
  [sym__nth_functional_notation] = "plain_value",
  [anon_sym_COLON] = ":",
  [sym_important] = "important",
  [anon_sym_and] = "and",
  [anon_sym_or] = "or",
  [anon_sym_only] = "only",
  [anon_sym_selector] = "selector",
  [aux_sym_color_value_token1] = "color_value_token1",
  [anon_sym_SQUOTE] = "'",
  [aux_sym_string_value_token1] = "string_content",
  [anon_sym_DQUOTE] = "\"",
  [aux_sym_string_value_token2] = "string_content",
  [sym_escape_sequence] = "escape_sequence",
  [aux_sym_integer_value_token1] = "integer_value_token1",
  [aux_sym_float_value_token1] = "float_value_token1",
  [sym_unit] = "unit",
  [anon_sym_LBRACK2] = "[",
  [anon_sym_DASH] = "-",
  [anon_sym_SLASH] = "/",
  [aux_sym_class_name_token1] = "identifier",
  [sym_identifier] = "identifier",
  [sym_at_keyword] = "at_keyword",
  [sym_js_comment] = "js_comment",
  [sym_comment] = "comment",
  [sym_plain_value] = "plain_value",
  [sym_important_value] = "important_value",
  [sym__descendant_operator] = "_descendant_operator",
  [sym__pseudo_class_selector_colon] = ":",
  [sym___error_recovery] = "__error_recovery",
  [sym_stylesheet] = "stylesheet",
  [sym_import_statement] = "import_statement",
  [sym_media_statement] = "media_statement",
  [sym_charset_statement] = "charset_statement",
  [sym_namespace_statement] = "namespace_statement",
  [sym_keyframes_statement] = "keyframes_statement",
  [sym_keyframe_block_list] = "keyframe_block_list",
  [sym_keyframe_block] = "keyframe_block",
  [sym_to] = "to",
  [sym_supports_statement] = "supports_statement",
  [sym_scope_statement] = "scope_statement",
  [sym_postcss_statement] = "postcss_statement",
  [sym_at_rule] = "at_rule",
  [sym_rule_set] = "rule_set",
  [sym_selectors] = "selectors",
  [sym_block] = "block",
  [sym__selector] = "_selector",
  [sym_universal_selector] = "universal_selector",
  [sym_class_selector] = "class_selector",
  [sym_pseudo_class_selector] = "pseudo_class_selector",
  [sym__nth_child_pseudo_class_selector] = "_nth_child_pseudo_class_selector",
  [sym_pseudo_element_selector] = "pseudo_element_selector",
  [sym_id_selector] = "id_selector",
  [sym_attribute_selector] = "attribute_selector",
  [sym_child_selector] = "child_selector",
  [sym_descendant_selector] = "descendant_selector",
  [sym_sibling_selector] = "sibling_selector",
  [sym_adjacent_sibling_selector] = "adjacent_sibling_selector",
  [sym_namespace_selector] = "namespace_selector",
  [sym_pseudo_class_arguments] = "arguments",
  [sym_pseudo_class_with_selector_arguments] = "arguments",
  [sym_pseudo_class_nth_child_arguments] = "arguments",
  [sym_pseudo_element_arguments] = "arguments",
  [sym_declaration] = "declaration",
  [sym_last_declaration] = "declaration",
  [sym__query] = "_query",
  [sym_feature_query] = "feature_query",
  [sym_parenthesized_query] = "parenthesized_query",
  [sym_binary_query] = "binary_query",
  [sym_unary_query] = "unary_query",
  [sym_selector_query] = "selector_query",
  [sym__value] = "_value",
  [sym_parenthesized_value] = "parenthesized_value",
  [sym_color_value] = "color_value",
  [sym_string_value] = "string_value",
  [sym_integer_value] = "integer_value",
  [sym_float_value] = "float_value",
  [sym_grid_value] = "grid_value",
  [sym_call_expression] = "call_expression",
  [sym_binary_expression] = "binary_expression",
  [sym_arguments] = "arguments",
  [sym_class_name] = "class_name",
  [aux_sym_stylesheet_repeat1] = "stylesheet_repeat1",
  [aux_sym_import_statement_repeat1] = "import_statement_repeat1",
  [aux_sym_keyframe_block_list_repeat1] = "keyframe_block_list_repeat1",
  [aux_sym_keyframe_block_repeat1] = "keyframe_block_repeat1",
  [aux_sym_postcss_statement_repeat1] = "postcss_statement_repeat1",
  [aux_sym_selectors_repeat1] = "selectors_repeat1",
  [aux_sym_block_repeat1] = "block_repeat1",
  [aux_sym_pseudo_class_arguments_repeat1] = "pseudo_class_arguments_repeat1",
  [aux_sym_pseudo_class_arguments_repeat2] = "pseudo_class_arguments_repeat2",
  [aux_sym_declaration_repeat1] = "declaration_repeat1",
  [aux_sym_string_value_repeat1] = "string_value_repeat1",
  [aux_sym_string_value_repeat2] = "string_value_repeat2",
  [aux_sym_grid_value_repeat1] = "grid_value_repeat1",
  [aux_sym_arguments_repeat1] = "arguments_repeat1",
  [aux_sym_class_name_repeat1] = "class_name_repeat1",
  [alias_sym_attribute_name] = "attribute_name",
  [alias_sym_feature_name] = "feature_name",
  [alias_sym_function_name] = "function_name",
  [alias_sym_id_name] = "id_name",
  [alias_sym_keyframes_name] = "keyframes_name",
  [alias_sym_keyword_query] = "keyword_query",
  [alias_sym_namespace_name] = "namespace_name",
  [alias_sym_property_name] = "property_name",
  [alias_sym_tag_name] = "tag_name",
};

static const TSSymbol ts_symbol_map[] = {
  [ts_builtin_sym_end] = ts_builtin_sym_end,
  [anon_sym_ATimport] = anon_sym_ATimport,
  [anon_sym_COMMA] = anon_sym_COMMA,
  [anon_sym_SEMI] = anon_sym_SEMI,
  [anon_sym_ATmedia] = anon_sym_ATmedia,
  [anon_sym_ATcharset] = anon_sym_ATcharset,
  [anon_sym_ATnamespace] = anon_sym_ATnamespace,
  [anon_sym_ATkeyframes] = anon_sym_ATkeyframes,
  [aux_sym_keyframes_statement_token1] = sym_at_keyword,
  [anon_sym_LBRACE] = anon_sym_LBRACE,
  [anon_sym_RBRACE] = anon_sym_RBRACE,
  [sym_from] = sym_from,
  [anon_sym_to] = anon_sym_to,
  [anon_sym_ATsupports] = anon_sym_ATsupports,
  [anon_sym_ATscope] = anon_sym_ATscope,
  [anon_sym_LPAREN] = anon_sym_LPAREN,
  [anon_sym_RPAREN] = anon_sym_RPAREN,
  [sym_nesting_selector] = sym_nesting_selector,
  [anon_sym_STAR] = anon_sym_STAR,
  [anon_sym_DOT] = anon_sym_DOT,
  [anon_sym_has] = sym_class_name,
  [anon_sym_not] = anon_sym_not,
  [anon_sym_is] = sym_class_name,
  [anon_sym_where] = sym_class_name,
  [anon_sym_host] = sym_class_name,
  [anon_sym_host_DASHcontext] = sym_class_name,
  [anon_sym_nth_DASHchild] = sym_class_name,
  [anon_sym_nth_DASHlast_DASHchild] = sym_class_name,
  [anon_sym_COLON_COLON] = anon_sym_COLON_COLON,
  [anon_sym_POUND] = anon_sym_POUND,
  [anon_sym_LBRACK] = anon_sym_LBRACK,
  [anon_sym_EQ] = anon_sym_EQ,
  [anon_sym_TILDE_EQ] = anon_sym_TILDE_EQ,
  [anon_sym_CARET_EQ] = anon_sym_CARET_EQ,
  [anon_sym_PIPE_EQ] = anon_sym_PIPE_EQ,
  [anon_sym_STAR_EQ] = anon_sym_STAR_EQ,
  [anon_sym_DOLLAR_EQ] = anon_sym_DOLLAR_EQ,
  [anon_sym_RBRACK] = anon_sym_RBRACK,
  [anon_sym_GT] = anon_sym_GT,
  [anon_sym_TILDE] = anon_sym_TILDE,
  [anon_sym_PLUS] = anon_sym_PLUS,
  [anon_sym_PIPE] = anon_sym_PIPE,
  [anon_sym_LPAREN2] = anon_sym_LPAREN,
  [anon_sym_even] = sym_plain_value,
  [anon_sym_odd] = sym_plain_value,
  [anon_sym_of] = anon_sym_of,
  [sym__nth_functional_notation] = sym_plain_value,
  [anon_sym_COLON] = anon_sym_COLON,
  [sym_important] = sym_important,
  [anon_sym_and] = anon_sym_and,
  [anon_sym_or] = anon_sym_or,
  [anon_sym_only] = anon_sym_only,
  [anon_sym_selector] = anon_sym_selector,
  [aux_sym_color_value_token1] = aux_sym_color_value_token1,
  [anon_sym_SQUOTE] = anon_sym_SQUOTE,
  [aux_sym_string_value_token1] = aux_sym_string_value_token1,
  [anon_sym_DQUOTE] = anon_sym_DQUOTE,
  [aux_sym_string_value_token2] = aux_sym_string_value_token1,
  [sym_escape_sequence] = sym_escape_sequence,
  [aux_sym_integer_value_token1] = aux_sym_integer_value_token1,
  [aux_sym_float_value_token1] = aux_sym_float_value_token1,
  [sym_unit] = sym_unit,
  [anon_sym_LBRACK2] = anon_sym_LBRACK,
  [anon_sym_DASH] = anon_sym_DASH,
  [anon_sym_SLASH] = anon_sym_SLASH,
  [aux_sym_class_name_token1] = sym_identifier,
  [sym_identifier] = sym_identifier,
  [sym_at_keyword] = sym_at_keyword,
  [sym_js_comment] = sym_js_comment,
  [sym_comment] = sym_comment,
  [sym_plain_value] = sym_plain_value,
  [sym_important_value] = sym_important_value,
  [sym__descendant_operator] = sym__descendant_operator,
  [sym__pseudo_class_selector_colon] = anon_sym_COLON,
  [sym___error_recovery] = sym___error_recovery,
  [sym_stylesheet] = sym_stylesheet,
  [sym_import_statement] = sym_import_statement,
  [sym_media_statement] = sym_media_statement,
  [sym_charset_statement] = sym_charset_statement,
  [sym_namespace_statement] = sym_namespace_statement,
  [sym_keyframes_statement] = sym_keyframes_statement,
  [sym_keyframe_block_list] = sym_keyframe_block_list,
  [sym_keyframe_block] = sym_keyframe_block,
  [sym_to] = sym_to,
  [sym_supports_statement] = sym_supports_statement,
  [sym_scope_statement] = sym_scope_statement,
  [sym_postcss_statement] = sym_postcss_statement,
  [sym_at_rule] = sym_at_rule,
  [sym_rule_set] = sym_rule_set,
  [sym_selectors] = sym_selectors,
  [sym_block] = sym_block,
  [sym__selector] = sym__selector,
  [sym_universal_selector] = sym_universal_selector,
  [sym_class_selector] = sym_class_selector,
  [sym_pseudo_class_selector] = sym_pseudo_class_selector,
  [sym__nth_child_pseudo_class_selector] = sym__nth_child_pseudo_class_selector,
  [sym_pseudo_element_selector] = sym_pseudo_element_selector,
  [sym_id_selector] = sym_id_selector,
  [sym_attribute_selector] = sym_attribute_selector,
  [sym_child_selector] = sym_child_selector,
  [sym_descendant_selector] = sym_descendant_selector,
  [sym_sibling_selector] = sym_sibling_selector,
  [sym_adjacent_sibling_selector] = sym_adjacent_sibling_selector,
  [sym_namespace_selector] = sym_namespace_selector,
  [sym_pseudo_class_arguments] = sym_arguments,
  [sym_pseudo_class_with_selector_arguments] = sym_arguments,
  [sym_pseudo_class_nth_child_arguments] = sym_arguments,
  [sym_pseudo_element_arguments] = sym_arguments,
  [sym_declaration] = sym_declaration,
  [sym_last_declaration] = sym_declaration,
  [sym__query] = sym__query,
  [sym_feature_query] = sym_feature_query,
  [sym_parenthesized_query] = sym_parenthesized_query,
  [sym_binary_query] = sym_binary_query,
  [sym_unary_query] = sym_unary_query,
  [sym_selector_query] = sym_selector_query,
  [sym__value] = sym__value,
  [sym_parenthesized_value] = sym_parenthesized_value,
  [sym_color_value] = sym_color_value,
  [sym_string_value] = sym_string_value,
  [sym_integer_value] = sym_integer_value,
  [sym_float_value] = sym_float_value,
  [sym_grid_value] = sym_grid_value,
  [sym_call_expression] = sym_call_expression,
  [sym_binary_expression] = sym_binary_expression,
  [sym_arguments] = sym_arguments,
  [sym_class_name] = sym_class_name,
  [aux_sym_stylesheet_repeat1] = aux_sym_stylesheet_repeat1,
  [aux_sym_import_statement_repeat1] = aux_sym_import_statement_repeat1,
  [aux_sym_keyframe_block_list_repeat1] = aux_sym_keyframe_block_list_repeat1,
  [aux_sym_keyframe_block_repeat1] = aux_sym_keyframe_block_repeat1,
  [aux_sym_postcss_statement_repeat1] = aux_sym_postcss_statement_repeat1,
  [aux_sym_selectors_repeat1] = aux_sym_selectors_repeat1,
  [aux_sym_block_repeat1] = aux_sym_block_repeat1,
  [aux_sym_pseudo_class_arguments_repeat1] = aux_sym_pseudo_class_arguments_repeat1,
  [aux_sym_pseudo_class_arguments_repeat2] = aux_sym_pseudo_class_arguments_repeat2,
  [aux_sym_declaration_repeat1] = aux_sym_declaration_repeat1,
  [aux_sym_string_value_repeat1] = aux_sym_string_value_repeat1,
  [aux_sym_string_value_repeat2] = aux_sym_string_value_repeat2,
  [aux_sym_grid_value_repeat1] = aux_sym_grid_value_repeat1,
  [aux_sym_arguments_repeat1] = aux_sym_arguments_repeat1,
  [aux_sym_class_name_repeat1] = aux_sym_class_name_repeat1,
  [alias_sym_attribute_name] = alias_sym_attribute_name,
  [alias_sym_feature_name] = alias_sym_feature_name,
  [alias_sym_function_name] = alias_sym_function_name,
  [alias_sym_id_name] = alias_sym_id_name,
  [alias_sym_keyframes_name] = alias_sym_keyframes_name,
  [alias_sym_keyword_query] = alias_sym_keyword_query,
  [alias_sym_namespace_name] = alias_sym_namespace_name,
  [alias_sym_property_name] = alias_sym_property_name,
  [alias_sym_tag_name] = alias_sym_tag_name,
};

static const TSSymbolMetadata ts_symbol_metadata[] = {
  [ts_builtin_sym_end] = {
    .visible = false,
    .named = true,
  },
  [anon_sym_ATimport] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COMMA] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SEMI] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_ATmedia] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_ATcharset] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_ATnamespace] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_ATkeyframes] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_keyframes_statement_token1] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_LBRACE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RBRACE] = {
    .visible = true,
    .named = false,
  },
  [sym_from] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_to] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_ATsupports] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_ATscope] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LPAREN] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RPAREN] = {
    .visible = true,
    .named = false,
  },
  [sym_nesting_selector] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_STAR] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DOT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_has] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_not] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_is] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_where] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_host] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_host_DASHcontext] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_nth_DASHchild] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_nth_DASHlast_DASHchild] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_COLON_COLON] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_POUND] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_TILDE_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_CARET_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PIPE_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_STAR_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DOLLAR_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RBRACK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_TILDE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PLUS] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PIPE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LPAREN2] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_even] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_odd] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_of] = {
    .visible = true,
    .named = false,
  },
  [sym__nth_functional_notation] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_COLON] = {
    .visible = true,
    .named = false,
  },
  [sym_important] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_and] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_or] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_only] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_selector] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_color_value_token1] = {
    .visible = false,
    .named = false,
  },
  [anon_sym_SQUOTE] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_string_value_token1] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_DQUOTE] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_string_value_token2] = {
    .visible = true,
    .named = true,
  },
  [sym_escape_sequence] = {
    .visible = true,
    .named = true,
  },
  [aux_sym_integer_value_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_float_value_token1] = {
    .visible = false,
    .named = false,
  },
  [sym_unit] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_LBRACK2] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SLASH] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_class_name_token1] = {
    .visible = true,
    .named = true,
  },
  [sym_identifier] = {
    .visible = true,
    .named = true,
  },
  [sym_at_keyword] = {
    .visible = true,
    .named = true,
  },
  [sym_js_comment] = {
    .visible = true,
    .named = true,
  },
  [sym_comment] = {
    .visible = true,
    .named = true,
  },
  [sym_plain_value] = {
    .visible = true,
    .named = true,
  },
  [sym_important_value] = {
    .visible = true,
    .named = true,
  },
  [sym__descendant_operator] = {
    .visible = false,
    .named = true,
  },
  [sym__pseudo_class_selector_colon] = {
    .visible = true,
    .named = false,
  },
  [sym___error_recovery] = {
    .visible = false,
    .named = true,
  },
  [sym_stylesheet] = {
    .visible = true,
    .named = true,
  },
  [sym_import_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_media_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_charset_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_namespace_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_keyframes_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_keyframe_block_list] = {
    .visible = true,
    .named = true,
  },
  [sym_keyframe_block] = {
    .visible = true,
    .named = true,
  },
  [sym_to] = {
    .visible = true,
    .named = true,
  },
  [sym_supports_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_scope_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_postcss_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_at_rule] = {
    .visible = true,
    .named = true,
  },
  [sym_rule_set] = {
    .visible = true,
    .named = true,
  },
  [sym_selectors] = {
    .visible = true,
    .named = true,
  },
  [sym_block] = {
    .visible = true,
    .named = true,
  },
  [sym__selector] = {
    .visible = false,
    .named = true,
  },
  [sym_universal_selector] = {
    .visible = true,
    .named = true,
  },
  [sym_class_selector] = {
    .visible = true,
    .named = true,
  },
  [sym_pseudo_class_selector] = {
    .visible = true,
    .named = true,
  },
  [sym__nth_child_pseudo_class_selector] = {
    .visible = false,
    .named = true,
  },
  [sym_pseudo_element_selector] = {
    .visible = true,
    .named = true,
  },
  [sym_id_selector] = {
    .visible = true,
    .named = true,
  },
  [sym_attribute_selector] = {
    .visible = true,
    .named = true,
  },
  [sym_child_selector] = {
    .visible = true,
    .named = true,
  },
  [sym_descendant_selector] = {
    .visible = true,
    .named = true,
  },
  [sym_sibling_selector] = {
    .visible = true,
    .named = true,
  },
  [sym_adjacent_sibling_selector] = {
    .visible = true,
    .named = true,
  },
  [sym_namespace_selector] = {
    .visible = true,
    .named = true,
  },
  [sym_pseudo_class_arguments] = {
    .visible = true,
    .named = true,
  },
  [sym_pseudo_class_with_selector_arguments] = {
    .visible = true,
    .named = true,
  },
  [sym_pseudo_class_nth_child_arguments] = {
    .visible = true,
    .named = true,
  },
  [sym_pseudo_element_arguments] = {
    .visible = true,
    .named = true,
  },
  [sym_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_last_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym__query] = {
    .visible = false,
    .named = true,
  },
  [sym_feature_query] = {
    .visible = true,
    .named = true,
  },
  [sym_parenthesized_query] = {
    .visible = true,
    .named = true,
  },
  [sym_binary_query] = {
    .visible = true,
    .named = true,
  },
  [sym_unary_query] = {
    .visible = true,
    .named = true,
  },
  [sym_selector_query] = {
    .visible = true,
    .named = true,
  },
  [sym__value] = {
    .visible = false,
    .named = true,
  },
  [sym_parenthesized_value] = {
    .visible = true,
    .named = true,
  },
  [sym_color_value] = {
    .visible = true,
    .named = true,
  },
  [sym_string_value] = {
    .visible = true,
    .named = true,
  },
  [sym_integer_value] = {
    .visible = true,
    .named = true,
  },
  [sym_float_value] = {
    .visible = true,
    .named = true,
  },
  [sym_grid_value] = {
    .visible = true,
    .named = true,
  },
  [sym_call_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_binary_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_arguments] = {
    .visible = true,
    .named = true,
  },
  [sym_class_name] = {
    .visible = true,
    .named = true,
  },
  [aux_sym_stylesheet_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_import_statement_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_keyframe_block_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_keyframe_block_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_postcss_statement_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_selectors_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_block_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_pseudo_class_arguments_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_pseudo_class_arguments_repeat2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_declaration_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_string_value_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_string_value_repeat2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_grid_value_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_arguments_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_class_name_repeat1] = {
    .visible = false,
    .named = false,
  },
  [alias_sym_attribute_name] = {
    .visible = true,
    .named = true,
  },
  [alias_sym_feature_name] = {
    .visible = true,
    .named = true,
  },
  [alias_sym_function_name] = {
    .visible = true,
    .named = true,
  },
  [alias_sym_id_name] = {
    .visible = true,
    .named = true,
  },
  [alias_sym_keyframes_name] = {
    .visible = true,
    .named = true,
  },
  [alias_sym_keyword_query] = {
    .visible = true,
    .named = true,
  },
  [alias_sym_namespace_name] = {
    .visible = true,
    .named = true,
  },
  [alias_sym_property_name] = {
    .visible = true,
    .named = true,
  },
  [alias_sym_tag_name] = {
    .visible = true,
    .named = true,
  },
};

static const TSSymbol ts_alias_sequences[PRODUCTION_ID_COUNT][MAX_ALIAS_SEQUENCE_LENGTH] = {
  [0] = {0},
  [1] = {
    [0] = alias_sym_tag_name,
  },
  [2] = {
    [1] = sym_class_name,
  },
  [3] = {
    [0] = sym_plain_value,
  },
  [4] = {
    [0] = alias_sym_keyword_query,
  },
  [5] = {
    [1] = alias_sym_tag_name,
  },
  [6] = {
    [1] = alias_sym_id_name,
  },
  [7] = {
    [0] = alias_sym_function_name,
  },
  [8] = {
    [1] = alias_sym_keyframes_name,
  },
  [9] = {
    [1] = alias_sym_attribute_name,
  },
  [10] = {
    [2] = sym_class_name,
  },
  [11] = {
    [2] = alias_sym_tag_name,
  },
  [12] = {
    [2] = alias_sym_id_name,
  },
  [13] = {
    [1] = alias_sym_namespace_name,
  },
  [14] = {
    [0] = alias_sym_property_name,
  },
  [15] = {
    [2] = alias_sym_attribute_name,
  },
  [16] = {
    [1] = alias_sym_feature_name,
  },
};

static const uint16_t ts_non_terminal_alias_map[] = {
  sym_namespace_selector, 2,
    sym_namespace_selector,
    alias_sym_attribute_name,
  0,
};

static const TSStateId ts_primary_state_ids[STATE_COUNT] = {
  [0] = 0,
  [1] = 1,
  [2] = 2,
  [3] = 3,
  [4] = 2,
  [5] = 3,
  [6] = 2,
  [7] = 3,
  [8] = 8,
  [9] = 9,
  [10] = 10,
  [11] = 11,
  [12] = 12,
  [13] = 13,
  [14] = 14,
  [15] = 15,
  [16] = 16,
  [17] = 15,
  [18] = 18,
  [19] = 19,
  [20] = 20,
  [21] = 21,
  [22] = 20,
  [23] = 23,
  [24] = 24,
  [25] = 25,
  [26] = 26,
  [27] = 27,
  [28] = 28,
  [29] = 29,
  [30] = 23,
  [31] = 31,
  [32] = 32,
  [33] = 33,
  [34] = 34,
  [35] = 35,
  [36] = 36,
  [37] = 37,
  [38] = 38,
  [39] = 37,
  [40] = 40,
  [41] = 41,
  [42] = 42,
  [43] = 43,
  [44] = 44,
  [45] = 45,
  [46] = 46,
  [47] = 47,
  [48] = 48,
  [49] = 49,
  [50] = 50,
  [51] = 51,
  [52] = 52,
  [53] = 53,
  [54] = 54,
  [55] = 55,
  [56] = 56,
  [57] = 57,
  [58] = 58,
  [59] = 59,
  [60] = 60,
  [61] = 61,
  [62] = 62,
  [63] = 63,
  [64] = 64,
  [65] = 65,
  [66] = 66,
  [67] = 67,
  [68] = 68,
  [69] = 69,
  [70] = 40,
  [71] = 47,
  [72] = 48,
  [73] = 41,
  [74] = 42,
  [75] = 43,
  [76] = 76,
  [77] = 44,
  [78] = 45,
  [79] = 46,
  [80] = 49,
  [81] = 53,
  [82] = 54,
  [83] = 55,
  [84] = 56,
  [85] = 58,
  [86] = 60,
  [87] = 61,
  [88] = 62,
  [89] = 63,
  [90] = 65,
  [91] = 66,
  [92] = 67,
  [93] = 68,
  [94] = 69,
  [95] = 95,
  [96] = 96,
  [97] = 97,
  [98] = 52,
  [99] = 52,
  [100] = 52,
  [101] = 76,
  [102] = 102,
  [103] = 103,
  [104] = 104,
  [105] = 105,
  [106] = 106,
  [107] = 107,
  [108] = 108,
  [109] = 109,
  [110] = 110,
  [111] = 111,
  [112] = 112,
  [113] = 113,
  [114] = 114,
  [115] = 115,
  [116] = 116,
  [117] = 117,
  [118] = 118,
  [119] = 117,
  [120] = 120,
  [121] = 121,
  [122] = 121,
  [123] = 121,
  [124] = 121,
  [125] = 125,
  [126] = 126,
  [127] = 127,
  [128] = 96,
  [129] = 129,
  [130] = 130,
  [131] = 95,
  [132] = 132,
  [133] = 133,
  [134] = 134,
  [135] = 135,
  [136] = 136,
  [137] = 137,
  [138] = 138,
  [139] = 139,
  [140] = 140,
  [141] = 141,
  [142] = 142,
  [143] = 143,
  [144] = 144,
  [145] = 145,
  [146] = 146,
  [147] = 147,
  [148] = 148,
  [149] = 149,
  [150] = 150,
  [151] = 151,
  [152] = 152,
  [153] = 153,
  [154] = 154,
  [155] = 155,
  [156] = 156,
  [157] = 157,
  [158] = 158,
  [159] = 159,
  [160] = 160,
  [161] = 161,
  [162] = 162,
  [163] = 163,
  [164] = 164,
  [165] = 165,
  [166] = 166,
  [167] = 167,
  [168] = 168,
  [169] = 169,
  [170] = 170,
  [171] = 171,
  [172] = 172,
  [173] = 173,
  [174] = 174,
  [175] = 175,
  [176] = 175,
  [177] = 177,
  [178] = 178,
  [179] = 175,
  [180] = 180,
  [181] = 175,
  [182] = 182,
  [183] = 169,
  [184] = 172,
  [185] = 180,
  [186] = 170,
  [187] = 172,
  [188] = 180,
  [189] = 172,
  [190] = 180,
  [191] = 182,
  [192] = 192,
  [193] = 193,
  [194] = 194,
  [195] = 195,
  [196] = 196,
  [197] = 197,
  [198] = 198,
  [199] = 199,
  [200] = 200,
  [201] = 201,
  [202] = 150,
  [203] = 95,
  [204] = 204,
  [205] = 96,
  [206] = 206,
  [207] = 207,
  [208] = 208,
  [209] = 173,
  [210] = 210,
  [211] = 211,
  [212] = 212,
  [213] = 213,
  [214] = 214,
  [215] = 215,
  [216] = 174,
  [217] = 217,
  [218] = 204,
  [219] = 95,
  [220] = 96,
  [221] = 221,
  [222] = 196,
  [223] = 206,
  [224] = 195,
  [225] = 225,
  [226] = 226,
  [227] = 197,
  [228] = 226,
  [229] = 229,
  [230] = 225,
  [231] = 198,
  [232] = 199,
  [233] = 221,
  [234] = 200,
  [235] = 193,
  [236] = 201,
  [237] = 194,
  [238] = 238,
  [239] = 239,
  [240] = 240,
  [241] = 241,
  [242] = 242,
  [243] = 243,
  [244] = 239,
  [245] = 245,
  [246] = 246,
  [247] = 247,
  [248] = 204,
  [249] = 150,
  [250] = 250,
  [251] = 251,
  [252] = 252,
  [253] = 252,
  [254] = 254,
  [255] = 255,
  [256] = 256,
  [257] = 257,
  [258] = 258,
  [259] = 173,
  [260] = 174,
  [261] = 261,
  [262] = 262,
  [263] = 263,
  [264] = 258,
  [265] = 254,
  [266] = 266,
  [267] = 261,
  [268] = 195,
  [269] = 201,
  [270] = 193,
  [271] = 96,
  [272] = 272,
  [273] = 196,
  [274] = 197,
  [275] = 150,
  [276] = 199,
  [277] = 200,
  [278] = 278,
  [279] = 206,
  [280] = 95,
  [281] = 198,
  [282] = 194,
  [283] = 174,
  [284] = 173,
  [285] = 204,
  [286] = 286,
  [287] = 201,
  [288] = 288,
  [289] = 95,
  [290] = 96,
  [291] = 195,
  [292] = 196,
  [293] = 197,
  [294] = 198,
  [295] = 295,
  [296] = 288,
  [297] = 295,
  [298] = 193,
  [299] = 194,
  [300] = 199,
  [301] = 200,
  [302] = 206,
  [303] = 303,
  [304] = 304,
  [305] = 303,
  [306] = 303,
  [307] = 304,
  [308] = 308,
  [309] = 303,
  [310] = 308,
  [311] = 311,
  [312] = 312,
  [313] = 313,
  [314] = 314,
  [315] = 315,
  [316] = 316,
  [317] = 313,
  [318] = 318,
  [319] = 319,
  [320] = 320,
  [321] = 321,
  [322] = 322,
  [323] = 320,
  [324] = 324,
  [325] = 325,
  [326] = 326,
  [327] = 327,
  [328] = 325,
  [329] = 329,
  [330] = 327,
  [331] = 321,
  [332] = 329,
  [333] = 320,
  [334] = 334,
  [335] = 335,
  [336] = 320,
  [337] = 337,
  [338] = 338,
  [339] = 339,
  [340] = 340,
  [341] = 341,
  [342] = 342,
  [343] = 337,
  [344] = 344,
  [345] = 345,
  [346] = 346,
  [347] = 347,
  [348] = 347,
  [349] = 349,
  [350] = 349,
  [351] = 337,
  [352] = 46,
  [353] = 353,
  [354] = 346,
  [355] = 338,
  [356] = 342,
  [357] = 357,
  [358] = 347,
  [359] = 349,
  [360] = 58,
  [361] = 338,
  [362] = 342,
  [363] = 337,
  [364] = 347,
  [365] = 349,
  [366] = 353,
  [367] = 65,
  [368] = 368,
  [369] = 338,
  [370] = 342,
  [371] = 347,
  [372] = 349,
  [373] = 342,
  [374] = 337,
  [375] = 347,
  [376] = 349,
  [377] = 377,
  [378] = 378,
  [379] = 342,
  [380] = 380,
  [381] = 337,
  [382] = 382,
  [383] = 383,
  [384] = 384,
  [385] = 385,
  [386] = 386,
  [387] = 387,
  [388] = 173,
  [389] = 389,
  [390] = 387,
  [391] = 391,
  [392] = 392,
  [393] = 385,
  [394] = 387,
  [395] = 387,
  [396] = 386,
  [397] = 173,
  [398] = 398,
  [399] = 399,
  [400] = 389,
  [401] = 401,
  [402] = 402,
  [403] = 403,
  [404] = 404,
  [405] = 404,
  [406] = 406,
  [407] = 407,
  [408] = 408,
  [409] = 409,
  [410] = 410,
  [411] = 411,
  [412] = 412,
  [413] = 413,
  [414] = 411,
  [415] = 415,
  [416] = 410,
  [417] = 417,
  [418] = 418,
  [419] = 419,
  [420] = 420,
  [421] = 421,
  [422] = 420,
  [423] = 420,
  [424] = 421,
  [425] = 425,
  [426] = 419,
  [427] = 427,
  [428] = 428,
  [429] = 429,
  [430] = 430,
  [431] = 425,
  [432] = 421,
  [433] = 433,
  [434] = 428,
  [435] = 419,
  [436] = 427,
  [437] = 437,
  [438] = 420,
  [439] = 439,
  [440] = 440,
  [441] = 437,
};

static const TSCharacterRange sym_plain_value_character_set_1[] = {
  {0, 0x08}, {0x0e, 0x1f}, {'"', '\''}, {'+', '+'}, {'-', ':'}, {'<', 'Z'}, {'\\', '\\'}, {'^', 'z'},
  {'|', '|'}, {'~', 0x10ffff},
};

static const TSCharacterRange sym_plain_value_character_set_2[] = {
  {0, 0x08}, {0x0e, 0x1f}, {'"', '\''}, {'*', '+'}, {'-', ':'}, {'<', 'Z'}, {'\\', '\\'}, {'^', 'z'},
  {'|', '|'}, {'~', 0x10ffff},
};

static bool ts_lex(TSLexer *lexer, TSStateId state) {
  START_LEXER();
  eof = lexer->eof(lexer);
  switch (state) {
    case 0:
      if (eof) ADVANCE(140);
      ADVANCE_MAP(
        '!', 84,
        '"', 255,
        '#', 180,
        '$', 55,
        '&', 157,
        '\'', 249,
        '(', 196,
        ')', 156,
        '*', 158,
        '+', 193,
        ',', 142,
        '-', 298,
        '.', 160,
        '/', 300,
        ':', 204,
        ';', 143,
        '=', 182,
        '>', 189,
        '@', 61,
        'E', 45,
        '[', 181,
        '\\', 131,
        ']', 188,
        '^', 56,
        'a', 95,
        'e', 43,
        'f', 109,
        'h', 58,
        'i', 112,
        'n', 100,
        'o', 71,
        's', 75,
        't', 101,
        'w', 82,
        '{', 149,
        '|', 195,
        '}', 150,
        '~', 191,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(137);
      if (('A' <= lookahead && lookahead <= 'F') ||
          ('b' <= lookahead && lookahead <= 'd')) ADVANCE(133);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(270);
      END_STATE();
    case 1:
      if (lookahead == '\n') SKIP(1);
      if (lookahead == '"') ADVANCE(255);
      if (lookahead == '/') ADVANCE(256);
      if (lookahead == '\\') ADVANCE(131);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(259);
      if (lookahead != 0) ADVANCE(260);
      END_STATE();
    case 2:
      if (lookahead == '\n') SKIP(2);
      if (lookahead == '\'') ADVANCE(249);
      if (lookahead == '/') ADVANCE(250);
      if (lookahead == '\\') ADVANCE(131);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(253);
      if (lookahead != 0) ADVANCE(254);
      END_STATE();
    case 3:
      ADVANCE_MAP(
        '!', 84,
        '"', 255,
        '#', 180,
        '%', 294,
        '\'', 249,
        '(', 155,
        '*', 158,
        '+', 193,
        '-', 297,
        '.', 128,
        '/', 301,
        ';', 143,
        '[', 295,
        '_', 306,
        'E', 290,
        'e', 290,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(7);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(268);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(291);
      if ((0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 4:
      ADVANCE_MAP(
        '!', 84,
        '"', 255,
        '#', 180,
        '\'', 249,
        '(', 196,
        '*', 158,
        '+', 193,
        ',', 142,
        '-', 297,
        '.', 128,
        '/', 301,
        ';', 143,
        '[', 295,
        '_', 306,
        'a', 314,
        'o', 318,
        '{', 149,
        'E', 303,
        'e', 303,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(6);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(268);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(323);
      if ((0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 5:
      ADVANCE_MAP(
        '!', 84,
        '"', 255,
        '#', 180,
        '\'', 249,
        '(', 196,
        '*', 158,
        '+', 193,
        '-', 297,
        '.', 128,
        '/', 301,
        ';', 143,
        '[', 295,
        '_', 306,
        'E', 303,
        'e', 303,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(7);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(268);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(323);
      if ((0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 6:
      ADVANCE_MAP(
        '!', 84,
        '"', 255,
        '#', 180,
        '\'', 249,
        '(', 155,
        '*', 158,
        '+', 193,
        ',', 142,
        '-', 297,
        '.', 128,
        '/', 301,
        ';', 143,
        '[', 295,
        '_', 306,
        'a', 314,
        'o', 318,
        '{', 149,
        'E', 303,
        'e', 303,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(6);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(268);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(323);
      if ((0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 7:
      ADVANCE_MAP(
        '!', 84,
        '"', 255,
        '#', 180,
        '\'', 249,
        '(', 155,
        '*', 158,
        '+', 193,
        '-', 297,
        '.', 128,
        '/', 301,
        ';', 143,
        '[', 295,
        '_', 306,
        'E', 303,
        'e', 303,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(7);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(268);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(323);
      if ((0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 8:
      ADVANCE_MAP(
        '!', 84,
        '"', 255,
        '#', 180,
        '\'', 249,
        '(', 155,
        '+', 50,
        '-', 49,
        '.', 128,
        '/', 33,
        ';', 143,
        '[', 295,
        '_', 306,
        'n', 315,
        'o', 313,
        's', 310,
        '{', 149,
        'E', 303,
        'e', 303,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(8);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(268);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(323);
      if ((0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 9:
      ADVANCE_MAP(
        '!', 84,
        '"', 255,
        '#', 180,
        '\'', 249,
        '(', 155,
        '+', 50,
        '-', 49,
        '.', 128,
        '/', 33,
        ';', 143,
        '[', 295,
        '_', 306,
        'E', 303,
        'e', 303,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(9);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(268);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(323);
      if ((0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 10:
      ADVANCE_MAP(
        '!', 85,
        '"', 255,
        '#', 180,
        '%', 294,
        '\'', 249,
        '(', 155,
        ')', 156,
        '*', 158,
        '+', 193,
        ',', 142,
        '-', 297,
        '.', 128,
        '/', 301,
        ';', 143,
        '[', 295,
        '_', 306,
        '}', 150,
        'E', 290,
        'e', 290,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(15);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(268);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(291);
      if ((0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 11:
      ADVANCE_MAP(
        '!', 85,
        '"', 255,
        '#', 180,
        '&', 157,
        '\'', 249,
        '(', 155,
        ')', 156,
        '*', 158,
        '+', 193,
        '-', 49,
        '.', 160,
        '/', 33,
        ':', 54,
        '>', 189,
        '[', 181,
        '_', 306,
        '|', 194,
        '~', 190,
        'E', 303,
        'e', 303,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(11);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(268);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(323);
      if ((0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 12:
      ADVANCE_MAP(
        '!', 85,
        '"', 255,
        '#', 180,
        '\'', 249,
        '(', 196,
        ')', 156,
        '*', 158,
        '+', 193,
        ',', 142,
        '-', 297,
        '.', 160,
        '/', 301,
        ':', 54,
        '>', 189,
        '[', 181,
        '_', 306,
        '|', 194,
        '~', 190,
        'E', 303,
        'e', 303,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(14);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(268);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(323);
      if ((0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 13:
      ADVANCE_MAP(
        '!', 85,
        '"', 255,
        '#', 180,
        '\'', 249,
        '(', 196,
        ')', 156,
        '*', 158,
        '+', 193,
        ',', 142,
        '-', 297,
        '.', 128,
        '/', 301,
        ';', 143,
        '[', 295,
        '_', 306,
        '}', 150,
        'E', 303,
        'e', 303,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(15);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(268);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(323);
      if ((0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 14:
      ADVANCE_MAP(
        '!', 85,
        '"', 255,
        '#', 180,
        '\'', 249,
        '(', 155,
        ')', 156,
        '*', 158,
        '+', 193,
        ',', 142,
        '-', 297,
        '.', 160,
        '/', 301,
        ':', 54,
        '>', 189,
        '[', 181,
        '_', 306,
        '|', 194,
        '~', 190,
        'E', 303,
        'e', 303,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(14);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(268);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(323);
      if ((0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 15:
      ADVANCE_MAP(
        '!', 85,
        '"', 255,
        '#', 180,
        '\'', 249,
        '(', 155,
        ')', 156,
        '*', 158,
        '+', 193,
        ',', 142,
        '-', 297,
        '.', 128,
        '/', 301,
        ';', 143,
        '[', 295,
        '_', 306,
        '}', 150,
        'E', 303,
        'e', 303,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(15);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(268);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(323);
      if ((0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 16:
      ADVANCE_MAP(
        '!', 85,
        '"', 255,
        '#', 180,
        '\'', 249,
        '(', 155,
        ')', 156,
        '+', 50,
        ',', 142,
        '-', 49,
        '.', 128,
        '/', 33,
        ';', 143,
        '[', 295,
        '_', 306,
        '}', 150,
        'E', 303,
        'e', 303,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(16);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(268);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(323);
      if ((0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 17:
      ADVANCE_MAP(
        '!', 85,
        '"', 255,
        '#', 180,
        '\'', 249,
        '(', 155,
        '+', 50,
        '-', 49,
        '.', 128,
        '/', 33,
        '[', 295,
        '_', 306,
        'n', 315,
        'o', 313,
        's', 310,
        'E', 303,
        'e', 303,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(17);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(268);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(323);
      if ((0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 18:
      ADVANCE_MAP(
        '#', 180,
        '$', 55,
        '(', 196,
        ')', 156,
        '*', 57,
        '+', 192,
        ',', 142,
        '.', 159,
        '/', 32,
        ':', 54,
        '=', 182,
        '>', 189,
        '[', 181,
        '\\', 131,
        ']', 188,
        '^', 56,
        '{', 149,
        '|', 195,
        '~', 191,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(19);
      if (('-' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= '_') ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(302);
      END_STATE();
    case 19:
      ADVANCE_MAP(
        '#', 180,
        '$', 55,
        ')', 156,
        '*', 57,
        '+', 192,
        ',', 142,
        '.', 159,
        '/', 32,
        ':', 54,
        '=', 182,
        '>', 189,
        '[', 181,
        '\\', 131,
        ']', 188,
        '^', 56,
        '{', 149,
        '|', 195,
        '~', 191,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(19);
      if (('-' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= '_') ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(302);
      END_STATE();
    case 20:
      ADVANCE_MAP(
        '%', 294,
        '(', 155,
        '*', 158,
        '+', 192,
        '-', 299,
        '/', 300,
        ';', 143,
        'n', 283,
        'o', 282,
        's', 279,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(24);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(289);
      if (lookahead == '_' ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 21:
      ADVANCE_MAP(
        '(', 196,
        ')', 156,
        '*', 158,
        '+', 192,
        ',', 142,
        '-', 296,
        '/', 300,
        ':', 203,
        ';', 143,
        ']', 188,
        'a', 94,
        'o', 79,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(26);
      END_STATE();
    case 22:
      ADVANCE_MAP(
        '(', 196,
        '*', 158,
        '+', 192,
        '-', 299,
        '/', 300,
        ';', 143,
        'n', 350,
        'o', 347,
        's', 335,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(24);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 23:
      ADVANCE_MAP(
        '(', 155,
        '*', 158,
        '+', 192,
        ',', 142,
        '-', 299,
        '/', 300,
        ';', 143,
        'n', 350,
        'o', 347,
        's', 335,
        '{', 149,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(23);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 24:
      ADVANCE_MAP(
        '(', 155,
        '*', 158,
        '+', 192,
        '-', 299,
        '/', 300,
        ';', 143,
        'n', 350,
        'o', 347,
        's', 335,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(24);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 25:
      ADVANCE_MAP(
        '(', 155,
        '-', 134,
        '/', 32,
        ';', 143,
        'n', 350,
        'o', 347,
        's', 335,
        '{', 149,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(25);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 26:
      ADVANCE_MAP(
        ')', 156,
        '*', 158,
        '+', 192,
        ',', 142,
        '-', 296,
        '/', 300,
        ':', 203,
        ';', 143,
        ']', 188,
        'a', 94,
        'o', 79,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(26);
      END_STATE();
    case 27:
      ADVANCE_MAP(
        ')', 156,
        '*', 158,
        '+', 192,
        ',', 142,
        '-', 296,
        '/', 300,
        ';', 143,
        ']', 188,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(27);
      END_STATE();
    case 28:
      ADVANCE_MAP(
        ')', 156,
        '*', 158,
        '+', 192,
        ',', 142,
        '-', 296,
        '/', 300,
        ';', 143,
        ']', 188,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(27);
      if (lookahead == '%' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(294);
      END_STATE();
    case 29:
      ADVANCE_MAP(
        ')', 156,
        ',', 142,
        '/', 32,
        ':', 203,
        ';', 143,
        'a', 94,
        'f', 108,
        'o', 79,
        't', 101,
        '{', 149,
        '}', 150,
        '+', 130,
        '-', 130,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(29);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(272);
      END_STATE();
    case 30:
      if (lookahead == ')') ADVANCE(156);
      if (lookahead == '/') ADVANCE(32);
      if (lookahead == 'o') ADVANCE(293);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(31);
      if (lookahead == '%' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(294);
      END_STATE();
    case 31:
      if (lookahead == ')') ADVANCE(156);
      if (lookahead == '/') ADVANCE(32);
      if (lookahead == 'o') ADVANCE(78);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(31);
      END_STATE();
    case 32:
      if (lookahead == '*') ADVANCE(35);
      if (lookahead == '/') ADVANCE(424);
      END_STATE();
    case 33:
      if (lookahead == '*') ADVANCE(35);
      if (lookahead == '/') ADVANCE(421);
      if ((!eof && set_contains(sym_plain_value_character_set_2, 10, lookahead))) ADVANCE(53);
      END_STATE();
    case 34:
      if (lookahead == '*') ADVANCE(34);
      if (lookahead == '/') ADVANCE(425);
      if (lookahead != 0) ADVANCE(35);
      END_STATE();
    case 35:
      if (lookahead == '*') ADVANCE(34);
      if (lookahead != 0) ADVANCE(35);
      END_STATE();
    case 36:
      if (lookahead == '+') ADVANCE(130);
      if (lookahead == '-') ADVANCE(97);
      if (lookahead == '/') ADVANCE(32);
      if (lookahead == 'e') ADVANCE(124);
      if (lookahead == 'n') ADVANCE(201);
      if (lookahead == 'o') ADVANCE(70);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(36);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(271);
      END_STATE();
    case 37:
      if (lookahead == ',') ADVANCE(142);
      if (lookahead == '/') ADVANCE(32);
      if (lookahead == '{') ADVANCE(149);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(37);
      END_STATE();
    case 38:
      if (lookahead == ',') ADVANCE(142);
      if (lookahead == '/') ADVANCE(32);
      if (lookahead == '{') ADVANCE(149);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(37);
      if (lookahead == '%' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(294);
      END_STATE();
    case 39:
      if (lookahead == '-') ADVANCE(63);
      END_STATE();
    case 40:
      if (lookahead == '-') ADVANCE(134);
      if (lookahead == '/') ADVANCE(32);
      if (lookahead == '\\') ADVANCE(131);
      if (lookahead == 'h') ADVANCE(324);
      if (lookahead == 'i') ADVANCE(355);
      if (lookahead == 'n') ADVANCE(349);
      if (lookahead == 'w') ADVANCE(339);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(40);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 41:
      if (lookahead == '-') ADVANCE(134);
      if (lookahead == '/') ADVANCE(32);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(42);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(133);
      if (('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(368);
      if (('G' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('g' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 42:
      if (lookahead == '-') ADVANCE(134);
      if (lookahead == '/') ADVANCE(32);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(42);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 43:
      if (lookahead == '-') ADVANCE(129);
      if (lookahead == 'v') ADVANCE(76);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(276);
      if (('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(132);
      END_STATE();
    case 44:
      if (lookahead == '-') ADVANCE(129);
      if (lookahead == 'v') ADVANCE(76);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(275);
      END_STATE();
    case 45:
      if (lookahead == '-') ADVANCE(129);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(276);
      if (('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(132);
      END_STATE();
    case 46:
      if (lookahead == '-') ADVANCE(129);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(275);
      END_STATE();
    case 47:
      if (lookahead == '-') ADVANCE(129);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(237);
      if (('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(246);
      END_STATE();
    case 48:
      if (lookahead == '-') ADVANCE(65);
      END_STATE();
    case 49:
      if (lookahead == '.') ADVANCE(128);
      if (lookahead == '/') ADVANCE(135);
      if (lookahead == '-' ||
          lookahead == '_') ADVANCE(306);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(303);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(268);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(323);
      if ((0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 50:
      if (lookahead == '.') ADVANCE(128);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(46);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(268);
      END_STATE();
    case 51:
      if (lookahead == '/') ADVANCE(32);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(51);
      END_STATE();
    case 52:
      if (lookahead == '/') ADVANCE(32);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(51);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(133);
      END_STATE();
    case 53:
      if (lookahead == '/') ADVANCE(135);
      if (lookahead == '-' ||
          lookahead == '_') ADVANCE(53);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(426);
      END_STATE();
    case 54:
      if (lookahead == ':') ADVANCE(179);
      END_STATE();
    case 55:
      if (lookahead == '=') ADVANCE(187);
      END_STATE();
    case 56:
      if (lookahead == '=') ADVANCE(184);
      END_STATE();
    case 57:
      if (lookahead == '=') ADVANCE(186);
      END_STATE();
    case 58:
      if (lookahead == 'a') ADVANCE(113);
      if (lookahead == 'o') ADVANCE(114);
      END_STATE();
    case 59:
      if (lookahead == 'a') ADVANCE(115);
      END_STATE();
    case 60:
      if (lookahead == 'a') ADVANCE(99);
      END_STATE();
    case 61:
      if (lookahead == 'c') ADVANCE(390);
      if (lookahead == 'i') ADVANCE(392);
      if (lookahead == 'k') ADVANCE(379);
      if (lookahead == 'm') ADVANCE(380);
      if (lookahead == 'n') ADVANCE(373);
      if (lookahead == 's') ADVANCE(376);
      if (lookahead == '-' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(419);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_') ADVANCE(420);
      END_STATE();
    case 62:
      if (lookahead == 'c') ADVANCE(103);
      END_STATE();
    case 63:
      if (lookahead == 'c') ADVANCE(81);
      if (lookahead == 'l') ADVANCE(59);
      END_STATE();
    case 64:
      if (lookahead == 'c') ADVANCE(122);
      END_STATE();
    case 65:
      if (lookahead == 'c') ADVANCE(83);
      END_STATE();
    case 66:
      if (lookahead == 'd') ADVANCE(207);
      END_STATE();
    case 67:
      if (lookahead == 'd') ADVANCE(198);
      END_STATE();
    case 68:
      if (lookahead == 'd') ADVANCE(175);
      END_STATE();
    case 69:
      if (lookahead == 'd') ADVANCE(177);
      END_STATE();
    case 70:
      if (lookahead == 'd') ADVANCE(67);
      END_STATE();
    case 71:
      if (lookahead == 'd') ADVANCE(67);
      if (lookahead == 'f') ADVANCE(199);
      if (lookahead == 'n') ADVANCE(88);
      if (lookahead == 'r') ADVANCE(209);
      END_STATE();
    case 72:
      if (lookahead == 'e') ADVANCE(64);
      END_STATE();
    case 73:
      if (lookahead == 'e') ADVANCE(169);
      END_STATE();
    case 74:
      if (lookahead == 'e') ADVANCE(125);
      END_STATE();
    case 75:
      if (lookahead == 'e') ADVANCE(90);
      END_STATE();
    case 76:
      if (lookahead == 'e') ADVANCE(96);
      END_STATE();
    case 77:
      if (lookahead == 'e') ADVANCE(110);
      END_STATE();
    case 78:
      if (lookahead == 'f') ADVANCE(199);
      END_STATE();
    case 79:
      if (lookahead == 'f') ADVANCE(199);
      if (lookahead == 'r') ADVANCE(209);
      END_STATE();
    case 80:
      if (lookahead == 'h') ADVANCE(39);
      END_STATE();
    case 81:
      if (lookahead == 'h') ADVANCE(86);
      END_STATE();
    case 82:
      if (lookahead == 'h') ADVANCE(77);
      END_STATE();
    case 83:
      if (lookahead == 'h') ADVANCE(87);
      END_STATE();
    case 84:
      if (lookahead == 'i') ADVANCE(429);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(436);
      END_STATE();
    case 85:
      if (lookahead == 'i') ADVANCE(93);
      END_STATE();
    case 86:
      if (lookahead == 'i') ADVANCE(89);
      END_STATE();
    case 87:
      if (lookahead == 'i') ADVANCE(91);
      END_STATE();
    case 88:
      if (lookahead == 'l') ADVANCE(126);
      END_STATE();
    case 89:
      if (lookahead == 'l') ADVANCE(68);
      END_STATE();
    case 90:
      if (lookahead == 'l') ADVANCE(72);
      END_STATE();
    case 91:
      if (lookahead == 'l') ADVANCE(69);
      END_STATE();
    case 92:
      if (lookahead == 'm') ADVANCE(151);
      END_STATE();
    case 93:
      if (lookahead == 'm') ADVANCE(106);
      END_STATE();
    case 94:
      if (lookahead == 'n') ADVANCE(66);
      END_STATE();
    case 95:
      if (lookahead == 'n') ADVANCE(66);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(132);
      END_STATE();
    case 96:
      if (lookahead == 'n') ADVANCE(197);
      END_STATE();
    case 97:
      if (lookahead == 'n') ADVANCE(201);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(271);
      END_STATE();
    case 98:
      if (lookahead == 'n') ADVANCE(123);
      END_STATE();
    case 99:
      if (lookahead == 'n') ADVANCE(119);
      END_STATE();
    case 100:
      if (lookahead == 'o') ADVANCE(116);
      if (lookahead == 't') ADVANCE(80);
      END_STATE();
    case 101:
      if (lookahead == 'o') ADVANCE(152);
      END_STATE();
    case 102:
      if (lookahead == 'o') ADVANCE(92);
      END_STATE();
    case 103:
      if (lookahead == 'o') ADVANCE(98);
      END_STATE();
    case 104:
      if (lookahead == 'o') ADVANCE(107);
      END_STATE();
    case 105:
      if (lookahead == 'o') ADVANCE(111);
      END_STATE();
    case 106:
      if (lookahead == 'p') ADVANCE(105);
      END_STATE();
    case 107:
      if (lookahead == 'r') ADVANCE(215);
      END_STATE();
    case 108:
      if (lookahead == 'r') ADVANCE(102);
      END_STATE();
    case 109:
      if (lookahead == 'r') ADVANCE(102);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(132);
      END_STATE();
    case 110:
      if (lookahead == 'r') ADVANCE(73);
      END_STATE();
    case 111:
      if (lookahead == 'r') ADVANCE(121);
      END_STATE();
    case 112:
      if (lookahead == 's') ADVANCE(167);
      END_STATE();
    case 113:
      if (lookahead == 's') ADVANCE(161);
      END_STATE();
    case 114:
      if (lookahead == 's') ADVANCE(117);
      END_STATE();
    case 115:
      if (lookahead == 's') ADVANCE(120);
      END_STATE();
    case 116:
      if (lookahead == 't') ADVANCE(163);
      END_STATE();
    case 117:
      if (lookahead == 't') ADVANCE(171);
      END_STATE();
    case 118:
      if (lookahead == 't') ADVANCE(173);
      END_STATE();
    case 119:
      if (lookahead == 't') ADVANCE(205);
      END_STATE();
    case 120:
      if (lookahead == 't') ADVANCE(48);
      END_STATE();
    case 121:
      if (lookahead == 't') ADVANCE(60);
      END_STATE();
    case 122:
      if (lookahead == 't') ADVANCE(104);
      END_STATE();
    case 123:
      if (lookahead == 't') ADVANCE(74);
      END_STATE();
    case 124:
      if (lookahead == 'v') ADVANCE(76);
      END_STATE();
    case 125:
      if (lookahead == 'x') ADVANCE(118);
      END_STATE();
    case 126:
      if (lookahead == 'y') ADVANCE(211);
      END_STATE();
    case 127:
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(127);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(202);
      END_STATE();
    case 128:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(274);
      END_STATE();
    case 129:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(275);
      END_STATE();
    case 130:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(272);
      END_STATE();
    case 131:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(267);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r') ADVANCE(261);
      END_STATE();
    case 132:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(246);
      END_STATE();
    case 133:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(132);
      END_STATE();
    case 134:
      if (lookahead == '-' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 135:
      if ((!eof && set_contains(sym_plain_value_character_set_1, 10, lookahead))) ADVANCE(53);
      END_STATE();
    case 136:
      if ((!eof && set_contains(sym_plain_value_character_set_1, 10, lookahead))) ADVANCE(426);
      END_STATE();
    case 137:
      if (eof) ADVANCE(140);
      ADVANCE_MAP(
        '!', 84,
        '"', 255,
        '#', 180,
        '$', 55,
        '&', 157,
        '\'', 249,
        '(', 155,
        ')', 156,
        '*', 158,
        '+', 193,
        ',', 142,
        '-', 298,
        '.', 160,
        '/', 300,
        ':', 204,
        ';', 143,
        '=', 182,
        '>', 189,
        '@', 61,
        'E', 46,
        '[', 181,
        '\\', 131,
        ']', 188,
        '^', 56,
        'a', 94,
        'e', 44,
        'f', 108,
        'h', 58,
        'i', 112,
        'n', 100,
        'o', 71,
        's', 75,
        't', 101,
        'w', 82,
        '{', 149,
        '|', 195,
        '}', 150,
        '~', 191,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(137);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(268);
      END_STATE();
    case 138:
      if (eof) ADVANCE(140);
      ADVANCE_MAP(
        '"', 255,
        '#', 180,
        '&', 157,
        '\'', 249,
        '(', 196,
        ')', 156,
        '*', 158,
        '+', 192,
        ',', 142,
        '-', 134,
        '.', 159,
        '/', 32,
        ':', 204,
        ';', 143,
        '>', 189,
        '@', 61,
        '[', 181,
        '\\', 131,
        ']', 188,
        '{', 149,
        '|', 194,
        '}', 150,
        '~', 190,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(139);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 139:
      if (eof) ADVANCE(140);
      ADVANCE_MAP(
        '"', 255,
        '#', 180,
        '&', 157,
        '\'', 249,
        ')', 156,
        '*', 158,
        '+', 192,
        ',', 142,
        '-', 134,
        '.', 159,
        '/', 32,
        ':', 204,
        ';', 143,
        '>', 189,
        '@', 61,
        '[', 181,
        '\\', 131,
        ']', 188,
        '{', 149,
        '|', 194,
        '}', 150,
        '~', 190,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(139);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 140:
      ACCEPT_TOKEN(ts_builtin_sym_end);
      END_STATE();
    case 141:
      ACCEPT_TOKEN(anon_sym_ATimport);
      if (lookahead == 'k') ADVANCE(387);
      if (lookahead == '-' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(419);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_') ADVANCE(420);
      END_STATE();
    case 142:
      ACCEPT_TOKEN(anon_sym_COMMA);
      END_STATE();
    case 143:
      ACCEPT_TOKEN(anon_sym_SEMI);
      END_STATE();
    case 144:
      ACCEPT_TOKEN(anon_sym_ATmedia);
      if (lookahead == 'k') ADVANCE(387);
      if (lookahead == '-' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(419);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_') ADVANCE(420);
      END_STATE();
    case 145:
      ACCEPT_TOKEN(anon_sym_ATcharset);
      if (lookahead == 'k') ADVANCE(387);
      if (lookahead == '-' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(419);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_') ADVANCE(420);
      END_STATE();
    case 146:
      ACCEPT_TOKEN(anon_sym_ATnamespace);
      if (lookahead == 'k') ADVANCE(387);
      if (lookahead == '-' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(419);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_') ADVANCE(420);
      END_STATE();
    case 147:
      ACCEPT_TOKEN(anon_sym_ATkeyframes);
      if (lookahead == 'k') ADVANCE(387);
      if (lookahead == '-' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(419);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_') ADVANCE(420);
      END_STATE();
    case 148:
      ACCEPT_TOKEN(aux_sym_keyframes_statement_token1);
      if (lookahead == 'k') ADVANCE(387);
      if (lookahead == '-' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(419);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_') ADVANCE(420);
      END_STATE();
    case 149:
      ACCEPT_TOKEN(anon_sym_LBRACE);
      END_STATE();
    case 150:
      ACCEPT_TOKEN(anon_sym_RBRACE);
      END_STATE();
    case 151:
      ACCEPT_TOKEN(sym_from);
      END_STATE();
    case 152:
      ACCEPT_TOKEN(anon_sym_to);
      END_STATE();
    case 153:
      ACCEPT_TOKEN(anon_sym_ATsupports);
      if (lookahead == 'k') ADVANCE(387);
      if (lookahead == '-' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(419);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_') ADVANCE(420);
      END_STATE();
    case 154:
      ACCEPT_TOKEN(anon_sym_ATscope);
      if (lookahead == 'k') ADVANCE(387);
      if (lookahead == '-' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(419);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_') ADVANCE(420);
      END_STATE();
    case 155:
      ACCEPT_TOKEN(anon_sym_LPAREN);
      END_STATE();
    case 156:
      ACCEPT_TOKEN(anon_sym_RPAREN);
      END_STATE();
    case 157:
      ACCEPT_TOKEN(sym_nesting_selector);
      END_STATE();
    case 158:
      ACCEPT_TOKEN(anon_sym_STAR);
      END_STATE();
    case 159:
      ACCEPT_TOKEN(anon_sym_DOT);
      END_STATE();
    case 160:
      ACCEPT_TOKEN(anon_sym_DOT);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(274);
      END_STATE();
    case 161:
      ACCEPT_TOKEN(anon_sym_has);
      END_STATE();
    case 162:
      ACCEPT_TOKEN(anon_sym_has);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 163:
      ACCEPT_TOKEN(anon_sym_not);
      END_STATE();
    case 164:
      ACCEPT_TOKEN(anon_sym_not);
      if (lookahead == '%') ADVANCE(294);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(289);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 165:
      ACCEPT_TOKEN(anon_sym_not);
      if (lookahead == '/') ADVANCE(136);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(323);
      if ((!eof && set_contains(sym_plain_value_character_set_2, 10, lookahead))) ADVANCE(426);
      END_STATE();
    case 166:
      ACCEPT_TOKEN(anon_sym_not);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 167:
      ACCEPT_TOKEN(anon_sym_is);
      END_STATE();
    case 168:
      ACCEPT_TOKEN(anon_sym_is);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 169:
      ACCEPT_TOKEN(anon_sym_where);
      END_STATE();
    case 170:
      ACCEPT_TOKEN(anon_sym_where);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 171:
      ACCEPT_TOKEN(anon_sym_host);
      if (lookahead == '-') ADVANCE(62);
      END_STATE();
    case 172:
      ACCEPT_TOKEN(anon_sym_host);
      if (lookahead == '-') ADVANCE(327);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 173:
      ACCEPT_TOKEN(anon_sym_host_DASHcontext);
      END_STATE();
    case 174:
      ACCEPT_TOKEN(anon_sym_host_DASHcontext);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 175:
      ACCEPT_TOKEN(anon_sym_nth_DASHchild);
      END_STATE();
    case 176:
      ACCEPT_TOKEN(anon_sym_nth_DASHchild);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 177:
      ACCEPT_TOKEN(anon_sym_nth_DASHlast_DASHchild);
      END_STATE();
    case 178:
      ACCEPT_TOKEN(anon_sym_nth_DASHlast_DASHchild);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 179:
      ACCEPT_TOKEN(anon_sym_COLON_COLON);
      END_STATE();
    case 180:
      ACCEPT_TOKEN(anon_sym_POUND);
      END_STATE();
    case 181:
      ACCEPT_TOKEN(anon_sym_LBRACK);
      END_STATE();
    case 182:
      ACCEPT_TOKEN(anon_sym_EQ);
      END_STATE();
    case 183:
      ACCEPT_TOKEN(anon_sym_TILDE_EQ);
      END_STATE();
    case 184:
      ACCEPT_TOKEN(anon_sym_CARET_EQ);
      END_STATE();
    case 185:
      ACCEPT_TOKEN(anon_sym_PIPE_EQ);
      END_STATE();
    case 186:
      ACCEPT_TOKEN(anon_sym_STAR_EQ);
      END_STATE();
    case 187:
      ACCEPT_TOKEN(anon_sym_DOLLAR_EQ);
      END_STATE();
    case 188:
      ACCEPT_TOKEN(anon_sym_RBRACK);
      END_STATE();
    case 189:
      ACCEPT_TOKEN(anon_sym_GT);
      END_STATE();
    case 190:
      ACCEPT_TOKEN(anon_sym_TILDE);
      END_STATE();
    case 191:
      ACCEPT_TOKEN(anon_sym_TILDE);
      if (lookahead == '=') ADVANCE(183);
      END_STATE();
    case 192:
      ACCEPT_TOKEN(anon_sym_PLUS);
      END_STATE();
    case 193:
      ACCEPT_TOKEN(anon_sym_PLUS);
      if (lookahead == '.') ADVANCE(128);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(46);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(268);
      END_STATE();
    case 194:
      ACCEPT_TOKEN(anon_sym_PIPE);
      END_STATE();
    case 195:
      ACCEPT_TOKEN(anon_sym_PIPE);
      if (lookahead == '=') ADVANCE(185);
      END_STATE();
    case 196:
      ACCEPT_TOKEN(anon_sym_LPAREN2);
      END_STATE();
    case 197:
      ACCEPT_TOKEN(anon_sym_even);
      END_STATE();
    case 198:
      ACCEPT_TOKEN(anon_sym_odd);
      END_STATE();
    case 199:
      ACCEPT_TOKEN(anon_sym_of);
      END_STATE();
    case 200:
      ACCEPT_TOKEN(anon_sym_of);
      if (lookahead == '%' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(294);
      END_STATE();
    case 201:
      ACCEPT_TOKEN(sym__nth_functional_notation);
      if (lookahead == '+') ADVANCE(127);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(201);
      END_STATE();
    case 202:
      ACCEPT_TOKEN(sym__nth_functional_notation);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(202);
      END_STATE();
    case 203:
      ACCEPT_TOKEN(anon_sym_COLON);
      END_STATE();
    case 204:
      ACCEPT_TOKEN(anon_sym_COLON);
      if (lookahead == ':') ADVANCE(179);
      END_STATE();
    case 205:
      ACCEPT_TOKEN(sym_important);
      END_STATE();
    case 206:
      ACCEPT_TOKEN(sym_important);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(436);
      END_STATE();
    case 207:
      ACCEPT_TOKEN(anon_sym_and);
      END_STATE();
    case 208:
      ACCEPT_TOKEN(anon_sym_and);
      if (lookahead == '/') ADVANCE(136);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(323);
      if ((!eof && set_contains(sym_plain_value_character_set_2, 10, lookahead))) ADVANCE(426);
      END_STATE();
    case 209:
      ACCEPT_TOKEN(anon_sym_or);
      END_STATE();
    case 210:
      ACCEPT_TOKEN(anon_sym_or);
      if (lookahead == '/') ADVANCE(136);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(323);
      if ((!eof && set_contains(sym_plain_value_character_set_2, 10, lookahead))) ADVANCE(426);
      END_STATE();
    case 211:
      ACCEPT_TOKEN(anon_sym_only);
      END_STATE();
    case 212:
      ACCEPT_TOKEN(anon_sym_only);
      if (lookahead == '%') ADVANCE(294);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(289);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 213:
      ACCEPT_TOKEN(anon_sym_only);
      if (lookahead == '/') ADVANCE(136);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(323);
      if ((!eof && set_contains(sym_plain_value_character_set_2, 10, lookahead))) ADVANCE(426);
      END_STATE();
    case 214:
      ACCEPT_TOKEN(anon_sym_only);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 215:
      ACCEPT_TOKEN(anon_sym_selector);
      END_STATE();
    case 216:
      ACCEPT_TOKEN(anon_sym_selector);
      if (lookahead == '%') ADVANCE(294);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(289);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 217:
      ACCEPT_TOKEN(anon_sym_selector);
      if (lookahead == '/') ADVANCE(136);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(323);
      if ((!eof && set_contains(sym_plain_value_character_set_2, 10, lookahead))) ADVANCE(426);
      END_STATE();
    case 218:
      ACCEPT_TOKEN(anon_sym_selector);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 219:
      ACCEPT_TOKEN(aux_sym_color_value_token1);
      END_STATE();
    case 220:
      ACCEPT_TOKEN(aux_sym_color_value_token1);
      if (lookahead == '-') ADVANCE(129);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(275);
      END_STATE();
    case 221:
      ACCEPT_TOKEN(aux_sym_color_value_token1);
      if (lookahead == '-') ADVANCE(129);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(232);
      if (('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(219);
      END_STATE();
    case 222:
      ACCEPT_TOKEN(aux_sym_color_value_token1);
      if (lookahead == '-') ADVANCE(129);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(233);
      if (('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(238);
      END_STATE();
    case 223:
      ACCEPT_TOKEN(aux_sym_color_value_token1);
      if (lookahead == '-') ADVANCE(129);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(234);
      if (('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(240);
      END_STATE();
    case 224:
      ACCEPT_TOKEN(aux_sym_color_value_token1);
      if (lookahead == '-') ADVANCE(129);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(235);
      if (('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(242);
      END_STATE();
    case 225:
      ACCEPT_TOKEN(aux_sym_color_value_token1);
      if (lookahead == '-') ADVANCE(129);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(236);
      if (('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(244);
      END_STATE();
    case 226:
      ACCEPT_TOKEN(aux_sym_color_value_token1);
      if (lookahead == '.') ADVANCE(128);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(46);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(268);
      END_STATE();
    case 227:
      ACCEPT_TOKEN(aux_sym_color_value_token1);
      if (lookahead == '.') ADVANCE(128);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(220);
      if (('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(219);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(226);
      END_STATE();
    case 228:
      ACCEPT_TOKEN(aux_sym_color_value_token1);
      if (lookahead == '.') ADVANCE(128);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(222);
      if (('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(240);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(229);
      END_STATE();
    case 229:
      ACCEPT_TOKEN(aux_sym_color_value_token1);
      if (lookahead == '.') ADVANCE(128);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(221);
      if (('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(238);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(227);
      END_STATE();
    case 230:
      ACCEPT_TOKEN(aux_sym_color_value_token1);
      if (lookahead == '.') ADVANCE(128);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(223);
      if (('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(242);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(228);
      END_STATE();
    case 231:
      ACCEPT_TOKEN(aux_sym_color_value_token1);
      if (lookahead == '.') ADVANCE(128);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(224);
      if (('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(244);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(230);
      END_STATE();
    case 232:
      ACCEPT_TOKEN(aux_sym_color_value_token1);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(275);
      END_STATE();
    case 233:
      ACCEPT_TOKEN(aux_sym_color_value_token1);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(232);
      if (('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(219);
      END_STATE();
    case 234:
      ACCEPT_TOKEN(aux_sym_color_value_token1);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(233);
      if (('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(238);
      END_STATE();
    case 235:
      ACCEPT_TOKEN(aux_sym_color_value_token1);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(234);
      if (('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(240);
      END_STATE();
    case 236:
      ACCEPT_TOKEN(aux_sym_color_value_token1);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(235);
      if (('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(242);
      END_STATE();
    case 237:
      ACCEPT_TOKEN(aux_sym_color_value_token1);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(236);
      if (('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(244);
      END_STATE();
    case 238:
      ACCEPT_TOKEN(aux_sym_color_value_token1);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(219);
      END_STATE();
    case 239:
      ACCEPT_TOKEN(aux_sym_color_value_token1);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(248);
      if (lookahead == '-' ||
          ('G' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('g' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 240:
      ACCEPT_TOKEN(aux_sym_color_value_token1);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(238);
      END_STATE();
    case 241:
      ACCEPT_TOKEN(aux_sym_color_value_token1);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(239);
      if (lookahead == '-' ||
          ('G' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('g' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 242:
      ACCEPT_TOKEN(aux_sym_color_value_token1);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(240);
      END_STATE();
    case 243:
      ACCEPT_TOKEN(aux_sym_color_value_token1);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(241);
      if (lookahead == '-' ||
          ('G' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('g' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 244:
      ACCEPT_TOKEN(aux_sym_color_value_token1);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(242);
      END_STATE();
    case 245:
      ACCEPT_TOKEN(aux_sym_color_value_token1);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(243);
      if (lookahead == '-' ||
          ('G' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('g' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 246:
      ACCEPT_TOKEN(aux_sym_color_value_token1);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(244);
      END_STATE();
    case 247:
      ACCEPT_TOKEN(aux_sym_color_value_token1);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(245);
      if (lookahead == '-' ||
          ('G' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('g' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 248:
      ACCEPT_TOKEN(aux_sym_color_value_token1);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 249:
      ACCEPT_TOKEN(anon_sym_SQUOTE);
      END_STATE();
    case 250:
      ACCEPT_TOKEN(aux_sym_string_value_token1);
      if (lookahead == '*') ADVANCE(252);
      if (lookahead == '/') ADVANCE(254);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\'' &&
          lookahead != '\\') ADVANCE(254);
      END_STATE();
    case 251:
      ACCEPT_TOKEN(aux_sym_string_value_token1);
      if (lookahead == '*') ADVANCE(251);
      if (lookahead == '/') ADVANCE(254);
      if (lookahead == '\n' ||
          lookahead == '\'' ||
          lookahead == '\\') ADVANCE(35);
      if (lookahead != 0) ADVANCE(252);
      END_STATE();
    case 252:
      ACCEPT_TOKEN(aux_sym_string_value_token1);
      if (lookahead == '*') ADVANCE(251);
      if (lookahead == '\n' ||
          lookahead == '\'' ||
          lookahead == '\\') ADVANCE(35);
      if (lookahead != 0) ADVANCE(252);
      END_STATE();
    case 253:
      ACCEPT_TOKEN(aux_sym_string_value_token1);
      if (lookahead == '/') ADVANCE(250);
      if (lookahead == '\t' ||
          (0x0b <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(253);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != '\'' &&
          lookahead != '\\') ADVANCE(254);
      END_STATE();
    case 254:
      ACCEPT_TOKEN(aux_sym_string_value_token1);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\'' &&
          lookahead != '\\') ADVANCE(254);
      END_STATE();
    case 255:
      ACCEPT_TOKEN(anon_sym_DQUOTE);
      END_STATE();
    case 256:
      ACCEPT_TOKEN(aux_sym_string_value_token2);
      if (lookahead == '*') ADVANCE(258);
      if (lookahead == '/') ADVANCE(260);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '"' &&
          lookahead != '\\') ADVANCE(260);
      END_STATE();
    case 257:
      ACCEPT_TOKEN(aux_sym_string_value_token2);
      if (lookahead == '*') ADVANCE(257);
      if (lookahead == '/') ADVANCE(260);
      if (lookahead == '\n' ||
          lookahead == '"' ||
          lookahead == '\\') ADVANCE(35);
      if (lookahead != 0) ADVANCE(258);
      END_STATE();
    case 258:
      ACCEPT_TOKEN(aux_sym_string_value_token2);
      if (lookahead == '*') ADVANCE(257);
      if (lookahead == '\n' ||
          lookahead == '"' ||
          lookahead == '\\') ADVANCE(35);
      if (lookahead != 0) ADVANCE(258);
      END_STATE();
    case 259:
      ACCEPT_TOKEN(aux_sym_string_value_token2);
      if (lookahead == '/') ADVANCE(256);
      if (lookahead == '\t' ||
          (0x0b <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(259);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != '"' &&
          lookahead != '\\') ADVANCE(260);
      END_STATE();
    case 260:
      ACCEPT_TOKEN(aux_sym_string_value_token2);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '"' &&
          lookahead != '\\') ADVANCE(260);
      END_STATE();
    case 261:
      ACCEPT_TOKEN(sym_escape_sequence);
      END_STATE();
    case 262:
      ACCEPT_TOKEN(sym_escape_sequence);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(261);
      END_STATE();
    case 263:
      ACCEPT_TOKEN(sym_escape_sequence);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(261);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(262);
      END_STATE();
    case 264:
      ACCEPT_TOKEN(sym_escape_sequence);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(261);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(263);
      END_STATE();
    case 265:
      ACCEPT_TOKEN(sym_escape_sequence);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(261);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(264);
      END_STATE();
    case 266:
      ACCEPT_TOKEN(sym_escape_sequence);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(261);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(265);
      END_STATE();
    case 267:
      ACCEPT_TOKEN(sym_escape_sequence);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(261);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(266);
      END_STATE();
    case 268:
      ACCEPT_TOKEN(aux_sym_integer_value_token1);
      if (lookahead == '.') ADVANCE(128);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(46);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(268);
      END_STATE();
    case 269:
      ACCEPT_TOKEN(aux_sym_integer_value_token1);
      if (lookahead == '.') ADVANCE(128);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(225);
      if (('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(246);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(231);
      END_STATE();
    case 270:
      ACCEPT_TOKEN(aux_sym_integer_value_token1);
      if (lookahead == '.') ADVANCE(128);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(47);
      if (('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(132);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(269);
      END_STATE();
    case 271:
      ACCEPT_TOKEN(aux_sym_integer_value_token1);
      if (lookahead == 'n') ADVANCE(201);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(271);
      END_STATE();
    case 272:
      ACCEPT_TOKEN(aux_sym_integer_value_token1);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(272);
      END_STATE();
    case 273:
      ACCEPT_TOKEN(aux_sym_float_value_token1);
      if (lookahead == '/') ADVANCE(136);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(273);
      if (lookahead == '-' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(323);
      if ((!eof && set_contains(sym_plain_value_character_set_2, 10, lookahead))) ADVANCE(426);
      END_STATE();
    case 274:
      ACCEPT_TOKEN(aux_sym_float_value_token1);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(46);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(274);
      END_STATE();
    case 275:
      ACCEPT_TOKEN(aux_sym_float_value_token1);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(275);
      END_STATE();
    case 276:
      ACCEPT_TOKEN(aux_sym_float_value_token1);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(237);
      if (('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(246);
      END_STATE();
    case 277:
      ACCEPT_TOKEN(sym_unit);
      if (lookahead == '%') ADVANCE(294);
      if (lookahead == 'c') ADVANCE(287);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(289);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 278:
      ACCEPT_TOKEN(sym_unit);
      if (lookahead == '%') ADVANCE(294);
      if (lookahead == 'e') ADVANCE(277);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(289);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 279:
      ACCEPT_TOKEN(sym_unit);
      if (lookahead == '%') ADVANCE(294);
      if (lookahead == 'e') ADVANCE(281);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(289);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 280:
      ACCEPT_TOKEN(sym_unit);
      if (lookahead == '%') ADVANCE(294);
      if (lookahead == 'l') ADVANCE(288);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(289);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 281:
      ACCEPT_TOKEN(sym_unit);
      if (lookahead == '%') ADVANCE(294);
      if (lookahead == 'l') ADVANCE(278);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(289);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 282:
      ACCEPT_TOKEN(sym_unit);
      if (lookahead == '%') ADVANCE(294);
      if (lookahead == 'n') ADVANCE(280);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(289);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 283:
      ACCEPT_TOKEN(sym_unit);
      if (lookahead == '%') ADVANCE(294);
      if (lookahead == 'o') ADVANCE(286);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(289);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 284:
      ACCEPT_TOKEN(sym_unit);
      if (lookahead == '%') ADVANCE(294);
      if (lookahead == 'o') ADVANCE(285);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(289);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 285:
      ACCEPT_TOKEN(sym_unit);
      if (lookahead == '%') ADVANCE(294);
      if (lookahead == 'r') ADVANCE(216);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(289);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 286:
      ACCEPT_TOKEN(sym_unit);
      if (lookahead == '%') ADVANCE(294);
      if (lookahead == 't') ADVANCE(164);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(289);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 287:
      ACCEPT_TOKEN(sym_unit);
      if (lookahead == '%') ADVANCE(294);
      if (lookahead == 't') ADVANCE(284);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(289);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 288:
      ACCEPT_TOKEN(sym_unit);
      if (lookahead == '%') ADVANCE(294);
      if (lookahead == 'y') ADVANCE(212);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(289);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 289:
      ACCEPT_TOKEN(sym_unit);
      if (lookahead == '%') ADVANCE(294);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(289);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 290:
      ACCEPT_TOKEN(sym_unit);
      if (lookahead == '%') ADVANCE(292);
      if (lookahead == '-') ADVANCE(322);
      if (lookahead == '/') ADVANCE(136);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(273);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(291);
      if (lookahead == '_' ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(323);
      if ((!eof && set_contains(sym_plain_value_character_set_2, 10, lookahead))) ADVANCE(426);
      END_STATE();
    case 291:
      ACCEPT_TOKEN(sym_unit);
      if (lookahead == '%') ADVANCE(292);
      if (lookahead == '/') ADVANCE(136);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(291);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(323);
      if ((!eof && set_contains(sym_plain_value_character_set_2, 10, lookahead))) ADVANCE(426);
      END_STATE();
    case 292:
      ACCEPT_TOKEN(sym_unit);
      if (lookahead == '/') ADVANCE(136);
      if (lookahead == '%' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(292);
      if ((!eof && set_contains(sym_plain_value_character_set_2, 10, lookahead))) ADVANCE(426);
      END_STATE();
    case 293:
      ACCEPT_TOKEN(sym_unit);
      if (lookahead == 'f') ADVANCE(200);
      if (lookahead == '%' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(294);
      END_STATE();
    case 294:
      ACCEPT_TOKEN(sym_unit);
      if (lookahead == '%' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(294);
      END_STATE();
    case 295:
      ACCEPT_TOKEN(anon_sym_LBRACK2);
      END_STATE();
    case 296:
      ACCEPT_TOKEN(anon_sym_DASH);
      END_STATE();
    case 297:
      ACCEPT_TOKEN(anon_sym_DASH);
      if (lookahead == '.') ADVANCE(128);
      if (lookahead == '/') ADVANCE(135);
      if (lookahead == '-' ||
          lookahead == '_') ADVANCE(306);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(303);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(268);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(323);
      if ((0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 298:
      ACCEPT_TOKEN(anon_sym_DASH);
      if (lookahead == '.') ADVANCE(128);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(46);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(268);
      END_STATE();
    case 299:
      ACCEPT_TOKEN(anon_sym_DASH);
      if (lookahead == '-' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 300:
      ACCEPT_TOKEN(anon_sym_SLASH);
      if (lookahead == '*') ADVANCE(35);
      END_STATE();
    case 301:
      ACCEPT_TOKEN(anon_sym_SLASH);
      if (lookahead == '*') ADVANCE(35);
      if (lookahead == '/') ADVANCE(421);
      if ((!eof && set_contains(sym_plain_value_character_set_2, 10, lookahead))) ADVANCE(53);
      END_STATE();
    case 302:
      ACCEPT_TOKEN(aux_sym_class_name_token1);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(302);
      END_STATE();
    case 303:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '-') ADVANCE(322);
      if (lookahead == '/') ADVANCE(136);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(273);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(323);
      if ((!eof && set_contains(sym_plain_value_character_set_2, 10, lookahead))) ADVANCE(426);
      END_STATE();
    case 304:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '-') ADVANCE(328);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 305:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '-') ADVANCE(329);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 306:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '/') ADVANCE(135);
      if (lookahead == '-' ||
          lookahead == '_') ADVANCE(306);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(323);
      if (('0' <= lookahead && lookahead <= '9') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 307:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '/') ADVANCE(136);
      if (lookahead == 'c') ADVANCE(320);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(323);
      if ((!eof && set_contains(sym_plain_value_character_set_2, 10, lookahead))) ADVANCE(426);
      END_STATE();
    case 308:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '/') ADVANCE(136);
      if (lookahead == 'd') ADVANCE(208);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(323);
      if ((!eof && set_contains(sym_plain_value_character_set_2, 10, lookahead))) ADVANCE(426);
      END_STATE();
    case 309:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '/') ADVANCE(136);
      if (lookahead == 'e') ADVANCE(307);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(323);
      if ((!eof && set_contains(sym_plain_value_character_set_2, 10, lookahead))) ADVANCE(426);
      END_STATE();
    case 310:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '/') ADVANCE(136);
      if (lookahead == 'e') ADVANCE(312);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(323);
      if ((!eof && set_contains(sym_plain_value_character_set_2, 10, lookahead))) ADVANCE(426);
      END_STATE();
    case 311:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '/') ADVANCE(136);
      if (lookahead == 'l') ADVANCE(321);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(323);
      if ((!eof && set_contains(sym_plain_value_character_set_2, 10, lookahead))) ADVANCE(426);
      END_STATE();
    case 312:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '/') ADVANCE(136);
      if (lookahead == 'l') ADVANCE(309);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(323);
      if ((!eof && set_contains(sym_plain_value_character_set_2, 10, lookahead))) ADVANCE(426);
      END_STATE();
    case 313:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '/') ADVANCE(136);
      if (lookahead == 'n') ADVANCE(311);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(323);
      if ((!eof && set_contains(sym_plain_value_character_set_2, 10, lookahead))) ADVANCE(426);
      END_STATE();
    case 314:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '/') ADVANCE(136);
      if (lookahead == 'n') ADVANCE(308);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(323);
      if ((!eof && set_contains(sym_plain_value_character_set_2, 10, lookahead))) ADVANCE(426);
      END_STATE();
    case 315:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '/') ADVANCE(136);
      if (lookahead == 'o') ADVANCE(319);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(323);
      if ((!eof && set_contains(sym_plain_value_character_set_2, 10, lookahead))) ADVANCE(426);
      END_STATE();
    case 316:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '/') ADVANCE(136);
      if (lookahead == 'o') ADVANCE(317);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(323);
      if ((!eof && set_contains(sym_plain_value_character_set_2, 10, lookahead))) ADVANCE(426);
      END_STATE();
    case 317:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '/') ADVANCE(136);
      if (lookahead == 'r') ADVANCE(217);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(323);
      if ((!eof && set_contains(sym_plain_value_character_set_2, 10, lookahead))) ADVANCE(426);
      END_STATE();
    case 318:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '/') ADVANCE(136);
      if (lookahead == 'r') ADVANCE(210);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(323);
      if ((!eof && set_contains(sym_plain_value_character_set_2, 10, lookahead))) ADVANCE(426);
      END_STATE();
    case 319:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '/') ADVANCE(136);
      if (lookahead == 't') ADVANCE(165);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(323);
      if ((!eof && set_contains(sym_plain_value_character_set_2, 10, lookahead))) ADVANCE(426);
      END_STATE();
    case 320:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '/') ADVANCE(136);
      if (lookahead == 't') ADVANCE(316);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(323);
      if ((!eof && set_contains(sym_plain_value_character_set_2, 10, lookahead))) ADVANCE(426);
      END_STATE();
    case 321:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '/') ADVANCE(136);
      if (lookahead == 'y') ADVANCE(213);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(323);
      if ((!eof && set_contains(sym_plain_value_character_set_2, 10, lookahead))) ADVANCE(426);
      END_STATE();
    case 322:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '/') ADVANCE(136);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(273);
      if (lookahead == '-' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(323);
      if ((!eof && set_contains(sym_plain_value_character_set_2, 10, lookahead))) ADVANCE(426);
      END_STATE();
    case 323:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '/') ADVANCE(136);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(323);
      if ((!eof && set_contains(sym_plain_value_character_set_2, 10, lookahead))) ADVANCE(426);
      END_STATE();
    case 324:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'a') ADVANCE(356);
      if (lookahead == 'o') ADVANCE(357);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 325:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'a') ADVANCE(358);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 326:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'c') ADVANCE(362);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 327:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'c') ADVANCE(352);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 328:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'c') ADVANCE(338);
      if (lookahead == 'l') ADVANCE(325);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 329:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'c') ADVANCE(340);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 330:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'd') ADVANCE(176);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 331:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'd') ADVANCE(178);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 332:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'e') ADVANCE(326);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 333:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'e') ADVANCE(170);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 334:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'e') ADVANCE(365);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 335:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'e') ADVANCE(345);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 336:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'e') ADVANCE(354);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 337:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'h') ADVANCE(304);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 338:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'h') ADVANCE(341);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 339:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'h') ADVANCE(336);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 340:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'h') ADVANCE(342);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 341:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'i') ADVANCE(344);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 342:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'i') ADVANCE(346);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 343:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'l') ADVANCE(366);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 344:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'l') ADVANCE(330);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 345:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'l') ADVANCE(332);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 346:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'l') ADVANCE(331);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 347:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'n') ADVANCE(343);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 348:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'n') ADVANCE(364);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 349:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'o') ADVANCE(359);
      if (lookahead == 't') ADVANCE(337);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 350:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'o') ADVANCE(359);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 351:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'o') ADVANCE(353);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 352:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'o') ADVANCE(348);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 353:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'r') ADVANCE(218);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 354:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'r') ADVANCE(333);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 355:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 's') ADVANCE(168);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 356:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 's') ADVANCE(162);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 357:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 's') ADVANCE(360);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 358:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 's') ADVANCE(363);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 359:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 't') ADVANCE(166);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 360:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 't') ADVANCE(172);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 361:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 't') ADVANCE(174);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 362:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 't') ADVANCE(351);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 363:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 't') ADVANCE(305);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 364:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 't') ADVANCE(334);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 365:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'x') ADVANCE(361);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 366:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'y') ADVANCE(214);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 367:
      ACCEPT_TOKEN(sym_identifier);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(247);
      if (lookahead == '-' ||
          ('G' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('g' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 368:
      ACCEPT_TOKEN(sym_identifier);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(367);
      if (lookahead == '-' ||
          ('G' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('g' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 369:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xa0 <= lookahead && lookahead <= 0xff)) ADVANCE(369);
      END_STATE();
    case 370:
      ACCEPT_TOKEN(sym_at_keyword);
      if (lookahead == 'a') ADVANCE(404);
      if (lookahead == 'k') ADVANCE(387);
      if (lookahead == '-' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(419);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_') ADVANCE(420);
      END_STATE();
    case 371:
      ACCEPT_TOKEN(sym_at_keyword);
      if (lookahead == 'a') ADVANCE(144);
      if (lookahead == 'k') ADVANCE(387);
      if (lookahead == '-' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(419);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_') ADVANCE(420);
      END_STATE();
    case 372:
      ACCEPT_TOKEN(sym_at_keyword);
      if (lookahead == 'a') ADVANCE(377);
      if (lookahead == 'k') ADVANCE(387);
      if (lookahead == '-' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(419);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_') ADVANCE(420);
      END_STATE();
    case 373:
      ACCEPT_TOKEN(sym_at_keyword);
      if (lookahead == 'a') ADVANCE(393);
      if (lookahead == 'k') ADVANCE(387);
      if (lookahead == '-' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(419);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_') ADVANCE(420);
      END_STATE();
    case 374:
      ACCEPT_TOKEN(sym_at_keyword);
      if (lookahead == 'a') ADVANCE(394);
      if (lookahead == 'k') ADVANCE(387);
      if (lookahead == '-' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(419);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_') ADVANCE(420);
      END_STATE();
    case 375:
      ACCEPT_TOKEN(sym_at_keyword);
      if (lookahead == 'a') ADVANCE(395);
      if (lookahead == 'k') ADVANCE(387);
      if (lookahead == '-' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(419);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_') ADVANCE(420);
      END_STATE();
    case 376:
      ACCEPT_TOKEN(sym_at_keyword);
      if (lookahead == 'c') ADVANCE(397);
      if (lookahead == 'k') ADVANCE(387);
      if (lookahead == 'u') ADVANCE(403);
      if (lookahead == '-' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(419);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_') ADVANCE(420);
      END_STATE();
    case 377:
      ACCEPT_TOKEN(sym_at_keyword);
      if (lookahead == 'c') ADVANCE(382);
      if (lookahead == 'k') ADVANCE(387);
      if (lookahead == '-' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(419);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_') ADVANCE(420);
      END_STATE();
    case 378:
      ACCEPT_TOKEN(sym_at_keyword);
      if (lookahead == 'd') ADVANCE(391);
      if (lookahead == 'k') ADVANCE(387);
      if (lookahead == '-' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(419);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_') ADVANCE(420);
      END_STATE();
    case 379:
      ACCEPT_TOKEN(sym_at_keyword);
      if (lookahead == 'e') ADVANCE(417);
      if (lookahead == 'k') ADVANCE(387);
      if (lookahead == '-' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(419);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_') ADVANCE(420);
      END_STATE();
    case 380:
      ACCEPT_TOKEN(sym_at_keyword);
      if (lookahead == 'e') ADVANCE(378);
      if (lookahead == 'k') ADVANCE(387);
      if (lookahead == '-' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(419);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_') ADVANCE(420);
      END_STATE();
    case 381:
      ACCEPT_TOKEN(sym_at_keyword);
      if (lookahead == 'e') ADVANCE(154);
      if (lookahead == 'k') ADVANCE(387);
      if (lookahead == '-' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(419);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_') ADVANCE(420);
      END_STATE();
    case 382:
      ACCEPT_TOKEN(sym_at_keyword);
      if (lookahead == 'e') ADVANCE(146);
      if (lookahead == 'k') ADVANCE(387);
      if (lookahead == '-' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(419);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_') ADVANCE(420);
      END_STATE();
    case 383:
      ACCEPT_TOKEN(sym_at_keyword);
      if (lookahead == 'e') ADVANCE(412);
      if (lookahead == 'k') ADVANCE(387);
      if (lookahead == '-' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(419);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_') ADVANCE(420);
      END_STATE();
    case 384:
      ACCEPT_TOKEN(sym_at_keyword);
      if (lookahead == 'e') ADVANCE(415);
      if (lookahead == 'k') ADVANCE(387);
      if (lookahead == '-' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(419);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_') ADVANCE(420);
      END_STATE();
    case 385:
      ACCEPT_TOKEN(sym_at_keyword);
      if (lookahead == 'e') ADVANCE(410);
      if (lookahead == 'k') ADVANCE(387);
      if (lookahead == '-' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(419);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_') ADVANCE(420);
      END_STATE();
    case 386:
      ACCEPT_TOKEN(sym_at_keyword);
      if (lookahead == 'e') ADVANCE(411);
      if (lookahead == 'k') ADVANCE(387);
      if (lookahead == '-' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(419);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_') ADVANCE(420);
      END_STATE();
    case 387:
      ACCEPT_TOKEN(sym_at_keyword);
      if (lookahead == 'e') ADVANCE(418);
      if (lookahead == 'k') ADVANCE(387);
      if (lookahead == '-' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(419);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_') ADVANCE(420);
      END_STATE();
    case 388:
      ACCEPT_TOKEN(sym_at_keyword);
      if (lookahead == 'f') ADVANCE(407);
      if (lookahead == 'k') ADVANCE(387);
      if (lookahead == '-' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(419);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_') ADVANCE(420);
      END_STATE();
    case 389:
      ACCEPT_TOKEN(sym_at_keyword);
      if (lookahead == 'f') ADVANCE(408);
      if (lookahead == 'k') ADVANCE(387);
      if (lookahead == '-' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(419);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_') ADVANCE(420);
      END_STATE();
    case 390:
      ACCEPT_TOKEN(sym_at_keyword);
      if (lookahead == 'h') ADVANCE(370);
      if (lookahead == 'k') ADVANCE(387);
      if (lookahead == '-' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(419);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_') ADVANCE(420);
      END_STATE();
    case 391:
      ACCEPT_TOKEN(sym_at_keyword);
      if (lookahead == 'i') ADVANCE(371);
      if (lookahead == 'k') ADVANCE(387);
      if (lookahead == '-' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(419);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_') ADVANCE(420);
      END_STATE();
    case 392:
      ACCEPT_TOKEN(sym_at_keyword);
      if (lookahead == 'k') ADVANCE(387);
      if (lookahead == 'm') ADVANCE(399);
      if (lookahead == '-' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(419);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_') ADVANCE(420);
      END_STATE();
    case 393:
      ACCEPT_TOKEN(sym_at_keyword);
      if (lookahead == 'k') ADVANCE(387);
      if (lookahead == 'm') ADVANCE(383);
      if (lookahead == '-' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(419);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_') ADVANCE(420);
      END_STATE();
    case 394:
      ACCEPT_TOKEN(sym_at_keyword);
      if (lookahead == 'k') ADVANCE(387);
      if (lookahead == 'm') ADVANCE(385);
      if (lookahead == '-' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(419);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_') ADVANCE(420);
      END_STATE();
    case 395:
      ACCEPT_TOKEN(sym_at_keyword);
      if (lookahead == 'k') ADVANCE(387);
      if (lookahead == 'm') ADVANCE(386);
      if (lookahead == '-' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(419);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_') ADVANCE(420);
      END_STATE();
    case 396:
      ACCEPT_TOKEN(sym_at_keyword);
      if (lookahead == 'k') ADVANCE(387);
      if (lookahead == 'o') ADVANCE(405);
      if (lookahead == '-' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(419);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_') ADVANCE(420);
      END_STATE();
    case 397:
      ACCEPT_TOKEN(sym_at_keyword);
      if (lookahead == 'k') ADVANCE(387);
      if (lookahead == 'o') ADVANCE(400);
      if (lookahead == '-' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(419);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_') ADVANCE(420);
      END_STATE();
    case 398:
      ACCEPT_TOKEN(sym_at_keyword);
      if (lookahead == 'k') ADVANCE(387);
      if (lookahead == 'o') ADVANCE(406);
      if (lookahead == '-' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(419);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_') ADVANCE(420);
      END_STATE();
    case 399:
      ACCEPT_TOKEN(sym_at_keyword);
      if (lookahead == 'k') ADVANCE(387);
      if (lookahead == 'p') ADVANCE(396);
      if (lookahead == '-' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(419);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_') ADVANCE(420);
      END_STATE();
    case 400:
      ACCEPT_TOKEN(sym_at_keyword);
      if (lookahead == 'k') ADVANCE(387);
      if (lookahead == 'p') ADVANCE(381);
      if (lookahead == '-' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(419);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_') ADVANCE(420);
      END_STATE();
    case 401:
      ACCEPT_TOKEN(sym_at_keyword);
      if (lookahead == 'k') ADVANCE(387);
      if (lookahead == 'p') ADVANCE(372);
      if (lookahead == '-' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(419);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_') ADVANCE(420);
      END_STATE();
    case 402:
      ACCEPT_TOKEN(sym_at_keyword);
      if (lookahead == 'k') ADVANCE(387);
      if (lookahead == 'p') ADVANCE(398);
      if (lookahead == '-' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(419);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_') ADVANCE(420);
      END_STATE();
    case 403:
      ACCEPT_TOKEN(sym_at_keyword);
      if (lookahead == 'k') ADVANCE(387);
      if (lookahead == 'p') ADVANCE(402);
      if (lookahead == '-' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(419);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_') ADVANCE(420);
      END_STATE();
    case 404:
      ACCEPT_TOKEN(sym_at_keyword);
      if (lookahead == 'k') ADVANCE(387);
      if (lookahead == 'r') ADVANCE(413);
      if (lookahead == '-' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(419);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_') ADVANCE(420);
      END_STATE();
    case 405:
      ACCEPT_TOKEN(sym_at_keyword);
      if (lookahead == 'k') ADVANCE(387);
      if (lookahead == 'r') ADVANCE(414);
      if (lookahead == '-' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(419);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_') ADVANCE(420);
      END_STATE();
    case 406:
      ACCEPT_TOKEN(sym_at_keyword);
      if (lookahead == 'k') ADVANCE(387);
      if (lookahead == 'r') ADVANCE(416);
      if (lookahead == '-' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(419);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_') ADVANCE(420);
      END_STATE();
    case 407:
      ACCEPT_TOKEN(sym_at_keyword);
      if (lookahead == 'k') ADVANCE(387);
      if (lookahead == 'r') ADVANCE(374);
      if (lookahead == '-' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(419);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_') ADVANCE(420);
      END_STATE();
    case 408:
      ACCEPT_TOKEN(sym_at_keyword);
      if (lookahead == 'k') ADVANCE(387);
      if (lookahead == 'r') ADVANCE(375);
      if (lookahead == '-' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(419);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_') ADVANCE(420);
      END_STATE();
    case 409:
      ACCEPT_TOKEN(sym_at_keyword);
      if (lookahead == 'k') ADVANCE(387);
      if (lookahead == 's') ADVANCE(153);
      if (lookahead == '-' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(419);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_') ADVANCE(420);
      END_STATE();
    case 410:
      ACCEPT_TOKEN(sym_at_keyword);
      if (lookahead == 'k') ADVANCE(387);
      if (lookahead == 's') ADVANCE(147);
      if (lookahead == '-' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(419);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_') ADVANCE(420);
      END_STATE();
    case 411:
      ACCEPT_TOKEN(sym_at_keyword);
      if (lookahead == 'k') ADVANCE(387);
      if (lookahead == 's') ADVANCE(148);
      if (lookahead == '-' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(419);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_') ADVANCE(420);
      END_STATE();
    case 412:
      ACCEPT_TOKEN(sym_at_keyword);
      if (lookahead == 'k') ADVANCE(387);
      if (lookahead == 's') ADVANCE(401);
      if (lookahead == '-' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(419);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_') ADVANCE(420);
      END_STATE();
    case 413:
      ACCEPT_TOKEN(sym_at_keyword);
      if (lookahead == 'k') ADVANCE(387);
      if (lookahead == 's') ADVANCE(384);
      if (lookahead == '-' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(419);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_') ADVANCE(420);
      END_STATE();
    case 414:
      ACCEPT_TOKEN(sym_at_keyword);
      if (lookahead == 'k') ADVANCE(387);
      if (lookahead == 't') ADVANCE(141);
      if (lookahead == '-' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(419);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_') ADVANCE(420);
      END_STATE();
    case 415:
      ACCEPT_TOKEN(sym_at_keyword);
      if (lookahead == 'k') ADVANCE(387);
      if (lookahead == 't') ADVANCE(145);
      if (lookahead == '-' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(419);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_') ADVANCE(420);
      END_STATE();
    case 416:
      ACCEPT_TOKEN(sym_at_keyword);
      if (lookahead == 'k') ADVANCE(387);
      if (lookahead == 't') ADVANCE(409);
      if (lookahead == '-' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(419);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_') ADVANCE(420);
      END_STATE();
    case 417:
      ACCEPT_TOKEN(sym_at_keyword);
      if (lookahead == 'k') ADVANCE(387);
      if (lookahead == 'y') ADVANCE(388);
      if (lookahead == '-' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(419);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_') ADVANCE(420);
      END_STATE();
    case 418:
      ACCEPT_TOKEN(sym_at_keyword);
      if (lookahead == 'k') ADVANCE(387);
      if (lookahead == 'y') ADVANCE(389);
      if (lookahead == '-' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(419);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_') ADVANCE(420);
      END_STATE();
    case 419:
      ACCEPT_TOKEN(sym_at_keyword);
      if (lookahead == 'k') ADVANCE(387);
      if (lookahead == '-' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(419);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_') ADVANCE(420);
      END_STATE();
    case 420:
      ACCEPT_TOKEN(sym_at_keyword);
      if (lookahead == '-' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(420);
      END_STATE();
    case 421:
      ACCEPT_TOKEN(sym_js_comment);
      if (lookahead == '/') ADVANCE(422);
      if (lookahead == '-' ||
          lookahead == '_') ADVANCE(421);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(427);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(424);
      END_STATE();
    case 422:
      ACCEPT_TOKEN(sym_js_comment);
      if (lookahead == '\t' ||
          (0x0b <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == '!' ||
          ('(' <= lookahead && lookahead <= '*') ||
          lookahead == ',' ||
          lookahead == ';' ||
          lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(424);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead)) ADVANCE(421);
      END_STATE();
    case 423:
      ACCEPT_TOKEN(sym_js_comment);
      if (lookahead == '\t' ||
          (0x0b <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == '!' ||
          ('(' <= lookahead && lookahead <= '*') ||
          lookahead == ',' ||
          lookahead == ';' ||
          lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(424);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead)) ADVANCE(427);
      END_STATE();
    case 424:
      ACCEPT_TOKEN(sym_js_comment);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(424);
      END_STATE();
    case 425:
      ACCEPT_TOKEN(sym_comment);
      END_STATE();
    case 426:
      ACCEPT_TOKEN(sym_plain_value);
      if (lookahead == '/') ADVANCE(136);
      if ((!eof && set_contains(sym_plain_value_character_set_2, 10, lookahead))) ADVANCE(426);
      END_STATE();
    case 427:
      ACCEPT_TOKEN(sym_plain_value);
      if (lookahead == '/') ADVANCE(423);
      if ((!eof && set_contains(sym_plain_value_character_set_2, 10, lookahead))) ADVANCE(427);
      END_STATE();
    case 428:
      ACCEPT_TOKEN(sym_important_value);
      if (lookahead == 'a') ADVANCE(430);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(436);
      END_STATE();
    case 429:
      ACCEPT_TOKEN(sym_important_value);
      if (lookahead == 'm') ADVANCE(432);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(436);
      END_STATE();
    case 430:
      ACCEPT_TOKEN(sym_important_value);
      if (lookahead == 'n') ADVANCE(435);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(436);
      END_STATE();
    case 431:
      ACCEPT_TOKEN(sym_important_value);
      if (lookahead == 'o') ADVANCE(433);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(436);
      END_STATE();
    case 432:
      ACCEPT_TOKEN(sym_important_value);
      if (lookahead == 'p') ADVANCE(431);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(436);
      END_STATE();
    case 433:
      ACCEPT_TOKEN(sym_important_value);
      if (lookahead == 'r') ADVANCE(434);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(436);
      END_STATE();
    case 434:
      ACCEPT_TOKEN(sym_important_value);
      if (lookahead == 't') ADVANCE(428);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(436);
      END_STATE();
    case 435:
      ACCEPT_TOKEN(sym_important_value);
      if (lookahead == 't') ADVANCE(206);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(436);
      END_STATE();
    case 436:
      ACCEPT_TOKEN(sym_important_value);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(436);
      END_STATE();
    default:
      return false;
  }
}

static const TSLexerMode ts_lex_modes[STATE_COUNT] = {
  [0] = {.lex_state = 0, .external_lex_state = 1},
  [1] = {.lex_state = 138, .external_lex_state = 2},
  [2] = {.lex_state = 138, .external_lex_state = 2},
  [3] = {.lex_state = 138, .external_lex_state = 2},
  [4] = {.lex_state = 138, .external_lex_state = 2},
  [5] = {.lex_state = 138, .external_lex_state = 2},
  [6] = {.lex_state = 138, .external_lex_state = 2},
  [7] = {.lex_state = 138, .external_lex_state = 2},
  [8] = {.lex_state = 138, .external_lex_state = 2},
  [9] = {.lex_state = 138, .external_lex_state = 2},
  [10] = {.lex_state = 138, .external_lex_state = 2},
  [11] = {.lex_state = 11, .external_lex_state = 2},
  [12] = {.lex_state = 11, .external_lex_state = 2},
  [13] = {.lex_state = 11, .external_lex_state = 2},
  [14] = {.lex_state = 8},
  [15] = {.lex_state = 17},
  [16] = {.lex_state = 138, .external_lex_state = 2},
  [17] = {.lex_state = 17},
  [18] = {.lex_state = 138, .external_lex_state = 2},
  [19] = {.lex_state = 138, .external_lex_state = 2},
  [20] = {.lex_state = 138, .external_lex_state = 2},
  [21] = {.lex_state = 138, .external_lex_state = 2},
  [22] = {.lex_state = 138, .external_lex_state = 2},
  [23] = {.lex_state = 138, .external_lex_state = 2},
  [24] = {.lex_state = 138, .external_lex_state = 2},
  [25] = {.lex_state = 138, .external_lex_state = 2},
  [26] = {.lex_state = 138, .external_lex_state = 2},
  [27] = {.lex_state = 138, .external_lex_state = 2},
  [28] = {.lex_state = 138, .external_lex_state = 2},
  [29] = {.lex_state = 15},
  [30] = {.lex_state = 138, .external_lex_state = 2},
  [31] = {.lex_state = 138, .external_lex_state = 2},
  [32] = {.lex_state = 138, .external_lex_state = 2},
  [33] = {.lex_state = 138, .external_lex_state = 2},
  [34] = {.lex_state = 138, .external_lex_state = 2},
  [35] = {.lex_state = 138, .external_lex_state = 2},
  [36] = {.lex_state = 138, .external_lex_state = 2},
  [37] = {.lex_state = 15},
  [38] = {.lex_state = 12, .external_lex_state = 3},
  [39] = {.lex_state = 15},
  [40] = {.lex_state = 138, .external_lex_state = 2},
  [41] = {.lex_state = 138, .external_lex_state = 2},
  [42] = {.lex_state = 138, .external_lex_state = 2},
  [43] = {.lex_state = 138, .external_lex_state = 2},
  [44] = {.lex_state = 138, .external_lex_state = 2},
  [45] = {.lex_state = 138, .external_lex_state = 2},
  [46] = {.lex_state = 138, .external_lex_state = 2},
  [47] = {.lex_state = 138, .external_lex_state = 2},
  [48] = {.lex_state = 138, .external_lex_state = 2},
  [49] = {.lex_state = 138, .external_lex_state = 2},
  [50] = {.lex_state = 18, .external_lex_state = 3},
  [51] = {.lex_state = 14, .external_lex_state = 3},
  [52] = {.lex_state = 16},
  [53] = {.lex_state = 138, .external_lex_state = 2},
  [54] = {.lex_state = 138, .external_lex_state = 2},
  [55] = {.lex_state = 138, .external_lex_state = 2},
  [56] = {.lex_state = 138, .external_lex_state = 2},
  [57] = {.lex_state = 138, .external_lex_state = 2},
  [58] = {.lex_state = 138, .external_lex_state = 2},
  [59] = {.lex_state = 18, .external_lex_state = 3},
  [60] = {.lex_state = 138, .external_lex_state = 2},
  [61] = {.lex_state = 138, .external_lex_state = 2},
  [62] = {.lex_state = 138, .external_lex_state = 2},
  [63] = {.lex_state = 138, .external_lex_state = 2},
  [64] = {.lex_state = 138, .external_lex_state = 2},
  [65] = {.lex_state = 138, .external_lex_state = 2},
  [66] = {.lex_state = 138, .external_lex_state = 2},
  [67] = {.lex_state = 138, .external_lex_state = 2},
  [68] = {.lex_state = 138, .external_lex_state = 2},
  [69] = {.lex_state = 138, .external_lex_state = 2},
  [70] = {.lex_state = 138, .external_lex_state = 2},
  [71] = {.lex_state = 138, .external_lex_state = 2},
  [72] = {.lex_state = 138, .external_lex_state = 2},
  [73] = {.lex_state = 138, .external_lex_state = 2},
  [74] = {.lex_state = 138, .external_lex_state = 2},
  [75] = {.lex_state = 138, .external_lex_state = 2},
  [76] = {.lex_state = 138, .external_lex_state = 2},
  [77] = {.lex_state = 138, .external_lex_state = 2},
  [78] = {.lex_state = 138, .external_lex_state = 2},
  [79] = {.lex_state = 138, .external_lex_state = 2},
  [80] = {.lex_state = 138, .external_lex_state = 2},
  [81] = {.lex_state = 138, .external_lex_state = 2},
  [82] = {.lex_state = 138, .external_lex_state = 2},
  [83] = {.lex_state = 138, .external_lex_state = 2},
  [84] = {.lex_state = 138, .external_lex_state = 2},
  [85] = {.lex_state = 138, .external_lex_state = 2},
  [86] = {.lex_state = 138, .external_lex_state = 2},
  [87] = {.lex_state = 138, .external_lex_state = 2},
  [88] = {.lex_state = 138, .external_lex_state = 2},
  [89] = {.lex_state = 138, .external_lex_state = 2},
  [90] = {.lex_state = 138, .external_lex_state = 2},
  [91] = {.lex_state = 138, .external_lex_state = 2},
  [92] = {.lex_state = 138, .external_lex_state = 2},
  [93] = {.lex_state = 138, .external_lex_state = 2},
  [94] = {.lex_state = 138, .external_lex_state = 2},
  [95] = {.lex_state = 14, .external_lex_state = 3},
  [96] = {.lex_state = 14, .external_lex_state = 3},
  [97] = {.lex_state = 18, .external_lex_state = 3},
  [98] = {.lex_state = 16},
  [99] = {.lex_state = 16},
  [100] = {.lex_state = 16},
  [101] = {.lex_state = 138, .external_lex_state = 2},
  [102] = {.lex_state = 16},
  [103] = {.lex_state = 16},
  [104] = {.lex_state = 16},
  [105] = {.lex_state = 16},
  [106] = {.lex_state = 16},
  [107] = {.lex_state = 16},
  [108] = {.lex_state = 16},
  [109] = {.lex_state = 18, .external_lex_state = 3},
  [110] = {.lex_state = 18, .external_lex_state = 3},
  [111] = {.lex_state = 9},
  [112] = {.lex_state = 18, .external_lex_state = 3},
  [113] = {.lex_state = 4},
  [114] = {.lex_state = 18, .external_lex_state = 3},
  [115] = {.lex_state = 9},
  [116] = {.lex_state = 18, .external_lex_state = 3},
  [117] = {.lex_state = 16},
  [118] = {.lex_state = 18, .external_lex_state = 3},
  [119] = {.lex_state = 16},
  [120] = {.lex_state = 16},
  [121] = {.lex_state = 16},
  [122] = {.lex_state = 16},
  [123] = {.lex_state = 16},
  [124] = {.lex_state = 16},
  [125] = {.lex_state = 18, .external_lex_state = 3},
  [126] = {.lex_state = 18, .external_lex_state = 3},
  [127] = {.lex_state = 18, .external_lex_state = 3},
  [128] = {.lex_state = 18, .external_lex_state = 3},
  [129] = {.lex_state = 16},
  [130] = {.lex_state = 18, .external_lex_state = 3},
  [131] = {.lex_state = 18, .external_lex_state = 3},
  [132] = {.lex_state = 18, .external_lex_state = 3},
  [133] = {.lex_state = 18, .external_lex_state = 3},
  [134] = {.lex_state = 18, .external_lex_state = 3},
  [135] = {.lex_state = 18, .external_lex_state = 3},
  [136] = {.lex_state = 18, .external_lex_state = 3},
  [137] = {.lex_state = 18, .external_lex_state = 3},
  [138] = {.lex_state = 18, .external_lex_state = 3},
  [139] = {.lex_state = 18, .external_lex_state = 3},
  [140] = {.lex_state = 18, .external_lex_state = 3},
  [141] = {.lex_state = 18, .external_lex_state = 3},
  [142] = {.lex_state = 18, .external_lex_state = 3},
  [143] = {.lex_state = 18, .external_lex_state = 3},
  [144] = {.lex_state = 18, .external_lex_state = 3},
  [145] = {.lex_state = 18, .external_lex_state = 3},
  [146] = {.lex_state = 18, .external_lex_state = 3},
  [147] = {.lex_state = 18, .external_lex_state = 3},
  [148] = {.lex_state = 18, .external_lex_state = 3},
  [149] = {.lex_state = 18, .external_lex_state = 3},
  [150] = {.lex_state = 13},
  [151] = {.lex_state = 18, .external_lex_state = 3},
  [152] = {.lex_state = 18, .external_lex_state = 3},
  [153] = {.lex_state = 18, .external_lex_state = 3},
  [154] = {.lex_state = 18, .external_lex_state = 3},
  [155] = {.lex_state = 18, .external_lex_state = 3},
  [156] = {.lex_state = 18, .external_lex_state = 3},
  [157] = {.lex_state = 18, .external_lex_state = 3},
  [158] = {.lex_state = 18, .external_lex_state = 3},
  [159] = {.lex_state = 18, .external_lex_state = 3},
  [160] = {.lex_state = 18, .external_lex_state = 3},
  [161] = {.lex_state = 18, .external_lex_state = 3},
  [162] = {.lex_state = 18, .external_lex_state = 3},
  [163] = {.lex_state = 18, .external_lex_state = 3},
  [164] = {.lex_state = 16},
  [165] = {.lex_state = 18, .external_lex_state = 3},
  [166] = {.lex_state = 18, .external_lex_state = 3},
  [167] = {.lex_state = 18, .external_lex_state = 3},
  [168] = {.lex_state = 16},
  [169] = {.lex_state = 16},
  [170] = {.lex_state = 16},
  [171] = {.lex_state = 16},
  [172] = {.lex_state = 16},
  [173] = {.lex_state = 10},
  [174] = {.lex_state = 10},
  [175] = {.lex_state = 16},
  [176] = {.lex_state = 16},
  [177] = {.lex_state = 16},
  [178] = {.lex_state = 16},
  [179] = {.lex_state = 16},
  [180] = {.lex_state = 16},
  [181] = {.lex_state = 16},
  [182] = {.lex_state = 16},
  [183] = {.lex_state = 16},
  [184] = {.lex_state = 16},
  [185] = {.lex_state = 16},
  [186] = {.lex_state = 16},
  [187] = {.lex_state = 16},
  [188] = {.lex_state = 16},
  [189] = {.lex_state = 16},
  [190] = {.lex_state = 16},
  [191] = {.lex_state = 16},
  [192] = {.lex_state = 16},
  [193] = {.lex_state = 15},
  [194] = {.lex_state = 15},
  [195] = {.lex_state = 15},
  [196] = {.lex_state = 15},
  [197] = {.lex_state = 15},
  [198] = {.lex_state = 15},
  [199] = {.lex_state = 15},
  [200] = {.lex_state = 15},
  [201] = {.lex_state = 15},
  [202] = {.lex_state = 5},
  [203] = {.lex_state = 15},
  [204] = {.lex_state = 15},
  [205] = {.lex_state = 15},
  [206] = {.lex_state = 15},
  [207] = {.lex_state = 18, .external_lex_state = 3},
  [208] = {.lex_state = 15},
  [209] = {.lex_state = 3},
  [210] = {.lex_state = 15},
  [211] = {.lex_state = 15},
  [212] = {.lex_state = 18, .external_lex_state = 3},
  [213] = {.lex_state = 18, .external_lex_state = 3},
  [214] = {.lex_state = 15},
  [215] = {.lex_state = 18, .external_lex_state = 3},
  [216] = {.lex_state = 3},
  [217] = {.lex_state = 15},
  [218] = {.lex_state = 7},
  [219] = {.lex_state = 7},
  [220] = {.lex_state = 7},
  [221] = {.lex_state = 23},
  [222] = {.lex_state = 7},
  [223] = {.lex_state = 7},
  [224] = {.lex_state = 7},
  [225] = {.lex_state = 15},
  [226] = {.lex_state = 15},
  [227] = {.lex_state = 7},
  [228] = {.lex_state = 15},
  [229] = {.lex_state = 7},
  [230] = {.lex_state = 15},
  [231] = {.lex_state = 7},
  [232] = {.lex_state = 7},
  [233] = {.lex_state = 23},
  [234] = {.lex_state = 7},
  [235] = {.lex_state = 7},
  [236] = {.lex_state = 7},
  [237] = {.lex_state = 7},
  [238] = {.lex_state = 25},
  [239] = {.lex_state = 138, .external_lex_state = 3},
  [240] = {.lex_state = 138, .external_lex_state = 3},
  [241] = {.lex_state = 138, .external_lex_state = 3},
  [242] = {.lex_state = 138, .external_lex_state = 3},
  [243] = {.lex_state = 138, .external_lex_state = 3},
  [244] = {.lex_state = 138, .external_lex_state = 3},
  [245] = {.lex_state = 138, .external_lex_state = 3},
  [246] = {.lex_state = 138, .external_lex_state = 3},
  [247] = {.lex_state = 40},
  [248] = {.lex_state = 23},
  [249] = {.lex_state = 22},
  [250] = {.lex_state = 40},
  [251] = {.lex_state = 138, .external_lex_state = 3},
  [252] = {.lex_state = 138, .external_lex_state = 3},
  [253] = {.lex_state = 138, .external_lex_state = 3},
  [254] = {.lex_state = 25},
  [255] = {.lex_state = 138, .external_lex_state = 3},
  [256] = {.lex_state = 25},
  [257] = {.lex_state = 25},
  [258] = {.lex_state = 25},
  [259] = {.lex_state = 20},
  [260] = {.lex_state = 20},
  [261] = {.lex_state = 138, .external_lex_state = 3},
  [262] = {.lex_state = 25},
  [263] = {.lex_state = 25},
  [264] = {.lex_state = 25},
  [265] = {.lex_state = 25},
  [266] = {.lex_state = 138, .external_lex_state = 3},
  [267] = {.lex_state = 138, .external_lex_state = 3},
  [268] = {.lex_state = 23},
  [269] = {.lex_state = 23},
  [270] = {.lex_state = 23},
  [271] = {.lex_state = 23},
  [272] = {.lex_state = 138, .external_lex_state = 3},
  [273] = {.lex_state = 23},
  [274] = {.lex_state = 23},
  [275] = {.lex_state = 21},
  [276] = {.lex_state = 23},
  [277] = {.lex_state = 23},
  [278] = {.lex_state = 21},
  [279] = {.lex_state = 23},
  [280] = {.lex_state = 23},
  [281] = {.lex_state = 23},
  [282] = {.lex_state = 23},
  [283] = {.lex_state = 28},
  [284] = {.lex_state = 28},
  [285] = {.lex_state = 21},
  [286] = {.lex_state = 29},
  [287] = {.lex_state = 21},
  [288] = {.lex_state = 29},
  [289] = {.lex_state = 21},
  [290] = {.lex_state = 21},
  [291] = {.lex_state = 21},
  [292] = {.lex_state = 21},
  [293] = {.lex_state = 21},
  [294] = {.lex_state = 21},
  [295] = {.lex_state = 29},
  [296] = {.lex_state = 29},
  [297] = {.lex_state = 29},
  [298] = {.lex_state = 21},
  [299] = {.lex_state = 21},
  [300] = {.lex_state = 21},
  [301] = {.lex_state = 21},
  [302] = {.lex_state = 21},
  [303] = {.lex_state = 21},
  [304] = {.lex_state = 138},
  [305] = {.lex_state = 21},
  [306] = {.lex_state = 21},
  [307] = {.lex_state = 138},
  [308] = {.lex_state = 29},
  [309] = {.lex_state = 21},
  [310] = {.lex_state = 29},
  [311] = {.lex_state = 29},
  [312] = {.lex_state = 29},
  [313] = {.lex_state = 29},
  [314] = {.lex_state = 29},
  [315] = {.lex_state = 29},
  [316] = {.lex_state = 29},
  [317] = {.lex_state = 29},
  [318] = {.lex_state = 21},
  [319] = {.lex_state = 29},
  [320] = {.lex_state = 21},
  [321] = {.lex_state = 21},
  [322] = {.lex_state = 36},
  [323] = {.lex_state = 21},
  [324] = {.lex_state = 29},
  [325] = {.lex_state = 138},
  [326] = {.lex_state = 21},
  [327] = {.lex_state = 29},
  [328] = {.lex_state = 138},
  [329] = {.lex_state = 138},
  [330] = {.lex_state = 29},
  [331] = {.lex_state = 21},
  [332] = {.lex_state = 138},
  [333] = {.lex_state = 21},
  [334] = {.lex_state = 29},
  [335] = {.lex_state = 21},
  [336] = {.lex_state = 21},
  [337] = {.lex_state = 1},
  [338] = {.lex_state = 138},
  [339] = {.lex_state = 29},
  [340] = {.lex_state = 138},
  [341] = {.lex_state = 138},
  [342] = {.lex_state = 2},
  [343] = {.lex_state = 1},
  [344] = {.lex_state = 138},
  [345] = {.lex_state = 29},
  [346] = {.lex_state = 29},
  [347] = {.lex_state = 2},
  [348] = {.lex_state = 2},
  [349] = {.lex_state = 1},
  [350] = {.lex_state = 1},
  [351] = {.lex_state = 1},
  [352] = {.lex_state = 29},
  [353] = {.lex_state = 138},
  [354] = {.lex_state = 29},
  [355] = {.lex_state = 138},
  [356] = {.lex_state = 2},
  [357] = {.lex_state = 29},
  [358] = {.lex_state = 2},
  [359] = {.lex_state = 1},
  [360] = {.lex_state = 29},
  [361] = {.lex_state = 138},
  [362] = {.lex_state = 2},
  [363] = {.lex_state = 1},
  [364] = {.lex_state = 2},
  [365] = {.lex_state = 1},
  [366] = {.lex_state = 138},
  [367] = {.lex_state = 29},
  [368] = {.lex_state = 138},
  [369] = {.lex_state = 138},
  [370] = {.lex_state = 2},
  [371] = {.lex_state = 2},
  [372] = {.lex_state = 1},
  [373] = {.lex_state = 2},
  [374] = {.lex_state = 1},
  [375] = {.lex_state = 2},
  [376] = {.lex_state = 1},
  [377] = {.lex_state = 2},
  [378] = {.lex_state = 1},
  [379] = {.lex_state = 2},
  [380] = {.lex_state = 138},
  [381] = {.lex_state = 1},
  [382] = {.lex_state = 29},
  [383] = {.lex_state = 138},
  [384] = {.lex_state = 138},
  [385] = {.lex_state = 25},
  [386] = {.lex_state = 138},
  [387] = {.lex_state = 138},
  [388] = {.lex_state = 38},
  [389] = {.lex_state = 29},
  [390] = {.lex_state = 138},
  [391] = {.lex_state = 138},
  [392] = {.lex_state = 138},
  [393] = {.lex_state = 25},
  [394] = {.lex_state = 138},
  [395] = {.lex_state = 138},
  [396] = {.lex_state = 138},
  [397] = {.lex_state = 30},
  [398] = {.lex_state = 138},
  [399] = {.lex_state = 138},
  [400] = {.lex_state = 29},
  [401] = {.lex_state = 138},
  [402] = {.lex_state = 138},
  [403] = {.lex_state = 138},
  [404] = {.lex_state = 138},
  [405] = {.lex_state = 138},
  [406] = {.lex_state = 138},
  [407] = {.lex_state = 138},
  [408] = {.lex_state = 138},
  [409] = {.lex_state = 29},
  [410] = {.lex_state = 138},
  [411] = {.lex_state = 138},
  [412] = {.lex_state = 138},
  [413] = {.lex_state = 138},
  [414] = {.lex_state = 138},
  [415] = {.lex_state = 41},
  [416] = {.lex_state = 138},
  [417] = {.lex_state = 138},
  [418] = {.lex_state = 138},
  [419] = {.lex_state = 138},
  [420] = {.lex_state = 52},
  [421] = {.lex_state = 138},
  [422] = {.lex_state = 52},
  [423] = {.lex_state = 52},
  [424] = {.lex_state = 138},
  [425] = {.lex_state = 138},
  [426] = {.lex_state = 138},
  [427] = {.lex_state = 138},
  [428] = {.lex_state = 138},
  [429] = {.lex_state = 138},
  [430] = {.lex_state = 25},
  [431] = {.lex_state = 138},
  [432] = {.lex_state = 138},
  [433] = {.lex_state = 138},
  [434] = {.lex_state = 138},
  [435] = {.lex_state = 138},
  [436] = {.lex_state = 138},
  [437] = {.lex_state = 25},
  [438] = {.lex_state = 52},
  [439] = {.lex_state = 138},
  [440] = {.lex_state = 138},
  [441] = {.lex_state = 25},
};

static const uint16_t ts_parse_table[LARGE_STATE_COUNT][SYMBOL_COUNT] = {
  [STATE(0)] = {
    [ts_builtin_sym_end] = ACTIONS(1),
    [anon_sym_ATimport] = ACTIONS(1),
    [anon_sym_COMMA] = ACTIONS(1),
    [anon_sym_SEMI] = ACTIONS(1),
    [anon_sym_ATmedia] = ACTIONS(1),
    [anon_sym_ATcharset] = ACTIONS(1),
    [anon_sym_ATnamespace] = ACTIONS(1),
    [anon_sym_ATkeyframes] = ACTIONS(1),
    [aux_sym_keyframes_statement_token1] = ACTIONS(1),
    [anon_sym_LBRACE] = ACTIONS(1),
    [anon_sym_RBRACE] = ACTIONS(1),
    [sym_from] = ACTIONS(1),
    [anon_sym_to] = ACTIONS(1),
    [anon_sym_ATsupports] = ACTIONS(1),
    [anon_sym_ATscope] = ACTIONS(1),
    [anon_sym_LPAREN] = ACTIONS(1),
    [anon_sym_RPAREN] = ACTIONS(1),
    [sym_nesting_selector] = ACTIONS(1),
    [anon_sym_STAR] = ACTIONS(1),
    [anon_sym_DOT] = ACTIONS(1),
    [anon_sym_has] = ACTIONS(1),
    [anon_sym_not] = ACTIONS(1),
    [anon_sym_is] = ACTIONS(1),
    [anon_sym_where] = ACTIONS(1),
    [anon_sym_host] = ACTIONS(1),
    [anon_sym_host_DASHcontext] = ACTIONS(1),
    [anon_sym_nth_DASHchild] = ACTIONS(1),
    [anon_sym_nth_DASHlast_DASHchild] = ACTIONS(1),
    [anon_sym_COLON_COLON] = ACTIONS(1),
    [anon_sym_POUND] = ACTIONS(1),
    [anon_sym_LBRACK] = ACTIONS(1),
    [anon_sym_EQ] = ACTIONS(1),
    [anon_sym_TILDE_EQ] = ACTIONS(1),
    [anon_sym_CARET_EQ] = ACTIONS(1),
    [anon_sym_PIPE_EQ] = ACTIONS(1),
    [anon_sym_DOLLAR_EQ] = ACTIONS(1),
    [anon_sym_RBRACK] = ACTIONS(1),
    [anon_sym_GT] = ACTIONS(1),
    [anon_sym_TILDE] = ACTIONS(1),
    [anon_sym_PLUS] = ACTIONS(1),
    [anon_sym_PIPE] = ACTIONS(1),
    [anon_sym_LPAREN2] = ACTIONS(1),
    [anon_sym_even] = ACTIONS(1),
    [anon_sym_odd] = ACTIONS(1),
    [anon_sym_of] = ACTIONS(1),
    [anon_sym_COLON] = ACTIONS(1),
    [sym_important] = ACTIONS(1),
    [anon_sym_and] = ACTIONS(1),
    [anon_sym_or] = ACTIONS(1),
    [anon_sym_only] = ACTIONS(1),
    [anon_sym_selector] = ACTIONS(1),
    [aux_sym_color_value_token1] = ACTIONS(1),
    [anon_sym_SQUOTE] = ACTIONS(1),
    [anon_sym_DQUOTE] = ACTIONS(1),
    [sym_escape_sequence] = ACTIONS(1),
    [aux_sym_integer_value_token1] = ACTIONS(1),
    [aux_sym_float_value_token1] = ACTIONS(1),
    [anon_sym_LBRACK2] = ACTIONS(1),
    [anon_sym_DASH] = ACTIONS(1),
    [anon_sym_SLASH] = ACTIONS(1),
    [sym_at_keyword] = ACTIONS(1),
    [sym_js_comment] = ACTIONS(3),
    [sym_comment] = ACTIONS(5),
    [sym_important_value] = ACTIONS(1),
    [sym__descendant_operator] = ACTIONS(1),
    [sym__pseudo_class_selector_colon] = ACTIONS(1),
    [sym___error_recovery] = ACTIONS(1),
  },
  [STATE(1)] = {
    [sym_stylesheet] = STATE(440),
    [sym_import_statement] = STATE(10),
    [sym_media_statement] = STATE(10),
    [sym_charset_statement] = STATE(10),
    [sym_namespace_statement] = STATE(10),
    [sym_keyframes_statement] = STATE(10),
    [sym_supports_statement] = STATE(10),
    [sym_scope_statement] = STATE(10),
    [sym_at_rule] = STATE(10),
    [sym_rule_set] = STATE(10),
    [sym_selectors] = STATE(410),
    [sym__selector] = STATE(243),
    [sym_universal_selector] = STATE(243),
    [sym_class_selector] = STATE(243),
    [sym_pseudo_class_selector] = STATE(243),
    [sym_pseudo_element_selector] = STATE(243),
    [sym_id_selector] = STATE(243),
    [sym_attribute_selector] = STATE(243),
    [sym_child_selector] = STATE(243),
    [sym_descendant_selector] = STATE(243),
    [sym_sibling_selector] = STATE(243),
    [sym_adjacent_sibling_selector] = STATE(243),
    [sym_namespace_selector] = STATE(243),
    [sym_declaration] = STATE(10),
    [sym_string_value] = STATE(243),
    [aux_sym_stylesheet_repeat1] = STATE(10),
    [ts_builtin_sym_end] = ACTIONS(7),
    [anon_sym_ATimport] = ACTIONS(9),
    [anon_sym_ATmedia] = ACTIONS(11),
    [anon_sym_ATcharset] = ACTIONS(13),
    [anon_sym_ATnamespace] = ACTIONS(15),
    [anon_sym_ATkeyframes] = ACTIONS(17),
    [aux_sym_keyframes_statement_token1] = ACTIONS(17),
    [anon_sym_ATsupports] = ACTIONS(19),
    [anon_sym_ATscope] = ACTIONS(21),
    [sym_nesting_selector] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_DOT] = ACTIONS(27),
    [anon_sym_COLON_COLON] = ACTIONS(29),
    [anon_sym_POUND] = ACTIONS(31),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_PLUS] = ACTIONS(39),
    [anon_sym_PIPE] = ACTIONS(41),
    [anon_sym_SQUOTE] = ACTIONS(43),
    [anon_sym_DQUOTE] = ACTIONS(45),
    [sym_identifier] = ACTIONS(47),
    [sym_at_keyword] = ACTIONS(49),
    [sym_js_comment] = ACTIONS(5),
    [sym_comment] = ACTIONS(5),
    [sym__pseudo_class_selector_colon] = ACTIONS(51),
  },
};

static const uint16_t ts_small_parse_table[] = {
  [0] = 28,
    ACTIONS(23), 1,
      sym_nesting_selector,
    ACTIONS(25), 1,
      anon_sym_STAR,
    ACTIONS(27), 1,
      anon_sym_DOT,
    ACTIONS(29), 1,
      anon_sym_COLON_COLON,
    ACTIONS(31), 1,
      anon_sym_POUND,
    ACTIONS(33), 1,
      anon_sym_LBRACK,
    ACTIONS(35), 1,
      anon_sym_GT,
    ACTIONS(37), 1,
      anon_sym_TILDE,
    ACTIONS(39), 1,
      anon_sym_PLUS,
    ACTIONS(41), 1,
      anon_sym_PIPE,
    ACTIONS(43), 1,
      anon_sym_SQUOTE,
    ACTIONS(45), 1,
      anon_sym_DQUOTE,
    ACTIONS(51), 1,
      sym__pseudo_class_selector_colon,
    ACTIONS(53), 1,
      anon_sym_ATimport,
    ACTIONS(55), 1,
      anon_sym_ATmedia,
    ACTIONS(57), 1,
      anon_sym_ATcharset,
    ACTIONS(59), 1,
      anon_sym_ATnamespace,
    ACTIONS(63), 1,
      anon_sym_RBRACE,
    ACTIONS(65), 1,
      anon_sym_ATsupports,
    ACTIONS(67), 1,
      anon_sym_ATscope,
    ACTIONS(69), 1,
      sym_identifier,
    ACTIONS(71), 1,
      sym_at_keyword,
    STATE(416), 1,
      sym_selectors,
    STATE(421), 1,
      sym_last_declaration,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(61), 2,
      anon_sym_ATkeyframes,
      aux_sym_keyframes_statement_token1,
    STATE(3), 12,
      sym_import_statement,
      sym_media_statement,
      sym_charset_statement,
      sym_namespace_statement,
      sym_keyframes_statement,
      sym_supports_statement,
      sym_scope_statement,
      sym_postcss_statement,
      sym_at_rule,
      sym_rule_set,
      sym_declaration,
      aux_sym_block_repeat1,
    STATE(243), 13,
      sym__selector,
      sym_universal_selector,
      sym_class_selector,
      sym_pseudo_class_selector,
      sym_pseudo_element_selector,
      sym_id_selector,
      sym_attribute_selector,
      sym_child_selector,
      sym_descendant_selector,
      sym_sibling_selector,
      sym_adjacent_sibling_selector,
      sym_namespace_selector,
      sym_string_value,
  [110] = 28,
    ACTIONS(23), 1,
      sym_nesting_selector,
    ACTIONS(25), 1,
      anon_sym_STAR,
    ACTIONS(27), 1,
      anon_sym_DOT,
    ACTIONS(29), 1,
      anon_sym_COLON_COLON,
    ACTIONS(31), 1,
      anon_sym_POUND,
    ACTIONS(33), 1,
      anon_sym_LBRACK,
    ACTIONS(35), 1,
      anon_sym_GT,
    ACTIONS(37), 1,
      anon_sym_TILDE,
    ACTIONS(39), 1,
      anon_sym_PLUS,
    ACTIONS(41), 1,
      anon_sym_PIPE,
    ACTIONS(43), 1,
      anon_sym_SQUOTE,
    ACTIONS(45), 1,
      anon_sym_DQUOTE,
    ACTIONS(51), 1,
      sym__pseudo_class_selector_colon,
    ACTIONS(53), 1,
      anon_sym_ATimport,
    ACTIONS(55), 1,
      anon_sym_ATmedia,
    ACTIONS(57), 1,
      anon_sym_ATcharset,
    ACTIONS(59), 1,
      anon_sym_ATnamespace,
    ACTIONS(65), 1,
      anon_sym_ATsupports,
    ACTIONS(67), 1,
      anon_sym_ATscope,
    ACTIONS(69), 1,
      sym_identifier,
    ACTIONS(71), 1,
      sym_at_keyword,
    ACTIONS(73), 1,
      anon_sym_RBRACE,
    STATE(416), 1,
      sym_selectors,
    STATE(419), 1,
      sym_last_declaration,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(61), 2,
      anon_sym_ATkeyframes,
      aux_sym_keyframes_statement_token1,
    STATE(8), 12,
      sym_import_statement,
      sym_media_statement,
      sym_charset_statement,
      sym_namespace_statement,
      sym_keyframes_statement,
      sym_supports_statement,
      sym_scope_statement,
      sym_postcss_statement,
      sym_at_rule,
      sym_rule_set,
      sym_declaration,
      aux_sym_block_repeat1,
    STATE(243), 13,
      sym__selector,
      sym_universal_selector,
      sym_class_selector,
      sym_pseudo_class_selector,
      sym_pseudo_element_selector,
      sym_id_selector,
      sym_attribute_selector,
      sym_child_selector,
      sym_descendant_selector,
      sym_sibling_selector,
      sym_adjacent_sibling_selector,
      sym_namespace_selector,
      sym_string_value,
  [220] = 28,
    ACTIONS(23), 1,
      sym_nesting_selector,
    ACTIONS(25), 1,
      anon_sym_STAR,
    ACTIONS(27), 1,
      anon_sym_DOT,
    ACTIONS(29), 1,
      anon_sym_COLON_COLON,
    ACTIONS(31), 1,
      anon_sym_POUND,
    ACTIONS(33), 1,
      anon_sym_LBRACK,
    ACTIONS(35), 1,
      anon_sym_GT,
    ACTIONS(37), 1,
      anon_sym_TILDE,
    ACTIONS(39), 1,
      anon_sym_PLUS,
    ACTIONS(41), 1,
      anon_sym_PIPE,
    ACTIONS(43), 1,
      anon_sym_SQUOTE,
    ACTIONS(45), 1,
      anon_sym_DQUOTE,
    ACTIONS(51), 1,
      sym__pseudo_class_selector_colon,
    ACTIONS(53), 1,
      anon_sym_ATimport,
    ACTIONS(55), 1,
      anon_sym_ATmedia,
    ACTIONS(57), 1,
      anon_sym_ATcharset,
    ACTIONS(59), 1,
      anon_sym_ATnamespace,
    ACTIONS(65), 1,
      anon_sym_ATsupports,
    ACTIONS(67), 1,
      anon_sym_ATscope,
    ACTIONS(69), 1,
      sym_identifier,
    ACTIONS(71), 1,
      sym_at_keyword,
    ACTIONS(75), 1,
      anon_sym_RBRACE,
    STATE(416), 1,
      sym_selectors,
    STATE(424), 1,
      sym_last_declaration,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(61), 2,
      anon_sym_ATkeyframes,
      aux_sym_keyframes_statement_token1,
    STATE(5), 12,
      sym_import_statement,
      sym_media_statement,
      sym_charset_statement,
      sym_namespace_statement,
      sym_keyframes_statement,
      sym_supports_statement,
      sym_scope_statement,
      sym_postcss_statement,
      sym_at_rule,
      sym_rule_set,
      sym_declaration,
      aux_sym_block_repeat1,
    STATE(243), 13,
      sym__selector,
      sym_universal_selector,
      sym_class_selector,
      sym_pseudo_class_selector,
      sym_pseudo_element_selector,
      sym_id_selector,
      sym_attribute_selector,
      sym_child_selector,
      sym_descendant_selector,
      sym_sibling_selector,
      sym_adjacent_sibling_selector,
      sym_namespace_selector,
      sym_string_value,
  [330] = 28,
    ACTIONS(23), 1,
      sym_nesting_selector,
    ACTIONS(25), 1,
      anon_sym_STAR,
    ACTIONS(27), 1,
      anon_sym_DOT,
    ACTIONS(29), 1,
      anon_sym_COLON_COLON,
    ACTIONS(31), 1,
      anon_sym_POUND,
    ACTIONS(33), 1,
      anon_sym_LBRACK,
    ACTIONS(35), 1,
      anon_sym_GT,
    ACTIONS(37), 1,
      anon_sym_TILDE,
    ACTIONS(39), 1,
      anon_sym_PLUS,
    ACTIONS(41), 1,
      anon_sym_PIPE,
    ACTIONS(43), 1,
      anon_sym_SQUOTE,
    ACTIONS(45), 1,
      anon_sym_DQUOTE,
    ACTIONS(51), 1,
      sym__pseudo_class_selector_colon,
    ACTIONS(53), 1,
      anon_sym_ATimport,
    ACTIONS(55), 1,
      anon_sym_ATmedia,
    ACTIONS(57), 1,
      anon_sym_ATcharset,
    ACTIONS(59), 1,
      anon_sym_ATnamespace,
    ACTIONS(65), 1,
      anon_sym_ATsupports,
    ACTIONS(67), 1,
      anon_sym_ATscope,
    ACTIONS(69), 1,
      sym_identifier,
    ACTIONS(71), 1,
      sym_at_keyword,
    ACTIONS(77), 1,
      anon_sym_RBRACE,
    STATE(416), 1,
      sym_selectors,
    STATE(435), 1,
      sym_last_declaration,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(61), 2,
      anon_sym_ATkeyframes,
      aux_sym_keyframes_statement_token1,
    STATE(8), 12,
      sym_import_statement,
      sym_media_statement,
      sym_charset_statement,
      sym_namespace_statement,
      sym_keyframes_statement,
      sym_supports_statement,
      sym_scope_statement,
      sym_postcss_statement,
      sym_at_rule,
      sym_rule_set,
      sym_declaration,
      aux_sym_block_repeat1,
    STATE(243), 13,
      sym__selector,
      sym_universal_selector,
      sym_class_selector,
      sym_pseudo_class_selector,
      sym_pseudo_element_selector,
      sym_id_selector,
      sym_attribute_selector,
      sym_child_selector,
      sym_descendant_selector,
      sym_sibling_selector,
      sym_adjacent_sibling_selector,
      sym_namespace_selector,
      sym_string_value,
  [440] = 28,
    ACTIONS(23), 1,
      sym_nesting_selector,
    ACTIONS(25), 1,
      anon_sym_STAR,
    ACTIONS(27), 1,
      anon_sym_DOT,
    ACTIONS(29), 1,
      anon_sym_COLON_COLON,
    ACTIONS(31), 1,
      anon_sym_POUND,
    ACTIONS(33), 1,
      anon_sym_LBRACK,
    ACTIONS(35), 1,
      anon_sym_GT,
    ACTIONS(37), 1,
      anon_sym_TILDE,
    ACTIONS(39), 1,
      anon_sym_PLUS,
    ACTIONS(41), 1,
      anon_sym_PIPE,
    ACTIONS(43), 1,
      anon_sym_SQUOTE,
    ACTIONS(45), 1,
      anon_sym_DQUOTE,
    ACTIONS(51), 1,
      sym__pseudo_class_selector_colon,
    ACTIONS(53), 1,
      anon_sym_ATimport,
    ACTIONS(55), 1,
      anon_sym_ATmedia,
    ACTIONS(57), 1,
      anon_sym_ATcharset,
    ACTIONS(59), 1,
      anon_sym_ATnamespace,
    ACTIONS(65), 1,
      anon_sym_ATsupports,
    ACTIONS(67), 1,
      anon_sym_ATscope,
    ACTIONS(69), 1,
      sym_identifier,
    ACTIONS(71), 1,
      sym_at_keyword,
    ACTIONS(79), 1,
      anon_sym_RBRACE,
    STATE(416), 1,
      sym_selectors,
    STATE(432), 1,
      sym_last_declaration,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(61), 2,
      anon_sym_ATkeyframes,
      aux_sym_keyframes_statement_token1,
    STATE(7), 12,
      sym_import_statement,
      sym_media_statement,
      sym_charset_statement,
      sym_namespace_statement,
      sym_keyframes_statement,
      sym_supports_statement,
      sym_scope_statement,
      sym_postcss_statement,
      sym_at_rule,
      sym_rule_set,
      sym_declaration,
      aux_sym_block_repeat1,
    STATE(243), 13,
      sym__selector,
      sym_universal_selector,
      sym_class_selector,
      sym_pseudo_class_selector,
      sym_pseudo_element_selector,
      sym_id_selector,
      sym_attribute_selector,
      sym_child_selector,
      sym_descendant_selector,
      sym_sibling_selector,
      sym_adjacent_sibling_selector,
      sym_namespace_selector,
      sym_string_value,
  [550] = 28,
    ACTIONS(23), 1,
      sym_nesting_selector,
    ACTIONS(25), 1,
      anon_sym_STAR,
    ACTIONS(27), 1,
      anon_sym_DOT,
    ACTIONS(29), 1,
      anon_sym_COLON_COLON,
    ACTIONS(31), 1,
      anon_sym_POUND,
    ACTIONS(33), 1,
      anon_sym_LBRACK,
    ACTIONS(35), 1,
      anon_sym_GT,
    ACTIONS(37), 1,
      anon_sym_TILDE,
    ACTIONS(39), 1,
      anon_sym_PLUS,
    ACTIONS(41), 1,
      anon_sym_PIPE,
    ACTIONS(43), 1,
      anon_sym_SQUOTE,
    ACTIONS(45), 1,
      anon_sym_DQUOTE,
    ACTIONS(51), 1,
      sym__pseudo_class_selector_colon,
    ACTIONS(53), 1,
      anon_sym_ATimport,
    ACTIONS(55), 1,
      anon_sym_ATmedia,
    ACTIONS(57), 1,
      anon_sym_ATcharset,
    ACTIONS(59), 1,
      anon_sym_ATnamespace,
    ACTIONS(65), 1,
      anon_sym_ATsupports,
    ACTIONS(67), 1,
      anon_sym_ATscope,
    ACTIONS(69), 1,
      sym_identifier,
    ACTIONS(71), 1,
      sym_at_keyword,
    ACTIONS(81), 1,
      anon_sym_RBRACE,
    STATE(416), 1,
      sym_selectors,
    STATE(426), 1,
      sym_last_declaration,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(61), 2,
      anon_sym_ATkeyframes,
      aux_sym_keyframes_statement_token1,
    STATE(8), 12,
      sym_import_statement,
      sym_media_statement,
      sym_charset_statement,
      sym_namespace_statement,
      sym_keyframes_statement,
      sym_supports_statement,
      sym_scope_statement,
      sym_postcss_statement,
      sym_at_rule,
      sym_rule_set,
      sym_declaration,
      aux_sym_block_repeat1,
    STATE(243), 13,
      sym__selector,
      sym_universal_selector,
      sym_class_selector,
      sym_pseudo_class_selector,
      sym_pseudo_element_selector,
      sym_id_selector,
      sym_attribute_selector,
      sym_child_selector,
      sym_descendant_selector,
      sym_sibling_selector,
      sym_adjacent_sibling_selector,
      sym_namespace_selector,
      sym_string_value,
  [660] = 27,
    ACTIONS(83), 1,
      anon_sym_ATimport,
    ACTIONS(86), 1,
      anon_sym_ATmedia,
    ACTIONS(89), 1,
      anon_sym_ATcharset,
    ACTIONS(92), 1,
      anon_sym_ATnamespace,
    ACTIONS(98), 1,
      anon_sym_RBRACE,
    ACTIONS(100), 1,
      anon_sym_ATsupports,
    ACTIONS(103), 1,
      anon_sym_ATscope,
    ACTIONS(106), 1,
      sym_nesting_selector,
    ACTIONS(109), 1,
      anon_sym_STAR,
    ACTIONS(112), 1,
      anon_sym_DOT,
    ACTIONS(115), 1,
      anon_sym_COLON_COLON,
    ACTIONS(118), 1,
      anon_sym_POUND,
    ACTIONS(121), 1,
      anon_sym_LBRACK,
    ACTIONS(124), 1,
      anon_sym_GT,
    ACTIONS(127), 1,
      anon_sym_TILDE,
    ACTIONS(130), 1,
      anon_sym_PLUS,
    ACTIONS(133), 1,
      anon_sym_PIPE,
    ACTIONS(136), 1,
      anon_sym_SQUOTE,
    ACTIONS(139), 1,
      anon_sym_DQUOTE,
    ACTIONS(142), 1,
      sym_identifier,
    ACTIONS(145), 1,
      sym_at_keyword,
    ACTIONS(148), 1,
      sym__pseudo_class_selector_colon,
    STATE(416), 1,
      sym_selectors,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(95), 2,
      anon_sym_ATkeyframes,
      aux_sym_keyframes_statement_token1,
    STATE(8), 12,
      sym_import_statement,
      sym_media_statement,
      sym_charset_statement,
      sym_namespace_statement,
      sym_keyframes_statement,
      sym_supports_statement,
      sym_scope_statement,
      sym_postcss_statement,
      sym_at_rule,
      sym_rule_set,
      sym_declaration,
      aux_sym_block_repeat1,
    STATE(243), 13,
      sym__selector,
      sym_universal_selector,
      sym_class_selector,
      sym_pseudo_class_selector,
      sym_pseudo_element_selector,
      sym_id_selector,
      sym_attribute_selector,
      sym_child_selector,
      sym_descendant_selector,
      sym_sibling_selector,
      sym_adjacent_sibling_selector,
      sym_namespace_selector,
      sym_string_value,
  [767] = 27,
    ACTIONS(151), 1,
      ts_builtin_sym_end,
    ACTIONS(153), 1,
      anon_sym_ATimport,
    ACTIONS(156), 1,
      anon_sym_ATmedia,
    ACTIONS(159), 1,
      anon_sym_ATcharset,
    ACTIONS(162), 1,
      anon_sym_ATnamespace,
    ACTIONS(168), 1,
      anon_sym_ATsupports,
    ACTIONS(171), 1,
      anon_sym_ATscope,
    ACTIONS(174), 1,
      sym_nesting_selector,
    ACTIONS(177), 1,
      anon_sym_STAR,
    ACTIONS(180), 1,
      anon_sym_DOT,
    ACTIONS(183), 1,
      anon_sym_COLON_COLON,
    ACTIONS(186), 1,
      anon_sym_POUND,
    ACTIONS(189), 1,
      anon_sym_LBRACK,
    ACTIONS(192), 1,
      anon_sym_GT,
    ACTIONS(195), 1,
      anon_sym_TILDE,
    ACTIONS(198), 1,
      anon_sym_PLUS,
    ACTIONS(201), 1,
      anon_sym_PIPE,
    ACTIONS(204), 1,
      anon_sym_SQUOTE,
    ACTIONS(207), 1,
      anon_sym_DQUOTE,
    ACTIONS(210), 1,
      sym_identifier,
    ACTIONS(213), 1,
      sym_at_keyword,
    ACTIONS(216), 1,
      sym__pseudo_class_selector_colon,
    STATE(410), 1,
      sym_selectors,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(165), 2,
      anon_sym_ATkeyframes,
      aux_sym_keyframes_statement_token1,
    STATE(9), 11,
      sym_import_statement,
      sym_media_statement,
      sym_charset_statement,
      sym_namespace_statement,
      sym_keyframes_statement,
      sym_supports_statement,
      sym_scope_statement,
      sym_at_rule,
      sym_rule_set,
      sym_declaration,
      aux_sym_stylesheet_repeat1,
    STATE(243), 13,
      sym__selector,
      sym_universal_selector,
      sym_class_selector,
      sym_pseudo_class_selector,
      sym_pseudo_element_selector,
      sym_id_selector,
      sym_attribute_selector,
      sym_child_selector,
      sym_descendant_selector,
      sym_sibling_selector,
      sym_adjacent_sibling_selector,
      sym_namespace_selector,
      sym_string_value,
  [873] = 27,
    ACTIONS(9), 1,
      anon_sym_ATimport,
    ACTIONS(11), 1,
      anon_sym_ATmedia,
    ACTIONS(13), 1,
      anon_sym_ATcharset,
    ACTIONS(15), 1,
      anon_sym_ATnamespace,
    ACTIONS(19), 1,
      anon_sym_ATsupports,
    ACTIONS(21), 1,
      anon_sym_ATscope,
    ACTIONS(23), 1,
      sym_nesting_selector,
    ACTIONS(25), 1,
      anon_sym_STAR,
    ACTIONS(27), 1,
      anon_sym_DOT,
    ACTIONS(29), 1,
      anon_sym_COLON_COLON,
    ACTIONS(31), 1,
      anon_sym_POUND,
    ACTIONS(33), 1,
      anon_sym_LBRACK,
    ACTIONS(35), 1,
      anon_sym_GT,
    ACTIONS(37), 1,
      anon_sym_TILDE,
    ACTIONS(39), 1,
      anon_sym_PLUS,
    ACTIONS(41), 1,
      anon_sym_PIPE,
    ACTIONS(43), 1,
      anon_sym_SQUOTE,
    ACTIONS(45), 1,
      anon_sym_DQUOTE,
    ACTIONS(47), 1,
      sym_identifier,
    ACTIONS(49), 1,
      sym_at_keyword,
    ACTIONS(51), 1,
      sym__pseudo_class_selector_colon,
    ACTIONS(219), 1,
      ts_builtin_sym_end,
    STATE(410), 1,
      sym_selectors,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(17), 2,
      anon_sym_ATkeyframes,
      aux_sym_keyframes_statement_token1,
    STATE(9), 11,
      sym_import_statement,
      sym_media_statement,
      sym_charset_statement,
      sym_namespace_statement,
      sym_keyframes_statement,
      sym_supports_statement,
      sym_scope_statement,
      sym_at_rule,
      sym_rule_set,
      sym_declaration,
      aux_sym_stylesheet_repeat1,
    STATE(243), 13,
      sym__selector,
      sym_universal_selector,
      sym_class_selector,
      sym_pseudo_class_selector,
      sym_pseudo_element_selector,
      sym_id_selector,
      sym_attribute_selector,
      sym_child_selector,
      sym_descendant_selector,
      sym_sibling_selector,
      sym_adjacent_sibling_selector,
      sym_namespace_selector,
      sym_string_value,
  [979] = 27,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(25), 1,
      anon_sym_STAR,
    ACTIONS(29), 1,
      anon_sym_COLON_COLON,
    ACTIONS(33), 1,
      anon_sym_LBRACK,
    ACTIONS(35), 1,
      anon_sym_GT,
    ACTIONS(37), 1,
      anon_sym_TILDE,
    ACTIONS(41), 1,
      anon_sym_PIPE,
    ACTIONS(51), 1,
      sym__pseudo_class_selector_colon,
    ACTIONS(221), 1,
      anon_sym_LPAREN,
    ACTIONS(223), 1,
      anon_sym_RPAREN,
    ACTIONS(225), 1,
      sym_nesting_selector,
    ACTIONS(227), 1,
      anon_sym_DOT,
    ACTIONS(229), 1,
      anon_sym_POUND,
    ACTIONS(231), 1,
      anon_sym_PLUS,
    ACTIONS(233), 1,
      sym_important,
    ACTIONS(235), 1,
      anon_sym_SQUOTE,
    ACTIONS(237), 1,
      anon_sym_DQUOTE,
    ACTIONS(239), 1,
      aux_sym_integer_value_token1,
    ACTIONS(241), 1,
      aux_sym_float_value_token1,
    ACTIONS(243), 1,
      anon_sym_LBRACK2,
    ACTIONS(245), 1,
      sym_identifier,
    ACTIONS(247), 1,
      sym_plain_value,
    STATE(51), 1,
      sym_string_value,
    STATE(106), 1,
      aux_sym_pseudo_class_arguments_repeat1,
    STATE(217), 8,
      sym__value,
      sym_parenthesized_value,
      sym_color_value,
      sym_integer_value,
      sym_float_value,
      sym_grid_value,
      sym_call_expression,
      sym_binary_expression,
    STATE(241), 12,
      sym__selector,
      sym_universal_selector,
      sym_class_selector,
      sym_pseudo_class_selector,
      sym_pseudo_element_selector,
      sym_id_selector,
      sym_attribute_selector,
      sym_child_selector,
      sym_descendant_selector,
      sym_sibling_selector,
      sym_adjacent_sibling_selector,
      sym_namespace_selector,
  [1079] = 27,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(25), 1,
      anon_sym_STAR,
    ACTIONS(29), 1,
      anon_sym_COLON_COLON,
    ACTIONS(33), 1,
      anon_sym_LBRACK,
    ACTIONS(35), 1,
      anon_sym_GT,
    ACTIONS(37), 1,
      anon_sym_TILDE,
    ACTIONS(41), 1,
      anon_sym_PIPE,
    ACTIONS(51), 1,
      sym__pseudo_class_selector_colon,
    ACTIONS(221), 1,
      anon_sym_LPAREN,
    ACTIONS(227), 1,
      anon_sym_DOT,
    ACTIONS(229), 1,
      anon_sym_POUND,
    ACTIONS(231), 1,
      anon_sym_PLUS,
    ACTIONS(233), 1,
      sym_important,
    ACTIONS(235), 1,
      anon_sym_SQUOTE,
    ACTIONS(237), 1,
      anon_sym_DQUOTE,
    ACTIONS(239), 1,
      aux_sym_integer_value_token1,
    ACTIONS(241), 1,
      aux_sym_float_value_token1,
    ACTIONS(243), 1,
      anon_sym_LBRACK2,
    ACTIONS(245), 1,
      sym_identifier,
    ACTIONS(247), 1,
      sym_plain_value,
    ACTIONS(249), 1,
      anon_sym_RPAREN,
    ACTIONS(251), 1,
      sym_nesting_selector,
    STATE(51), 1,
      sym_string_value,
    STATE(103), 1,
      aux_sym_pseudo_class_arguments_repeat1,
    STATE(217), 8,
      sym__value,
      sym_parenthesized_value,
      sym_color_value,
      sym_integer_value,
      sym_float_value,
      sym_grid_value,
      sym_call_expression,
      sym_binary_expression,
    STATE(246), 12,
      sym__selector,
      sym_universal_selector,
      sym_class_selector,
      sym_pseudo_class_selector,
      sym_pseudo_element_selector,
      sym_id_selector,
      sym_attribute_selector,
      sym_child_selector,
      sym_descendant_selector,
      sym_sibling_selector,
      sym_adjacent_sibling_selector,
      sym_namespace_selector,
  [1179] = 26,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(25), 1,
      anon_sym_STAR,
    ACTIONS(29), 1,
      anon_sym_COLON_COLON,
    ACTIONS(33), 1,
      anon_sym_LBRACK,
    ACTIONS(35), 1,
      anon_sym_GT,
    ACTIONS(37), 1,
      anon_sym_TILDE,
    ACTIONS(41), 1,
      anon_sym_PIPE,
    ACTIONS(51), 1,
      sym__pseudo_class_selector_colon,
    ACTIONS(221), 1,
      anon_sym_LPAREN,
    ACTIONS(227), 1,
      anon_sym_DOT,
    ACTIONS(229), 1,
      anon_sym_POUND,
    ACTIONS(231), 1,
      anon_sym_PLUS,
    ACTIONS(233), 1,
      sym_important,
    ACTIONS(235), 1,
      anon_sym_SQUOTE,
    ACTIONS(237), 1,
      anon_sym_DQUOTE,
    ACTIONS(239), 1,
      aux_sym_integer_value_token1,
    ACTIONS(241), 1,
      aux_sym_float_value_token1,
    ACTIONS(243), 1,
      anon_sym_LBRACK2,
    ACTIONS(245), 1,
      sym_identifier,
    ACTIONS(247), 1,
      sym_plain_value,
    ACTIONS(253), 1,
      sym_nesting_selector,
    STATE(51), 1,
      sym_string_value,
    STATE(108), 1,
      aux_sym_pseudo_class_arguments_repeat1,
    STATE(217), 8,
      sym__value,
      sym_parenthesized_value,
      sym_color_value,
      sym_integer_value,
      sym_float_value,
      sym_grid_value,
      sym_call_expression,
      sym_binary_expression,
    STATE(251), 12,
      sym__selector,
      sym_universal_selector,
      sym_class_selector,
      sym_pseudo_class_selector,
      sym_pseudo_element_selector,
      sym_id_selector,
      sym_attribute_selector,
      sym_child_selector,
      sym_descendant_selector,
      sym_sibling_selector,
      sym_adjacent_sibling_selector,
      sym_namespace_selector,
  [1276] = 20,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(255), 1,
      anon_sym_SEMI,
    ACTIONS(257), 1,
      anon_sym_LBRACE,
    ACTIONS(259), 1,
      anon_sym_LPAREN,
    ACTIONS(263), 1,
      anon_sym_POUND,
    ACTIONS(267), 1,
      anon_sym_selector,
    ACTIONS(269), 1,
      anon_sym_SQUOTE,
    ACTIONS(271), 1,
      anon_sym_DQUOTE,
    ACTIONS(273), 1,
      aux_sym_integer_value_token1,
    ACTIONS(275), 1,
      aux_sym_float_value_token1,
    ACTIONS(277), 1,
      anon_sym_LBRACK2,
    ACTIONS(279), 1,
      sym_identifier,
    ACTIONS(281), 1,
      sym_important_value,
    STATE(71), 1,
      sym_block,
    STATE(115), 1,
      aux_sym_postcss_statement_repeat1,
    ACTIONS(261), 2,
      anon_sym_not,
      anon_sym_only,
    ACTIONS(265), 2,
      sym_important,
      sym_plain_value,
    STATE(308), 6,
      sym__query,
      sym_feature_query,
      sym_parenthesized_query,
      sym_binary_query,
      sym_unary_query,
      sym_selector_query,
    STATE(229), 9,
      sym__value,
      sym_parenthesized_value,
      sym_color_value,
      sym_string_value,
      sym_integer_value,
      sym_float_value,
      sym_grid_value,
      sym_call_expression,
      sym_binary_expression,
  [1352] = 16,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(267), 1,
      anon_sym_selector,
    ACTIONS(283), 1,
      anon_sym_LPAREN,
    ACTIONS(285), 1,
      anon_sym_POUND,
    ACTIONS(287), 1,
      sym_important,
    ACTIONS(289), 1,
      anon_sym_SQUOTE,
    ACTIONS(291), 1,
      anon_sym_DQUOTE,
    ACTIONS(293), 1,
      aux_sym_integer_value_token1,
    ACTIONS(295), 1,
      aux_sym_float_value_token1,
    ACTIONS(297), 1,
      anon_sym_LBRACK2,
    ACTIONS(299), 1,
      sym_identifier,
    ACTIONS(301), 1,
      sym_plain_value,
    ACTIONS(261), 2,
      anon_sym_not,
      anon_sym_only,
    STATE(382), 6,
      sym__query,
      sym_feature_query,
      sym_parenthesized_query,
      sym_binary_query,
      sym_unary_query,
      sym_selector_query,
    STATE(333), 9,
      sym__value,
      sym_parenthesized_value,
      sym_color_value,
      sym_string_value,
      sym_integer_value,
      sym_float_value,
      sym_grid_value,
      sym_call_expression,
      sym_binary_expression,
  [1415] = 17,
    ACTIONS(25), 1,
      anon_sym_STAR,
    ACTIONS(27), 1,
      anon_sym_DOT,
    ACTIONS(29), 1,
      anon_sym_COLON_COLON,
    ACTIONS(31), 1,
      anon_sym_POUND,
    ACTIONS(33), 1,
      anon_sym_LBRACK,
    ACTIONS(35), 1,
      anon_sym_GT,
    ACTIONS(37), 1,
      anon_sym_TILDE,
    ACTIONS(39), 1,
      anon_sym_PLUS,
    ACTIONS(41), 1,
      anon_sym_PIPE,
    ACTIONS(43), 1,
      anon_sym_SQUOTE,
    ACTIONS(45), 1,
      anon_sym_DQUOTE,
    ACTIONS(51), 1,
      sym__pseudo_class_selector_colon,
    ACTIONS(303), 1,
      anon_sym_RPAREN,
    ACTIONS(305), 1,
      sym_nesting_selector,
    ACTIONS(307), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    STATE(240), 13,
      sym__selector,
      sym_universal_selector,
      sym_class_selector,
      sym_pseudo_class_selector,
      sym_pseudo_element_selector,
      sym_id_selector,
      sym_attribute_selector,
      sym_child_selector,
      sym_descendant_selector,
      sym_sibling_selector,
      sym_adjacent_sibling_selector,
      sym_namespace_selector,
      sym_string_value,
  [1480] = 16,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(267), 1,
      anon_sym_selector,
    ACTIONS(283), 1,
      anon_sym_LPAREN,
    ACTIONS(285), 1,
      anon_sym_POUND,
    ACTIONS(289), 1,
      anon_sym_SQUOTE,
    ACTIONS(291), 1,
      anon_sym_DQUOTE,
    ACTIONS(293), 1,
      aux_sym_integer_value_token1,
    ACTIONS(295), 1,
      aux_sym_float_value_token1,
    ACTIONS(297), 1,
      anon_sym_LBRACK2,
    ACTIONS(299), 1,
      sym_identifier,
    ACTIONS(309), 1,
      sym_important,
    ACTIONS(311), 1,
      sym_plain_value,
    ACTIONS(261), 2,
      anon_sym_not,
      anon_sym_only,
    STATE(382), 6,
      sym__query,
      sym_feature_query,
      sym_parenthesized_query,
      sym_binary_query,
      sym_unary_query,
      sym_selector_query,
    STATE(323), 9,
      sym__value,
      sym_parenthesized_value,
      sym_color_value,
      sym_string_value,
      sym_integer_value,
      sym_float_value,
      sym_grid_value,
      sym_call_expression,
      sym_binary_expression,
  [1543] = 16,
    ACTIONS(25), 1,
      anon_sym_STAR,
    ACTIONS(27), 1,
      anon_sym_DOT,
    ACTIONS(29), 1,
      anon_sym_COLON_COLON,
    ACTIONS(31), 1,
      anon_sym_POUND,
    ACTIONS(33), 1,
      anon_sym_LBRACK,
    ACTIONS(35), 1,
      anon_sym_GT,
    ACTIONS(37), 1,
      anon_sym_TILDE,
    ACTIONS(39), 1,
      anon_sym_PLUS,
    ACTIONS(41), 1,
      anon_sym_PIPE,
    ACTIONS(43), 1,
      anon_sym_SQUOTE,
    ACTIONS(45), 1,
      anon_sym_DQUOTE,
    ACTIONS(51), 1,
      sym__pseudo_class_selector_colon,
    ACTIONS(307), 1,
      sym_identifier,
    ACTIONS(313), 1,
      sym_nesting_selector,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    STATE(255), 13,
      sym__selector,
      sym_universal_selector,
      sym_class_selector,
      sym_pseudo_class_selector,
      sym_pseudo_element_selector,
      sym_id_selector,
      sym_attribute_selector,
      sym_child_selector,
      sym_descendant_selector,
      sym_sibling_selector,
      sym_adjacent_sibling_selector,
      sym_namespace_selector,
      sym_string_value,
  [1605] = 17,
    ACTIONS(25), 1,
      anon_sym_STAR,
    ACTIONS(27), 1,
      anon_sym_DOT,
    ACTIONS(29), 1,
      anon_sym_COLON_COLON,
    ACTIONS(31), 1,
      anon_sym_POUND,
    ACTIONS(33), 1,
      anon_sym_LBRACK,
    ACTIONS(35), 1,
      anon_sym_GT,
    ACTIONS(37), 1,
      anon_sym_TILDE,
    ACTIONS(39), 1,
      anon_sym_PLUS,
    ACTIONS(41), 1,
      anon_sym_PIPE,
    ACTIONS(43), 1,
      anon_sym_SQUOTE,
    ACTIONS(45), 1,
      anon_sym_DQUOTE,
    ACTIONS(51), 1,
      sym__pseudo_class_selector_colon,
    ACTIONS(315), 1,
      sym_nesting_selector,
    ACTIONS(317), 1,
      sym_identifier,
    STATE(213), 1,
      sym_namespace_selector,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    STATE(272), 12,
      sym__selector,
      sym_universal_selector,
      sym_class_selector,
      sym_pseudo_class_selector,
      sym_pseudo_element_selector,
      sym_id_selector,
      sym_attribute_selector,
      sym_child_selector,
      sym_descendant_selector,
      sym_sibling_selector,
      sym_adjacent_sibling_selector,
      sym_string_value,
  [1669] = 16,
    ACTIONS(25), 1,
      anon_sym_STAR,
    ACTIONS(27), 1,
      anon_sym_DOT,
    ACTIONS(29), 1,
      anon_sym_COLON_COLON,
    ACTIONS(31), 1,
      anon_sym_POUND,
    ACTIONS(33), 1,
      anon_sym_LBRACK,
    ACTIONS(35), 1,
      anon_sym_GT,
    ACTIONS(37), 1,
      anon_sym_TILDE,
    ACTIONS(39), 1,
      anon_sym_PLUS,
    ACTIONS(41), 1,
      anon_sym_PIPE,
    ACTIONS(43), 1,
      anon_sym_SQUOTE,
    ACTIONS(45), 1,
      anon_sym_DQUOTE,
    ACTIONS(51), 1,
      sym__pseudo_class_selector_colon,
    ACTIONS(307), 1,
      sym_identifier,
    ACTIONS(319), 1,
      sym_nesting_selector,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    STATE(267), 13,
      sym__selector,
      sym_universal_selector,
      sym_class_selector,
      sym_pseudo_class_selector,
      sym_pseudo_element_selector,
      sym_id_selector,
      sym_attribute_selector,
      sym_child_selector,
      sym_descendant_selector,
      sym_sibling_selector,
      sym_adjacent_sibling_selector,
      sym_namespace_selector,
      sym_string_value,
  [1731] = 16,
    ACTIONS(25), 1,
      anon_sym_STAR,
    ACTIONS(27), 1,
      anon_sym_DOT,
    ACTIONS(29), 1,
      anon_sym_COLON_COLON,
    ACTIONS(31), 1,
      anon_sym_POUND,
    ACTIONS(33), 1,
      anon_sym_LBRACK,
    ACTIONS(35), 1,
      anon_sym_GT,
    ACTIONS(37), 1,
      anon_sym_TILDE,
    ACTIONS(39), 1,
      anon_sym_PLUS,
    ACTIONS(41), 1,
      anon_sym_PIPE,
    ACTIONS(43), 1,
      anon_sym_SQUOTE,
    ACTIONS(45), 1,
      anon_sym_DQUOTE,
    ACTIONS(51), 1,
      sym__pseudo_class_selector_colon,
    ACTIONS(307), 1,
      sym_identifier,
    ACTIONS(321), 1,
      sym_nesting_selector,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    STATE(159), 13,
      sym__selector,
      sym_universal_selector,
      sym_class_selector,
      sym_pseudo_class_selector,
      sym_pseudo_element_selector,
      sym_id_selector,
      sym_attribute_selector,
      sym_child_selector,
      sym_descendant_selector,
      sym_sibling_selector,
      sym_adjacent_sibling_selector,
      sym_namespace_selector,
      sym_string_value,
  [1793] = 16,
    ACTIONS(25), 1,
      anon_sym_STAR,
    ACTIONS(27), 1,
      anon_sym_DOT,
    ACTIONS(29), 1,
      anon_sym_COLON_COLON,
    ACTIONS(31), 1,
      anon_sym_POUND,
    ACTIONS(33), 1,
      anon_sym_LBRACK,
    ACTIONS(35), 1,
      anon_sym_GT,
    ACTIONS(37), 1,
      anon_sym_TILDE,
    ACTIONS(39), 1,
      anon_sym_PLUS,
    ACTIONS(41), 1,
      anon_sym_PIPE,
    ACTIONS(43), 1,
      anon_sym_SQUOTE,
    ACTIONS(45), 1,
      anon_sym_DQUOTE,
    ACTIONS(51), 1,
      sym__pseudo_class_selector_colon,
    ACTIONS(307), 1,
      sym_identifier,
    ACTIONS(323), 1,
      sym_nesting_selector,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    STATE(261), 13,
      sym__selector,
      sym_universal_selector,
      sym_class_selector,
      sym_pseudo_class_selector,
      sym_pseudo_element_selector,
      sym_id_selector,
      sym_attribute_selector,
      sym_child_selector,
      sym_descendant_selector,
      sym_sibling_selector,
      sym_adjacent_sibling_selector,
      sym_namespace_selector,
      sym_string_value,
  [1855] = 16,
    ACTIONS(25), 1,
      anon_sym_STAR,
    ACTIONS(27), 1,
      anon_sym_DOT,
    ACTIONS(29), 1,
      anon_sym_COLON_COLON,
    ACTIONS(31), 1,
      anon_sym_POUND,
    ACTIONS(33), 1,
      anon_sym_LBRACK,
    ACTIONS(35), 1,
      anon_sym_GT,
    ACTIONS(37), 1,
      anon_sym_TILDE,
    ACTIONS(39), 1,
      anon_sym_PLUS,
    ACTIONS(41), 1,
      anon_sym_PIPE,
    ACTIONS(43), 1,
      anon_sym_SQUOTE,
    ACTIONS(45), 1,
      anon_sym_DQUOTE,
    ACTIONS(51), 1,
      sym__pseudo_class_selector_colon,
    ACTIONS(307), 1,
      sym_identifier,
    ACTIONS(325), 1,
      sym_nesting_selector,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    STATE(253), 13,
      sym__selector,
      sym_universal_selector,
      sym_class_selector,
      sym_pseudo_class_selector,
      sym_pseudo_element_selector,
      sym_id_selector,
      sym_attribute_selector,
      sym_child_selector,
      sym_descendant_selector,
      sym_sibling_selector,
      sym_adjacent_sibling_selector,
      sym_namespace_selector,
      sym_string_value,
  [1917] = 16,
    ACTIONS(25), 1,
      anon_sym_STAR,
    ACTIONS(27), 1,
      anon_sym_DOT,
    ACTIONS(29), 1,
      anon_sym_COLON_COLON,
    ACTIONS(31), 1,
      anon_sym_POUND,
    ACTIONS(33), 1,
      anon_sym_LBRACK,
    ACTIONS(35), 1,
      anon_sym_GT,
    ACTIONS(37), 1,
      anon_sym_TILDE,
    ACTIONS(39), 1,
      anon_sym_PLUS,
    ACTIONS(41), 1,
      anon_sym_PIPE,
    ACTIONS(43), 1,
      anon_sym_SQUOTE,
    ACTIONS(45), 1,
      anon_sym_DQUOTE,
    ACTIONS(51), 1,
      sym__pseudo_class_selector_colon,
    ACTIONS(307), 1,
      sym_identifier,
    ACTIONS(327), 1,
      sym_nesting_selector,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    STATE(132), 13,
      sym__selector,
      sym_universal_selector,
      sym_class_selector,
      sym_pseudo_class_selector,
      sym_pseudo_element_selector,
      sym_id_selector,
      sym_attribute_selector,
      sym_child_selector,
      sym_descendant_selector,
      sym_sibling_selector,
      sym_adjacent_sibling_selector,
      sym_namespace_selector,
      sym_string_value,
  [1979] = 16,
    ACTIONS(25), 1,
      anon_sym_STAR,
    ACTIONS(27), 1,
      anon_sym_DOT,
    ACTIONS(29), 1,
      anon_sym_COLON_COLON,
    ACTIONS(31), 1,
      anon_sym_POUND,
    ACTIONS(33), 1,
      anon_sym_LBRACK,
    ACTIONS(35), 1,
      anon_sym_GT,
    ACTIONS(37), 1,
      anon_sym_TILDE,
    ACTIONS(39), 1,
      anon_sym_PLUS,
    ACTIONS(41), 1,
      anon_sym_PIPE,
    ACTIONS(43), 1,
      anon_sym_SQUOTE,
    ACTIONS(45), 1,
      anon_sym_DQUOTE,
    ACTIONS(51), 1,
      sym__pseudo_class_selector_colon,
    ACTIONS(307), 1,
      sym_identifier,
    ACTIONS(329), 1,
      sym_nesting_selector,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    STATE(245), 13,
      sym__selector,
      sym_universal_selector,
      sym_class_selector,
      sym_pseudo_class_selector,
      sym_pseudo_element_selector,
      sym_id_selector,
      sym_attribute_selector,
      sym_child_selector,
      sym_descendant_selector,
      sym_sibling_selector,
      sym_adjacent_sibling_selector,
      sym_namespace_selector,
      sym_string_value,
  [2041] = 17,
    ACTIONS(25), 1,
      anon_sym_STAR,
    ACTIONS(27), 1,
      anon_sym_DOT,
    ACTIONS(29), 1,
      anon_sym_COLON_COLON,
    ACTIONS(31), 1,
      anon_sym_POUND,
    ACTIONS(33), 1,
      anon_sym_LBRACK,
    ACTIONS(35), 1,
      anon_sym_GT,
    ACTIONS(37), 1,
      anon_sym_TILDE,
    ACTIONS(39), 1,
      anon_sym_PLUS,
    ACTIONS(41), 1,
      anon_sym_PIPE,
    ACTIONS(43), 1,
      anon_sym_SQUOTE,
    ACTIONS(45), 1,
      anon_sym_DQUOTE,
    ACTIONS(51), 1,
      sym__pseudo_class_selector_colon,
    ACTIONS(315), 1,
      sym_nesting_selector,
    ACTIONS(331), 1,
      sym_identifier,
    STATE(215), 1,
      sym_namespace_selector,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    STATE(272), 12,
      sym__selector,
      sym_universal_selector,
      sym_class_selector,
      sym_pseudo_class_selector,
      sym_pseudo_element_selector,
      sym_id_selector,
      sym_attribute_selector,
      sym_child_selector,
      sym_descendant_selector,
      sym_sibling_selector,
      sym_adjacent_sibling_selector,
      sym_string_value,
  [2105] = 16,
    ACTIONS(25), 1,
      anon_sym_STAR,
    ACTIONS(27), 1,
      anon_sym_DOT,
    ACTIONS(29), 1,
      anon_sym_COLON_COLON,
    ACTIONS(31), 1,
      anon_sym_POUND,
    ACTIONS(33), 1,
      anon_sym_LBRACK,
    ACTIONS(35), 1,
      anon_sym_GT,
    ACTIONS(37), 1,
      anon_sym_TILDE,
    ACTIONS(39), 1,
      anon_sym_PLUS,
    ACTIONS(41), 1,
      anon_sym_PIPE,
    ACTIONS(43), 1,
      anon_sym_SQUOTE,
    ACTIONS(45), 1,
      anon_sym_DQUOTE,
    ACTIONS(51), 1,
      sym__pseudo_class_selector_colon,
    ACTIONS(307), 1,
      sym_identifier,
    ACTIONS(333), 1,
      sym_nesting_selector,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    STATE(266), 13,
      sym__selector,
      sym_universal_selector,
      sym_class_selector,
      sym_pseudo_class_selector,
      sym_pseudo_element_selector,
      sym_id_selector,
      sym_attribute_selector,
      sym_child_selector,
      sym_descendant_selector,
      sym_sibling_selector,
      sym_adjacent_sibling_selector,
      sym_namespace_selector,
      sym_string_value,
  [2167] = 16,
    ACTIONS(25), 1,
      anon_sym_STAR,
    ACTIONS(27), 1,
      anon_sym_DOT,
    ACTIONS(29), 1,
      anon_sym_COLON_COLON,
    ACTIONS(31), 1,
      anon_sym_POUND,
    ACTIONS(33), 1,
      anon_sym_LBRACK,
    ACTIONS(35), 1,
      anon_sym_GT,
    ACTIONS(37), 1,
      anon_sym_TILDE,
    ACTIONS(39), 1,
      anon_sym_PLUS,
    ACTIONS(41), 1,
      anon_sym_PIPE,
    ACTIONS(43), 1,
      anon_sym_SQUOTE,
    ACTIONS(45), 1,
      anon_sym_DQUOTE,
    ACTIONS(51), 1,
      sym__pseudo_class_selector_colon,
    ACTIONS(307), 1,
      sym_identifier,
    ACTIONS(335), 1,
      sym_nesting_selector,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    STATE(166), 13,
      sym__selector,
      sym_universal_selector,
      sym_class_selector,
      sym_pseudo_class_selector,
      sym_pseudo_element_selector,
      sym_id_selector,
      sym_attribute_selector,
      sym_child_selector,
      sym_descendant_selector,
      sym_sibling_selector,
      sym_adjacent_sibling_selector,
      sym_namespace_selector,
      sym_string_value,
  [2229] = 19,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(221), 1,
      anon_sym_LPAREN,
    ACTIONS(239), 1,
      aux_sym_integer_value_token1,
    ACTIONS(241), 1,
      aux_sym_float_value_token1,
    ACTIONS(337), 1,
      anon_sym_COMMA,
    ACTIONS(339), 1,
      anon_sym_SEMI,
    ACTIONS(341), 1,
      anon_sym_RBRACE,
    ACTIONS(343), 1,
      anon_sym_STAR,
    ACTIONS(345), 1,
      anon_sym_POUND,
    ACTIONS(349), 1,
      sym_important,
    ACTIONS(351), 1,
      anon_sym_SQUOTE,
    ACTIONS(353), 1,
      anon_sym_DQUOTE,
    ACTIONS(355), 1,
      anon_sym_LBRACK2,
    ACTIONS(357), 1,
      sym_identifier,
    ACTIONS(359), 1,
      sym_plain_value,
    STATE(105), 1,
      aux_sym_declaration_repeat1,
    ACTIONS(347), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
    STATE(214), 9,
      sym__value,
      sym_parenthesized_value,
      sym_color_value,
      sym_string_value,
      sym_integer_value,
      sym_float_value,
      sym_grid_value,
      sym_call_expression,
      sym_binary_expression,
  [2297] = 16,
    ACTIONS(25), 1,
      anon_sym_STAR,
    ACTIONS(27), 1,
      anon_sym_DOT,
    ACTIONS(29), 1,
      anon_sym_COLON_COLON,
    ACTIONS(31), 1,
      anon_sym_POUND,
    ACTIONS(33), 1,
      anon_sym_LBRACK,
    ACTIONS(35), 1,
      anon_sym_GT,
    ACTIONS(37), 1,
      anon_sym_TILDE,
    ACTIONS(39), 1,
      anon_sym_PLUS,
    ACTIONS(41), 1,
      anon_sym_PIPE,
    ACTIONS(43), 1,
      anon_sym_SQUOTE,
    ACTIONS(45), 1,
      anon_sym_DQUOTE,
    ACTIONS(51), 1,
      sym__pseudo_class_selector_colon,
    ACTIONS(307), 1,
      sym_identifier,
    ACTIONS(361), 1,
      sym_nesting_selector,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    STATE(252), 13,
      sym__selector,
      sym_universal_selector,
      sym_class_selector,
      sym_pseudo_class_selector,
      sym_pseudo_element_selector,
      sym_id_selector,
      sym_attribute_selector,
      sym_child_selector,
      sym_descendant_selector,
      sym_sibling_selector,
      sym_adjacent_sibling_selector,
      sym_namespace_selector,
      sym_string_value,
  [2359] = 16,
    ACTIONS(25), 1,
      anon_sym_STAR,
    ACTIONS(27), 1,
      anon_sym_DOT,
    ACTIONS(29), 1,
      anon_sym_COLON_COLON,
    ACTIONS(31), 1,
      anon_sym_POUND,
    ACTIONS(33), 1,
      anon_sym_LBRACK,
    ACTIONS(35), 1,
      anon_sym_GT,
    ACTIONS(37), 1,
      anon_sym_TILDE,
    ACTIONS(39), 1,
      anon_sym_PLUS,
    ACTIONS(41), 1,
      anon_sym_PIPE,
    ACTIONS(43), 1,
      anon_sym_SQUOTE,
    ACTIONS(45), 1,
      anon_sym_DQUOTE,
    ACTIONS(51), 1,
      sym__pseudo_class_selector_colon,
    ACTIONS(307), 1,
      sym_identifier,
    ACTIONS(363), 1,
      sym_nesting_selector,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    STATE(141), 13,
      sym__selector,
      sym_universal_selector,
      sym_class_selector,
      sym_pseudo_class_selector,
      sym_pseudo_element_selector,
      sym_id_selector,
      sym_attribute_selector,
      sym_child_selector,
      sym_descendant_selector,
      sym_sibling_selector,
      sym_adjacent_sibling_selector,
      sym_namespace_selector,
      sym_string_value,
  [2421] = 16,
    ACTIONS(25), 1,
      anon_sym_STAR,
    ACTIONS(27), 1,
      anon_sym_DOT,
    ACTIONS(29), 1,
      anon_sym_COLON_COLON,
    ACTIONS(31), 1,
      anon_sym_POUND,
    ACTIONS(33), 1,
      anon_sym_LBRACK,
    ACTIONS(35), 1,
      anon_sym_GT,
    ACTIONS(37), 1,
      anon_sym_TILDE,
    ACTIONS(39), 1,
      anon_sym_PLUS,
    ACTIONS(41), 1,
      anon_sym_PIPE,
    ACTIONS(43), 1,
      anon_sym_SQUOTE,
    ACTIONS(45), 1,
      anon_sym_DQUOTE,
    ACTIONS(51), 1,
      sym__pseudo_class_selector_colon,
    ACTIONS(307), 1,
      sym_identifier,
    ACTIONS(365), 1,
      sym_nesting_selector,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    STATE(137), 13,
      sym__selector,
      sym_universal_selector,
      sym_class_selector,
      sym_pseudo_class_selector,
      sym_pseudo_element_selector,
      sym_id_selector,
      sym_attribute_selector,
      sym_child_selector,
      sym_descendant_selector,
      sym_sibling_selector,
      sym_adjacent_sibling_selector,
      sym_namespace_selector,
      sym_string_value,
  [2483] = 16,
    ACTIONS(25), 1,
      anon_sym_STAR,
    ACTIONS(27), 1,
      anon_sym_DOT,
    ACTIONS(29), 1,
      anon_sym_COLON_COLON,
    ACTIONS(31), 1,
      anon_sym_POUND,
    ACTIONS(33), 1,
      anon_sym_LBRACK,
    ACTIONS(35), 1,
      anon_sym_GT,
    ACTIONS(37), 1,
      anon_sym_TILDE,
    ACTIONS(39), 1,
      anon_sym_PLUS,
    ACTIONS(41), 1,
      anon_sym_PIPE,
    ACTIONS(43), 1,
      anon_sym_SQUOTE,
    ACTIONS(45), 1,
      anon_sym_DQUOTE,
    ACTIONS(51), 1,
      sym__pseudo_class_selector_colon,
    ACTIONS(307), 1,
      sym_identifier,
    ACTIONS(367), 1,
      sym_nesting_selector,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    STATE(142), 13,
      sym__selector,
      sym_universal_selector,
      sym_class_selector,
      sym_pseudo_class_selector,
      sym_pseudo_element_selector,
      sym_id_selector,
      sym_attribute_selector,
      sym_child_selector,
      sym_descendant_selector,
      sym_sibling_selector,
      sym_adjacent_sibling_selector,
      sym_namespace_selector,
      sym_string_value,
  [2545] = 16,
    ACTIONS(25), 1,
      anon_sym_STAR,
    ACTIONS(27), 1,
      anon_sym_DOT,
    ACTIONS(29), 1,
      anon_sym_COLON_COLON,
    ACTIONS(31), 1,
      anon_sym_POUND,
    ACTIONS(33), 1,
      anon_sym_LBRACK,
    ACTIONS(35), 1,
      anon_sym_GT,
    ACTIONS(37), 1,
      anon_sym_TILDE,
    ACTIONS(39), 1,
      anon_sym_PLUS,
    ACTIONS(41), 1,
      anon_sym_PIPE,
    ACTIONS(43), 1,
      anon_sym_SQUOTE,
    ACTIONS(45), 1,
      anon_sym_DQUOTE,
    ACTIONS(51), 1,
      sym__pseudo_class_selector_colon,
    ACTIONS(307), 1,
      sym_identifier,
    ACTIONS(369), 1,
      sym_nesting_selector,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    STATE(161), 13,
      sym__selector,
      sym_universal_selector,
      sym_class_selector,
      sym_pseudo_class_selector,
      sym_pseudo_element_selector,
      sym_id_selector,
      sym_attribute_selector,
      sym_child_selector,
      sym_descendant_selector,
      sym_sibling_selector,
      sym_adjacent_sibling_selector,
      sym_namespace_selector,
      sym_string_value,
  [2607] = 16,
    ACTIONS(25), 1,
      anon_sym_STAR,
    ACTIONS(27), 1,
      anon_sym_DOT,
    ACTIONS(29), 1,
      anon_sym_COLON_COLON,
    ACTIONS(31), 1,
      anon_sym_POUND,
    ACTIONS(33), 1,
      anon_sym_LBRACK,
    ACTIONS(35), 1,
      anon_sym_GT,
    ACTIONS(37), 1,
      anon_sym_TILDE,
    ACTIONS(39), 1,
      anon_sym_PLUS,
    ACTIONS(41), 1,
      anon_sym_PIPE,
    ACTIONS(43), 1,
      anon_sym_SQUOTE,
    ACTIONS(45), 1,
      anon_sym_DQUOTE,
    ACTIONS(51), 1,
      sym__pseudo_class_selector_colon,
    ACTIONS(307), 1,
      sym_identifier,
    ACTIONS(371), 1,
      sym_nesting_selector,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    STATE(130), 13,
      sym__selector,
      sym_universal_selector,
      sym_class_selector,
      sym_pseudo_class_selector,
      sym_pseudo_element_selector,
      sym_id_selector,
      sym_attribute_selector,
      sym_child_selector,
      sym_descendant_selector,
      sym_sibling_selector,
      sym_adjacent_sibling_selector,
      sym_namespace_selector,
      sym_string_value,
  [2669] = 16,
    ACTIONS(25), 1,
      anon_sym_STAR,
    ACTIONS(27), 1,
      anon_sym_DOT,
    ACTIONS(29), 1,
      anon_sym_COLON_COLON,
    ACTIONS(31), 1,
      anon_sym_POUND,
    ACTIONS(33), 1,
      anon_sym_LBRACK,
    ACTIONS(35), 1,
      anon_sym_GT,
    ACTIONS(37), 1,
      anon_sym_TILDE,
    ACTIONS(39), 1,
      anon_sym_PLUS,
    ACTIONS(41), 1,
      anon_sym_PIPE,
    ACTIONS(43), 1,
      anon_sym_SQUOTE,
    ACTIONS(45), 1,
      anon_sym_DQUOTE,
    ACTIONS(51), 1,
      sym__pseudo_class_selector_colon,
    ACTIONS(307), 1,
      sym_identifier,
    ACTIONS(373), 1,
      sym_nesting_selector,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    STATE(163), 13,
      sym__selector,
      sym_universal_selector,
      sym_class_selector,
      sym_pseudo_class_selector,
      sym_pseudo_element_selector,
      sym_id_selector,
      sym_attribute_selector,
      sym_child_selector,
      sym_descendant_selector,
      sym_sibling_selector,
      sym_adjacent_sibling_selector,
      sym_namespace_selector,
      sym_string_value,
  [2731] = 18,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(221), 1,
      anon_sym_LPAREN,
    ACTIONS(239), 1,
      aux_sym_integer_value_token1,
    ACTIONS(241), 1,
      aux_sym_float_value_token1,
    ACTIONS(337), 1,
      anon_sym_COMMA,
    ACTIONS(343), 1,
      anon_sym_STAR,
    ACTIONS(345), 1,
      anon_sym_POUND,
    ACTIONS(351), 1,
      anon_sym_SQUOTE,
    ACTIONS(353), 1,
      anon_sym_DQUOTE,
    ACTIONS(355), 1,
      anon_sym_LBRACK2,
    ACTIONS(357), 1,
      sym_identifier,
    ACTIONS(359), 1,
      sym_plain_value,
    ACTIONS(375), 1,
      anon_sym_SEMI,
    ACTIONS(377), 1,
      sym_important,
    STATE(117), 1,
      aux_sym_declaration_repeat1,
    ACTIONS(347), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
    STATE(214), 9,
      sym__value,
      sym_parenthesized_value,
      sym_color_value,
      sym_string_value,
      sym_integer_value,
      sym_float_value,
      sym_grid_value,
      sym_call_expression,
      sym_binary_expression,
  [2796] = 8,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(387), 1,
      anon_sym_LPAREN2,
    STATE(193), 1,
      sym_arguments,
    ACTIONS(385), 2,
      anon_sym_DOT,
      anon_sym_PLUS,
    ACTIONS(383), 4,
      anon_sym_STAR,
      sym_important,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
    ACTIONS(381), 8,
      anon_sym_LPAREN,
      aux_sym_integer_value_token1,
      aux_sym_float_value_token1,
      anon_sym_LBRACK2,
      anon_sym_DASH,
      anon_sym_SLASH,
      sym_identifier,
      sym_plain_value,
    ACTIONS(379), 10,
      sym__descendant_operator,
      sym__pseudo_class_selector_colon,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_TILDE,
      anon_sym_PIPE,
  [2841] = 18,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(221), 1,
      anon_sym_LPAREN,
    ACTIONS(239), 1,
      aux_sym_integer_value_token1,
    ACTIONS(241), 1,
      aux_sym_float_value_token1,
    ACTIONS(337), 1,
      anon_sym_COMMA,
    ACTIONS(339), 1,
      anon_sym_SEMI,
    ACTIONS(343), 1,
      anon_sym_STAR,
    ACTIONS(345), 1,
      anon_sym_POUND,
    ACTIONS(351), 1,
      anon_sym_SQUOTE,
    ACTIONS(353), 1,
      anon_sym_DQUOTE,
    ACTIONS(355), 1,
      anon_sym_LBRACK2,
    ACTIONS(357), 1,
      sym_identifier,
    ACTIONS(359), 1,
      sym_plain_value,
    ACTIONS(389), 1,
      sym_important,
    STATE(119), 1,
      aux_sym_declaration_repeat1,
    ACTIONS(347), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
    STATE(214), 9,
      sym__value,
      sym_parenthesized_value,
      sym_color_value,
      sym_string_value,
      sym_integer_value,
      sym_float_value,
      sym_grid_value,
      sym_call_expression,
      sym_binary_expression,
  [2906] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(393), 9,
      anon_sym_ATimport,
      anon_sym_ATmedia,
      anon_sym_ATcharset,
      anon_sym_ATnamespace,
      anon_sym_ATkeyframes,
      aux_sym_keyframes_statement_token1,
      anon_sym_ATsupports,
      anon_sym_ATscope,
      sym_at_keyword,
    ACTIONS(391), 15,
      sym__pseudo_class_selector_colon,
      ts_builtin_sym_end,
      sym_nesting_selector,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_TILDE,
      anon_sym_PLUS,
      anon_sym_PIPE,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_identifier,
  [2939] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(397), 9,
      anon_sym_ATimport,
      anon_sym_ATmedia,
      anon_sym_ATcharset,
      anon_sym_ATnamespace,
      anon_sym_ATkeyframes,
      aux_sym_keyframes_statement_token1,
      anon_sym_ATsupports,
      anon_sym_ATscope,
      sym_at_keyword,
    ACTIONS(395), 15,
      sym__pseudo_class_selector_colon,
      ts_builtin_sym_end,
      sym_nesting_selector,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_TILDE,
      anon_sym_PLUS,
      anon_sym_PIPE,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_identifier,
  [2972] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(401), 9,
      anon_sym_ATimport,
      anon_sym_ATmedia,
      anon_sym_ATcharset,
      anon_sym_ATnamespace,
      anon_sym_ATkeyframes,
      aux_sym_keyframes_statement_token1,
      anon_sym_ATsupports,
      anon_sym_ATscope,
      sym_at_keyword,
    ACTIONS(399), 15,
      sym__pseudo_class_selector_colon,
      ts_builtin_sym_end,
      sym_nesting_selector,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_TILDE,
      anon_sym_PLUS,
      anon_sym_PIPE,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_identifier,
  [3005] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(405), 9,
      anon_sym_ATimport,
      anon_sym_ATmedia,
      anon_sym_ATcharset,
      anon_sym_ATnamespace,
      anon_sym_ATkeyframes,
      aux_sym_keyframes_statement_token1,
      anon_sym_ATsupports,
      anon_sym_ATscope,
      sym_at_keyword,
    ACTIONS(403), 15,
      sym__pseudo_class_selector_colon,
      ts_builtin_sym_end,
      sym_nesting_selector,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_TILDE,
      anon_sym_PLUS,
      anon_sym_PIPE,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_identifier,
  [3038] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(409), 9,
      anon_sym_ATimport,
      anon_sym_ATmedia,
      anon_sym_ATcharset,
      anon_sym_ATnamespace,
      anon_sym_ATkeyframes,
      aux_sym_keyframes_statement_token1,
      anon_sym_ATsupports,
      anon_sym_ATscope,
      sym_at_keyword,
    ACTIONS(407), 15,
      sym__pseudo_class_selector_colon,
      ts_builtin_sym_end,
      sym_nesting_selector,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_TILDE,
      anon_sym_PLUS,
      anon_sym_PIPE,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_identifier,
  [3071] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(413), 9,
      anon_sym_ATimport,
      anon_sym_ATmedia,
      anon_sym_ATcharset,
      anon_sym_ATnamespace,
      anon_sym_ATkeyframes,
      aux_sym_keyframes_statement_token1,
      anon_sym_ATsupports,
      anon_sym_ATscope,
      sym_at_keyword,
    ACTIONS(411), 15,
      sym__pseudo_class_selector_colon,
      ts_builtin_sym_end,
      sym_nesting_selector,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_TILDE,
      anon_sym_PLUS,
      anon_sym_PIPE,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_identifier,
  [3104] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(417), 9,
      anon_sym_ATimport,
      anon_sym_ATmedia,
      anon_sym_ATcharset,
      anon_sym_ATnamespace,
      anon_sym_ATkeyframes,
      aux_sym_keyframes_statement_token1,
      anon_sym_ATsupports,
      anon_sym_ATscope,
      sym_at_keyword,
    ACTIONS(415), 15,
      sym__pseudo_class_selector_colon,
      ts_builtin_sym_end,
      sym_nesting_selector,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_TILDE,
      anon_sym_PLUS,
      anon_sym_PIPE,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_identifier,
  [3137] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(421), 9,
      anon_sym_ATimport,
      anon_sym_ATmedia,
      anon_sym_ATcharset,
      anon_sym_ATnamespace,
      anon_sym_ATkeyframes,
      aux_sym_keyframes_statement_token1,
      anon_sym_ATsupports,
      anon_sym_ATscope,
      sym_at_keyword,
    ACTIONS(419), 15,
      sym__pseudo_class_selector_colon,
      ts_builtin_sym_end,
      sym_nesting_selector,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_TILDE,
      anon_sym_PLUS,
      anon_sym_PIPE,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_identifier,
  [3170] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(425), 9,
      anon_sym_ATimport,
      anon_sym_ATmedia,
      anon_sym_ATcharset,
      anon_sym_ATnamespace,
      anon_sym_ATkeyframes,
      aux_sym_keyframes_statement_token1,
      anon_sym_ATsupports,
      anon_sym_ATscope,
      sym_at_keyword,
    ACTIONS(423), 15,
      sym__pseudo_class_selector_colon,
      ts_builtin_sym_end,
      sym_nesting_selector,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_TILDE,
      anon_sym_PLUS,
      anon_sym_PIPE,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_identifier,
  [3203] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(429), 9,
      anon_sym_ATimport,
      anon_sym_ATmedia,
      anon_sym_ATcharset,
      anon_sym_ATnamespace,
      anon_sym_ATkeyframes,
      aux_sym_keyframes_statement_token1,
      anon_sym_ATsupports,
      anon_sym_ATscope,
      sym_at_keyword,
    ACTIONS(427), 15,
      sym__pseudo_class_selector_colon,
      ts_builtin_sym_end,
      sym_nesting_selector,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_TILDE,
      anon_sym_PLUS,
      anon_sym_PIPE,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_identifier,
  [3236] = 5,
    STATE(50), 1,
      aux_sym_class_name_repeat1,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(433), 2,
      anon_sym_TILDE,
      anon_sym_PIPE,
    ACTIONS(435), 2,
      sym_escape_sequence,
      aux_sym_class_name_token1,
    ACTIONS(431), 19,
      sym__descendant_operator,
      sym__pseudo_class_selector_colon,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_RPAREN,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_TILDE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_STAR_EQ,
      anon_sym_DOLLAR_EQ,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_PLUS,
      anon_sym_LPAREN2,
  [3273] = 6,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(442), 2,
      anon_sym_DOT,
      anon_sym_PLUS,
    ACTIONS(440), 5,
      anon_sym_LPAREN,
      anon_sym_STAR,
      sym_important,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
    ACTIONS(444), 7,
      aux_sym_integer_value_token1,
      aux_sym_float_value_token1,
      anon_sym_LBRACK2,
      anon_sym_DASH,
      anon_sym_SLASH,
      sym_identifier,
      sym_plain_value,
    ACTIONS(438), 10,
      sym__descendant_operator,
      sym__pseudo_class_selector_colon,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_TILDE,
      anon_sym_PIPE,
  [3312] = 17,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(221), 1,
      anon_sym_LPAREN,
    ACTIONS(233), 1,
      sym_important,
    ACTIONS(239), 1,
      aux_sym_integer_value_token1,
    ACTIONS(241), 1,
      aux_sym_float_value_token1,
    ACTIONS(247), 1,
      sym_plain_value,
    ACTIONS(345), 1,
      anon_sym_POUND,
    ACTIONS(351), 1,
      anon_sym_SQUOTE,
    ACTIONS(353), 1,
      anon_sym_DQUOTE,
    ACTIONS(355), 1,
      anon_sym_LBRACK2,
    ACTIONS(357), 1,
      sym_identifier,
    ACTIONS(448), 1,
      anon_sym_RPAREN,
    STATE(107), 1,
      aux_sym_pseudo_class_arguments_repeat1,
    STATE(338), 1,
      aux_sym_arguments_repeat1,
    ACTIONS(446), 2,
      anon_sym_COMMA,
      anon_sym_SEMI,
    STATE(217), 9,
      sym__value,
      sym_parenthesized_value,
      sym_color_value,
      sym_string_value,
      sym_integer_value,
      sym_float_value,
      sym_grid_value,
      sym_call_expression,
      sym_binary_expression,
  [3373] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(452), 9,
      anon_sym_ATimport,
      anon_sym_ATmedia,
      anon_sym_ATcharset,
      anon_sym_ATnamespace,
      anon_sym_ATkeyframes,
      aux_sym_keyframes_statement_token1,
      anon_sym_ATsupports,
      anon_sym_ATscope,
      sym_at_keyword,
    ACTIONS(450), 15,
      sym__pseudo_class_selector_colon,
      ts_builtin_sym_end,
      sym_nesting_selector,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_TILDE,
      anon_sym_PLUS,
      anon_sym_PIPE,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_identifier,
  [3406] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(456), 9,
      anon_sym_ATimport,
      anon_sym_ATmedia,
      anon_sym_ATcharset,
      anon_sym_ATnamespace,
      anon_sym_ATkeyframes,
      aux_sym_keyframes_statement_token1,
      anon_sym_ATsupports,
      anon_sym_ATscope,
      sym_at_keyword,
    ACTIONS(454), 15,
      sym__pseudo_class_selector_colon,
      ts_builtin_sym_end,
      sym_nesting_selector,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_TILDE,
      anon_sym_PLUS,
      anon_sym_PIPE,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_identifier,
  [3439] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(460), 9,
      anon_sym_ATimport,
      anon_sym_ATmedia,
      anon_sym_ATcharset,
      anon_sym_ATnamespace,
      anon_sym_ATkeyframes,
      aux_sym_keyframes_statement_token1,
      anon_sym_ATsupports,
      anon_sym_ATscope,
      sym_at_keyword,
    ACTIONS(458), 15,
      sym__pseudo_class_selector_colon,
      ts_builtin_sym_end,
      sym_nesting_selector,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_TILDE,
      anon_sym_PLUS,
      anon_sym_PIPE,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_identifier,
  [3472] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(464), 9,
      anon_sym_ATimport,
      anon_sym_ATmedia,
      anon_sym_ATcharset,
      anon_sym_ATnamespace,
      anon_sym_ATkeyframes,
      aux_sym_keyframes_statement_token1,
      anon_sym_ATsupports,
      anon_sym_ATscope,
      sym_at_keyword,
    ACTIONS(462), 15,
      sym__pseudo_class_selector_colon,
      ts_builtin_sym_end,
      sym_nesting_selector,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_TILDE,
      anon_sym_PLUS,
      anon_sym_PIPE,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_identifier,
  [3505] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(421), 9,
      anon_sym_ATimport,
      anon_sym_ATmedia,
      anon_sym_ATcharset,
      anon_sym_ATnamespace,
      anon_sym_ATkeyframes,
      aux_sym_keyframes_statement_token1,
      anon_sym_ATsupports,
      anon_sym_ATscope,
      sym_at_keyword,
    ACTIONS(419), 15,
      sym__pseudo_class_selector_colon,
      anon_sym_RBRACE,
      sym_nesting_selector,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_TILDE,
      anon_sym_PLUS,
      anon_sym_PIPE,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_identifier,
  [3538] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(468), 9,
      anon_sym_ATimport,
      anon_sym_ATmedia,
      anon_sym_ATcharset,
      anon_sym_ATnamespace,
      anon_sym_ATkeyframes,
      aux_sym_keyframes_statement_token1,
      anon_sym_ATsupports,
      anon_sym_ATscope,
      sym_at_keyword,
    ACTIONS(466), 15,
      sym__pseudo_class_selector_colon,
      ts_builtin_sym_end,
      sym_nesting_selector,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_TILDE,
      anon_sym_PLUS,
      anon_sym_PIPE,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_identifier,
  [3571] = 5,
    STATE(97), 1,
      aux_sym_class_name_repeat1,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(472), 2,
      anon_sym_TILDE,
      anon_sym_PIPE,
    ACTIONS(474), 2,
      sym_escape_sequence,
      aux_sym_class_name_token1,
    ACTIONS(470), 19,
      sym__descendant_operator,
      sym__pseudo_class_selector_colon,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_RPAREN,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_TILDE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_STAR_EQ,
      anon_sym_DOLLAR_EQ,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_PLUS,
      anon_sym_LPAREN2,
  [3608] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(478), 9,
      anon_sym_ATimport,
      anon_sym_ATmedia,
      anon_sym_ATcharset,
      anon_sym_ATnamespace,
      anon_sym_ATkeyframes,
      aux_sym_keyframes_statement_token1,
      anon_sym_ATsupports,
      anon_sym_ATscope,
      sym_at_keyword,
    ACTIONS(476), 15,
      sym__pseudo_class_selector_colon,
      ts_builtin_sym_end,
      sym_nesting_selector,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_TILDE,
      anon_sym_PLUS,
      anon_sym_PIPE,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_identifier,
  [3641] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(482), 9,
      anon_sym_ATimport,
      anon_sym_ATmedia,
      anon_sym_ATcharset,
      anon_sym_ATnamespace,
      anon_sym_ATkeyframes,
      aux_sym_keyframes_statement_token1,
      anon_sym_ATsupports,
      anon_sym_ATscope,
      sym_at_keyword,
    ACTIONS(480), 15,
      sym__pseudo_class_selector_colon,
      ts_builtin_sym_end,
      sym_nesting_selector,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_TILDE,
      anon_sym_PLUS,
      anon_sym_PIPE,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_identifier,
  [3674] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(486), 9,
      anon_sym_ATimport,
      anon_sym_ATmedia,
      anon_sym_ATcharset,
      anon_sym_ATnamespace,
      anon_sym_ATkeyframes,
      aux_sym_keyframes_statement_token1,
      anon_sym_ATsupports,
      anon_sym_ATscope,
      sym_at_keyword,
    ACTIONS(484), 15,
      sym__pseudo_class_selector_colon,
      ts_builtin_sym_end,
      sym_nesting_selector,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_TILDE,
      anon_sym_PLUS,
      anon_sym_PIPE,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_identifier,
  [3707] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(490), 9,
      anon_sym_ATimport,
      anon_sym_ATmedia,
      anon_sym_ATcharset,
      anon_sym_ATnamespace,
      anon_sym_ATkeyframes,
      aux_sym_keyframes_statement_token1,
      anon_sym_ATsupports,
      anon_sym_ATscope,
      sym_at_keyword,
    ACTIONS(488), 15,
      sym__pseudo_class_selector_colon,
      ts_builtin_sym_end,
      sym_nesting_selector,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_TILDE,
      anon_sym_PLUS,
      anon_sym_PIPE,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_identifier,
  [3740] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(492), 9,
      anon_sym_ATimport,
      anon_sym_ATmedia,
      anon_sym_ATcharset,
      anon_sym_ATnamespace,
      anon_sym_ATkeyframes,
      aux_sym_keyframes_statement_token1,
      anon_sym_ATsupports,
      anon_sym_ATscope,
      sym_at_keyword,
    ACTIONS(494), 15,
      sym__pseudo_class_selector_colon,
      anon_sym_RBRACE,
      sym_nesting_selector,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_TILDE,
      anon_sym_PLUS,
      anon_sym_PIPE,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_identifier,
  [3773] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(498), 9,
      anon_sym_ATimport,
      anon_sym_ATmedia,
      anon_sym_ATcharset,
      anon_sym_ATnamespace,
      anon_sym_ATkeyframes,
      aux_sym_keyframes_statement_token1,
      anon_sym_ATsupports,
      anon_sym_ATscope,
      sym_at_keyword,
    ACTIONS(496), 15,
      sym__pseudo_class_selector_colon,
      ts_builtin_sym_end,
      sym_nesting_selector,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_TILDE,
      anon_sym_PLUS,
      anon_sym_PIPE,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_identifier,
  [3806] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(502), 9,
      anon_sym_ATimport,
      anon_sym_ATmedia,
      anon_sym_ATcharset,
      anon_sym_ATnamespace,
      anon_sym_ATkeyframes,
      aux_sym_keyframes_statement_token1,
      anon_sym_ATsupports,
      anon_sym_ATscope,
      sym_at_keyword,
    ACTIONS(500), 15,
      sym__pseudo_class_selector_colon,
      ts_builtin_sym_end,
      sym_nesting_selector,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_TILDE,
      anon_sym_PLUS,
      anon_sym_PIPE,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_identifier,
  [3839] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(506), 9,
      anon_sym_ATimport,
      anon_sym_ATmedia,
      anon_sym_ATcharset,
      anon_sym_ATnamespace,
      anon_sym_ATkeyframes,
      aux_sym_keyframes_statement_token1,
      anon_sym_ATsupports,
      anon_sym_ATscope,
      sym_at_keyword,
    ACTIONS(504), 15,
      sym__pseudo_class_selector_colon,
      ts_builtin_sym_end,
      sym_nesting_selector,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_TILDE,
      anon_sym_PLUS,
      anon_sym_PIPE,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_identifier,
  [3872] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(510), 9,
      anon_sym_ATimport,
      anon_sym_ATmedia,
      anon_sym_ATcharset,
      anon_sym_ATnamespace,
      anon_sym_ATkeyframes,
      aux_sym_keyframes_statement_token1,
      anon_sym_ATsupports,
      anon_sym_ATscope,
      sym_at_keyword,
    ACTIONS(508), 15,
      sym__pseudo_class_selector_colon,
      ts_builtin_sym_end,
      sym_nesting_selector,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_TILDE,
      anon_sym_PLUS,
      anon_sym_PIPE,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_identifier,
  [3905] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(514), 9,
      anon_sym_ATimport,
      anon_sym_ATmedia,
      anon_sym_ATcharset,
      anon_sym_ATnamespace,
      anon_sym_ATkeyframes,
      aux_sym_keyframes_statement_token1,
      anon_sym_ATsupports,
      anon_sym_ATscope,
      sym_at_keyword,
    ACTIONS(512), 15,
      sym__pseudo_class_selector_colon,
      ts_builtin_sym_end,
      sym_nesting_selector,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_TILDE,
      anon_sym_PLUS,
      anon_sym_PIPE,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_identifier,
  [3938] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(393), 9,
      anon_sym_ATimport,
      anon_sym_ATmedia,
      anon_sym_ATcharset,
      anon_sym_ATnamespace,
      anon_sym_ATkeyframes,
      aux_sym_keyframes_statement_token1,
      anon_sym_ATsupports,
      anon_sym_ATscope,
      sym_at_keyword,
    ACTIONS(391), 15,
      sym__pseudo_class_selector_colon,
      anon_sym_RBRACE,
      sym_nesting_selector,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_TILDE,
      anon_sym_PLUS,
      anon_sym_PIPE,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_identifier,
  [3971] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(421), 9,
      anon_sym_ATimport,
      anon_sym_ATmedia,
      anon_sym_ATcharset,
      anon_sym_ATnamespace,
      anon_sym_ATkeyframes,
      aux_sym_keyframes_statement_token1,
      anon_sym_ATsupports,
      anon_sym_ATscope,
      sym_at_keyword,
    ACTIONS(419), 15,
      sym__pseudo_class_selector_colon,
      anon_sym_RBRACE,
      sym_nesting_selector,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_TILDE,
      anon_sym_PLUS,
      anon_sym_PIPE,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_identifier,
  [4004] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(425), 9,
      anon_sym_ATimport,
      anon_sym_ATmedia,
      anon_sym_ATcharset,
      anon_sym_ATnamespace,
      anon_sym_ATkeyframes,
      aux_sym_keyframes_statement_token1,
      anon_sym_ATsupports,
      anon_sym_ATscope,
      sym_at_keyword,
    ACTIONS(423), 15,
      sym__pseudo_class_selector_colon,
      anon_sym_RBRACE,
      sym_nesting_selector,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_TILDE,
      anon_sym_PLUS,
      anon_sym_PIPE,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_identifier,
  [4037] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(397), 9,
      anon_sym_ATimport,
      anon_sym_ATmedia,
      anon_sym_ATcharset,
      anon_sym_ATnamespace,
      anon_sym_ATkeyframes,
      aux_sym_keyframes_statement_token1,
      anon_sym_ATsupports,
      anon_sym_ATscope,
      sym_at_keyword,
    ACTIONS(395), 15,
      sym__pseudo_class_selector_colon,
      anon_sym_RBRACE,
      sym_nesting_selector,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_TILDE,
      anon_sym_PLUS,
      anon_sym_PIPE,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_identifier,
  [4070] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(401), 9,
      anon_sym_ATimport,
      anon_sym_ATmedia,
      anon_sym_ATcharset,
      anon_sym_ATnamespace,
      anon_sym_ATkeyframes,
      aux_sym_keyframes_statement_token1,
      anon_sym_ATsupports,
      anon_sym_ATscope,
      sym_at_keyword,
    ACTIONS(399), 15,
      sym__pseudo_class_selector_colon,
      anon_sym_RBRACE,
      sym_nesting_selector,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_TILDE,
      anon_sym_PLUS,
      anon_sym_PIPE,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_identifier,
  [4103] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(405), 9,
      anon_sym_ATimport,
      anon_sym_ATmedia,
      anon_sym_ATcharset,
      anon_sym_ATnamespace,
      anon_sym_ATkeyframes,
      aux_sym_keyframes_statement_token1,
      anon_sym_ATsupports,
      anon_sym_ATscope,
      sym_at_keyword,
    ACTIONS(403), 15,
      sym__pseudo_class_selector_colon,
      anon_sym_RBRACE,
      sym_nesting_selector,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_TILDE,
      anon_sym_PLUS,
      anon_sym_PIPE,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_identifier,
  [4136] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(516), 9,
      anon_sym_ATimport,
      anon_sym_ATmedia,
      anon_sym_ATcharset,
      anon_sym_ATnamespace,
      anon_sym_ATkeyframes,
      aux_sym_keyframes_statement_token1,
      anon_sym_ATsupports,
      anon_sym_ATscope,
      sym_at_keyword,
    ACTIONS(518), 15,
      sym__pseudo_class_selector_colon,
      anon_sym_RBRACE,
      sym_nesting_selector,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_TILDE,
      anon_sym_PLUS,
      anon_sym_PIPE,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_identifier,
  [4169] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(409), 9,
      anon_sym_ATimport,
      anon_sym_ATmedia,
      anon_sym_ATcharset,
      anon_sym_ATnamespace,
      anon_sym_ATkeyframes,
      aux_sym_keyframes_statement_token1,
      anon_sym_ATsupports,
      anon_sym_ATscope,
      sym_at_keyword,
    ACTIONS(407), 15,
      sym__pseudo_class_selector_colon,
      anon_sym_RBRACE,
      sym_nesting_selector,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_TILDE,
      anon_sym_PLUS,
      anon_sym_PIPE,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_identifier,
  [4202] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(413), 9,
      anon_sym_ATimport,
      anon_sym_ATmedia,
      anon_sym_ATcharset,
      anon_sym_ATnamespace,
      anon_sym_ATkeyframes,
      aux_sym_keyframes_statement_token1,
      anon_sym_ATsupports,
      anon_sym_ATscope,
      sym_at_keyword,
    ACTIONS(411), 15,
      sym__pseudo_class_selector_colon,
      anon_sym_RBRACE,
      sym_nesting_selector,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_TILDE,
      anon_sym_PLUS,
      anon_sym_PIPE,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_identifier,
  [4235] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(417), 9,
      anon_sym_ATimport,
      anon_sym_ATmedia,
      anon_sym_ATcharset,
      anon_sym_ATnamespace,
      anon_sym_ATkeyframes,
      aux_sym_keyframes_statement_token1,
      anon_sym_ATsupports,
      anon_sym_ATscope,
      sym_at_keyword,
    ACTIONS(415), 15,
      sym__pseudo_class_selector_colon,
      anon_sym_RBRACE,
      sym_nesting_selector,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_TILDE,
      anon_sym_PLUS,
      anon_sym_PIPE,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_identifier,
  [4268] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(429), 9,
      anon_sym_ATimport,
      anon_sym_ATmedia,
      anon_sym_ATcharset,
      anon_sym_ATnamespace,
      anon_sym_ATkeyframes,
      aux_sym_keyframes_statement_token1,
      anon_sym_ATsupports,
      anon_sym_ATscope,
      sym_at_keyword,
    ACTIONS(427), 15,
      sym__pseudo_class_selector_colon,
      anon_sym_RBRACE,
      sym_nesting_selector,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_TILDE,
      anon_sym_PLUS,
      anon_sym_PIPE,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_identifier,
  [4301] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(452), 9,
      anon_sym_ATimport,
      anon_sym_ATmedia,
      anon_sym_ATcharset,
      anon_sym_ATnamespace,
      anon_sym_ATkeyframes,
      aux_sym_keyframes_statement_token1,
      anon_sym_ATsupports,
      anon_sym_ATscope,
      sym_at_keyword,
    ACTIONS(450), 15,
      sym__pseudo_class_selector_colon,
      anon_sym_RBRACE,
      sym_nesting_selector,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_TILDE,
      anon_sym_PLUS,
      anon_sym_PIPE,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_identifier,
  [4334] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(456), 9,
      anon_sym_ATimport,
      anon_sym_ATmedia,
      anon_sym_ATcharset,
      anon_sym_ATnamespace,
      anon_sym_ATkeyframes,
      aux_sym_keyframes_statement_token1,
      anon_sym_ATsupports,
      anon_sym_ATscope,
      sym_at_keyword,
    ACTIONS(454), 15,
      sym__pseudo_class_selector_colon,
      anon_sym_RBRACE,
      sym_nesting_selector,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_TILDE,
      anon_sym_PLUS,
      anon_sym_PIPE,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_identifier,
  [4367] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(460), 9,
      anon_sym_ATimport,
      anon_sym_ATmedia,
      anon_sym_ATcharset,
      anon_sym_ATnamespace,
      anon_sym_ATkeyframes,
      aux_sym_keyframes_statement_token1,
      anon_sym_ATsupports,
      anon_sym_ATscope,
      sym_at_keyword,
    ACTIONS(458), 15,
      sym__pseudo_class_selector_colon,
      anon_sym_RBRACE,
      sym_nesting_selector,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_TILDE,
      anon_sym_PLUS,
      anon_sym_PIPE,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_identifier,
  [4400] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(464), 9,
      anon_sym_ATimport,
      anon_sym_ATmedia,
      anon_sym_ATcharset,
      anon_sym_ATnamespace,
      anon_sym_ATkeyframes,
      aux_sym_keyframes_statement_token1,
      anon_sym_ATsupports,
      anon_sym_ATscope,
      sym_at_keyword,
    ACTIONS(462), 15,
      sym__pseudo_class_selector_colon,
      anon_sym_RBRACE,
      sym_nesting_selector,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_TILDE,
      anon_sym_PLUS,
      anon_sym_PIPE,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_identifier,
  [4433] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(468), 9,
      anon_sym_ATimport,
      anon_sym_ATmedia,
      anon_sym_ATcharset,
      anon_sym_ATnamespace,
      anon_sym_ATkeyframes,
      aux_sym_keyframes_statement_token1,
      anon_sym_ATsupports,
      anon_sym_ATscope,
      sym_at_keyword,
    ACTIONS(466), 15,
      sym__pseudo_class_selector_colon,
      anon_sym_RBRACE,
      sym_nesting_selector,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_TILDE,
      anon_sym_PLUS,
      anon_sym_PIPE,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_identifier,
  [4466] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(478), 9,
      anon_sym_ATimport,
      anon_sym_ATmedia,
      anon_sym_ATcharset,
      anon_sym_ATnamespace,
      anon_sym_ATkeyframes,
      aux_sym_keyframes_statement_token1,
      anon_sym_ATsupports,
      anon_sym_ATscope,
      sym_at_keyword,
    ACTIONS(476), 15,
      sym__pseudo_class_selector_colon,
      anon_sym_RBRACE,
      sym_nesting_selector,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_TILDE,
      anon_sym_PLUS,
      anon_sym_PIPE,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_identifier,
  [4499] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(482), 9,
      anon_sym_ATimport,
      anon_sym_ATmedia,
      anon_sym_ATcharset,
      anon_sym_ATnamespace,
      anon_sym_ATkeyframes,
      aux_sym_keyframes_statement_token1,
      anon_sym_ATsupports,
      anon_sym_ATscope,
      sym_at_keyword,
    ACTIONS(480), 15,
      sym__pseudo_class_selector_colon,
      anon_sym_RBRACE,
      sym_nesting_selector,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_TILDE,
      anon_sym_PLUS,
      anon_sym_PIPE,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_identifier,
  [4532] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(486), 9,
      anon_sym_ATimport,
      anon_sym_ATmedia,
      anon_sym_ATcharset,
      anon_sym_ATnamespace,
      anon_sym_ATkeyframes,
      aux_sym_keyframes_statement_token1,
      anon_sym_ATsupports,
      anon_sym_ATscope,
      sym_at_keyword,
    ACTIONS(484), 15,
      sym__pseudo_class_selector_colon,
      anon_sym_RBRACE,
      sym_nesting_selector,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_TILDE,
      anon_sym_PLUS,
      anon_sym_PIPE,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_identifier,
  [4565] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(490), 9,
      anon_sym_ATimport,
      anon_sym_ATmedia,
      anon_sym_ATcharset,
      anon_sym_ATnamespace,
      anon_sym_ATkeyframes,
      aux_sym_keyframes_statement_token1,
      anon_sym_ATsupports,
      anon_sym_ATscope,
      sym_at_keyword,
    ACTIONS(488), 15,
      sym__pseudo_class_selector_colon,
      anon_sym_RBRACE,
      sym_nesting_selector,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_TILDE,
      anon_sym_PLUS,
      anon_sym_PIPE,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_identifier,
  [4598] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(498), 9,
      anon_sym_ATimport,
      anon_sym_ATmedia,
      anon_sym_ATcharset,
      anon_sym_ATnamespace,
      anon_sym_ATkeyframes,
      aux_sym_keyframes_statement_token1,
      anon_sym_ATsupports,
      anon_sym_ATscope,
      sym_at_keyword,
    ACTIONS(496), 15,
      sym__pseudo_class_selector_colon,
      anon_sym_RBRACE,
      sym_nesting_selector,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_TILDE,
      anon_sym_PLUS,
      anon_sym_PIPE,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_identifier,
  [4631] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(502), 9,
      anon_sym_ATimport,
      anon_sym_ATmedia,
      anon_sym_ATcharset,
      anon_sym_ATnamespace,
      anon_sym_ATkeyframes,
      aux_sym_keyframes_statement_token1,
      anon_sym_ATsupports,
      anon_sym_ATscope,
      sym_at_keyword,
    ACTIONS(500), 15,
      sym__pseudo_class_selector_colon,
      anon_sym_RBRACE,
      sym_nesting_selector,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_TILDE,
      anon_sym_PLUS,
      anon_sym_PIPE,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_identifier,
  [4664] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(506), 9,
      anon_sym_ATimport,
      anon_sym_ATmedia,
      anon_sym_ATcharset,
      anon_sym_ATnamespace,
      anon_sym_ATkeyframes,
      aux_sym_keyframes_statement_token1,
      anon_sym_ATsupports,
      anon_sym_ATscope,
      sym_at_keyword,
    ACTIONS(504), 15,
      sym__pseudo_class_selector_colon,
      anon_sym_RBRACE,
      sym_nesting_selector,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_TILDE,
      anon_sym_PLUS,
      anon_sym_PIPE,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_identifier,
  [4697] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(510), 9,
      anon_sym_ATimport,
      anon_sym_ATmedia,
      anon_sym_ATcharset,
      anon_sym_ATnamespace,
      anon_sym_ATkeyframes,
      aux_sym_keyframes_statement_token1,
      anon_sym_ATsupports,
      anon_sym_ATscope,
      sym_at_keyword,
    ACTIONS(508), 15,
      sym__pseudo_class_selector_colon,
      anon_sym_RBRACE,
      sym_nesting_selector,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_TILDE,
      anon_sym_PLUS,
      anon_sym_PIPE,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_identifier,
  [4730] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(514), 9,
      anon_sym_ATimport,
      anon_sym_ATmedia,
      anon_sym_ATcharset,
      anon_sym_ATnamespace,
      anon_sym_ATkeyframes,
      aux_sym_keyframes_statement_token1,
      anon_sym_ATsupports,
      anon_sym_ATscope,
      sym_at_keyword,
    ACTIONS(512), 15,
      sym__pseudo_class_selector_colon,
      anon_sym_RBRACE,
      sym_nesting_selector,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_TILDE,
      anon_sym_PLUS,
      anon_sym_PIPE,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_identifier,
  [4763] = 4,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(522), 9,
      anon_sym_DOT,
      anon_sym_PLUS,
      aux_sym_integer_value_token1,
      aux_sym_float_value_token1,
      anon_sym_LBRACK2,
      anon_sym_DASH,
      anon_sym_SLASH,
      sym_identifier,
      sym_plain_value,
    ACTIONS(520), 15,
      sym__descendant_operator,
      sym__pseudo_class_selector_colon,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_STAR,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_TILDE,
      anon_sym_PIPE,
      sym_important,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
  [4798] = 4,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(526), 9,
      anon_sym_DOT,
      anon_sym_PLUS,
      aux_sym_integer_value_token1,
      aux_sym_float_value_token1,
      anon_sym_LBRACK2,
      anon_sym_DASH,
      anon_sym_SLASH,
      sym_identifier,
      sym_plain_value,
    ACTIONS(524), 15,
      sym__descendant_operator,
      sym__pseudo_class_selector_colon,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_STAR,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_TILDE,
      anon_sym_PIPE,
      sym_important,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
  [4833] = 5,
    STATE(50), 1,
      aux_sym_class_name_repeat1,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(530), 2,
      anon_sym_TILDE,
      anon_sym_PIPE,
    ACTIONS(532), 2,
      sym_escape_sequence,
      aux_sym_class_name_token1,
    ACTIONS(528), 19,
      sym__descendant_operator,
      sym__pseudo_class_selector_colon,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_RPAREN,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_TILDE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_STAR_EQ,
      anon_sym_DOLLAR_EQ,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_PLUS,
      anon_sym_LPAREN2,
  [4870] = 17,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(221), 1,
      anon_sym_LPAREN,
    ACTIONS(233), 1,
      sym_important,
    ACTIONS(239), 1,
      aux_sym_integer_value_token1,
    ACTIONS(241), 1,
      aux_sym_float_value_token1,
    ACTIONS(247), 1,
      sym_plain_value,
    ACTIONS(345), 1,
      anon_sym_POUND,
    ACTIONS(351), 1,
      anon_sym_SQUOTE,
    ACTIONS(353), 1,
      anon_sym_DQUOTE,
    ACTIONS(355), 1,
      anon_sym_LBRACK2,
    ACTIONS(357), 1,
      sym_identifier,
    ACTIONS(534), 1,
      anon_sym_RPAREN,
    STATE(107), 1,
      aux_sym_pseudo_class_arguments_repeat1,
    STATE(355), 1,
      aux_sym_arguments_repeat1,
    ACTIONS(446), 2,
      anon_sym_COMMA,
      anon_sym_SEMI,
    STATE(217), 9,
      sym__value,
      sym_parenthesized_value,
      sym_color_value,
      sym_string_value,
      sym_integer_value,
      sym_float_value,
      sym_grid_value,
      sym_call_expression,
      sym_binary_expression,
  [4931] = 17,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(221), 1,
      anon_sym_LPAREN,
    ACTIONS(233), 1,
      sym_important,
    ACTIONS(239), 1,
      aux_sym_integer_value_token1,
    ACTIONS(241), 1,
      aux_sym_float_value_token1,
    ACTIONS(247), 1,
      sym_plain_value,
    ACTIONS(345), 1,
      anon_sym_POUND,
    ACTIONS(351), 1,
      anon_sym_SQUOTE,
    ACTIONS(353), 1,
      anon_sym_DQUOTE,
    ACTIONS(355), 1,
      anon_sym_LBRACK2,
    ACTIONS(357), 1,
      sym_identifier,
    ACTIONS(536), 1,
      anon_sym_RPAREN,
    STATE(107), 1,
      aux_sym_pseudo_class_arguments_repeat1,
    STATE(361), 1,
      aux_sym_arguments_repeat1,
    ACTIONS(446), 2,
      anon_sym_COMMA,
      anon_sym_SEMI,
    STATE(217), 9,
      sym__value,
      sym_parenthesized_value,
      sym_color_value,
      sym_string_value,
      sym_integer_value,
      sym_float_value,
      sym_grid_value,
      sym_call_expression,
      sym_binary_expression,
  [4992] = 17,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(221), 1,
      anon_sym_LPAREN,
    ACTIONS(233), 1,
      sym_important,
    ACTIONS(239), 1,
      aux_sym_integer_value_token1,
    ACTIONS(241), 1,
      aux_sym_float_value_token1,
    ACTIONS(247), 1,
      sym_plain_value,
    ACTIONS(345), 1,
      anon_sym_POUND,
    ACTIONS(351), 1,
      anon_sym_SQUOTE,
    ACTIONS(353), 1,
      anon_sym_DQUOTE,
    ACTIONS(355), 1,
      anon_sym_LBRACK2,
    ACTIONS(357), 1,
      sym_identifier,
    ACTIONS(538), 1,
      anon_sym_RPAREN,
    STATE(107), 1,
      aux_sym_pseudo_class_arguments_repeat1,
    STATE(369), 1,
      aux_sym_arguments_repeat1,
    ACTIONS(446), 2,
      anon_sym_COMMA,
      anon_sym_SEMI,
    STATE(217), 9,
      sym__value,
      sym_parenthesized_value,
      sym_color_value,
      sym_string_value,
      sym_integer_value,
      sym_float_value,
      sym_grid_value,
      sym_call_expression,
      sym_binary_expression,
  [5053] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(516), 9,
      anon_sym_ATimport,
      anon_sym_ATmedia,
      anon_sym_ATcharset,
      anon_sym_ATnamespace,
      anon_sym_ATkeyframes,
      aux_sym_keyframes_statement_token1,
      anon_sym_ATsupports,
      anon_sym_ATscope,
      sym_at_keyword,
    ACTIONS(518), 15,
      sym__pseudo_class_selector_colon,
      ts_builtin_sym_end,
      sym_nesting_selector,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_TILDE,
      anon_sym_PLUS,
      anon_sym_PIPE,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_identifier,
  [5086] = 16,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(540), 1,
      anon_sym_COMMA,
    ACTIONS(545), 1,
      anon_sym_LPAREN,
    ACTIONS(548), 1,
      anon_sym_POUND,
    ACTIONS(551), 1,
      sym_important,
    ACTIONS(554), 1,
      anon_sym_SQUOTE,
    ACTIONS(557), 1,
      anon_sym_DQUOTE,
    ACTIONS(560), 1,
      aux_sym_integer_value_token1,
    ACTIONS(563), 1,
      aux_sym_float_value_token1,
    ACTIONS(566), 1,
      anon_sym_LBRACK2,
    ACTIONS(569), 1,
      sym_identifier,
    ACTIONS(572), 1,
      sym_plain_value,
    STATE(102), 1,
      aux_sym_declaration_repeat1,
    ACTIONS(543), 2,
      anon_sym_SEMI,
      anon_sym_RBRACE,
    STATE(214), 9,
      sym__value,
      sym_parenthesized_value,
      sym_color_value,
      sym_string_value,
      sym_integer_value,
      sym_float_value,
      sym_grid_value,
      sym_call_expression,
      sym_binary_expression,
  [5144] = 17,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(221), 1,
      anon_sym_LPAREN,
    ACTIONS(233), 1,
      sym_important,
    ACTIONS(239), 1,
      aux_sym_integer_value_token1,
    ACTIONS(241), 1,
      aux_sym_float_value_token1,
    ACTIONS(247), 1,
      sym_plain_value,
    ACTIONS(345), 1,
      anon_sym_POUND,
    ACTIONS(351), 1,
      anon_sym_SQUOTE,
    ACTIONS(353), 1,
      anon_sym_DQUOTE,
    ACTIONS(355), 1,
      anon_sym_LBRACK2,
    ACTIONS(357), 1,
      sym_identifier,
    ACTIONS(575), 1,
      anon_sym_COMMA,
    ACTIONS(577), 1,
      anon_sym_RPAREN,
    STATE(107), 1,
      aux_sym_pseudo_class_arguments_repeat1,
    STATE(401), 1,
      aux_sym_pseudo_class_arguments_repeat2,
    STATE(217), 9,
      sym__value,
      sym_parenthesized_value,
      sym_color_value,
      sym_string_value,
      sym_integer_value,
      sym_float_value,
      sym_grid_value,
      sym_call_expression,
      sym_binary_expression,
  [5204] = 15,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(221), 1,
      anon_sym_LPAREN,
    ACTIONS(233), 1,
      sym_important,
    ACTIONS(239), 1,
      aux_sym_integer_value_token1,
    ACTIONS(241), 1,
      aux_sym_float_value_token1,
    ACTIONS(247), 1,
      sym_plain_value,
    ACTIONS(345), 1,
      anon_sym_POUND,
    ACTIONS(351), 1,
      anon_sym_SQUOTE,
    ACTIONS(353), 1,
      anon_sym_DQUOTE,
    ACTIONS(355), 1,
      anon_sym_LBRACK2,
    ACTIONS(357), 1,
      sym_identifier,
    STATE(107), 1,
      aux_sym_pseudo_class_arguments_repeat1,
    ACTIONS(579), 3,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_RPAREN,
    STATE(217), 9,
      sym__value,
      sym_parenthesized_value,
      sym_color_value,
      sym_string_value,
      sym_integer_value,
      sym_float_value,
      sym_grid_value,
      sym_call_expression,
      sym_binary_expression,
  [5260] = 17,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(221), 1,
      anon_sym_LPAREN,
    ACTIONS(239), 1,
      aux_sym_integer_value_token1,
    ACTIONS(241), 1,
      aux_sym_float_value_token1,
    ACTIONS(337), 1,
      anon_sym_COMMA,
    ACTIONS(345), 1,
      anon_sym_POUND,
    ACTIONS(351), 1,
      anon_sym_SQUOTE,
    ACTIONS(353), 1,
      anon_sym_DQUOTE,
    ACTIONS(355), 1,
      anon_sym_LBRACK2,
    ACTIONS(357), 1,
      sym_identifier,
    ACTIONS(359), 1,
      sym_plain_value,
    ACTIONS(581), 1,
      anon_sym_SEMI,
    ACTIONS(583), 1,
      anon_sym_RBRACE,
    ACTIONS(585), 1,
      sym_important,
    STATE(102), 1,
      aux_sym_declaration_repeat1,
    STATE(214), 9,
      sym__value,
      sym_parenthesized_value,
      sym_color_value,
      sym_string_value,
      sym_integer_value,
      sym_float_value,
      sym_grid_value,
      sym_call_expression,
      sym_binary_expression,
  [5320] = 17,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(221), 1,
      anon_sym_LPAREN,
    ACTIONS(233), 1,
      sym_important,
    ACTIONS(239), 1,
      aux_sym_integer_value_token1,
    ACTIONS(241), 1,
      aux_sym_float_value_token1,
    ACTIONS(247), 1,
      sym_plain_value,
    ACTIONS(345), 1,
      anon_sym_POUND,
    ACTIONS(351), 1,
      anon_sym_SQUOTE,
    ACTIONS(353), 1,
      anon_sym_DQUOTE,
    ACTIONS(355), 1,
      anon_sym_LBRACK2,
    ACTIONS(357), 1,
      sym_identifier,
    ACTIONS(575), 1,
      anon_sym_COMMA,
    ACTIONS(587), 1,
      anon_sym_RPAREN,
    STATE(107), 1,
      aux_sym_pseudo_class_arguments_repeat1,
    STATE(384), 1,
      aux_sym_pseudo_class_arguments_repeat2,
    STATE(217), 9,
      sym__value,
      sym_parenthesized_value,
      sym_color_value,
      sym_string_value,
      sym_integer_value,
      sym_float_value,
      sym_grid_value,
      sym_call_expression,
      sym_binary_expression,
  [5380] = 15,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(591), 1,
      anon_sym_LPAREN,
    ACTIONS(594), 1,
      anon_sym_POUND,
    ACTIONS(597), 1,
      sym_important,
    ACTIONS(600), 1,
      anon_sym_SQUOTE,
    ACTIONS(603), 1,
      anon_sym_DQUOTE,
    ACTIONS(606), 1,
      aux_sym_integer_value_token1,
    ACTIONS(609), 1,
      aux_sym_float_value_token1,
    ACTIONS(612), 1,
      anon_sym_LBRACK2,
    ACTIONS(615), 1,
      sym_identifier,
    ACTIONS(618), 1,
      sym_plain_value,
    STATE(107), 1,
      aux_sym_pseudo_class_arguments_repeat1,
    ACTIONS(589), 3,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_RPAREN,
    STATE(217), 9,
      sym__value,
      sym_parenthesized_value,
      sym_color_value,
      sym_string_value,
      sym_integer_value,
      sym_float_value,
      sym_grid_value,
      sym_call_expression,
      sym_binary_expression,
  [5436] = 15,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(221), 1,
      anon_sym_LPAREN,
    ACTIONS(233), 1,
      sym_important,
    ACTIONS(239), 1,
      aux_sym_integer_value_token1,
    ACTIONS(241), 1,
      aux_sym_float_value_token1,
    ACTIONS(247), 1,
      sym_plain_value,
    ACTIONS(345), 1,
      anon_sym_POUND,
    ACTIONS(351), 1,
      anon_sym_SQUOTE,
    ACTIONS(353), 1,
      anon_sym_DQUOTE,
    ACTIONS(355), 1,
      anon_sym_LBRACK2,
    ACTIONS(357), 1,
      sym_identifier,
    STATE(107), 1,
      aux_sym_pseudo_class_arguments_repeat1,
    ACTIONS(621), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
    STATE(217), 9,
      sym__value,
      sym_parenthesized_value,
      sym_color_value,
      sym_string_value,
      sym_integer_value,
      sym_float_value,
      sym_grid_value,
      sym_call_expression,
      sym_binary_expression,
  [5491] = 5,
    ACTIONS(627), 1,
      anon_sym_LPAREN2,
    STATE(144), 1,
      sym_pseudo_class_with_selector_arguments,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(625), 2,
      anon_sym_TILDE,
      anon_sym_PIPE,
    ACTIONS(623), 18,
      sym__descendant_operator,
      sym__pseudo_class_selector_colon,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_RPAREN,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_TILDE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_STAR_EQ,
      anon_sym_DOLLAR_EQ,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_PLUS,
  [5526] = 5,
    ACTIONS(633), 1,
      anon_sym_LPAREN2,
    STATE(125), 1,
      sym_pseudo_class_arguments,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(631), 2,
      anon_sym_TILDE,
      anon_sym_PIPE,
    ACTIONS(629), 18,
      sym__descendant_operator,
      sym__pseudo_class_selector_colon,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_RPAREN,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_TILDE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_STAR_EQ,
      anon_sym_DOLLAR_EQ,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_PLUS,
  [5561] = 15,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(635), 1,
      anon_sym_SEMI,
    ACTIONS(637), 1,
      anon_sym_LPAREN,
    ACTIONS(640), 1,
      anon_sym_POUND,
    ACTIONS(646), 1,
      anon_sym_SQUOTE,
    ACTIONS(649), 1,
      anon_sym_DQUOTE,
    ACTIONS(652), 1,
      aux_sym_integer_value_token1,
    ACTIONS(655), 1,
      aux_sym_float_value_token1,
    ACTIONS(658), 1,
      anon_sym_LBRACK2,
    ACTIONS(661), 1,
      sym_identifier,
    ACTIONS(664), 1,
      sym_important_value,
    STATE(111), 1,
      aux_sym_postcss_statement_repeat1,
    ACTIONS(643), 2,
      sym_important,
      sym_plain_value,
    STATE(229), 9,
      sym__value,
      sym_parenthesized_value,
      sym_color_value,
      sym_string_value,
      sym_integer_value,
      sym_float_value,
      sym_grid_value,
      sym_call_expression,
      sym_binary_expression,
  [5616] = 5,
    ACTIONS(627), 1,
      anon_sym_LPAREN2,
    STATE(153), 1,
      sym_pseudo_class_with_selector_arguments,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(669), 2,
      anon_sym_TILDE,
      anon_sym_PIPE,
    ACTIONS(667), 18,
      sym__descendant_operator,
      sym__pseudo_class_selector_colon,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_RPAREN,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_TILDE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_STAR_EQ,
      anon_sym_DOLLAR_EQ,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_PLUS,
  [5651] = 8,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(673), 1,
      anon_sym_LPAREN2,
    STATE(235), 1,
      sym_arguments,
    ACTIONS(675), 2,
      anon_sym_and,
      anon_sym_or,
    ACTIONS(671), 3,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_LBRACE,
    ACTIONS(383), 5,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_LBRACK2,
    ACTIONS(381), 10,
      anon_sym_LPAREN,
      anon_sym_PLUS,
      sym_important,
      aux_sym_integer_value_token1,
      aux_sym_float_value_token1,
      anon_sym_DASH,
      anon_sym_SLASH,
      sym_identifier,
      sym_plain_value,
      sym_important_value,
  [5692] = 5,
    ACTIONS(633), 1,
      anon_sym_LPAREN2,
    STATE(167), 1,
      sym_pseudo_class_arguments,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(679), 2,
      anon_sym_TILDE,
      anon_sym_PIPE,
    ACTIONS(677), 18,
      sym__descendant_operator,
      sym__pseudo_class_selector_colon,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_RPAREN,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_TILDE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_STAR_EQ,
      anon_sym_DOLLAR_EQ,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_PLUS,
  [5727] = 15,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(263), 1,
      anon_sym_POUND,
    ACTIONS(269), 1,
      anon_sym_SQUOTE,
    ACTIONS(271), 1,
      anon_sym_DQUOTE,
    ACTIONS(273), 1,
      aux_sym_integer_value_token1,
    ACTIONS(275), 1,
      aux_sym_float_value_token1,
    ACTIONS(277), 1,
      anon_sym_LBRACK2,
    ACTIONS(681), 1,
      anon_sym_SEMI,
    ACTIONS(683), 1,
      anon_sym_LPAREN,
    ACTIONS(685), 1,
      sym_identifier,
    ACTIONS(687), 1,
      sym_important_value,
    STATE(111), 1,
      aux_sym_postcss_statement_repeat1,
    ACTIONS(265), 2,
      sym_important,
      sym_plain_value,
    STATE(229), 9,
      sym__value,
      sym_parenthesized_value,
      sym_color_value,
      sym_string_value,
      sym_integer_value,
      sym_float_value,
      sym_grid_value,
      sym_call_expression,
      sym_binary_expression,
  [5782] = 5,
    ACTIONS(693), 1,
      anon_sym_LPAREN2,
    STATE(146), 1,
      sym_pseudo_element_arguments,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(691), 2,
      anon_sym_TILDE,
      anon_sym_PIPE,
    ACTIONS(689), 18,
      sym__descendant_operator,
      sym__pseudo_class_selector_colon,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_RPAREN,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_TILDE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_STAR_EQ,
      anon_sym_DOLLAR_EQ,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_PLUS,
  [5817] = 16,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(221), 1,
      anon_sym_LPAREN,
    ACTIONS(239), 1,
      aux_sym_integer_value_token1,
    ACTIONS(241), 1,
      aux_sym_float_value_token1,
    ACTIONS(337), 1,
      anon_sym_COMMA,
    ACTIONS(345), 1,
      anon_sym_POUND,
    ACTIONS(351), 1,
      anon_sym_SQUOTE,
    ACTIONS(353), 1,
      anon_sym_DQUOTE,
    ACTIONS(355), 1,
      anon_sym_LBRACK2,
    ACTIONS(357), 1,
      sym_identifier,
    ACTIONS(359), 1,
      sym_plain_value,
    ACTIONS(695), 1,
      anon_sym_SEMI,
    ACTIONS(697), 1,
      sym_important,
    STATE(102), 1,
      aux_sym_declaration_repeat1,
    STATE(214), 9,
      sym__value,
      sym_parenthesized_value,
      sym_color_value,
      sym_string_value,
      sym_integer_value,
      sym_float_value,
      sym_grid_value,
      sym_call_expression,
      sym_binary_expression,
  [5874] = 5,
    ACTIONS(693), 1,
      anon_sym_LPAREN2,
    STATE(145), 1,
      sym_pseudo_element_arguments,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(701), 2,
      anon_sym_TILDE,
      anon_sym_PIPE,
    ACTIONS(699), 18,
      sym__descendant_operator,
      sym__pseudo_class_selector_colon,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_RPAREN,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_TILDE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_STAR_EQ,
      anon_sym_DOLLAR_EQ,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_PLUS,
  [5909] = 16,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(221), 1,
      anon_sym_LPAREN,
    ACTIONS(239), 1,
      aux_sym_integer_value_token1,
    ACTIONS(241), 1,
      aux_sym_float_value_token1,
    ACTIONS(337), 1,
      anon_sym_COMMA,
    ACTIONS(345), 1,
      anon_sym_POUND,
    ACTIONS(351), 1,
      anon_sym_SQUOTE,
    ACTIONS(353), 1,
      anon_sym_DQUOTE,
    ACTIONS(355), 1,
      anon_sym_LBRACK2,
    ACTIONS(357), 1,
      sym_identifier,
    ACTIONS(359), 1,
      sym_plain_value,
    ACTIONS(581), 1,
      anon_sym_SEMI,
    ACTIONS(703), 1,
      sym_important,
    STATE(102), 1,
      aux_sym_declaration_repeat1,
    STATE(214), 9,
      sym__value,
      sym_parenthesized_value,
      sym_color_value,
      sym_string_value,
      sym_integer_value,
      sym_float_value,
      sym_grid_value,
      sym_call_expression,
      sym_binary_expression,
  [5966] = 15,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(221), 1,
      anon_sym_LPAREN,
    ACTIONS(233), 1,
      sym_important,
    ACTIONS(239), 1,
      aux_sym_integer_value_token1,
    ACTIONS(241), 1,
      aux_sym_float_value_token1,
    ACTIONS(247), 1,
      sym_plain_value,
    ACTIONS(345), 1,
      anon_sym_POUND,
    ACTIONS(351), 1,
      anon_sym_SQUOTE,
    ACTIONS(353), 1,
      anon_sym_DQUOTE,
    ACTIONS(355), 1,
      anon_sym_LBRACK2,
    ACTIONS(357), 1,
      sym_identifier,
    ACTIONS(705), 1,
      anon_sym_RPAREN,
    STATE(107), 1,
      aux_sym_pseudo_class_arguments_repeat1,
    STATE(217), 9,
      sym__value,
      sym_parenthesized_value,
      sym_color_value,
      sym_string_value,
      sym_integer_value,
      sym_float_value,
      sym_grid_value,
      sym_call_expression,
      sym_binary_expression,
  [6020] = 15,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(221), 1,
      anon_sym_LPAREN,
    ACTIONS(233), 1,
      sym_important,
    ACTIONS(239), 1,
      aux_sym_integer_value_token1,
    ACTIONS(241), 1,
      aux_sym_float_value_token1,
    ACTIONS(247), 1,
      sym_plain_value,
    ACTIONS(345), 1,
      anon_sym_POUND,
    ACTIONS(351), 1,
      anon_sym_SQUOTE,
    ACTIONS(353), 1,
      anon_sym_DQUOTE,
    ACTIONS(355), 1,
      anon_sym_LBRACK2,
    ACTIONS(357), 1,
      sym_identifier,
    ACTIONS(707), 1,
      anon_sym_RPAREN,
    STATE(98), 1,
      aux_sym_pseudo_class_arguments_repeat1,
    STATE(217), 9,
      sym__value,
      sym_parenthesized_value,
      sym_color_value,
      sym_string_value,
      sym_integer_value,
      sym_float_value,
      sym_grid_value,
      sym_call_expression,
      sym_binary_expression,
  [6074] = 15,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(221), 1,
      anon_sym_LPAREN,
    ACTIONS(233), 1,
      sym_important,
    ACTIONS(239), 1,
      aux_sym_integer_value_token1,
    ACTIONS(241), 1,
      aux_sym_float_value_token1,
    ACTIONS(247), 1,
      sym_plain_value,
    ACTIONS(345), 1,
      anon_sym_POUND,
    ACTIONS(351), 1,
      anon_sym_SQUOTE,
    ACTIONS(353), 1,
      anon_sym_DQUOTE,
    ACTIONS(355), 1,
      anon_sym_LBRACK2,
    ACTIONS(357), 1,
      sym_identifier,
    ACTIONS(709), 1,
      anon_sym_RPAREN,
    STATE(99), 1,
      aux_sym_pseudo_class_arguments_repeat1,
    STATE(217), 9,
      sym__value,
      sym_parenthesized_value,
      sym_color_value,
      sym_string_value,
      sym_integer_value,
      sym_float_value,
      sym_grid_value,
      sym_call_expression,
      sym_binary_expression,
  [6128] = 15,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(221), 1,
      anon_sym_LPAREN,
    ACTIONS(233), 1,
      sym_important,
    ACTIONS(239), 1,
      aux_sym_integer_value_token1,
    ACTIONS(241), 1,
      aux_sym_float_value_token1,
    ACTIONS(247), 1,
      sym_plain_value,
    ACTIONS(345), 1,
      anon_sym_POUND,
    ACTIONS(351), 1,
      anon_sym_SQUOTE,
    ACTIONS(353), 1,
      anon_sym_DQUOTE,
    ACTIONS(355), 1,
      anon_sym_LBRACK2,
    ACTIONS(357), 1,
      sym_identifier,
    ACTIONS(711), 1,
      anon_sym_RPAREN,
    STATE(100), 1,
      aux_sym_pseudo_class_arguments_repeat1,
    STATE(217), 9,
      sym__value,
      sym_parenthesized_value,
      sym_color_value,
      sym_string_value,
      sym_integer_value,
      sym_float_value,
      sym_grid_value,
      sym_call_expression,
      sym_binary_expression,
  [6182] = 15,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(221), 1,
      anon_sym_LPAREN,
    ACTIONS(233), 1,
      sym_important,
    ACTIONS(239), 1,
      aux_sym_integer_value_token1,
    ACTIONS(241), 1,
      aux_sym_float_value_token1,
    ACTIONS(247), 1,
      sym_plain_value,
    ACTIONS(345), 1,
      anon_sym_POUND,
    ACTIONS(351), 1,
      anon_sym_SQUOTE,
    ACTIONS(353), 1,
      anon_sym_DQUOTE,
    ACTIONS(355), 1,
      anon_sym_LBRACK2,
    ACTIONS(357), 1,
      sym_identifier,
    ACTIONS(713), 1,
      anon_sym_RPAREN,
    STATE(52), 1,
      aux_sym_pseudo_class_arguments_repeat1,
    STATE(217), 9,
      sym__value,
      sym_parenthesized_value,
      sym_color_value,
      sym_string_value,
      sym_integer_value,
      sym_float_value,
      sym_grid_value,
      sym_call_expression,
      sym_binary_expression,
  [6236] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(679), 2,
      anon_sym_TILDE,
      anon_sym_PIPE,
    ACTIONS(677), 18,
      sym__descendant_operator,
      sym__pseudo_class_selector_colon,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_RPAREN,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_TILDE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_STAR_EQ,
      anon_sym_DOLLAR_EQ,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_PLUS,
  [6265] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(717), 2,
      anon_sym_TILDE,
      anon_sym_PIPE,
    ACTIONS(715), 18,
      sym__descendant_operator,
      sym__pseudo_class_selector_colon,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_RPAREN,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_TILDE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_STAR_EQ,
      anon_sym_DOLLAR_EQ,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_PLUS,
  [6294] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(721), 2,
      anon_sym_TILDE,
      anon_sym_PIPE,
    ACTIONS(719), 18,
      sym__descendant_operator,
      sym__pseudo_class_selector_colon,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_RPAREN,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_TILDE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_STAR_EQ,
      anon_sym_DOLLAR_EQ,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_PLUS,
  [6323] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(526), 2,
      anon_sym_TILDE,
      anon_sym_PIPE,
    ACTIONS(524), 18,
      sym__descendant_operator,
      sym__pseudo_class_selector_colon,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_RPAREN,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_TILDE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_STAR_EQ,
      anon_sym_DOLLAR_EQ,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_PLUS,
  [6352] = 14,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(221), 1,
      anon_sym_LPAREN,
    ACTIONS(233), 1,
      sym_important,
    ACTIONS(239), 1,
      aux_sym_integer_value_token1,
    ACTIONS(241), 1,
      aux_sym_float_value_token1,
    ACTIONS(247), 1,
      sym_plain_value,
    ACTIONS(345), 1,
      anon_sym_POUND,
    ACTIONS(351), 1,
      anon_sym_SQUOTE,
    ACTIONS(353), 1,
      anon_sym_DQUOTE,
    ACTIONS(355), 1,
      anon_sym_LBRACK2,
    ACTIONS(357), 1,
      sym_identifier,
    STATE(120), 1,
      aux_sym_pseudo_class_arguments_repeat1,
    STATE(217), 9,
      sym__value,
      sym_parenthesized_value,
      sym_color_value,
      sym_string_value,
      sym_integer_value,
      sym_float_value,
      sym_grid_value,
      sym_call_expression,
      sym_binary_expression,
  [6403] = 4,
    ACTIONS(725), 1,
      anon_sym_DOT,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(727), 2,
      anon_sym_TILDE,
      anon_sym_PIPE,
    ACTIONS(723), 17,
      sym__descendant_operator,
      sym__pseudo_class_selector_colon,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_RPAREN,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_TILDE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_STAR_EQ,
      anon_sym_DOLLAR_EQ,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_PLUS,
  [6434] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(522), 2,
      anon_sym_TILDE,
      anon_sym_PIPE,
    ACTIONS(520), 18,
      sym__descendant_operator,
      sym__pseudo_class_selector_colon,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_RPAREN,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_TILDE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_STAR_EQ,
      anon_sym_DOLLAR_EQ,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_PLUS,
  [6463] = 4,
    ACTIONS(725), 1,
      anon_sym_DOT,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(731), 2,
      anon_sym_TILDE,
      anon_sym_PIPE,
    ACTIONS(729), 17,
      sym__descendant_operator,
      sym__pseudo_class_selector_colon,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_RPAREN,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_TILDE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_STAR_EQ,
      anon_sym_DOLLAR_EQ,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_PLUS,
  [6494] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(735), 2,
      anon_sym_TILDE,
      anon_sym_PIPE,
    ACTIONS(733), 18,
      sym__descendant_operator,
      sym__pseudo_class_selector_colon,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_RPAREN,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_TILDE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_STAR_EQ,
      anon_sym_DOLLAR_EQ,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_PLUS,
  [6523] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(739), 2,
      anon_sym_TILDE,
      anon_sym_PIPE,
    ACTIONS(737), 18,
      sym__descendant_operator,
      sym__pseudo_class_selector_colon,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_RPAREN,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_TILDE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_STAR_EQ,
      anon_sym_DOLLAR_EQ,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_PLUS,
  [6552] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(743), 2,
      anon_sym_TILDE,
      anon_sym_PIPE,
    ACTIONS(741), 18,
      sym__descendant_operator,
      sym__pseudo_class_selector_colon,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_RPAREN,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_TILDE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_STAR_EQ,
      anon_sym_DOLLAR_EQ,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_PLUS,
  [6581] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(747), 2,
      anon_sym_TILDE,
      anon_sym_PIPE,
    ACTIONS(745), 18,
      sym__descendant_operator,
      sym__pseudo_class_selector_colon,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_RPAREN,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_TILDE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_STAR_EQ,
      anon_sym_DOLLAR_EQ,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_PLUS,
  [6610] = 4,
    ACTIONS(725), 1,
      anon_sym_DOT,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(751), 2,
      anon_sym_TILDE,
      anon_sym_PIPE,
    ACTIONS(749), 17,
      sym__descendant_operator,
      sym__pseudo_class_selector_colon,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_RPAREN,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_TILDE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_STAR_EQ,
      anon_sym_DOLLAR_EQ,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_PLUS,
  [6641] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(755), 2,
      anon_sym_TILDE,
      anon_sym_PIPE,
    ACTIONS(753), 18,
      sym__descendant_operator,
      sym__pseudo_class_selector_colon,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_RPAREN,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_TILDE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_STAR_EQ,
      anon_sym_DOLLAR_EQ,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_PLUS,
  [6670] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(759), 2,
      anon_sym_TILDE,
      anon_sym_PIPE,
    ACTIONS(757), 18,
      sym__descendant_operator,
      sym__pseudo_class_selector_colon,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_RPAREN,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_TILDE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_STAR_EQ,
      anon_sym_DOLLAR_EQ,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_PLUS,
  [6699] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(763), 2,
      anon_sym_TILDE,
      anon_sym_PIPE,
    ACTIONS(761), 18,
      sym__descendant_operator,
      sym__pseudo_class_selector_colon,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_RPAREN,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_TILDE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_STAR_EQ,
      anon_sym_DOLLAR_EQ,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_PLUS,
  [6728] = 4,
    ACTIONS(725), 1,
      anon_sym_DOT,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(767), 2,
      anon_sym_TILDE,
      anon_sym_PIPE,
    ACTIONS(765), 17,
      sym__descendant_operator,
      sym__pseudo_class_selector_colon,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_RPAREN,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_TILDE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_STAR_EQ,
      anon_sym_DOLLAR_EQ,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_PLUS,
  [6759] = 4,
    ACTIONS(725), 1,
      anon_sym_DOT,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(771), 2,
      anon_sym_TILDE,
      anon_sym_PIPE,
    ACTIONS(769), 17,
      sym__descendant_operator,
      sym__pseudo_class_selector_colon,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_RPAREN,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_TILDE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_STAR_EQ,
      anon_sym_DOLLAR_EQ,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_PLUS,
  [6790] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(775), 2,
      anon_sym_TILDE,
      anon_sym_PIPE,
    ACTIONS(773), 18,
      sym__descendant_operator,
      sym__pseudo_class_selector_colon,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_RPAREN,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_TILDE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_STAR_EQ,
      anon_sym_DOLLAR_EQ,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_PLUS,
  [6819] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(779), 2,
      anon_sym_TILDE,
      anon_sym_PIPE,
    ACTIONS(777), 18,
      sym__descendant_operator,
      sym__pseudo_class_selector_colon,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_RPAREN,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_TILDE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_STAR_EQ,
      anon_sym_DOLLAR_EQ,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_PLUS,
  [6848] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(783), 2,
      anon_sym_TILDE,
      anon_sym_PIPE,
    ACTIONS(781), 18,
      sym__descendant_operator,
      sym__pseudo_class_selector_colon,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_RPAREN,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_TILDE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_STAR_EQ,
      anon_sym_DOLLAR_EQ,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_PLUS,
  [6877] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(787), 2,
      anon_sym_TILDE,
      anon_sym_PIPE,
    ACTIONS(785), 18,
      sym__descendant_operator,
      sym__pseudo_class_selector_colon,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_RPAREN,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_TILDE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_STAR_EQ,
      anon_sym_DOLLAR_EQ,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_PLUS,
  [6906] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(631), 2,
      anon_sym_TILDE,
      anon_sym_PIPE,
    ACTIONS(629), 18,
      sym__descendant_operator,
      sym__pseudo_class_selector_colon,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_RPAREN,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_TILDE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_STAR_EQ,
      anon_sym_DOLLAR_EQ,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_PLUS,
  [6935] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(791), 2,
      anon_sym_TILDE,
      anon_sym_PIPE,
    ACTIONS(789), 18,
      sym__descendant_operator,
      sym__pseudo_class_selector_colon,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_RPAREN,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_TILDE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_STAR_EQ,
      anon_sym_DOLLAR_EQ,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_PLUS,
  [6964] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(795), 2,
      anon_sym_TILDE,
      anon_sym_PIPE,
    ACTIONS(793), 18,
      sym__descendant_operator,
      sym__pseudo_class_selector_colon,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_RPAREN,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_TILDE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_STAR_EQ,
      anon_sym_DOLLAR_EQ,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_PLUS,
  [6993] = 6,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(387), 1,
      anon_sym_LPAREN2,
    STATE(193), 1,
      sym_arguments,
    ACTIONS(381), 8,
      anon_sym_LPAREN,
      anon_sym_PLUS,
      aux_sym_integer_value_token1,
      aux_sym_float_value_token1,
      anon_sym_DASH,
      anon_sym_SLASH,
      sym_identifier,
      sym_plain_value,
    ACTIONS(383), 10,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_STAR,
      anon_sym_POUND,
      sym_important,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_LBRACK2,
  [7028] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(799), 2,
      anon_sym_TILDE,
      anon_sym_PIPE,
    ACTIONS(797), 18,
      sym__descendant_operator,
      sym__pseudo_class_selector_colon,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_RPAREN,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_TILDE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_STAR_EQ,
      anon_sym_DOLLAR_EQ,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_PLUS,
  [7057] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(803), 2,
      anon_sym_TILDE,
      anon_sym_PIPE,
    ACTIONS(801), 18,
      sym__descendant_operator,
      sym__pseudo_class_selector_colon,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_RPAREN,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_TILDE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_STAR_EQ,
      anon_sym_DOLLAR_EQ,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_PLUS,
  [7086] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(807), 2,
      anon_sym_TILDE,
      anon_sym_PIPE,
    ACTIONS(805), 18,
      sym__descendant_operator,
      sym__pseudo_class_selector_colon,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_RPAREN,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_TILDE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_STAR_EQ,
      anon_sym_DOLLAR_EQ,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_PLUS,
  [7115] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(385), 2,
      anon_sym_TILDE,
      anon_sym_PIPE,
    ACTIONS(379), 18,
      sym__descendant_operator,
      sym__pseudo_class_selector_colon,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_RPAREN,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_TILDE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_STAR_EQ,
      anon_sym_DOLLAR_EQ,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_PLUS,
  [7144] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(811), 2,
      anon_sym_TILDE,
      anon_sym_PIPE,
    ACTIONS(809), 18,
      sym__descendant_operator,
      sym__pseudo_class_selector_colon,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_RPAREN,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_TILDE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_STAR_EQ,
      anon_sym_DOLLAR_EQ,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_PLUS,
  [7173] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(815), 2,
      anon_sym_TILDE,
      anon_sym_PIPE,
    ACTIONS(813), 18,
      sym__descendant_operator,
      sym__pseudo_class_selector_colon,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_RPAREN,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_TILDE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_STAR_EQ,
      anon_sym_DOLLAR_EQ,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_PLUS,
  [7202] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(819), 2,
      anon_sym_TILDE,
      anon_sym_PIPE,
    ACTIONS(817), 18,
      sym__descendant_operator,
      sym__pseudo_class_selector_colon,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_RPAREN,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_TILDE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_STAR_EQ,
      anon_sym_DOLLAR_EQ,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_PLUS,
  [7231] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(823), 2,
      anon_sym_TILDE,
      anon_sym_PIPE,
    ACTIONS(821), 18,
      sym__descendant_operator,
      sym__pseudo_class_selector_colon,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_RPAREN,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_TILDE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_STAR_EQ,
      anon_sym_DOLLAR_EQ,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_PLUS,
  [7260] = 4,
    ACTIONS(725), 1,
      anon_sym_DOT,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(827), 2,
      anon_sym_TILDE,
      anon_sym_PIPE,
    ACTIONS(825), 17,
      sym__descendant_operator,
      sym__pseudo_class_selector_colon,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_RPAREN,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_TILDE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_STAR_EQ,
      anon_sym_DOLLAR_EQ,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_PLUS,
  [7291] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(831), 2,
      anon_sym_TILDE,
      anon_sym_PIPE,
    ACTIONS(829), 18,
      sym__descendant_operator,
      sym__pseudo_class_selector_colon,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_RPAREN,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_TILDE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_STAR_EQ,
      anon_sym_DOLLAR_EQ,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_PLUS,
  [7320] = 4,
    ACTIONS(725), 1,
      anon_sym_DOT,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(835), 2,
      anon_sym_TILDE,
      anon_sym_PIPE,
    ACTIONS(833), 17,
      sym__descendant_operator,
      sym__pseudo_class_selector_colon,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_RPAREN,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_TILDE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_STAR_EQ,
      anon_sym_DOLLAR_EQ,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_PLUS,
  [7351] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(839), 2,
      anon_sym_TILDE,
      anon_sym_PIPE,
    ACTIONS(837), 18,
      sym__descendant_operator,
      sym__pseudo_class_selector_colon,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_RPAREN,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_TILDE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_STAR_EQ,
      anon_sym_DOLLAR_EQ,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_PLUS,
  [7380] = 4,
    ACTIONS(725), 1,
      anon_sym_DOT,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(843), 2,
      anon_sym_TILDE,
      anon_sym_PIPE,
    ACTIONS(841), 17,
      sym__descendant_operator,
      sym__pseudo_class_selector_colon,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_RPAREN,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_TILDE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_STAR_EQ,
      anon_sym_DOLLAR_EQ,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_PLUS,
  [7411] = 14,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(221), 1,
      anon_sym_LPAREN,
    ACTIONS(233), 1,
      sym_important,
    ACTIONS(239), 1,
      aux_sym_integer_value_token1,
    ACTIONS(241), 1,
      aux_sym_float_value_token1,
    ACTIONS(247), 1,
      sym_plain_value,
    ACTIONS(345), 1,
      anon_sym_POUND,
    ACTIONS(351), 1,
      anon_sym_SQUOTE,
    ACTIONS(353), 1,
      anon_sym_DQUOTE,
    ACTIONS(355), 1,
      anon_sym_LBRACK2,
    ACTIONS(357), 1,
      sym_identifier,
    STATE(104), 1,
      aux_sym_pseudo_class_arguments_repeat1,
    STATE(217), 9,
      sym__value,
      sym_parenthesized_value,
      sym_color_value,
      sym_string_value,
      sym_integer_value,
      sym_float_value,
      sym_grid_value,
      sym_call_expression,
      sym_binary_expression,
  [7462] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(847), 2,
      anon_sym_TILDE,
      anon_sym_PIPE,
    ACTIONS(845), 18,
      sym__descendant_operator,
      sym__pseudo_class_selector_colon,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_RPAREN,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_TILDE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_STAR_EQ,
      anon_sym_DOLLAR_EQ,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_PLUS,
  [7491] = 4,
    ACTIONS(725), 1,
      anon_sym_DOT,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(851), 2,
      anon_sym_TILDE,
      anon_sym_PIPE,
    ACTIONS(849), 17,
      sym__descendant_operator,
      sym__pseudo_class_selector_colon,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_RPAREN,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_TILDE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_STAR_EQ,
      anon_sym_DOLLAR_EQ,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_PLUS,
  [7522] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(855), 2,
      anon_sym_TILDE,
      anon_sym_PIPE,
    ACTIONS(853), 18,
      sym__descendant_operator,
      sym__pseudo_class_selector_colon,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_RPAREN,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_TILDE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_STAR_EQ,
      anon_sym_DOLLAR_EQ,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_PLUS,
  [7551] = 13,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(285), 1,
      anon_sym_POUND,
    ACTIONS(289), 1,
      anon_sym_SQUOTE,
    ACTIONS(291), 1,
      anon_sym_DQUOTE,
    ACTIONS(293), 1,
      aux_sym_integer_value_token1,
    ACTIONS(295), 1,
      aux_sym_float_value_token1,
    ACTIONS(297), 1,
      anon_sym_LBRACK2,
    ACTIONS(857), 1,
      anon_sym_LPAREN,
    ACTIONS(859), 1,
      sym_important,
    ACTIONS(861), 1,
      sym_identifier,
    ACTIONS(863), 1,
      sym_plain_value,
    STATE(326), 9,
      sym__value,
      sym_parenthesized_value,
      sym_color_value,
      sym_string_value,
      sym_integer_value,
      sym_float_value,
      sym_grid_value,
      sym_call_expression,
      sym_binary_expression,
  [7599] = 13,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(285), 1,
      anon_sym_POUND,
    ACTIONS(289), 1,
      anon_sym_SQUOTE,
    ACTIONS(291), 1,
      anon_sym_DQUOTE,
    ACTIONS(293), 1,
      aux_sym_integer_value_token1,
    ACTIONS(295), 1,
      aux_sym_float_value_token1,
    ACTIONS(297), 1,
      anon_sym_LBRACK2,
    ACTIONS(857), 1,
      anon_sym_LPAREN,
    ACTIONS(861), 1,
      sym_identifier,
    ACTIONS(865), 1,
      sym_important,
    ACTIONS(867), 1,
      sym_plain_value,
    STATE(331), 9,
      sym__value,
      sym_parenthesized_value,
      sym_color_value,
      sym_string_value,
      sym_integer_value,
      sym_float_value,
      sym_grid_value,
      sym_call_expression,
      sym_binary_expression,
  [7647] = 13,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(221), 1,
      anon_sym_LPAREN,
    ACTIONS(239), 1,
      aux_sym_integer_value_token1,
    ACTIONS(241), 1,
      aux_sym_float_value_token1,
    ACTIONS(345), 1,
      anon_sym_POUND,
    ACTIONS(351), 1,
      anon_sym_SQUOTE,
    ACTIONS(353), 1,
      anon_sym_DQUOTE,
    ACTIONS(355), 1,
      anon_sym_LBRACK2,
    ACTIONS(357), 1,
      sym_identifier,
    ACTIONS(869), 1,
      sym_important,
    ACTIONS(871), 1,
      sym_plain_value,
    STATE(37), 9,
      sym__value,
      sym_parenthesized_value,
      sym_color_value,
      sym_string_value,
      sym_integer_value,
      sym_float_value,
      sym_grid_value,
      sym_call_expression,
      sym_binary_expression,
  [7695] = 13,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(221), 1,
      anon_sym_LPAREN,
    ACTIONS(239), 1,
      aux_sym_integer_value_token1,
    ACTIONS(241), 1,
      aux_sym_float_value_token1,
    ACTIONS(345), 1,
      anon_sym_POUND,
    ACTIONS(351), 1,
      anon_sym_SQUOTE,
    ACTIONS(353), 1,
      anon_sym_DQUOTE,
    ACTIONS(355), 1,
      anon_sym_LBRACK2,
    ACTIONS(357), 1,
      sym_identifier,
    ACTIONS(873), 1,
      sym_important,
    ACTIONS(875), 1,
      sym_plain_value,
    STATE(210), 9,
      sym__value,
      sym_parenthesized_value,
      sym_color_value,
      sym_string_value,
      sym_integer_value,
      sym_float_value,
      sym_grid_value,
      sym_call_expression,
      sym_binary_expression,
  [7743] = 13,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(285), 1,
      anon_sym_POUND,
    ACTIONS(289), 1,
      anon_sym_SQUOTE,
    ACTIONS(291), 1,
      anon_sym_DQUOTE,
    ACTIONS(293), 1,
      aux_sym_integer_value_token1,
    ACTIONS(295), 1,
      aux_sym_float_value_token1,
    ACTIONS(297), 1,
      anon_sym_LBRACK2,
    ACTIONS(857), 1,
      anon_sym_LPAREN,
    ACTIONS(861), 1,
      sym_identifier,
    ACTIONS(877), 1,
      sym_important,
    ACTIONS(879), 1,
      sym_plain_value,
    STATE(320), 9,
      sym__value,
      sym_parenthesized_value,
      sym_color_value,
      sym_string_value,
      sym_integer_value,
      sym_float_value,
      sym_grid_value,
      sym_call_expression,
      sym_binary_expression,
  [7791] = 5,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(885), 1,
      sym_unit,
    ACTIONS(883), 7,
      anon_sym_PLUS,
      aux_sym_integer_value_token1,
      aux_sym_float_value_token1,
      anon_sym_DASH,
      anon_sym_SLASH,
      sym_identifier,
      sym_plain_value,
    ACTIONS(881), 11,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_STAR,
      anon_sym_POUND,
      sym_important,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_LBRACK2,
  [7823] = 5,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(891), 1,
      sym_unit,
    ACTIONS(889), 7,
      anon_sym_PLUS,
      aux_sym_integer_value_token1,
      aux_sym_float_value_token1,
      anon_sym_DASH,
      anon_sym_SLASH,
      sym_identifier,
      sym_plain_value,
    ACTIONS(887), 11,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_STAR,
      anon_sym_POUND,
      sym_important,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_LBRACK2,
  [7855] = 13,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(221), 1,
      anon_sym_LPAREN,
    ACTIONS(239), 1,
      aux_sym_integer_value_token1,
    ACTIONS(241), 1,
      aux_sym_float_value_token1,
    ACTIONS(345), 1,
      anon_sym_POUND,
    ACTIONS(351), 1,
      anon_sym_SQUOTE,
    ACTIONS(353), 1,
      anon_sym_DQUOTE,
    ACTIONS(355), 1,
      anon_sym_LBRACK2,
    ACTIONS(357), 1,
      sym_identifier,
    ACTIONS(893), 1,
      sym_important,
    ACTIONS(895), 1,
      sym_plain_value,
    STATE(198), 9,
      sym__value,
      sym_parenthesized_value,
      sym_color_value,
      sym_string_value,
      sym_integer_value,
      sym_float_value,
      sym_grid_value,
      sym_call_expression,
      sym_binary_expression,
  [7903] = 13,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(263), 1,
      anon_sym_POUND,
    ACTIONS(269), 1,
      anon_sym_SQUOTE,
    ACTIONS(271), 1,
      anon_sym_DQUOTE,
    ACTIONS(273), 1,
      aux_sym_integer_value_token1,
    ACTIONS(275), 1,
      aux_sym_float_value_token1,
    ACTIONS(277), 1,
      anon_sym_LBRACK2,
    ACTIONS(683), 1,
      anon_sym_LPAREN,
    ACTIONS(685), 1,
      sym_identifier,
    ACTIONS(897), 1,
      sym_important,
    ACTIONS(899), 1,
      sym_plain_value,
    STATE(231), 9,
      sym__value,
      sym_parenthesized_value,
      sym_color_value,
      sym_string_value,
      sym_integer_value,
      sym_float_value,
      sym_grid_value,
      sym_call_expression,
      sym_binary_expression,
  [7951] = 13,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(285), 1,
      anon_sym_POUND,
    ACTIONS(289), 1,
      anon_sym_SQUOTE,
    ACTIONS(291), 1,
      anon_sym_DQUOTE,
    ACTIONS(293), 1,
      aux_sym_integer_value_token1,
    ACTIONS(295), 1,
      aux_sym_float_value_token1,
    ACTIONS(297), 1,
      anon_sym_LBRACK2,
    ACTIONS(857), 1,
      anon_sym_LPAREN,
    ACTIONS(861), 1,
      sym_identifier,
    ACTIONS(901), 1,
      sym_important,
    ACTIONS(903), 1,
      sym_plain_value,
    STATE(335), 9,
      sym__value,
      sym_parenthesized_value,
      sym_color_value,
      sym_string_value,
      sym_integer_value,
      sym_float_value,
      sym_grid_value,
      sym_call_expression,
      sym_binary_expression,
  [7999] = 13,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(285), 1,
      anon_sym_POUND,
    ACTIONS(289), 1,
      anon_sym_SQUOTE,
    ACTIONS(291), 1,
      anon_sym_DQUOTE,
    ACTIONS(293), 1,
      aux_sym_integer_value_token1,
    ACTIONS(295), 1,
      aux_sym_float_value_token1,
    ACTIONS(297), 1,
      anon_sym_LBRACK2,
    ACTIONS(857), 1,
      anon_sym_LPAREN,
    ACTIONS(861), 1,
      sym_identifier,
    ACTIONS(905), 1,
      sym_important,
    ACTIONS(907), 1,
      sym_plain_value,
    STATE(318), 9,
      sym__value,
      sym_parenthesized_value,
      sym_color_value,
      sym_string_value,
      sym_integer_value,
      sym_float_value,
      sym_grid_value,
      sym_call_expression,
      sym_binary_expression,
  [8047] = 13,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(909), 1,
      anon_sym_LPAREN,
    ACTIONS(911), 1,
      anon_sym_POUND,
    ACTIONS(913), 1,
      sym_important,
    ACTIONS(915), 1,
      anon_sym_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      aux_sym_integer_value_token1,
    ACTIONS(921), 1,
      aux_sym_float_value_token1,
    ACTIONS(923), 1,
      anon_sym_LBRACK2,
    ACTIONS(925), 1,
      sym_identifier,
    ACTIONS(927), 1,
      sym_plain_value,
    STATE(281), 9,
      sym__value,
      sym_parenthesized_value,
      sym_color_value,
      sym_string_value,
      sym_integer_value,
      sym_float_value,
      sym_grid_value,
      sym_call_expression,
      sym_binary_expression,
  [8095] = 13,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(285), 1,
      anon_sym_POUND,
    ACTIONS(289), 1,
      anon_sym_SQUOTE,
    ACTIONS(291), 1,
      anon_sym_DQUOTE,
    ACTIONS(293), 1,
      aux_sym_integer_value_token1,
    ACTIONS(295), 1,
      aux_sym_float_value_token1,
    ACTIONS(297), 1,
      anon_sym_LBRACK2,
    ACTIONS(857), 1,
      anon_sym_LPAREN,
    ACTIONS(861), 1,
      sym_identifier,
    ACTIONS(929), 1,
      sym_important,
    ACTIONS(931), 1,
      sym_plain_value,
    STATE(303), 9,
      sym__value,
      sym_parenthesized_value,
      sym_color_value,
      sym_string_value,
      sym_integer_value,
      sym_float_value,
      sym_grid_value,
      sym_call_expression,
      sym_binary_expression,
  [8143] = 13,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(285), 1,
      anon_sym_POUND,
    ACTIONS(289), 1,
      anon_sym_SQUOTE,
    ACTIONS(291), 1,
      anon_sym_DQUOTE,
    ACTIONS(293), 1,
      aux_sym_integer_value_token1,
    ACTIONS(295), 1,
      aux_sym_float_value_token1,
    ACTIONS(297), 1,
      anon_sym_LBRACK2,
    ACTIONS(857), 1,
      anon_sym_LPAREN,
    ACTIONS(861), 1,
      sym_identifier,
    ACTIONS(933), 1,
      sym_important,
    ACTIONS(935), 1,
      sym_plain_value,
    STATE(294), 9,
      sym__value,
      sym_parenthesized_value,
      sym_color_value,
      sym_string_value,
      sym_integer_value,
      sym_float_value,
      sym_grid_value,
      sym_call_expression,
      sym_binary_expression,
  [8191] = 13,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(909), 1,
      anon_sym_LPAREN,
    ACTIONS(911), 1,
      anon_sym_POUND,
    ACTIONS(915), 1,
      anon_sym_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      aux_sym_integer_value_token1,
    ACTIONS(921), 1,
      aux_sym_float_value_token1,
    ACTIONS(923), 1,
      anon_sym_LBRACK2,
    ACTIONS(925), 1,
      sym_identifier,
    ACTIONS(937), 1,
      sym_important,
    ACTIONS(939), 1,
      sym_plain_value,
    STATE(221), 9,
      sym__value,
      sym_parenthesized_value,
      sym_color_value,
      sym_string_value,
      sym_integer_value,
      sym_float_value,
      sym_grid_value,
      sym_call_expression,
      sym_binary_expression,
  [8239] = 13,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(285), 1,
      anon_sym_POUND,
    ACTIONS(289), 1,
      anon_sym_SQUOTE,
    ACTIONS(291), 1,
      anon_sym_DQUOTE,
    ACTIONS(293), 1,
      aux_sym_integer_value_token1,
    ACTIONS(295), 1,
      aux_sym_float_value_token1,
    ACTIONS(297), 1,
      anon_sym_LBRACK2,
    ACTIONS(857), 1,
      anon_sym_LPAREN,
    ACTIONS(861), 1,
      sym_identifier,
    ACTIONS(941), 1,
      sym_important,
    ACTIONS(943), 1,
      sym_plain_value,
    STATE(321), 9,
      sym__value,
      sym_parenthesized_value,
      sym_color_value,
      sym_string_value,
      sym_integer_value,
      sym_float_value,
      sym_grid_value,
      sym_call_expression,
      sym_binary_expression,
  [8287] = 13,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(285), 1,
      anon_sym_POUND,
    ACTIONS(287), 1,
      sym_important,
    ACTIONS(289), 1,
      anon_sym_SQUOTE,
    ACTIONS(291), 1,
      anon_sym_DQUOTE,
    ACTIONS(293), 1,
      aux_sym_integer_value_token1,
    ACTIONS(295), 1,
      aux_sym_float_value_token1,
    ACTIONS(297), 1,
      anon_sym_LBRACK2,
    ACTIONS(301), 1,
      sym_plain_value,
    ACTIONS(857), 1,
      anon_sym_LPAREN,
    ACTIONS(861), 1,
      sym_identifier,
    STATE(333), 9,
      sym__value,
      sym_parenthesized_value,
      sym_color_value,
      sym_string_value,
      sym_integer_value,
      sym_float_value,
      sym_grid_value,
      sym_call_expression,
      sym_binary_expression,
  [8335] = 13,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(285), 1,
      anon_sym_POUND,
    ACTIONS(289), 1,
      anon_sym_SQUOTE,
    ACTIONS(291), 1,
      anon_sym_DQUOTE,
    ACTIONS(293), 1,
      aux_sym_integer_value_token1,
    ACTIONS(295), 1,
      aux_sym_float_value_token1,
    ACTIONS(297), 1,
      anon_sym_LBRACK2,
    ACTIONS(857), 1,
      anon_sym_LPAREN,
    ACTIONS(861), 1,
      sym_identifier,
    ACTIONS(945), 1,
      sym_important,
    ACTIONS(947), 1,
      sym_plain_value,
    STATE(309), 9,
      sym__value,
      sym_parenthesized_value,
      sym_color_value,
      sym_string_value,
      sym_integer_value,
      sym_float_value,
      sym_grid_value,
      sym_call_expression,
      sym_binary_expression,
  [8383] = 13,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(221), 1,
      anon_sym_LPAREN,
    ACTIONS(239), 1,
      aux_sym_integer_value_token1,
    ACTIONS(241), 1,
      aux_sym_float_value_token1,
    ACTIONS(345), 1,
      anon_sym_POUND,
    ACTIONS(351), 1,
      anon_sym_SQUOTE,
    ACTIONS(353), 1,
      anon_sym_DQUOTE,
    ACTIONS(355), 1,
      anon_sym_LBRACK2,
    ACTIONS(357), 1,
      sym_identifier,
    ACTIONS(949), 1,
      sym_important,
    ACTIONS(951), 1,
      sym_plain_value,
    STATE(39), 9,
      sym__value,
      sym_parenthesized_value,
      sym_color_value,
      sym_string_value,
      sym_integer_value,
      sym_float_value,
      sym_grid_value,
      sym_call_expression,
      sym_binary_expression,
  [8431] = 13,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(285), 1,
      anon_sym_POUND,
    ACTIONS(289), 1,
      anon_sym_SQUOTE,
    ACTIONS(291), 1,
      anon_sym_DQUOTE,
    ACTIONS(293), 1,
      aux_sym_integer_value_token1,
    ACTIONS(295), 1,
      aux_sym_float_value_token1,
    ACTIONS(297), 1,
      anon_sym_LBRACK2,
    ACTIONS(857), 1,
      anon_sym_LPAREN,
    ACTIONS(861), 1,
      sym_identifier,
    ACTIONS(953), 1,
      sym_important,
    ACTIONS(955), 1,
      sym_plain_value,
    STATE(336), 9,
      sym__value,
      sym_parenthesized_value,
      sym_color_value,
      sym_string_value,
      sym_integer_value,
      sym_float_value,
      sym_grid_value,
      sym_call_expression,
      sym_binary_expression,
  [8479] = 13,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(285), 1,
      anon_sym_POUND,
    ACTIONS(289), 1,
      anon_sym_SQUOTE,
    ACTIONS(291), 1,
      anon_sym_DQUOTE,
    ACTIONS(293), 1,
      aux_sym_integer_value_token1,
    ACTIONS(295), 1,
      aux_sym_float_value_token1,
    ACTIONS(297), 1,
      anon_sym_LBRACK2,
    ACTIONS(857), 1,
      anon_sym_LPAREN,
    ACTIONS(861), 1,
      sym_identifier,
    ACTIONS(957), 1,
      sym_important,
    ACTIONS(959), 1,
      sym_plain_value,
    STATE(306), 9,
      sym__value,
      sym_parenthesized_value,
      sym_color_value,
      sym_string_value,
      sym_integer_value,
      sym_float_value,
      sym_grid_value,
      sym_call_expression,
      sym_binary_expression,
  [8527] = 13,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(285), 1,
      anon_sym_POUND,
    ACTIONS(289), 1,
      anon_sym_SQUOTE,
    ACTIONS(291), 1,
      anon_sym_DQUOTE,
    ACTIONS(293), 1,
      aux_sym_integer_value_token1,
    ACTIONS(295), 1,
      aux_sym_float_value_token1,
    ACTIONS(297), 1,
      anon_sym_LBRACK2,
    ACTIONS(309), 1,
      sym_important,
    ACTIONS(311), 1,
      sym_plain_value,
    ACTIONS(857), 1,
      anon_sym_LPAREN,
    ACTIONS(861), 1,
      sym_identifier,
    STATE(323), 9,
      sym__value,
      sym_parenthesized_value,
      sym_color_value,
      sym_string_value,
      sym_integer_value,
      sym_float_value,
      sym_grid_value,
      sym_call_expression,
      sym_binary_expression,
  [8575] = 13,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(285), 1,
      anon_sym_POUND,
    ACTIONS(289), 1,
      anon_sym_SQUOTE,
    ACTIONS(291), 1,
      anon_sym_DQUOTE,
    ACTIONS(293), 1,
      aux_sym_integer_value_token1,
    ACTIONS(295), 1,
      aux_sym_float_value_token1,
    ACTIONS(297), 1,
      anon_sym_LBRACK2,
    ACTIONS(857), 1,
      anon_sym_LPAREN,
    ACTIONS(861), 1,
      sym_identifier,
    ACTIONS(961), 1,
      sym_important,
    ACTIONS(963), 1,
      sym_plain_value,
    STATE(305), 9,
      sym__value,
      sym_parenthesized_value,
      sym_color_value,
      sym_string_value,
      sym_integer_value,
      sym_float_value,
      sym_grid_value,
      sym_call_expression,
      sym_binary_expression,
  [8623] = 13,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(909), 1,
      anon_sym_LPAREN,
    ACTIONS(911), 1,
      anon_sym_POUND,
    ACTIONS(915), 1,
      anon_sym_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      aux_sym_integer_value_token1,
    ACTIONS(921), 1,
      aux_sym_float_value_token1,
    ACTIONS(923), 1,
      anon_sym_LBRACK2,
    ACTIONS(925), 1,
      sym_identifier,
    ACTIONS(965), 1,
      sym_important,
    ACTIONS(967), 1,
      sym_plain_value,
    STATE(233), 9,
      sym__value,
      sym_parenthesized_value,
      sym_color_value,
      sym_string_value,
      sym_integer_value,
      sym_float_value,
      sym_grid_value,
      sym_call_expression,
      sym_binary_expression,
  [8671] = 13,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(221), 1,
      anon_sym_LPAREN,
    ACTIONS(239), 1,
      aux_sym_integer_value_token1,
    ACTIONS(241), 1,
      aux_sym_float_value_token1,
    ACTIONS(345), 1,
      anon_sym_POUND,
    ACTIONS(351), 1,
      anon_sym_SQUOTE,
    ACTIONS(353), 1,
      anon_sym_DQUOTE,
    ACTIONS(355), 1,
      anon_sym_LBRACK2,
    ACTIONS(357), 1,
      sym_identifier,
    ACTIONS(969), 1,
      sym_important,
    ACTIONS(971), 1,
      sym_plain_value,
    STATE(29), 9,
      sym__value,
      sym_parenthesized_value,
      sym_color_value,
      sym_string_value,
      sym_integer_value,
      sym_float_value,
      sym_grid_value,
      sym_call_expression,
      sym_binary_expression,
  [8719] = 4,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(975), 7,
      anon_sym_PLUS,
      aux_sym_integer_value_token1,
      aux_sym_float_value_token1,
      anon_sym_DASH,
      anon_sym_SLASH,
      sym_identifier,
      sym_plain_value,
    ACTIONS(973), 11,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_STAR,
      anon_sym_POUND,
      sym_important,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_LBRACK2,
  [8748] = 4,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(979), 7,
      anon_sym_PLUS,
      aux_sym_integer_value_token1,
      aux_sym_float_value_token1,
      anon_sym_DASH,
      anon_sym_SLASH,
      sym_identifier,
      sym_plain_value,
    ACTIONS(977), 11,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_STAR,
      anon_sym_POUND,
      sym_important,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_LBRACK2,
  [8777] = 4,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(983), 7,
      anon_sym_PLUS,
      aux_sym_integer_value_token1,
      aux_sym_float_value_token1,
      anon_sym_DASH,
      anon_sym_SLASH,
      sym_identifier,
      sym_plain_value,
    ACTIONS(981), 11,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_STAR,
      anon_sym_POUND,
      sym_important,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_LBRACK2,
  [8806] = 4,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(987), 7,
      anon_sym_PLUS,
      aux_sym_integer_value_token1,
      aux_sym_float_value_token1,
      anon_sym_DASH,
      anon_sym_SLASH,
      sym_identifier,
      sym_plain_value,
    ACTIONS(985), 11,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_STAR,
      anon_sym_POUND,
      sym_important,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_LBRACK2,
  [8835] = 4,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(991), 7,
      anon_sym_PLUS,
      aux_sym_integer_value_token1,
      aux_sym_float_value_token1,
      anon_sym_DASH,
      anon_sym_SLASH,
      sym_identifier,
      sym_plain_value,
    ACTIONS(989), 11,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_STAR,
      anon_sym_POUND,
      sym_important,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_LBRACK2,
  [8864] = 4,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(995), 7,
      anon_sym_PLUS,
      aux_sym_integer_value_token1,
      aux_sym_float_value_token1,
      anon_sym_DASH,
      anon_sym_SLASH,
      sym_identifier,
      sym_plain_value,
    ACTIONS(993), 11,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_STAR,
      anon_sym_POUND,
      sym_important,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_LBRACK2,
  [8893] = 4,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(999), 7,
      anon_sym_PLUS,
      aux_sym_integer_value_token1,
      aux_sym_float_value_token1,
      anon_sym_DASH,
      anon_sym_SLASH,
      sym_identifier,
      sym_plain_value,
    ACTIONS(997), 11,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_STAR,
      anon_sym_POUND,
      sym_important,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_LBRACK2,
  [8922] = 4,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1003), 7,
      anon_sym_PLUS,
      aux_sym_integer_value_token1,
      aux_sym_float_value_token1,
      anon_sym_DASH,
      anon_sym_SLASH,
      sym_identifier,
      sym_plain_value,
    ACTIONS(1001), 11,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_STAR,
      anon_sym_POUND,
      sym_important,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_LBRACK2,
  [8951] = 4,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1007), 7,
      anon_sym_PLUS,
      aux_sym_integer_value_token1,
      aux_sym_float_value_token1,
      anon_sym_DASH,
      anon_sym_SLASH,
      sym_identifier,
      sym_plain_value,
    ACTIONS(1005), 11,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_STAR,
      anon_sym_POUND,
      sym_important,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_LBRACK2,
  [8980] = 6,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(673), 1,
      anon_sym_LPAREN2,
    STATE(235), 1,
      sym_arguments,
    ACTIONS(383), 6,
      anon_sym_SEMI,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_LBRACK2,
    ACTIONS(381), 10,
      anon_sym_LPAREN,
      anon_sym_PLUS,
      sym_important,
      aux_sym_integer_value_token1,
      aux_sym_float_value_token1,
      anon_sym_DASH,
      anon_sym_SLASH,
      sym_identifier,
      sym_plain_value,
      sym_important_value,
  [9013] = 4,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(522), 7,
      anon_sym_PLUS,
      aux_sym_integer_value_token1,
      aux_sym_float_value_token1,
      anon_sym_DASH,
      anon_sym_SLASH,
      sym_identifier,
      sym_plain_value,
    ACTIONS(520), 11,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_STAR,
      anon_sym_POUND,
      sym_important,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_LBRACK2,
  [9042] = 4,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1011), 7,
      anon_sym_PLUS,
      aux_sym_integer_value_token1,
      aux_sym_float_value_token1,
      anon_sym_DASH,
      anon_sym_SLASH,
      sym_identifier,
      sym_plain_value,
    ACTIONS(1009), 11,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_STAR,
      anon_sym_POUND,
      sym_important,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_LBRACK2,
  [9071] = 4,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(526), 7,
      anon_sym_PLUS,
      aux_sym_integer_value_token1,
      aux_sym_float_value_token1,
      anon_sym_DASH,
      anon_sym_SLASH,
      sym_identifier,
      sym_plain_value,
    ACTIONS(524), 11,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_STAR,
      anon_sym_POUND,
      sym_important,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_LBRACK2,
  [9100] = 4,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1015), 7,
      anon_sym_PLUS,
      aux_sym_integer_value_token1,
      aux_sym_float_value_token1,
      anon_sym_DASH,
      anon_sym_SLASH,
      sym_identifier,
      sym_plain_value,
    ACTIONS(1013), 11,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_STAR,
      anon_sym_POUND,
      sym_important,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_LBRACK2,
  [9129] = 5,
    ACTIONS(1019), 1,
      anon_sym_RBRACK,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(385), 2,
      anon_sym_TILDE,
      anon_sym_PIPE,
    ACTIONS(1017), 6,
      anon_sym_EQ,
      anon_sym_TILDE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_STAR_EQ,
      anon_sym_DOLLAR_EQ,
    ACTIONS(379), 8,
      sym__descendant_operator,
      sym__pseudo_class_selector_colon,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_PLUS,
  [9159] = 6,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(581), 1,
      anon_sym_SEMI,
    ACTIONS(583), 1,
      anon_sym_RBRACE,
    ACTIONS(444), 7,
      anon_sym_PLUS,
      aux_sym_integer_value_token1,
      aux_sym_float_value_token1,
      anon_sym_DASH,
      anon_sym_SLASH,
      sym_identifier,
      sym_plain_value,
    ACTIONS(440), 8,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_STAR,
      anon_sym_POUND,
      sym_important,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_LBRACK2,
  [9191] = 5,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1021), 1,
      sym_unit,
    ACTIONS(881), 7,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_LBRACK2,
    ACTIONS(883), 9,
      anon_sym_PLUS,
      sym_important,
      aux_sym_integer_value_token1,
      aux_sym_float_value_token1,
      anon_sym_DASH,
      anon_sym_SLASH,
      sym_identifier,
      sym_plain_value,
      sym_important_value,
  [9221] = 6,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(343), 1,
      anon_sym_STAR,
    ACTIONS(347), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
    ACTIONS(1023), 4,
      aux_sym_integer_value_token1,
      aux_sym_float_value_token1,
      sym_identifier,
      sym_plain_value,
    ACTIONS(543), 9,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_POUND,
      sym_important,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_LBRACK2,
  [9253] = 6,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1025), 1,
      anon_sym_SEMI,
    ACTIONS(1027), 1,
      anon_sym_RBRACE,
    ACTIONS(444), 7,
      anon_sym_PLUS,
      aux_sym_integer_value_token1,
      aux_sym_float_value_token1,
      anon_sym_DASH,
      anon_sym_SLASH,
      sym_identifier,
      sym_plain_value,
    ACTIONS(440), 8,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_STAR,
      anon_sym_POUND,
      sym_important,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_LBRACK2,
  [9285] = 5,
    ACTIONS(1031), 1,
      anon_sym_RBRACK,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(385), 2,
      anon_sym_TILDE,
      anon_sym_PIPE,
    ACTIONS(1029), 6,
      anon_sym_EQ,
      anon_sym_TILDE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_STAR_EQ,
      anon_sym_DOLLAR_EQ,
    ACTIONS(379), 8,
      sym__descendant_operator,
      sym__pseudo_class_selector_colon,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_PLUS,
  [9315] = 5,
    ACTIONS(1031), 1,
      anon_sym_RBRACK,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(442), 2,
      anon_sym_TILDE,
      anon_sym_PIPE,
    ACTIONS(1029), 6,
      anon_sym_EQ,
      anon_sym_TILDE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_STAR_EQ,
      anon_sym_DOLLAR_EQ,
    ACTIONS(438), 8,
      sym__descendant_operator,
      sym__pseudo_class_selector_colon,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_PLUS,
  [9345] = 6,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(343), 1,
      anon_sym_STAR,
    ACTIONS(347), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
    ACTIONS(1035), 4,
      aux_sym_integer_value_token1,
      aux_sym_float_value_token1,
      sym_identifier,
      sym_plain_value,
    ACTIONS(1033), 9,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_POUND,
      sym_important,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_LBRACK2,
  [9377] = 5,
    ACTIONS(1019), 1,
      anon_sym_RBRACK,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(442), 2,
      anon_sym_TILDE,
      anon_sym_PIPE,
    ACTIONS(1017), 6,
      anon_sym_EQ,
      anon_sym_TILDE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_STAR_EQ,
      anon_sym_DOLLAR_EQ,
    ACTIONS(438), 8,
      sym__descendant_operator,
      sym__pseudo_class_selector_colon,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_PLUS,
  [9407] = 5,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1037), 1,
      sym_unit,
    ACTIONS(887), 7,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_LBRACK2,
    ACTIONS(889), 9,
      anon_sym_PLUS,
      sym_important,
      aux_sym_integer_value_token1,
      aux_sym_float_value_token1,
      anon_sym_DASH,
      anon_sym_SLASH,
      sym_identifier,
      sym_plain_value,
      sym_important_value,
  [9437] = 6,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(343), 1,
      anon_sym_STAR,
    ACTIONS(347), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
    ACTIONS(1041), 4,
      aux_sym_integer_value_token1,
      aux_sym_float_value_token1,
      sym_identifier,
      sym_plain_value,
    ACTIONS(1039), 9,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_POUND,
      sym_important,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_LBRACK2,
  [9469] = 4,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1009), 7,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_LBRACK2,
    ACTIONS(1011), 9,
      anon_sym_PLUS,
      sym_important,
      aux_sym_integer_value_token1,
      aux_sym_float_value_token1,
      anon_sym_DASH,
      anon_sym_SLASH,
      sym_identifier,
      sym_plain_value,
      sym_important_value,
  [9496] = 4,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(520), 7,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_LBRACK2,
    ACTIONS(522), 9,
      anon_sym_PLUS,
      sym_important,
      aux_sym_integer_value_token1,
      aux_sym_float_value_token1,
      anon_sym_DASH,
      anon_sym_SLASH,
      sym_identifier,
      sym_plain_value,
      sym_important_value,
  [9523] = 4,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(524), 7,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_LBRACK2,
    ACTIONS(526), 9,
      anon_sym_PLUS,
      sym_important,
      aux_sym_integer_value_token1,
      aux_sym_float_value_token1,
      anon_sym_DASH,
      anon_sym_SLASH,
      sym_identifier,
      sym_plain_value,
      sym_important_value,
  [9550] = 10,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(267), 1,
      anon_sym_selector,
    ACTIONS(1043), 1,
      anon_sym_SEMI,
    ACTIONS(1045), 1,
      anon_sym_LPAREN,
    ACTIONS(1051), 1,
      sym_identifier,
    ACTIONS(261), 2,
      anon_sym_not,
      anon_sym_only,
    ACTIONS(1047), 2,
      anon_sym_STAR,
      anon_sym_PLUS,
    ACTIONS(1049), 2,
      anon_sym_DASH,
      anon_sym_SLASH,
    STATE(327), 6,
      sym__query,
      sym_feature_query,
      sym_parenthesized_query,
      sym_binary_query,
      sym_unary_query,
      sym_selector_query,
  [9589] = 4,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(985), 7,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_LBRACK2,
    ACTIONS(987), 9,
      anon_sym_PLUS,
      sym_important,
      aux_sym_integer_value_token1,
      aux_sym_float_value_token1,
      anon_sym_DASH,
      anon_sym_SLASH,
      sym_identifier,
      sym_plain_value,
      sym_important_value,
  [9616] = 4,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1013), 7,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_LBRACK2,
    ACTIONS(1015), 9,
      anon_sym_PLUS,
      sym_important,
      aux_sym_integer_value_token1,
      aux_sym_float_value_token1,
      anon_sym_DASH,
      anon_sym_SLASH,
      sym_identifier,
      sym_plain_value,
      sym_important_value,
  [9643] = 4,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(981), 7,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_LBRACK2,
    ACTIONS(983), 9,
      anon_sym_PLUS,
      sym_important,
      aux_sym_integer_value_token1,
      aux_sym_float_value_token1,
      anon_sym_DASH,
      anon_sym_SLASH,
      sym_identifier,
      sym_plain_value,
      sym_important_value,
  [9670] = 5,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(581), 1,
      anon_sym_SEMI,
    ACTIONS(444), 7,
      anon_sym_PLUS,
      aux_sym_integer_value_token1,
      aux_sym_float_value_token1,
      anon_sym_DASH,
      anon_sym_SLASH,
      sym_identifier,
      sym_plain_value,
    ACTIONS(440), 8,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_STAR,
      anon_sym_POUND,
      sym_important,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_LBRACK2,
  [9699] = 5,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1025), 1,
      anon_sym_SEMI,
    ACTIONS(444), 7,
      anon_sym_PLUS,
      aux_sym_integer_value_token1,
      aux_sym_float_value_token1,
      anon_sym_DASH,
      anon_sym_SLASH,
      sym_identifier,
      sym_plain_value,
    ACTIONS(440), 8,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_STAR,
      anon_sym_POUND,
      sym_important,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_LBRACK2,
  [9728] = 4,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(989), 7,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_LBRACK2,
    ACTIONS(991), 9,
      anon_sym_PLUS,
      sym_important,
      aux_sym_integer_value_token1,
      aux_sym_float_value_token1,
      anon_sym_DASH,
      anon_sym_SLASH,
      sym_identifier,
      sym_plain_value,
      sym_important_value,
  [9755] = 5,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1053), 1,
      anon_sym_SEMI,
    ACTIONS(444), 7,
      anon_sym_PLUS,
      aux_sym_integer_value_token1,
      aux_sym_float_value_token1,
      anon_sym_DASH,
      anon_sym_SLASH,
      sym_identifier,
      sym_plain_value,
    ACTIONS(440), 8,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_STAR,
      anon_sym_POUND,
      sym_important,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_LBRACK2,
  [9784] = 6,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1057), 1,
      anon_sym_STAR,
    ACTIONS(1059), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
    ACTIONS(1055), 6,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_POUND,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_LBRACK2,
    ACTIONS(1061), 6,
      sym_important,
      aux_sym_integer_value_token1,
      aux_sym_float_value_token1,
      sym_identifier,
      sym_plain_value,
      sym_important_value,
  [9815] = 5,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(695), 1,
      anon_sym_SEMI,
    ACTIONS(444), 7,
      anon_sym_PLUS,
      aux_sym_integer_value_token1,
      aux_sym_float_value_token1,
      anon_sym_DASH,
      anon_sym_SLASH,
      sym_identifier,
      sym_plain_value,
    ACTIONS(440), 8,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_STAR,
      anon_sym_POUND,
      sym_important,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_LBRACK2,
  [9844] = 4,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(993), 7,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_LBRACK2,
    ACTIONS(995), 9,
      anon_sym_PLUS,
      sym_important,
      aux_sym_integer_value_token1,
      aux_sym_float_value_token1,
      anon_sym_DASH,
      anon_sym_SLASH,
      sym_identifier,
      sym_plain_value,
      sym_important_value,
  [9871] = 4,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(997), 7,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_LBRACK2,
    ACTIONS(999), 9,
      anon_sym_PLUS,
      sym_important,
      aux_sym_integer_value_token1,
      aux_sym_float_value_token1,
      anon_sym_DASH,
      anon_sym_SLASH,
      sym_identifier,
      sym_plain_value,
      sym_important_value,
  [9898] = 10,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(267), 1,
      anon_sym_selector,
    ACTIONS(1045), 1,
      anon_sym_LPAREN,
    ACTIONS(1051), 1,
      sym_identifier,
    ACTIONS(1063), 1,
      anon_sym_SEMI,
    ACTIONS(261), 2,
      anon_sym_not,
      anon_sym_only,
    ACTIONS(1047), 2,
      anon_sym_STAR,
      anon_sym_PLUS,
    ACTIONS(1049), 2,
      anon_sym_DASH,
      anon_sym_SLASH,
    STATE(330), 6,
      sym__query,
      sym_feature_query,
      sym_parenthesized_query,
      sym_binary_query,
      sym_unary_query,
      sym_selector_query,
  [9937] = 4,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1001), 7,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_LBRACK2,
    ACTIONS(1003), 9,
      anon_sym_PLUS,
      sym_important,
      aux_sym_integer_value_token1,
      aux_sym_float_value_token1,
      anon_sym_DASH,
      anon_sym_SLASH,
      sym_identifier,
      sym_plain_value,
      sym_important_value,
  [9964] = 4,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(973), 7,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_LBRACK2,
    ACTIONS(975), 9,
      anon_sym_PLUS,
      sym_important,
      aux_sym_integer_value_token1,
      aux_sym_float_value_token1,
      anon_sym_DASH,
      anon_sym_SLASH,
      sym_identifier,
      sym_plain_value,
      sym_important_value,
  [9991] = 4,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1005), 7,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_LBRACK2,
    ACTIONS(1007), 9,
      anon_sym_PLUS,
      sym_important,
      aux_sym_integer_value_token1,
      aux_sym_float_value_token1,
      anon_sym_DASH,
      anon_sym_SLASH,
      sym_identifier,
      sym_plain_value,
      sym_important_value,
  [10018] = 4,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(977), 7,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_LBRACK2,
    ACTIONS(979), 9,
      anon_sym_PLUS,
      sym_important,
      aux_sym_integer_value_token1,
      aux_sym_float_value_token1,
      anon_sym_DASH,
      anon_sym_SLASH,
      sym_identifier,
      sym_plain_value,
      sym_important_value,
  [10045] = 9,
    ACTIONS(267), 1,
      anon_sym_selector,
    ACTIONS(1045), 1,
      anon_sym_LPAREN,
    ACTIONS(1051), 1,
      sym_identifier,
    ACTIONS(1065), 1,
      anon_sym_SEMI,
    ACTIONS(1067), 1,
      anon_sym_LBRACE,
    STATE(47), 1,
      sym_block,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(261), 2,
      anon_sym_not,
      anon_sym_only,
    STATE(310), 6,
      sym__query,
      sym_feature_query,
      sym_parenthesized_query,
      sym_binary_query,
      sym_unary_query,
      sym_selector_query,
  [10080] = 3,
    ACTIONS(1069), 1,
      anon_sym_COLON,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(379), 12,
      sym__descendant_operator,
      sym__pseudo_class_selector_colon,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_TILDE,
      anon_sym_PLUS,
      anon_sym_PIPE,
  [10102] = 14,
    ACTIONS(725), 1,
      anon_sym_DOT,
    ACTIONS(1071), 1,
      anon_sym_COMMA,
    ACTIONS(1073), 1,
      anon_sym_RPAREN,
    ACTIONS(1075), 1,
      anon_sym_COLON_COLON,
    ACTIONS(1077), 1,
      anon_sym_POUND,
    ACTIONS(1079), 1,
      anon_sym_LBRACK,
    ACTIONS(1081), 1,
      anon_sym_GT,
    ACTIONS(1083), 1,
      anon_sym_TILDE,
    ACTIONS(1085), 1,
      anon_sym_PLUS,
    ACTIONS(1087), 1,
      anon_sym_PIPE,
    ACTIONS(1089), 1,
      sym__descendant_operator,
    ACTIONS(1091), 1,
      sym__pseudo_class_selector_colon,
    STATE(398), 1,
      aux_sym_selectors_repeat1,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
  [10146] = 14,
    ACTIONS(575), 1,
      anon_sym_COMMA,
    ACTIONS(587), 1,
      anon_sym_RPAREN,
    ACTIONS(725), 1,
      anon_sym_DOT,
    ACTIONS(1075), 1,
      anon_sym_COLON_COLON,
    ACTIONS(1077), 1,
      anon_sym_POUND,
    ACTIONS(1079), 1,
      anon_sym_LBRACK,
    ACTIONS(1081), 1,
      anon_sym_GT,
    ACTIONS(1083), 1,
      anon_sym_TILDE,
    ACTIONS(1085), 1,
      anon_sym_PLUS,
    ACTIONS(1087), 1,
      anon_sym_PIPE,
    ACTIONS(1089), 1,
      sym__descendant_operator,
    ACTIONS(1091), 1,
      sym__pseudo_class_selector_colon,
    STATE(384), 1,
      aux_sym_pseudo_class_arguments_repeat2,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
  [10190] = 3,
    ACTIONS(1093), 1,
      anon_sym_COLON,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(379), 12,
      sym__descendant_operator,
      sym__pseudo_class_selector_colon,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_TILDE,
      anon_sym_PLUS,
      anon_sym_PIPE,
  [10212] = 14,
    ACTIONS(725), 1,
      anon_sym_DOT,
    ACTIONS(1071), 1,
      anon_sym_COMMA,
    ACTIONS(1075), 1,
      anon_sym_COLON_COLON,
    ACTIONS(1077), 1,
      anon_sym_POUND,
    ACTIONS(1079), 1,
      anon_sym_LBRACK,
    ACTIONS(1081), 1,
      anon_sym_GT,
    ACTIONS(1083), 1,
      anon_sym_TILDE,
    ACTIONS(1085), 1,
      anon_sym_PLUS,
    ACTIONS(1087), 1,
      anon_sym_PIPE,
    ACTIONS(1089), 1,
      sym__descendant_operator,
    ACTIONS(1091), 1,
      sym__pseudo_class_selector_colon,
    ACTIONS(1095), 1,
      anon_sym_LBRACE,
    STATE(403), 1,
      aux_sym_selectors_repeat1,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
  [10256] = 3,
    ACTIONS(1097), 1,
      anon_sym_COLON,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(379), 12,
      sym__descendant_operator,
      sym__pseudo_class_selector_colon,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_DOT,
      anon_sym_COLON_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_TILDE,
      anon_sym_PLUS,
      anon_sym_PIPE,
  [10278] = 12,
    ACTIONS(725), 1,
      anon_sym_DOT,
    ACTIONS(1075), 1,
      anon_sym_COLON_COLON,
    ACTIONS(1077), 1,
      anon_sym_POUND,
    ACTIONS(1079), 1,
      anon_sym_LBRACK,
    ACTIONS(1081), 1,
      anon_sym_GT,
    ACTIONS(1083), 1,
      anon_sym_TILDE,
    ACTIONS(1085), 1,
      anon_sym_PLUS,
    ACTIONS(1087), 1,
      anon_sym_PIPE,
    ACTIONS(1089), 1,
      sym__descendant_operator,
    ACTIONS(1091), 1,
      sym__pseudo_class_selector_colon,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(1099), 3,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_RPAREN,
  [10318] = 14,
    ACTIONS(575), 1,
      anon_sym_COMMA,
    ACTIONS(577), 1,
      anon_sym_RPAREN,
    ACTIONS(725), 1,
      anon_sym_DOT,
    ACTIONS(1075), 1,
      anon_sym_COLON_COLON,
    ACTIONS(1077), 1,
      anon_sym_POUND,
    ACTIONS(1079), 1,
      anon_sym_LBRACK,
    ACTIONS(1081), 1,
      anon_sym_GT,
    ACTIONS(1083), 1,
      anon_sym_TILDE,
    ACTIONS(1085), 1,
      anon_sym_PLUS,
    ACTIONS(1087), 1,
      anon_sym_PIPE,
    ACTIONS(1089), 1,
      sym__descendant_operator,
    ACTIONS(1091), 1,
      sym__pseudo_class_selector_colon,
    STATE(401), 1,
      aux_sym_pseudo_class_arguments_repeat2,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
  [10362] = 8,
    ACTIONS(1103), 1,
      anon_sym_host,
    ACTIONS(1107), 1,
      sym_escape_sequence,
    ACTIONS(1109), 1,
      sym_identifier,
    STATE(114), 1,
      sym_class_name,
    STATE(125), 1,
      sym__nth_child_pseudo_class_selector,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(1105), 2,
      anon_sym_nth_DASHchild,
      anon_sym_nth_DASHlast_DASHchild,
    ACTIONS(1101), 5,
      anon_sym_has,
      anon_sym_not,
      anon_sym_is,
      anon_sym_where,
      anon_sym_host_DASHcontext,
  [10393] = 4,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1009), 6,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_LPAREN,
      anon_sym_STAR,
      anon_sym_PLUS,
    ACTIONS(1011), 6,
      anon_sym_not,
      anon_sym_only,
      anon_sym_selector,
      anon_sym_DASH,
      anon_sym_SLASH,
      sym_identifier,
  [10416] = 6,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1111), 1,
      anon_sym_LPAREN2,
    STATE(270), 1,
      sym_arguments,
    ACTIONS(383), 3,
      anon_sym_SEMI,
      anon_sym_STAR,
      anon_sym_PLUS,
    ACTIONS(381), 7,
      anon_sym_LPAREN,
      anon_sym_not,
      anon_sym_only,
      anon_sym_selector,
      anon_sym_DASH,
      anon_sym_SLASH,
      sym_identifier,
  [10443] = 8,
    ACTIONS(1107), 1,
      sym_escape_sequence,
    ACTIONS(1109), 1,
      sym_identifier,
    ACTIONS(1115), 1,
      anon_sym_host,
    STATE(110), 1,
      sym_class_name,
    STATE(147), 1,
      sym__nth_child_pseudo_class_selector,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(1105), 2,
      anon_sym_nth_DASHchild,
      anon_sym_nth_DASHlast_DASHchild,
    ACTIONS(1113), 5,
      anon_sym_has,
      anon_sym_not,
      anon_sym_is,
      anon_sym_where,
      anon_sym_host_DASHcontext,
  [10474] = 12,
    ACTIONS(725), 1,
      anon_sym_DOT,
    ACTIONS(1075), 1,
      anon_sym_COLON_COLON,
    ACTIONS(1077), 1,
      anon_sym_POUND,
    ACTIONS(1079), 1,
      anon_sym_LBRACK,
    ACTIONS(1081), 1,
      anon_sym_GT,
    ACTIONS(1083), 1,
      anon_sym_TILDE,
    ACTIONS(1085), 1,
      anon_sym_PLUS,
    ACTIONS(1087), 1,
      anon_sym_PIPE,
    ACTIONS(1089), 1,
      sym__descendant_operator,
    ACTIONS(1091), 1,
      sym__pseudo_class_selector_colon,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(621), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [10513] = 12,
    ACTIONS(725), 1,
      anon_sym_DOT,
    ACTIONS(1075), 1,
      anon_sym_COLON_COLON,
    ACTIONS(1077), 1,
      anon_sym_POUND,
    ACTIONS(1079), 1,
      anon_sym_LBRACK,
    ACTIONS(1081), 1,
      anon_sym_GT,
    ACTIONS(1083), 1,
      anon_sym_TILDE,
    ACTIONS(1085), 1,
      anon_sym_PLUS,
    ACTIONS(1087), 1,
      anon_sym_PIPE,
    ACTIONS(1089), 1,
      sym__descendant_operator,
    ACTIONS(1091), 1,
      sym__pseudo_class_selector_colon,
    ACTIONS(1117), 1,
      anon_sym_RPAREN,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
  [10551] = 12,
    ACTIONS(725), 1,
      anon_sym_DOT,
    ACTIONS(1075), 1,
      anon_sym_COLON_COLON,
    ACTIONS(1077), 1,
      anon_sym_POUND,
    ACTIONS(1079), 1,
      anon_sym_LBRACK,
    ACTIONS(1081), 1,
      anon_sym_GT,
    ACTIONS(1083), 1,
      anon_sym_TILDE,
    ACTIONS(1085), 1,
      anon_sym_PLUS,
    ACTIONS(1087), 1,
      anon_sym_PIPE,
    ACTIONS(1089), 1,
      sym__descendant_operator,
    ACTIONS(1091), 1,
      sym__pseudo_class_selector_colon,
    ACTIONS(1119), 1,
      anon_sym_RPAREN,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
  [10589] = 6,
    ACTIONS(267), 1,
      anon_sym_selector,
    ACTIONS(1045), 1,
      anon_sym_LPAREN,
    ACTIONS(1051), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(261), 2,
      anon_sym_not,
      anon_sym_only,
    STATE(354), 6,
      sym__query,
      sym_feature_query,
      sym_parenthesized_query,
      sym_binary_query,
      sym_unary_query,
      sym_selector_query,
  [10615] = 12,
    ACTIONS(725), 1,
      anon_sym_DOT,
    ACTIONS(1075), 1,
      anon_sym_COLON_COLON,
    ACTIONS(1077), 1,
      anon_sym_POUND,
    ACTIONS(1079), 1,
      anon_sym_LBRACK,
    ACTIONS(1081), 1,
      anon_sym_GT,
    ACTIONS(1083), 1,
      anon_sym_TILDE,
    ACTIONS(1085), 1,
      anon_sym_PLUS,
    ACTIONS(1087), 1,
      anon_sym_PIPE,
    ACTIONS(1089), 1,
      sym__descendant_operator,
    ACTIONS(1091), 1,
      sym__pseudo_class_selector_colon,
    ACTIONS(1121), 1,
      anon_sym_RPAREN,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
  [10653] = 6,
    ACTIONS(267), 1,
      anon_sym_selector,
    ACTIONS(1045), 1,
      anon_sym_LPAREN,
    ACTIONS(1051), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(261), 2,
      anon_sym_not,
      anon_sym_only,
    STATE(334), 6,
      sym__query,
      sym_feature_query,
      sym_parenthesized_query,
      sym_binary_query,
      sym_unary_query,
      sym_selector_query,
  [10679] = 6,
    ACTIONS(267), 1,
      anon_sym_selector,
    ACTIONS(1045), 1,
      anon_sym_LPAREN,
    ACTIONS(1051), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(261), 2,
      anon_sym_not,
      anon_sym_only,
    STATE(311), 6,
      sym__query,
      sym_feature_query,
      sym_parenthesized_query,
      sym_binary_query,
      sym_unary_query,
      sym_selector_query,
  [10705] = 6,
    ACTIONS(267), 1,
      anon_sym_selector,
    ACTIONS(1045), 1,
      anon_sym_LPAREN,
    ACTIONS(1051), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(261), 2,
      anon_sym_not,
      anon_sym_only,
    STATE(317), 6,
      sym__query,
      sym_feature_query,
      sym_parenthesized_query,
      sym_binary_query,
      sym_unary_query,
      sym_selector_query,
  [10731] = 5,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1123), 1,
      sym_unit,
    ACTIONS(881), 4,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_STAR,
      anon_sym_PLUS,
    ACTIONS(883), 6,
      anon_sym_not,
      anon_sym_only,
      anon_sym_selector,
      anon_sym_DASH,
      anon_sym_SLASH,
      sym_identifier,
  [10755] = 5,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1125), 1,
      sym_unit,
    ACTIONS(887), 4,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_STAR,
      anon_sym_PLUS,
    ACTIONS(889), 6,
      anon_sym_not,
      anon_sym_only,
      anon_sym_selector,
      anon_sym_DASH,
      anon_sym_SLASH,
      sym_identifier,
  [10779] = 12,
    ACTIONS(725), 1,
      anon_sym_DOT,
    ACTIONS(1075), 1,
      anon_sym_COLON_COLON,
    ACTIONS(1077), 1,
      anon_sym_POUND,
    ACTIONS(1079), 1,
      anon_sym_LBRACK,
    ACTIONS(1081), 1,
      anon_sym_GT,
    ACTIONS(1083), 1,
      anon_sym_TILDE,
    ACTIONS(1085), 1,
      anon_sym_PLUS,
    ACTIONS(1087), 1,
      anon_sym_PIPE,
    ACTIONS(1089), 1,
      sym__descendant_operator,
    ACTIONS(1091), 1,
      sym__pseudo_class_selector_colon,
    ACTIONS(1127), 1,
      anon_sym_RPAREN,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
  [10817] = 6,
    ACTIONS(267), 1,
      anon_sym_selector,
    ACTIONS(1045), 1,
      anon_sym_LPAREN,
    ACTIONS(1129), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(261), 2,
      anon_sym_not,
      anon_sym_only,
    STATE(382), 6,
      sym__query,
      sym_feature_query,
      sym_parenthesized_query,
      sym_binary_query,
      sym_unary_query,
      sym_selector_query,
  [10843] = 6,
    ACTIONS(267), 1,
      anon_sym_selector,
    ACTIONS(1045), 1,
      anon_sym_LPAREN,
    ACTIONS(1051), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(261), 2,
      anon_sym_not,
      anon_sym_only,
    STATE(314), 6,
      sym__query,
      sym_feature_query,
      sym_parenthesized_query,
      sym_binary_query,
      sym_unary_query,
      sym_selector_query,
  [10869] = 6,
    ACTIONS(267), 1,
      anon_sym_selector,
    ACTIONS(1045), 1,
      anon_sym_LPAREN,
    ACTIONS(1051), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(261), 2,
      anon_sym_not,
      anon_sym_only,
    STATE(313), 6,
      sym__query,
      sym_feature_query,
      sym_parenthesized_query,
      sym_binary_query,
      sym_unary_query,
      sym_selector_query,
  [10895] = 6,
    ACTIONS(267), 1,
      anon_sym_selector,
    ACTIONS(1045), 1,
      anon_sym_LPAREN,
    ACTIONS(1051), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(261), 2,
      anon_sym_not,
      anon_sym_only,
    STATE(346), 6,
      sym__query,
      sym_feature_query,
      sym_parenthesized_query,
      sym_binary_query,
      sym_unary_query,
      sym_selector_query,
  [10921] = 12,
    ACTIONS(725), 1,
      anon_sym_DOT,
    ACTIONS(1075), 1,
      anon_sym_COLON_COLON,
    ACTIONS(1077), 1,
      anon_sym_POUND,
    ACTIONS(1079), 1,
      anon_sym_LBRACK,
    ACTIONS(1081), 1,
      anon_sym_GT,
    ACTIONS(1083), 1,
      anon_sym_TILDE,
    ACTIONS(1085), 1,
      anon_sym_PLUS,
    ACTIONS(1087), 1,
      anon_sym_PIPE,
    ACTIONS(1089), 1,
      sym__descendant_operator,
    ACTIONS(1091), 1,
      sym__pseudo_class_selector_colon,
    ACTIONS(1131), 1,
      anon_sym_RPAREN,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
  [10959] = 12,
    ACTIONS(725), 1,
      anon_sym_DOT,
    ACTIONS(1075), 1,
      anon_sym_COLON_COLON,
    ACTIONS(1077), 1,
      anon_sym_POUND,
    ACTIONS(1079), 1,
      anon_sym_LBRACK,
    ACTIONS(1081), 1,
      anon_sym_GT,
    ACTIONS(1083), 1,
      anon_sym_TILDE,
    ACTIONS(1085), 1,
      anon_sym_PLUS,
    ACTIONS(1087), 1,
      anon_sym_PIPE,
    ACTIONS(1089), 1,
      sym__descendant_operator,
    ACTIONS(1091), 1,
      sym__pseudo_class_selector_colon,
    ACTIONS(1133), 1,
      anon_sym_RPAREN,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
  [10997] = 4,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(981), 4,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_STAR,
      anon_sym_PLUS,
    ACTIONS(983), 6,
      anon_sym_not,
      anon_sym_only,
      anon_sym_selector,
      anon_sym_DASH,
      anon_sym_SLASH,
      sym_identifier,
  [11018] = 4,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1005), 4,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_STAR,
      anon_sym_PLUS,
    ACTIONS(1007), 6,
      anon_sym_not,
      anon_sym_only,
      anon_sym_selector,
      anon_sym_DASH,
      anon_sym_SLASH,
      sym_identifier,
  [11039] = 4,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(973), 4,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_STAR,
      anon_sym_PLUS,
    ACTIONS(975), 6,
      anon_sym_not,
      anon_sym_only,
      anon_sym_selector,
      anon_sym_DASH,
      anon_sym_SLASH,
      sym_identifier,
  [11060] = 4,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(524), 4,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_STAR,
      anon_sym_PLUS,
    ACTIONS(526), 6,
      anon_sym_not,
      anon_sym_only,
      anon_sym_selector,
      anon_sym_DASH,
      anon_sym_SLASH,
      sym_identifier,
  [11081] = 11,
    ACTIONS(725), 1,
      anon_sym_DOT,
    ACTIONS(1075), 1,
      anon_sym_COLON_COLON,
    ACTIONS(1077), 1,
      anon_sym_POUND,
    ACTIONS(1079), 1,
      anon_sym_LBRACK,
    ACTIONS(1081), 1,
      anon_sym_GT,
    ACTIONS(1083), 1,
      anon_sym_TILDE,
    ACTIONS(1085), 1,
      anon_sym_PLUS,
    ACTIONS(1087), 1,
      anon_sym_PIPE,
    ACTIONS(1089), 1,
      sym__descendant_operator,
    ACTIONS(1091), 1,
      sym__pseudo_class_selector_colon,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
  [11116] = 4,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(985), 4,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_STAR,
      anon_sym_PLUS,
    ACTIONS(987), 6,
      anon_sym_not,
      anon_sym_only,
      anon_sym_selector,
      anon_sym_DASH,
      anon_sym_SLASH,
      sym_identifier,
  [11137] = 4,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(989), 4,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_STAR,
      anon_sym_PLUS,
    ACTIONS(991), 6,
      anon_sym_not,
      anon_sym_only,
      anon_sym_selector,
      anon_sym_DASH,
      anon_sym_SLASH,
      sym_identifier,
  [11158] = 6,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(381), 1,
      anon_sym_SLASH,
    ACTIONS(1135), 1,
      anon_sym_LPAREN2,
    STATE(298), 1,
      sym_arguments,
    ACTIONS(383), 7,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_STAR,
      anon_sym_RBRACK,
      anon_sym_PLUS,
      anon_sym_DASH,
  [11183] = 4,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(997), 4,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_STAR,
      anon_sym_PLUS,
    ACTIONS(999), 6,
      anon_sym_not,
      anon_sym_only,
      anon_sym_selector,
      anon_sym_DASH,
      anon_sym_SLASH,
      sym_identifier,
  [11204] = 4,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1001), 4,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_STAR,
      anon_sym_PLUS,
    ACTIONS(1003), 6,
      anon_sym_not,
      anon_sym_only,
      anon_sym_selector,
      anon_sym_DASH,
      anon_sym_SLASH,
      sym_identifier,
  [11225] = 8,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(381), 1,
      anon_sym_SLASH,
    ACTIONS(1135), 1,
      anon_sym_LPAREN2,
    ACTIONS(1137), 1,
      anon_sym_COLON,
    STATE(298), 1,
      sym_arguments,
    ACTIONS(383), 3,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(671), 3,
      anon_sym_RPAREN,
      anon_sym_and,
      anon_sym_or,
  [11254] = 4,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1013), 4,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_STAR,
      anon_sym_PLUS,
    ACTIONS(1015), 6,
      anon_sym_not,
      anon_sym_only,
      anon_sym_selector,
      anon_sym_DASH,
      anon_sym_SLASH,
      sym_identifier,
  [11275] = 4,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(520), 4,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_STAR,
      anon_sym_PLUS,
    ACTIONS(522), 6,
      anon_sym_not,
      anon_sym_only,
      anon_sym_selector,
      anon_sym_DASH,
      anon_sym_SLASH,
      sym_identifier,
  [11296] = 4,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(993), 4,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_STAR,
      anon_sym_PLUS,
    ACTIONS(995), 6,
      anon_sym_not,
      anon_sym_only,
      anon_sym_selector,
      anon_sym_DASH,
      anon_sym_SLASH,
      sym_identifier,
  [11317] = 4,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(977), 4,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_STAR,
      anon_sym_PLUS,
    ACTIONS(979), 6,
      anon_sym_not,
      anon_sym_only,
      anon_sym_selector,
      anon_sym_DASH,
      anon_sym_SLASH,
      sym_identifier,
  [11338] = 5,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(889), 1,
      anon_sym_SLASH,
    ACTIONS(1139), 1,
      sym_unit,
    ACTIONS(887), 7,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_STAR,
      anon_sym_RBRACK,
      anon_sym_PLUS,
      anon_sym_DASH,
  [11360] = 5,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(883), 1,
      anon_sym_SLASH,
    ACTIONS(1141), 1,
      sym_unit,
    ACTIONS(881), 7,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_STAR,
      anon_sym_RBRACK,
      anon_sym_PLUS,
      anon_sym_DASH,
  [11382] = 4,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1011), 1,
      anon_sym_SLASH,
    ACTIONS(1009), 8,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_STAR,
      anon_sym_RBRACK,
      anon_sym_PLUS,
      anon_sym_of,
      anon_sym_DASH,
  [11402] = 7,
    ACTIONS(1143), 1,
      anon_sym_RBRACE,
    ACTIONS(1145), 1,
      sym_from,
    ACTIONS(1148), 1,
      anon_sym_to,
    ACTIONS(1151), 1,
      aux_sym_integer_value_token1,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    STATE(286), 2,
      sym_keyframe_block,
      aux_sym_keyframe_block_list_repeat1,
    STATE(341), 2,
      sym_to,
      sym_integer_value,
  [11427] = 4,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1007), 1,
      anon_sym_SLASH,
    ACTIONS(1005), 7,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_STAR,
      anon_sym_RBRACK,
      anon_sym_PLUS,
      anon_sym_DASH,
  [11446] = 7,
    ACTIONS(1154), 1,
      anon_sym_RBRACE,
    ACTIONS(1156), 1,
      sym_from,
    ACTIONS(1158), 1,
      anon_sym_to,
    ACTIONS(1160), 1,
      aux_sym_integer_value_token1,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    STATE(295), 2,
      sym_keyframe_block,
      aux_sym_keyframe_block_list_repeat1,
    STATE(341), 2,
      sym_to,
      sym_integer_value,
  [11471] = 4,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(522), 1,
      anon_sym_SLASH,
    ACTIONS(520), 7,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_STAR,
      anon_sym_RBRACK,
      anon_sym_PLUS,
      anon_sym_DASH,
  [11490] = 4,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(526), 1,
      anon_sym_SLASH,
    ACTIONS(524), 7,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_STAR,
      anon_sym_RBRACK,
      anon_sym_PLUS,
      anon_sym_DASH,
  [11509] = 4,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(983), 1,
      anon_sym_SLASH,
    ACTIONS(981), 7,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_STAR,
      anon_sym_RBRACK,
      anon_sym_PLUS,
      anon_sym_DASH,
  [11528] = 4,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(987), 1,
      anon_sym_SLASH,
    ACTIONS(985), 7,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_STAR,
      anon_sym_RBRACK,
      anon_sym_PLUS,
      anon_sym_DASH,
  [11547] = 4,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(991), 1,
      anon_sym_SLASH,
    ACTIONS(989), 7,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_STAR,
      anon_sym_RBRACK,
      anon_sym_PLUS,
      anon_sym_DASH,
  [11566] = 4,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(995), 1,
      anon_sym_SLASH,
    ACTIONS(993), 7,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_STAR,
      anon_sym_RBRACK,
      anon_sym_PLUS,
      anon_sym_DASH,
  [11585] = 7,
    ACTIONS(1156), 1,
      sym_from,
    ACTIONS(1158), 1,
      anon_sym_to,
    ACTIONS(1160), 1,
      aux_sym_integer_value_token1,
    ACTIONS(1162), 1,
      anon_sym_RBRACE,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    STATE(286), 2,
      sym_keyframe_block,
      aux_sym_keyframe_block_list_repeat1,
    STATE(341), 2,
      sym_to,
      sym_integer_value,
  [11610] = 7,
    ACTIONS(1156), 1,
      sym_from,
    ACTIONS(1158), 1,
      anon_sym_to,
    ACTIONS(1160), 1,
      aux_sym_integer_value_token1,
    ACTIONS(1164), 1,
      anon_sym_RBRACE,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    STATE(297), 2,
      sym_keyframe_block,
      aux_sym_keyframe_block_list_repeat1,
    STATE(341), 2,
      sym_to,
      sym_integer_value,
  [11635] = 7,
    ACTIONS(1156), 1,
      sym_from,
    ACTIONS(1158), 1,
      anon_sym_to,
    ACTIONS(1160), 1,
      aux_sym_integer_value_token1,
    ACTIONS(1166), 1,
      anon_sym_RBRACE,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    STATE(286), 2,
      sym_keyframe_block,
      aux_sym_keyframe_block_list_repeat1,
    STATE(341), 2,
      sym_to,
      sym_integer_value,
  [11660] = 4,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(975), 1,
      anon_sym_SLASH,
    ACTIONS(973), 7,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_STAR,
      anon_sym_RBRACK,
      anon_sym_PLUS,
      anon_sym_DASH,
  [11679] = 4,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(979), 1,
      anon_sym_SLASH,
    ACTIONS(977), 7,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_STAR,
      anon_sym_RBRACK,
      anon_sym_PLUS,
      anon_sym_DASH,
  [11698] = 4,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(999), 1,
      anon_sym_SLASH,
    ACTIONS(997), 7,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_STAR,
      anon_sym_RBRACK,
      anon_sym_PLUS,
      anon_sym_DASH,
  [11717] = 4,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1003), 1,
      anon_sym_SLASH,
    ACTIONS(1001), 7,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_STAR,
      anon_sym_RBRACK,
      anon_sym_PLUS,
      anon_sym_DASH,
  [11736] = 4,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1015), 1,
      anon_sym_SLASH,
    ACTIONS(1013), 7,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_STAR,
      anon_sym_RBRACK,
      anon_sym_PLUS,
      anon_sym_DASH,
  [11755] = 7,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1168), 1,
      anon_sym_COMMA,
    ACTIONS(1172), 1,
      anon_sym_RBRACK,
    ACTIONS(1174), 1,
      anon_sym_SLASH,
    STATE(394), 1,
      aux_sym_grid_value_repeat1,
    ACTIONS(1170), 3,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_DASH,
  [11779] = 7,
    ACTIONS(915), 1,
      anon_sym_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(1111), 1,
      anon_sym_LPAREN2,
    ACTIONS(1176), 1,
      sym_identifier,
    STATE(270), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    STATE(427), 2,
      sym_string_value,
      sym_call_expression,
  [11803] = 7,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1168), 1,
      anon_sym_COMMA,
    ACTIONS(1174), 1,
      anon_sym_SLASH,
    ACTIONS(1178), 1,
      anon_sym_RBRACK,
    STATE(390), 1,
      aux_sym_grid_value_repeat1,
    ACTIONS(1170), 3,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_DASH,
  [11827] = 7,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1168), 1,
      anon_sym_COMMA,
    ACTIONS(1174), 1,
      anon_sym_SLASH,
    ACTIONS(1180), 1,
      anon_sym_RBRACK,
    STATE(387), 1,
      aux_sym_grid_value_repeat1,
    ACTIONS(1170), 3,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_DASH,
  [11851] = 7,
    ACTIONS(915), 1,
      anon_sym_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(1111), 1,
      anon_sym_LPAREN2,
    ACTIONS(1176), 1,
      sym_identifier,
    STATE(270), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    STATE(436), 2,
      sym_string_value,
      sym_call_expression,
  [11875] = 7,
    ACTIONS(257), 1,
      anon_sym_LBRACE,
    ACTIONS(1182), 1,
      anon_sym_COMMA,
    ACTIONS(1184), 1,
      anon_sym_SEMI,
    STATE(80), 1,
      sym_block,
    STATE(332), 1,
      aux_sym_import_statement_repeat1,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(1186), 2,
      anon_sym_and,
      anon_sym_or,
  [11899] = 7,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1168), 1,
      anon_sym_COMMA,
    ACTIONS(1174), 1,
      anon_sym_SLASH,
    ACTIONS(1188), 1,
      anon_sym_RBRACK,
    STATE(395), 1,
      aux_sym_grid_value_repeat1,
    ACTIONS(1170), 3,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_DASH,
  [11923] = 7,
    ACTIONS(1067), 1,
      anon_sym_LBRACE,
    ACTIONS(1182), 1,
      anon_sym_COMMA,
    ACTIONS(1190), 1,
      anon_sym_SEMI,
    STATE(49), 1,
      sym_block,
    STATE(329), 1,
      aux_sym_import_statement_repeat1,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(1186), 2,
      anon_sym_and,
      anon_sym_or,
  [11947] = 2,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(1192), 6,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_RPAREN,
      anon_sym_and,
      anon_sym_or,
  [11960] = 2,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(671), 6,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_RPAREN,
      anon_sym_and,
      anon_sym_or,
  [11973] = 6,
    ACTIONS(257), 1,
      anon_sym_LBRACE,
    ACTIONS(1182), 1,
      anon_sym_COMMA,
    STATE(74), 1,
      sym_block,
    STATE(353), 1,
      aux_sym_import_statement_repeat1,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(1186), 2,
      anon_sym_and,
      anon_sym_or,
  [11994] = 2,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(1194), 6,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_RPAREN,
      anon_sym_and,
      anon_sym_or,
  [12007] = 2,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(1196), 6,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_RPAREN,
      anon_sym_and,
      anon_sym_or,
  [12020] = 2,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(1198), 6,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_RPAREN,
      anon_sym_and,
      anon_sym_or,
  [12033] = 6,
    ACTIONS(1067), 1,
      anon_sym_LBRACE,
    ACTIONS(1182), 1,
      anon_sym_COMMA,
    STATE(42), 1,
      sym_block,
    STATE(366), 1,
      aux_sym_import_statement_repeat1,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(1186), 2,
      anon_sym_and,
      anon_sym_or,
  [12054] = 5,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1174), 1,
      anon_sym_SLASH,
    ACTIONS(1200), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
    ACTIONS(1170), 3,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_DASH,
  [12073] = 2,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(1202), 6,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_RPAREN,
      anon_sym_and,
      anon_sym_or,
  [12086] = 5,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1174), 1,
      anon_sym_SLASH,
    ACTIONS(1204), 1,
      anon_sym_RPAREN,
    ACTIONS(1170), 3,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_DASH,
  [12104] = 5,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1174), 1,
      anon_sym_SLASH,
    ACTIONS(1206), 1,
      anon_sym_SEMI,
    ACTIONS(1170), 3,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_DASH,
  [12122] = 4,
    ACTIONS(1210), 1,
      aux_sym_integer_value_token1,
    STATE(409), 1,
      sym_integer_value,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(1208), 3,
      anon_sym_even,
      anon_sym_odd,
      sym__nth_functional_notation,
  [12138] = 5,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1174), 1,
      anon_sym_SLASH,
    ACTIONS(1212), 1,
      anon_sym_RPAREN,
    ACTIONS(1170), 3,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_DASH,
  [12156] = 5,
    ACTIONS(1158), 1,
      anon_sym_to,
    ACTIONS(1160), 1,
      aux_sym_integer_value_token1,
    ACTIONS(1214), 1,
      sym_from,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    STATE(417), 2,
      sym_to,
      sym_integer_value,
  [12174] = 5,
    ACTIONS(915), 1,
      anon_sym_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(1216), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    STATE(434), 2,
      sym_string_value,
      sym_call_expression,
  [12192] = 5,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1174), 1,
      anon_sym_SLASH,
    ACTIONS(1218), 1,
      anon_sym_RBRACK,
    ACTIONS(1170), 3,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_DASH,
  [12210] = 5,
    ACTIONS(1182), 1,
      anon_sym_COMMA,
    ACTIONS(1220), 1,
      anon_sym_SEMI,
    STATE(396), 1,
      aux_sym_import_statement_repeat1,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(1186), 2,
      anon_sym_and,
      anon_sym_or,
  [12228] = 5,
    ACTIONS(915), 1,
      anon_sym_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(1222), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    STATE(428), 2,
      sym_string_value,
      sym_call_expression,
  [12246] = 6,
    ACTIONS(1067), 1,
      anon_sym_LBRACE,
    ACTIONS(1182), 1,
      anon_sym_COMMA,
    ACTIONS(1224), 1,
      anon_sym_SEMI,
    STATE(61), 1,
      sym_block,
    STATE(344), 1,
      aux_sym_import_statement_repeat1,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
  [12266] = 5,
    ACTIONS(1182), 1,
      anon_sym_COMMA,
    ACTIONS(1226), 1,
      anon_sym_SEMI,
    STATE(386), 1,
      aux_sym_import_statement_repeat1,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(1186), 2,
      anon_sym_and,
      anon_sym_or,
  [12284] = 5,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1174), 1,
      anon_sym_SLASH,
    ACTIONS(1228), 1,
      anon_sym_SEMI,
    ACTIONS(1170), 3,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_DASH,
  [12302] = 6,
    ACTIONS(257), 1,
      anon_sym_LBRACE,
    ACTIONS(1182), 1,
      anon_sym_COMMA,
    ACTIONS(1230), 1,
      anon_sym_SEMI,
    STATE(87), 1,
      sym_block,
    STATE(344), 1,
      aux_sym_import_statement_repeat1,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
  [12322] = 5,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1174), 1,
      anon_sym_SLASH,
    ACTIONS(1232), 1,
      anon_sym_RPAREN,
    ACTIONS(1170), 3,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_DASH,
  [12340] = 3,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(1186), 2,
      anon_sym_and,
      anon_sym_or,
    ACTIONS(1234), 3,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_LBRACE,
  [12354] = 5,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1174), 1,
      anon_sym_SLASH,
    ACTIONS(1236), 1,
      anon_sym_RBRACK,
    ACTIONS(1170), 3,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_DASH,
  [12372] = 5,
    ACTIONS(3), 1,
      sym_js_comment,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1174), 1,
      anon_sym_SLASH,
    ACTIONS(1238), 1,
      anon_sym_RPAREN,
    ACTIONS(1170), 3,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_DASH,
  [12390] = 4,
    ACTIONS(1240), 1,
      anon_sym_DQUOTE,
    STATE(372), 1,
      aux_sym_string_value_repeat2,
    ACTIONS(3), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(1242), 2,
      aux_sym_string_value_token2,
      sym_escape_sequence,
  [12405] = 4,
    ACTIONS(1244), 1,
      anon_sym_RPAREN,
    STATE(380), 1,
      aux_sym_arguments_repeat1,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(446), 2,
      anon_sym_COMMA,
      anon_sym_SEMI,
  [12420] = 2,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(1246), 4,
      anon_sym_RBRACE,
      sym_from,
      anon_sym_to,
      aux_sym_integer_value_token1,
  [12431] = 5,
    ACTIONS(1248), 1,
      anon_sym_COMMA,
    ACTIONS(1250), 1,
      anon_sym_LBRACE,
    STATE(357), 1,
      sym_block,
    STATE(392), 1,
      aux_sym_keyframe_block_repeat1,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
  [12448] = 5,
    ACTIONS(1248), 1,
      anon_sym_COMMA,
    ACTIONS(1250), 1,
      anon_sym_LBRACE,
    STATE(339), 1,
      sym_block,
    STATE(340), 1,
      aux_sym_keyframe_block_repeat1,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
  [12465] = 4,
    ACTIONS(1252), 1,
      anon_sym_SQUOTE,
    STATE(348), 1,
      aux_sym_string_value_repeat1,
    ACTIONS(3), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(1254), 2,
      aux_sym_string_value_token1,
      sym_escape_sequence,
  [12480] = 4,
    ACTIONS(1252), 1,
      anon_sym_DQUOTE,
    STATE(349), 1,
      aux_sym_string_value_repeat2,
    ACTIONS(3), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(1256), 2,
      aux_sym_string_value_token2,
      sym_escape_sequence,
  [12495] = 4,
    ACTIONS(1258), 1,
      anon_sym_COMMA,
    STATE(344), 1,
      aux_sym_import_statement_repeat1,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(1234), 2,
      anon_sym_SEMI,
      anon_sym_LBRACE,
  [12510] = 3,
    ACTIONS(1137), 1,
      anon_sym_COLON,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(671), 3,
      anon_sym_RPAREN,
      anon_sym_and,
      anon_sym_or,
  [12523] = 4,
    ACTIONS(257), 1,
      anon_sym_LBRACE,
    STATE(78), 1,
      sym_block,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(1186), 2,
      anon_sym_and,
      anon_sym_or,
  [12538] = 4,
    ACTIONS(1261), 1,
      anon_sym_SQUOTE,
    STATE(377), 1,
      aux_sym_string_value_repeat1,
    ACTIONS(3), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(1263), 2,
      aux_sym_string_value_token1,
      sym_escape_sequence,
  [12553] = 4,
    ACTIONS(1265), 1,
      anon_sym_SQUOTE,
    STATE(377), 1,
      aux_sym_string_value_repeat1,
    ACTIONS(3), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(1263), 2,
      aux_sym_string_value_token1,
      sym_escape_sequence,
  [12568] = 4,
    ACTIONS(1265), 1,
      anon_sym_DQUOTE,
    STATE(378), 1,
      aux_sym_string_value_repeat2,
    ACTIONS(3), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(1267), 2,
      aux_sym_string_value_token2,
      sym_escape_sequence,
  [12583] = 4,
    ACTIONS(1261), 1,
      anon_sym_DQUOTE,
    STATE(378), 1,
      aux_sym_string_value_repeat2,
    ACTIONS(3), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(1267), 2,
      aux_sym_string_value_token2,
      sym_escape_sequence,
  [12598] = 4,
    ACTIONS(1269), 1,
      anon_sym_DQUOTE,
    STATE(350), 1,
      aux_sym_string_value_repeat2,
    ACTIONS(3), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(1271), 2,
      aux_sym_string_value_token2,
      sym_escape_sequence,
  [12613] = 2,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(415), 4,
      anon_sym_RBRACE,
      sym_from,
      anon_sym_to,
      aux_sym_integer_value_token1,
  [12624] = 5,
    ACTIONS(257), 1,
      anon_sym_LBRACE,
    ACTIONS(1182), 1,
      anon_sym_COMMA,
    STATE(82), 1,
      sym_block,
    STATE(344), 1,
      aux_sym_import_statement_repeat1,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
  [12641] = 4,
    ACTIONS(1067), 1,
      anon_sym_LBRACE,
    STATE(45), 1,
      sym_block,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(1186), 2,
      anon_sym_and,
      anon_sym_or,
  [12656] = 4,
    ACTIONS(1273), 1,
      anon_sym_RPAREN,
    STATE(380), 1,
      aux_sym_arguments_repeat1,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(446), 2,
      anon_sym_COMMA,
      anon_sym_SEMI,
  [12671] = 4,
    ACTIONS(1275), 1,
      anon_sym_SQUOTE,
    STATE(358), 1,
      aux_sym_string_value_repeat1,
    ACTIONS(3), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(1277), 2,
      aux_sym_string_value_token1,
      sym_escape_sequence,
  [12686] = 2,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(1279), 4,
      anon_sym_RBRACE,
      sym_from,
      anon_sym_to,
      aux_sym_integer_value_token1,
  [12697] = 4,
    ACTIONS(1281), 1,
      anon_sym_SQUOTE,
    STATE(377), 1,
      aux_sym_string_value_repeat1,
    ACTIONS(3), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(1263), 2,
      aux_sym_string_value_token1,
      sym_escape_sequence,
  [12712] = 4,
    ACTIONS(1281), 1,
      anon_sym_DQUOTE,
    STATE(378), 1,
      aux_sym_string_value_repeat2,
    ACTIONS(3), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(1267), 2,
      aux_sym_string_value_token2,
      sym_escape_sequence,
  [12727] = 2,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(466), 4,
      anon_sym_RBRACE,
      sym_from,
      anon_sym_to,
      aux_sym_integer_value_token1,
  [12738] = 4,
    ACTIONS(1283), 1,
      anon_sym_RPAREN,
    STATE(380), 1,
      aux_sym_arguments_repeat1,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(446), 2,
      anon_sym_COMMA,
      anon_sym_SEMI,
  [12753] = 4,
    ACTIONS(1285), 1,
      anon_sym_SQUOTE,
    STATE(364), 1,
      aux_sym_string_value_repeat1,
    ACTIONS(3), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(1287), 2,
      aux_sym_string_value_token1,
      sym_escape_sequence,
  [12768] = 4,
    ACTIONS(1285), 1,
      anon_sym_DQUOTE,
    STATE(365), 1,
      aux_sym_string_value_repeat2,
    ACTIONS(3), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(1289), 2,
      aux_sym_string_value_token2,
      sym_escape_sequence,
  [12783] = 4,
    ACTIONS(1291), 1,
      anon_sym_SQUOTE,
    STATE(377), 1,
      aux_sym_string_value_repeat1,
    ACTIONS(3), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(1263), 2,
      aux_sym_string_value_token1,
      sym_escape_sequence,
  [12798] = 4,
    ACTIONS(1291), 1,
      anon_sym_DQUOTE,
    STATE(378), 1,
      aux_sym_string_value_repeat2,
    ACTIONS(3), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(1267), 2,
      aux_sym_string_value_token2,
      sym_escape_sequence,
  [12813] = 5,
    ACTIONS(1067), 1,
      anon_sym_LBRACE,
    ACTIONS(1182), 1,
      anon_sym_COMMA,
    STATE(54), 1,
      sym_block,
    STATE(344), 1,
      aux_sym_import_statement_repeat1,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
  [12830] = 2,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(496), 4,
      anon_sym_RBRACE,
      sym_from,
      anon_sym_to,
      aux_sym_integer_value_token1,
  [12841] = 4,
    ACTIONS(1293), 1,
      anon_sym_COMMA,
    STATE(368), 1,
      aux_sym_selectors_repeat1,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(1099), 2,
      anon_sym_LBRACE,
      anon_sym_RPAREN,
  [12856] = 4,
    ACTIONS(1296), 1,
      anon_sym_RPAREN,
    STATE(380), 1,
      aux_sym_arguments_repeat1,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(446), 2,
      anon_sym_COMMA,
      anon_sym_SEMI,
  [12871] = 4,
    ACTIONS(1240), 1,
      anon_sym_SQUOTE,
    STATE(371), 1,
      aux_sym_string_value_repeat1,
    ACTIONS(3), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(1298), 2,
      aux_sym_string_value_token1,
      sym_escape_sequence,
  [12886] = 4,
    ACTIONS(1300), 1,
      anon_sym_SQUOTE,
    STATE(377), 1,
      aux_sym_string_value_repeat1,
    ACTIONS(3), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(1263), 2,
      aux_sym_string_value_token1,
      sym_escape_sequence,
  [12901] = 4,
    ACTIONS(1300), 1,
      anon_sym_DQUOTE,
    STATE(378), 1,
      aux_sym_string_value_repeat2,
    ACTIONS(3), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(1267), 2,
      aux_sym_string_value_token2,
      sym_escape_sequence,
  [12916] = 4,
    ACTIONS(1302), 1,
      anon_sym_SQUOTE,
    STATE(375), 1,
      aux_sym_string_value_repeat1,
    ACTIONS(3), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(1304), 2,
      aux_sym_string_value_token1,
      sym_escape_sequence,
  [12931] = 4,
    ACTIONS(1302), 1,
      anon_sym_DQUOTE,
    STATE(376), 1,
      aux_sym_string_value_repeat2,
    ACTIONS(3), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(1306), 2,
      aux_sym_string_value_token2,
      sym_escape_sequence,
  [12946] = 4,
    ACTIONS(1308), 1,
      anon_sym_SQUOTE,
    STATE(377), 1,
      aux_sym_string_value_repeat1,
    ACTIONS(3), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(1263), 2,
      aux_sym_string_value_token1,
      sym_escape_sequence,
  [12961] = 4,
    ACTIONS(1308), 1,
      anon_sym_DQUOTE,
    STATE(378), 1,
      aux_sym_string_value_repeat2,
    ACTIONS(3), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(1267), 2,
      aux_sym_string_value_token2,
      sym_escape_sequence,
  [12976] = 4,
    ACTIONS(1310), 1,
      anon_sym_SQUOTE,
    STATE(377), 1,
      aux_sym_string_value_repeat1,
    ACTIONS(3), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(1312), 2,
      aux_sym_string_value_token1,
      sym_escape_sequence,
  [12991] = 4,
    ACTIONS(1315), 1,
      anon_sym_DQUOTE,
    STATE(378), 1,
      aux_sym_string_value_repeat2,
    ACTIONS(3), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(1317), 2,
      aux_sym_string_value_token2,
      sym_escape_sequence,
  [13006] = 4,
    ACTIONS(1269), 1,
      anon_sym_SQUOTE,
    STATE(347), 1,
      aux_sym_string_value_repeat1,
    ACTIONS(3), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(1320), 2,
      aux_sym_string_value_token1,
      sym_escape_sequence,
  [13021] = 4,
    ACTIONS(579), 1,
      anon_sym_RPAREN,
    STATE(380), 1,
      aux_sym_arguments_repeat1,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(1322), 2,
      anon_sym_COMMA,
      anon_sym_SEMI,
  [13036] = 4,
    ACTIONS(1275), 1,
      anon_sym_DQUOTE,
    STATE(359), 1,
      aux_sym_string_value_repeat2,
    ACTIONS(3), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(1325), 2,
      aux_sym_string_value_token2,
      sym_escape_sequence,
  [13051] = 3,
    ACTIONS(1327), 1,
      anon_sym_RPAREN,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(1186), 2,
      anon_sym_and,
      anon_sym_or,
  [13063] = 4,
    ACTIONS(1200), 1,
      anon_sym_RBRACK,
    ACTIONS(1329), 1,
      anon_sym_COMMA,
    STATE(383), 1,
      aux_sym_grid_value_repeat1,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
  [13077] = 4,
    ACTIONS(575), 1,
      anon_sym_COMMA,
    ACTIONS(1332), 1,
      anon_sym_RPAREN,
    STATE(402), 1,
      aux_sym_pseudo_class_arguments_repeat2,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
  [13091] = 4,
    ACTIONS(257), 1,
      anon_sym_LBRACE,
    ACTIONS(1334), 1,
      anon_sym_LPAREN,
    STATE(70), 1,
      sym_block,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
  [13105] = 4,
    ACTIONS(1182), 1,
      anon_sym_COMMA,
    ACTIONS(1336), 1,
      anon_sym_SEMI,
    STATE(344), 1,
      aux_sym_import_statement_repeat1,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
  [13119] = 4,
    ACTIONS(1168), 1,
      anon_sym_COMMA,
    ACTIONS(1338), 1,
      anon_sym_RBRACK,
    STATE(383), 1,
      aux_sym_grid_value_repeat1,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
  [13133] = 3,
    ACTIONS(1340), 1,
      sym_unit,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(881), 2,
      anon_sym_COMMA,
      anon_sym_LBRACE,
  [13145] = 4,
    ACTIONS(1067), 1,
      anon_sym_LBRACE,
    ACTIONS(1342), 1,
      anon_sym_to,
    STATE(66), 1,
      sym_block,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
  [13159] = 4,
    ACTIONS(1168), 1,
      anon_sym_COMMA,
    ACTIONS(1344), 1,
      anon_sym_RBRACK,
    STATE(383), 1,
      aux_sym_grid_value_repeat1,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
  [13173] = 3,
    STATE(152), 1,
      sym_class_name,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(1107), 2,
      sym_escape_sequence,
      sym_identifier,
  [13185] = 4,
    ACTIONS(1346), 1,
      anon_sym_COMMA,
    ACTIONS(1349), 1,
      anon_sym_LBRACE,
    STATE(392), 1,
      aux_sym_keyframe_block_repeat1,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
  [13199] = 4,
    ACTIONS(1067), 1,
      anon_sym_LBRACE,
    ACTIONS(1351), 1,
      anon_sym_LPAREN,
    STATE(40), 1,
      sym_block,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
  [13213] = 4,
    ACTIONS(1168), 1,
      anon_sym_COMMA,
    ACTIONS(1353), 1,
      anon_sym_RBRACK,
    STATE(383), 1,
      aux_sym_grid_value_repeat1,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
  [13227] = 4,
    ACTIONS(1168), 1,
      anon_sym_COMMA,
    ACTIONS(1355), 1,
      anon_sym_RBRACK,
    STATE(383), 1,
      aux_sym_grid_value_repeat1,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
  [13241] = 4,
    ACTIONS(1182), 1,
      anon_sym_COMMA,
    ACTIONS(1357), 1,
      anon_sym_SEMI,
    STATE(344), 1,
      aux_sym_import_statement_repeat1,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
  [13255] = 4,
    ACTIONS(881), 1,
      anon_sym_RPAREN,
    ACTIONS(883), 1,
      anon_sym_of,
    ACTIONS(1359), 1,
      sym_unit,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
  [13269] = 4,
    ACTIONS(1071), 1,
      anon_sym_COMMA,
    ACTIONS(1361), 1,
      anon_sym_RPAREN,
    STATE(368), 1,
      aux_sym_selectors_repeat1,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
  [13283] = 3,
    STATE(138), 1,
      sym_class_name,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(1107), 2,
      sym_escape_sequence,
      sym_identifier,
  [13295] = 4,
    ACTIONS(257), 1,
      anon_sym_LBRACE,
    ACTIONS(1363), 1,
      anon_sym_to,
    STATE(91), 1,
      sym_block,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
  [13309] = 4,
    ACTIONS(575), 1,
      anon_sym_COMMA,
    ACTIONS(1365), 1,
      anon_sym_RPAREN,
    STATE(402), 1,
      aux_sym_pseudo_class_arguments_repeat2,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
  [13323] = 4,
    ACTIONS(621), 1,
      anon_sym_RPAREN,
    ACTIONS(1367), 1,
      anon_sym_COMMA,
    STATE(402), 1,
      aux_sym_pseudo_class_arguments_repeat2,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
  [13337] = 4,
    ACTIONS(1071), 1,
      anon_sym_COMMA,
    ACTIONS(1370), 1,
      anon_sym_LBRACE,
    STATE(368), 1,
      aux_sym_selectors_repeat1,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
  [13351] = 3,
    ACTIONS(257), 1,
      anon_sym_LBRACE,
    STATE(94), 1,
      sym_block,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
  [13362] = 3,
    ACTIONS(1067), 1,
      anon_sym_LBRACE,
    STATE(69), 1,
      sym_block,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
  [13373] = 3,
    ACTIONS(627), 1,
      anon_sym_LPAREN2,
    STATE(153), 1,
      sym_pseudo_class_with_selector_arguments,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
  [13384] = 3,
    ACTIONS(1372), 1,
      anon_sym_LPAREN2,
    STATE(155), 1,
      sym_pseudo_class_nth_child_arguments,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
  [13395] = 2,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(1374), 2,
      anon_sym_COMMA,
      anon_sym_LBRACE,
  [13404] = 3,
    ACTIONS(1376), 1,
      anon_sym_RPAREN,
    ACTIONS(1378), 1,
      anon_sym_of,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
  [13415] = 3,
    ACTIONS(1067), 1,
      anon_sym_LBRACE,
    STATE(48), 1,
      sym_block,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
  [13426] = 3,
    ACTIONS(1380), 1,
      anon_sym_LBRACE,
    STATE(44), 1,
      sym_keyframe_block_list,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
  [13437] = 3,
    ACTIONS(627), 1,
      anon_sym_LPAREN2,
    STATE(144), 1,
      sym_pseudo_class_with_selector_arguments,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
  [13448] = 3,
    ACTIONS(1111), 1,
      anon_sym_LPAREN2,
    STATE(270), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
  [13459] = 3,
    ACTIONS(1382), 1,
      anon_sym_LBRACE,
    STATE(77), 1,
      sym_keyframe_block_list,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
  [13470] = 3,
    ACTIONS(1384), 1,
      aux_sym_color_value_token1,
    ACTIONS(1386), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
  [13481] = 3,
    ACTIONS(257), 1,
      anon_sym_LBRACE,
    STATE(72), 1,
      sym_block,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
  [13492] = 2,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
    ACTIONS(1349), 2,
      anon_sym_COMMA,
      anon_sym_LBRACE,
  [13501] = 2,
    ACTIONS(1388), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
  [13509] = 2,
    ACTIONS(1390), 1,
      anon_sym_RBRACE,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
  [13517] = 2,
    ACTIONS(1392), 1,
      aux_sym_color_value_token1,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
  [13525] = 2,
    ACTIONS(73), 1,
      anon_sym_RBRACE,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
  [13533] = 2,
    ACTIONS(1394), 1,
      aux_sym_color_value_token1,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
  [13541] = 2,
    ACTIONS(1396), 1,
      aux_sym_color_value_token1,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
  [13549] = 2,
    ACTIONS(77), 1,
      anon_sym_RBRACE,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
  [13557] = 2,
    ACTIONS(1398), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
  [13565] = 2,
    ACTIONS(1400), 1,
      anon_sym_RBRACE,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
  [13573] = 2,
    ACTIONS(1402), 1,
      anon_sym_SEMI,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
  [13581] = 2,
    ACTIONS(1404), 1,
      anon_sym_SEMI,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
  [13589] = 2,
    ACTIONS(1406), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
  [13597] = 2,
    ACTIONS(1408), 1,
      anon_sym_LPAREN,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
  [13605] = 2,
    ACTIONS(1410), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
  [13613] = 2,
    ACTIONS(81), 1,
      anon_sym_RBRACE,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
  [13621] = 2,
    ACTIONS(1412), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
  [13629] = 2,
    ACTIONS(1414), 1,
      anon_sym_SEMI,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
  [13637] = 2,
    ACTIONS(1416), 1,
      anon_sym_RBRACE,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
  [13645] = 2,
    ACTIONS(1418), 1,
      anon_sym_SEMI,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
  [13653] = 2,
    ACTIONS(1420), 1,
      anon_sym_LPAREN,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
  [13661] = 2,
    ACTIONS(1422), 1,
      aux_sym_color_value_token1,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
  [13669] = 2,
    ACTIONS(1424), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
  [13677] = 2,
    ACTIONS(1426), 1,
      ts_builtin_sym_end,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
  [13685] = 2,
    ACTIONS(1428), 1,
      anon_sym_LPAREN,
    ACTIONS(5), 2,
      sym_js_comment,
      sym_comment,
};

static const uint32_t ts_small_parse_table_map[] = {
  [SMALL_STATE(2)] = 0,
  [SMALL_STATE(3)] = 110,
  [SMALL_STATE(4)] = 220,
  [SMALL_STATE(5)] = 330,
  [SMALL_STATE(6)] = 440,
  [SMALL_STATE(7)] = 550,
  [SMALL_STATE(8)] = 660,
  [SMALL_STATE(9)] = 767,
  [SMALL_STATE(10)] = 873,
  [SMALL_STATE(11)] = 979,
  [SMALL_STATE(12)] = 1079,
  [SMALL_STATE(13)] = 1179,
  [SMALL_STATE(14)] = 1276,
  [SMALL_STATE(15)] = 1352,
  [SMALL_STATE(16)] = 1415,
  [SMALL_STATE(17)] = 1480,
  [SMALL_STATE(18)] = 1543,
  [SMALL_STATE(19)] = 1605,
  [SMALL_STATE(20)] = 1669,
  [SMALL_STATE(21)] = 1731,
  [SMALL_STATE(22)] = 1793,
  [SMALL_STATE(23)] = 1855,
  [SMALL_STATE(24)] = 1917,
  [SMALL_STATE(25)] = 1979,
  [SMALL_STATE(26)] = 2041,
  [SMALL_STATE(27)] = 2105,
  [SMALL_STATE(28)] = 2167,
  [SMALL_STATE(29)] = 2229,
  [SMALL_STATE(30)] = 2297,
  [SMALL_STATE(31)] = 2359,
  [SMALL_STATE(32)] = 2421,
  [SMALL_STATE(33)] = 2483,
  [SMALL_STATE(34)] = 2545,
  [SMALL_STATE(35)] = 2607,
  [SMALL_STATE(36)] = 2669,
  [SMALL_STATE(37)] = 2731,
  [SMALL_STATE(38)] = 2796,
  [SMALL_STATE(39)] = 2841,
  [SMALL_STATE(40)] = 2906,
  [SMALL_STATE(41)] = 2939,
  [SMALL_STATE(42)] = 2972,
  [SMALL_STATE(43)] = 3005,
  [SMALL_STATE(44)] = 3038,
  [SMALL_STATE(45)] = 3071,
  [SMALL_STATE(46)] = 3104,
  [SMALL_STATE(47)] = 3137,
  [SMALL_STATE(48)] = 3170,
  [SMALL_STATE(49)] = 3203,
  [SMALL_STATE(50)] = 3236,
  [SMALL_STATE(51)] = 3273,
  [SMALL_STATE(52)] = 3312,
  [SMALL_STATE(53)] = 3373,
  [SMALL_STATE(54)] = 3406,
  [SMALL_STATE(55)] = 3439,
  [SMALL_STATE(56)] = 3472,
  [SMALL_STATE(57)] = 3505,
  [SMALL_STATE(58)] = 3538,
  [SMALL_STATE(59)] = 3571,
  [SMALL_STATE(60)] = 3608,
  [SMALL_STATE(61)] = 3641,
  [SMALL_STATE(62)] = 3674,
  [SMALL_STATE(63)] = 3707,
  [SMALL_STATE(64)] = 3740,
  [SMALL_STATE(65)] = 3773,
  [SMALL_STATE(66)] = 3806,
  [SMALL_STATE(67)] = 3839,
  [SMALL_STATE(68)] = 3872,
  [SMALL_STATE(69)] = 3905,
  [SMALL_STATE(70)] = 3938,
  [SMALL_STATE(71)] = 3971,
  [SMALL_STATE(72)] = 4004,
  [SMALL_STATE(73)] = 4037,
  [SMALL_STATE(74)] = 4070,
  [SMALL_STATE(75)] = 4103,
  [SMALL_STATE(76)] = 4136,
  [SMALL_STATE(77)] = 4169,
  [SMALL_STATE(78)] = 4202,
  [SMALL_STATE(79)] = 4235,
  [SMALL_STATE(80)] = 4268,
  [SMALL_STATE(81)] = 4301,
  [SMALL_STATE(82)] = 4334,
  [SMALL_STATE(83)] = 4367,
  [SMALL_STATE(84)] = 4400,
  [SMALL_STATE(85)] = 4433,
  [SMALL_STATE(86)] = 4466,
  [SMALL_STATE(87)] = 4499,
  [SMALL_STATE(88)] = 4532,
  [SMALL_STATE(89)] = 4565,
  [SMALL_STATE(90)] = 4598,
  [SMALL_STATE(91)] = 4631,
  [SMALL_STATE(92)] = 4664,
  [SMALL_STATE(93)] = 4697,
  [SMALL_STATE(94)] = 4730,
  [SMALL_STATE(95)] = 4763,
  [SMALL_STATE(96)] = 4798,
  [SMALL_STATE(97)] = 4833,
  [SMALL_STATE(98)] = 4870,
  [SMALL_STATE(99)] = 4931,
  [SMALL_STATE(100)] = 4992,
  [SMALL_STATE(101)] = 5053,
  [SMALL_STATE(102)] = 5086,
  [SMALL_STATE(103)] = 5144,
  [SMALL_STATE(104)] = 5204,
  [SMALL_STATE(105)] = 5260,
  [SMALL_STATE(106)] = 5320,
  [SMALL_STATE(107)] = 5380,
  [SMALL_STATE(108)] = 5436,
  [SMALL_STATE(109)] = 5491,
  [SMALL_STATE(110)] = 5526,
  [SMALL_STATE(111)] = 5561,
  [SMALL_STATE(112)] = 5616,
  [SMALL_STATE(113)] = 5651,
  [SMALL_STATE(114)] = 5692,
  [SMALL_STATE(115)] = 5727,
  [SMALL_STATE(116)] = 5782,
  [SMALL_STATE(117)] = 5817,
  [SMALL_STATE(118)] = 5874,
  [SMALL_STATE(119)] = 5909,
  [SMALL_STATE(120)] = 5966,
  [SMALL_STATE(121)] = 6020,
  [SMALL_STATE(122)] = 6074,
  [SMALL_STATE(123)] = 6128,
  [SMALL_STATE(124)] = 6182,
  [SMALL_STATE(125)] = 6236,
  [SMALL_STATE(126)] = 6265,
  [SMALL_STATE(127)] = 6294,
  [SMALL_STATE(128)] = 6323,
  [SMALL_STATE(129)] = 6352,
  [SMALL_STATE(130)] = 6403,
  [SMALL_STATE(131)] = 6434,
  [SMALL_STATE(132)] = 6463,
  [SMALL_STATE(133)] = 6494,
  [SMALL_STATE(134)] = 6523,
  [SMALL_STATE(135)] = 6552,
  [SMALL_STATE(136)] = 6581,
  [SMALL_STATE(137)] = 6610,
  [SMALL_STATE(138)] = 6641,
  [SMALL_STATE(139)] = 6670,
  [SMALL_STATE(140)] = 6699,
  [SMALL_STATE(141)] = 6728,
  [SMALL_STATE(142)] = 6759,
  [SMALL_STATE(143)] = 6790,
  [SMALL_STATE(144)] = 6819,
  [SMALL_STATE(145)] = 6848,
  [SMALL_STATE(146)] = 6877,
  [SMALL_STATE(147)] = 6906,
  [SMALL_STATE(148)] = 6935,
  [SMALL_STATE(149)] = 6964,
  [SMALL_STATE(150)] = 6993,
  [SMALL_STATE(151)] = 7028,
  [SMALL_STATE(152)] = 7057,
  [SMALL_STATE(153)] = 7086,
  [SMALL_STATE(154)] = 7115,
  [SMALL_STATE(155)] = 7144,
  [SMALL_STATE(156)] = 7173,
  [SMALL_STATE(157)] = 7202,
  [SMALL_STATE(158)] = 7231,
  [SMALL_STATE(159)] = 7260,
  [SMALL_STATE(160)] = 7291,
  [SMALL_STATE(161)] = 7320,
  [SMALL_STATE(162)] = 7351,
  [SMALL_STATE(163)] = 7380,
  [SMALL_STATE(164)] = 7411,
  [SMALL_STATE(165)] = 7462,
  [SMALL_STATE(166)] = 7491,
  [SMALL_STATE(167)] = 7522,
  [SMALL_STATE(168)] = 7551,
  [SMALL_STATE(169)] = 7599,
  [SMALL_STATE(170)] = 7647,
  [SMALL_STATE(171)] = 7695,
  [SMALL_STATE(172)] = 7743,
  [SMALL_STATE(173)] = 7791,
  [SMALL_STATE(174)] = 7823,
  [SMALL_STATE(175)] = 7855,
  [SMALL_STATE(176)] = 7903,
  [SMALL_STATE(177)] = 7951,
  [SMALL_STATE(178)] = 7999,
  [SMALL_STATE(179)] = 8047,
  [SMALL_STATE(180)] = 8095,
  [SMALL_STATE(181)] = 8143,
  [SMALL_STATE(182)] = 8191,
  [SMALL_STATE(183)] = 8239,
  [SMALL_STATE(184)] = 8287,
  [SMALL_STATE(185)] = 8335,
  [SMALL_STATE(186)] = 8383,
  [SMALL_STATE(187)] = 8431,
  [SMALL_STATE(188)] = 8479,
  [SMALL_STATE(189)] = 8527,
  [SMALL_STATE(190)] = 8575,
  [SMALL_STATE(191)] = 8623,
  [SMALL_STATE(192)] = 8671,
  [SMALL_STATE(193)] = 8719,
  [SMALL_STATE(194)] = 8748,
  [SMALL_STATE(195)] = 8777,
  [SMALL_STATE(196)] = 8806,
  [SMALL_STATE(197)] = 8835,
  [SMALL_STATE(198)] = 8864,
  [SMALL_STATE(199)] = 8893,
  [SMALL_STATE(200)] = 8922,
  [SMALL_STATE(201)] = 8951,
  [SMALL_STATE(202)] = 8980,
  [SMALL_STATE(203)] = 9013,
  [SMALL_STATE(204)] = 9042,
  [SMALL_STATE(205)] = 9071,
  [SMALL_STATE(206)] = 9100,
  [SMALL_STATE(207)] = 9129,
  [SMALL_STATE(208)] = 9159,
  [SMALL_STATE(209)] = 9191,
  [SMALL_STATE(210)] = 9221,
  [SMALL_STATE(211)] = 9253,
  [SMALL_STATE(212)] = 9285,
  [SMALL_STATE(213)] = 9315,
  [SMALL_STATE(214)] = 9345,
  [SMALL_STATE(215)] = 9377,
  [SMALL_STATE(216)] = 9407,
  [SMALL_STATE(217)] = 9437,
  [SMALL_STATE(218)] = 9469,
  [SMALL_STATE(219)] = 9496,
  [SMALL_STATE(220)] = 9523,
  [SMALL_STATE(221)] = 9550,
  [SMALL_STATE(222)] = 9589,
  [SMALL_STATE(223)] = 9616,
  [SMALL_STATE(224)] = 9643,
  [SMALL_STATE(225)] = 9670,
  [SMALL_STATE(226)] = 9699,
  [SMALL_STATE(227)] = 9728,
  [SMALL_STATE(228)] = 9755,
  [SMALL_STATE(229)] = 9784,
  [SMALL_STATE(230)] = 9815,
  [SMALL_STATE(231)] = 9844,
  [SMALL_STATE(232)] = 9871,
  [SMALL_STATE(233)] = 9898,
  [SMALL_STATE(234)] = 9937,
  [SMALL_STATE(235)] = 9964,
  [SMALL_STATE(236)] = 9991,
  [SMALL_STATE(237)] = 10018,
  [SMALL_STATE(238)] = 10045,
  [SMALL_STATE(239)] = 10080,
  [SMALL_STATE(240)] = 10102,
  [SMALL_STATE(241)] = 10146,
  [SMALL_STATE(242)] = 10190,
  [SMALL_STATE(243)] = 10212,
  [SMALL_STATE(244)] = 10256,
  [SMALL_STATE(245)] = 10278,
  [SMALL_STATE(246)] = 10318,
  [SMALL_STATE(247)] = 10362,
  [SMALL_STATE(248)] = 10393,
  [SMALL_STATE(249)] = 10416,
  [SMALL_STATE(250)] = 10443,
  [SMALL_STATE(251)] = 10474,
  [SMALL_STATE(252)] = 10513,
  [SMALL_STATE(253)] = 10551,
  [SMALL_STATE(254)] = 10589,
  [SMALL_STATE(255)] = 10615,
  [SMALL_STATE(256)] = 10653,
  [SMALL_STATE(257)] = 10679,
  [SMALL_STATE(258)] = 10705,
  [SMALL_STATE(259)] = 10731,
  [SMALL_STATE(260)] = 10755,
  [SMALL_STATE(261)] = 10779,
  [SMALL_STATE(262)] = 10817,
  [SMALL_STATE(263)] = 10843,
  [SMALL_STATE(264)] = 10869,
  [SMALL_STATE(265)] = 10895,
  [SMALL_STATE(266)] = 10921,
  [SMALL_STATE(267)] = 10959,
  [SMALL_STATE(268)] = 10997,
  [SMALL_STATE(269)] = 11018,
  [SMALL_STATE(270)] = 11039,
  [SMALL_STATE(271)] = 11060,
  [SMALL_STATE(272)] = 11081,
  [SMALL_STATE(273)] = 11116,
  [SMALL_STATE(274)] = 11137,
  [SMALL_STATE(275)] = 11158,
  [SMALL_STATE(276)] = 11183,
  [SMALL_STATE(277)] = 11204,
  [SMALL_STATE(278)] = 11225,
  [SMALL_STATE(279)] = 11254,
  [SMALL_STATE(280)] = 11275,
  [SMALL_STATE(281)] = 11296,
  [SMALL_STATE(282)] = 11317,
  [SMALL_STATE(283)] = 11338,
  [SMALL_STATE(284)] = 11360,
  [SMALL_STATE(285)] = 11382,
  [SMALL_STATE(286)] = 11402,
  [SMALL_STATE(287)] = 11427,
  [SMALL_STATE(288)] = 11446,
  [SMALL_STATE(289)] = 11471,
  [SMALL_STATE(290)] = 11490,
  [SMALL_STATE(291)] = 11509,
  [SMALL_STATE(292)] = 11528,
  [SMALL_STATE(293)] = 11547,
  [SMALL_STATE(294)] = 11566,
  [SMALL_STATE(295)] = 11585,
  [SMALL_STATE(296)] = 11610,
  [SMALL_STATE(297)] = 11635,
  [SMALL_STATE(298)] = 11660,
  [SMALL_STATE(299)] = 11679,
  [SMALL_STATE(300)] = 11698,
  [SMALL_STATE(301)] = 11717,
  [SMALL_STATE(302)] = 11736,
  [SMALL_STATE(303)] = 11755,
  [SMALL_STATE(304)] = 11779,
  [SMALL_STATE(305)] = 11803,
  [SMALL_STATE(306)] = 11827,
  [SMALL_STATE(307)] = 11851,
  [SMALL_STATE(308)] = 11875,
  [SMALL_STATE(309)] = 11899,
  [SMALL_STATE(310)] = 11923,
  [SMALL_STATE(311)] = 11947,
  [SMALL_STATE(312)] = 11960,
  [SMALL_STATE(313)] = 11973,
  [SMALL_STATE(314)] = 11994,
  [SMALL_STATE(315)] = 12007,
  [SMALL_STATE(316)] = 12020,
  [SMALL_STATE(317)] = 12033,
  [SMALL_STATE(318)] = 12054,
  [SMALL_STATE(319)] = 12073,
  [SMALL_STATE(320)] = 12086,
  [SMALL_STATE(321)] = 12104,
  [SMALL_STATE(322)] = 12122,
  [SMALL_STATE(323)] = 12138,
  [SMALL_STATE(324)] = 12156,
  [SMALL_STATE(325)] = 12174,
  [SMALL_STATE(326)] = 12192,
  [SMALL_STATE(327)] = 12210,
  [SMALL_STATE(328)] = 12228,
  [SMALL_STATE(329)] = 12246,
  [SMALL_STATE(330)] = 12266,
  [SMALL_STATE(331)] = 12284,
  [SMALL_STATE(332)] = 12302,
  [SMALL_STATE(333)] = 12322,
  [SMALL_STATE(334)] = 12340,
  [SMALL_STATE(335)] = 12354,
  [SMALL_STATE(336)] = 12372,
  [SMALL_STATE(337)] = 12390,
  [SMALL_STATE(338)] = 12405,
  [SMALL_STATE(339)] = 12420,
  [SMALL_STATE(340)] = 12431,
  [SMALL_STATE(341)] = 12448,
  [SMALL_STATE(342)] = 12465,
  [SMALL_STATE(343)] = 12480,
  [SMALL_STATE(344)] = 12495,
  [SMALL_STATE(345)] = 12510,
  [SMALL_STATE(346)] = 12523,
  [SMALL_STATE(347)] = 12538,
  [SMALL_STATE(348)] = 12553,
  [SMALL_STATE(349)] = 12568,
  [SMALL_STATE(350)] = 12583,
  [SMALL_STATE(351)] = 12598,
  [SMALL_STATE(352)] = 12613,
  [SMALL_STATE(353)] = 12624,
  [SMALL_STATE(354)] = 12641,
  [SMALL_STATE(355)] = 12656,
  [SMALL_STATE(356)] = 12671,
  [SMALL_STATE(357)] = 12686,
  [SMALL_STATE(358)] = 12697,
  [SMALL_STATE(359)] = 12712,
  [SMALL_STATE(360)] = 12727,
  [SMALL_STATE(361)] = 12738,
  [SMALL_STATE(362)] = 12753,
  [SMALL_STATE(363)] = 12768,
  [SMALL_STATE(364)] = 12783,
  [SMALL_STATE(365)] = 12798,
  [SMALL_STATE(366)] = 12813,
  [SMALL_STATE(367)] = 12830,
  [SMALL_STATE(368)] = 12841,
  [SMALL_STATE(369)] = 12856,
  [SMALL_STATE(370)] = 12871,
  [SMALL_STATE(371)] = 12886,
  [SMALL_STATE(372)] = 12901,
  [SMALL_STATE(373)] = 12916,
  [SMALL_STATE(374)] = 12931,
  [SMALL_STATE(375)] = 12946,
  [SMALL_STATE(376)] = 12961,
  [SMALL_STATE(377)] = 12976,
  [SMALL_STATE(378)] = 12991,
  [SMALL_STATE(379)] = 13006,
  [SMALL_STATE(380)] = 13021,
  [SMALL_STATE(381)] = 13036,
  [SMALL_STATE(382)] = 13051,
  [SMALL_STATE(383)] = 13063,
  [SMALL_STATE(384)] = 13077,
  [SMALL_STATE(385)] = 13091,
  [SMALL_STATE(386)] = 13105,
  [SMALL_STATE(387)] = 13119,
  [SMALL_STATE(388)] = 13133,
  [SMALL_STATE(389)] = 13145,
  [SMALL_STATE(390)] = 13159,
  [SMALL_STATE(391)] = 13173,
  [SMALL_STATE(392)] = 13185,
  [SMALL_STATE(393)] = 13199,
  [SMALL_STATE(394)] = 13213,
  [SMALL_STATE(395)] = 13227,
  [SMALL_STATE(396)] = 13241,
  [SMALL_STATE(397)] = 13255,
  [SMALL_STATE(398)] = 13269,
  [SMALL_STATE(399)] = 13283,
  [SMALL_STATE(400)] = 13295,
  [SMALL_STATE(401)] = 13309,
  [SMALL_STATE(402)] = 13323,
  [SMALL_STATE(403)] = 13337,
  [SMALL_STATE(404)] = 13351,
  [SMALL_STATE(405)] = 13362,
  [SMALL_STATE(406)] = 13373,
  [SMALL_STATE(407)] = 13384,
  [SMALL_STATE(408)] = 13395,
  [SMALL_STATE(409)] = 13404,
  [SMALL_STATE(410)] = 13415,
  [SMALL_STATE(411)] = 13426,
  [SMALL_STATE(412)] = 13437,
  [SMALL_STATE(413)] = 13448,
  [SMALL_STATE(414)] = 13459,
  [SMALL_STATE(415)] = 13470,
  [SMALL_STATE(416)] = 13481,
  [SMALL_STATE(417)] = 13492,
  [SMALL_STATE(418)] = 13501,
  [SMALL_STATE(419)] = 13509,
  [SMALL_STATE(420)] = 13517,
  [SMALL_STATE(421)] = 13525,
  [SMALL_STATE(422)] = 13533,
  [SMALL_STATE(423)] = 13541,
  [SMALL_STATE(424)] = 13549,
  [SMALL_STATE(425)] = 13557,
  [SMALL_STATE(426)] = 13565,
  [SMALL_STATE(427)] = 13573,
  [SMALL_STATE(428)] = 13581,
  [SMALL_STATE(429)] = 13589,
  [SMALL_STATE(430)] = 13597,
  [SMALL_STATE(431)] = 13605,
  [SMALL_STATE(432)] = 13613,
  [SMALL_STATE(433)] = 13621,
  [SMALL_STATE(434)] = 13629,
  [SMALL_STATE(435)] = 13637,
  [SMALL_STATE(436)] = 13645,
  [SMALL_STATE(437)] = 13653,
  [SMALL_STATE(438)] = 13661,
  [SMALL_STATE(439)] = 13669,
  [SMALL_STATE(440)] = 13677,
  [SMALL_STATE(441)] = 13685,
};

static const TSParseActionEntry ts_parse_actions[] = {
  [0] = {.entry = {.count = 0, .reusable = false}},
  [1] = {.entry = {.count = 1, .reusable = false}}, RECOVER(),
  [3] = {.entry = {.count = 1, .reusable = false}}, SHIFT_EXTRA(),
  [5] = {.entry = {.count = 1, .reusable = true}}, SHIFT_EXTRA(),
  [7] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_stylesheet, 0, 0, 0),
  [9] = {.entry = {.count = 1, .reusable = false}}, SHIFT(191),
  [11] = {.entry = {.count = 1, .reusable = false}}, SHIFT(258),
  [13] = {.entry = {.count = 1, .reusable = false}}, SHIFT(169),
  [15] = {.entry = {.count = 1, .reusable = false}}, SHIFT(325),
  [17] = {.entry = {.count = 1, .reusable = false}}, SHIFT(431),
  [19] = {.entry = {.count = 1, .reusable = false}}, SHIFT(254),
  [21] = {.entry = {.count = 1, .reusable = false}}, SHIFT(393),
  [23] = {.entry = {.count = 1, .reusable = true}}, SHIFT(243),
  [25] = {.entry = {.count = 1, .reusable = true}}, SHIFT(134),
  [27] = {.entry = {.count = 1, .reusable = true}}, SHIFT(391),
  [29] = {.entry = {.count = 1, .reusable = true}}, SHIFT(439),
  [31] = {.entry = {.count = 1, .reusable = true}}, SHIFT(433),
  [33] = {.entry = {.count = 1, .reusable = true}}, SHIFT(19),
  [35] = {.entry = {.count = 1, .reusable = true}}, SHIFT(36),
  [37] = {.entry = {.count = 1, .reusable = true}}, SHIFT(21),
  [39] = {.entry = {.count = 1, .reusable = true}}, SHIFT(34),
  [41] = {.entry = {.count = 1, .reusable = true}}, SHIFT(35),
  [43] = {.entry = {.count = 1, .reusable = true}}, SHIFT(379),
  [45] = {.entry = {.count = 1, .reusable = true}}, SHIFT(351),
  [47] = {.entry = {.count = 1, .reusable = true}}, SHIFT(239),
  [49] = {.entry = {.count = 1, .reusable = false}}, SHIFT(238),
  [51] = {.entry = {.count = 1, .reusable = true}}, SHIFT(250),
  [53] = {.entry = {.count = 1, .reusable = false}}, SHIFT(182),
  [55] = {.entry = {.count = 1, .reusable = false}}, SHIFT(264),
  [57] = {.entry = {.count = 1, .reusable = false}}, SHIFT(183),
  [59] = {.entry = {.count = 1, .reusable = false}}, SHIFT(328),
  [61] = {.entry = {.count = 1, .reusable = false}}, SHIFT(425),
  [63] = {.entry = {.count = 1, .reusable = true}}, SHIFT(46),
  [65] = {.entry = {.count = 1, .reusable = false}}, SHIFT(265),
  [67] = {.entry = {.count = 1, .reusable = false}}, SHIFT(385),
  [69] = {.entry = {.count = 1, .reusable = true}}, SHIFT(242),
  [71] = {.entry = {.count = 1, .reusable = false}}, SHIFT(14),
  [73] = {.entry = {.count = 1, .reusable = true}}, SHIFT(58),
  [75] = {.entry = {.count = 1, .reusable = true}}, SHIFT(352),
  [77] = {.entry = {.count = 1, .reusable = true}}, SHIFT(360),
  [79] = {.entry = {.count = 1, .reusable = true}}, SHIFT(79),
  [81] = {.entry = {.count = 1, .reusable = true}}, SHIFT(85),
  [83] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_block_repeat1, 2, 0, 0), SHIFT_REPEAT(182),
  [86] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_block_repeat1, 2, 0, 0), SHIFT_REPEAT(264),
  [89] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_block_repeat1, 2, 0, 0), SHIFT_REPEAT(183),
  [92] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_block_repeat1, 2, 0, 0), SHIFT_REPEAT(328),
  [95] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_block_repeat1, 2, 0, 0), SHIFT_REPEAT(425),
  [98] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_block_repeat1, 2, 0, 0),
  [100] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_block_repeat1, 2, 0, 0), SHIFT_REPEAT(265),
  [103] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_block_repeat1, 2, 0, 0), SHIFT_REPEAT(385),
  [106] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_block_repeat1, 2, 0, 0), SHIFT_REPEAT(243),
  [109] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_block_repeat1, 2, 0, 0), SHIFT_REPEAT(134),
  [112] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_block_repeat1, 2, 0, 0), SHIFT_REPEAT(391),
  [115] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_block_repeat1, 2, 0, 0), SHIFT_REPEAT(439),
  [118] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_block_repeat1, 2, 0, 0), SHIFT_REPEAT(433),
  [121] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_block_repeat1, 2, 0, 0), SHIFT_REPEAT(19),
  [124] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_block_repeat1, 2, 0, 0), SHIFT_REPEAT(36),
  [127] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_block_repeat1, 2, 0, 0), SHIFT_REPEAT(21),
  [130] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_block_repeat1, 2, 0, 0), SHIFT_REPEAT(34),
  [133] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_block_repeat1, 2, 0, 0), SHIFT_REPEAT(35),
  [136] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_block_repeat1, 2, 0, 0), SHIFT_REPEAT(379),
  [139] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_block_repeat1, 2, 0, 0), SHIFT_REPEAT(351),
  [142] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_block_repeat1, 2, 0, 0), SHIFT_REPEAT(244),
  [145] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_block_repeat1, 2, 0, 0), SHIFT_REPEAT(14),
  [148] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_block_repeat1, 2, 0, 0), SHIFT_REPEAT(250),
  [151] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_stylesheet_repeat1, 2, 0, 0),
  [153] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_stylesheet_repeat1, 2, 0, 0), SHIFT_REPEAT(191),
  [156] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_stylesheet_repeat1, 2, 0, 0), SHIFT_REPEAT(258),
  [159] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_stylesheet_repeat1, 2, 0, 0), SHIFT_REPEAT(169),
  [162] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_stylesheet_repeat1, 2, 0, 0), SHIFT_REPEAT(325),
  [165] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_stylesheet_repeat1, 2, 0, 0), SHIFT_REPEAT(431),
  [168] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_stylesheet_repeat1, 2, 0, 0), SHIFT_REPEAT(254),
  [171] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_stylesheet_repeat1, 2, 0, 0), SHIFT_REPEAT(393),
  [174] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_stylesheet_repeat1, 2, 0, 0), SHIFT_REPEAT(243),
  [177] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_stylesheet_repeat1, 2, 0, 0), SHIFT_REPEAT(134),
  [180] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_stylesheet_repeat1, 2, 0, 0), SHIFT_REPEAT(391),
  [183] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_stylesheet_repeat1, 2, 0, 0), SHIFT_REPEAT(439),
  [186] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_stylesheet_repeat1, 2, 0, 0), SHIFT_REPEAT(433),
  [189] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_stylesheet_repeat1, 2, 0, 0), SHIFT_REPEAT(19),
  [192] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_stylesheet_repeat1, 2, 0, 0), SHIFT_REPEAT(36),
  [195] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_stylesheet_repeat1, 2, 0, 0), SHIFT_REPEAT(21),
  [198] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_stylesheet_repeat1, 2, 0, 0), SHIFT_REPEAT(34),
  [201] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_stylesheet_repeat1, 2, 0, 0), SHIFT_REPEAT(35),
  [204] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_stylesheet_repeat1, 2, 0, 0), SHIFT_REPEAT(379),
  [207] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_stylesheet_repeat1, 2, 0, 0), SHIFT_REPEAT(351),
  [210] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_stylesheet_repeat1, 2, 0, 0), SHIFT_REPEAT(239),
  [213] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_stylesheet_repeat1, 2, 0, 0), SHIFT_REPEAT(238),
  [216] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_stylesheet_repeat1, 2, 0, 0), SHIFT_REPEAT(250),
  [219] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_stylesheet, 1, 0, 0),
  [221] = {.entry = {.count = 1, .reusable = true}}, SHIFT(187),
  [223] = {.entry = {.count = 1, .reusable = true}}, SHIFT(139),
  [225] = {.entry = {.count = 1, .reusable = true}}, SHIFT(241),
  [227] = {.entry = {.count = 1, .reusable = false}}, SHIFT(391),
  [229] = {.entry = {.count = 1, .reusable = true}}, SHIFT(415),
  [231] = {.entry = {.count = 1, .reusable = false}}, SHIFT(34),
  [233] = {.entry = {.count = 1, .reusable = true}}, SHIFT(217),
  [235] = {.entry = {.count = 1, .reusable = true}}, SHIFT(370),
  [237] = {.entry = {.count = 1, .reusable = true}}, SHIFT(337),
  [239] = {.entry = {.count = 1, .reusable = false}}, SHIFT(173),
  [241] = {.entry = {.count = 1, .reusable = false}}, SHIFT(174),
  [243] = {.entry = {.count = 1, .reusable = false}}, SHIFT(188),
  [245] = {.entry = {.count = 1, .reusable = false}}, SHIFT(38),
  [247] = {.entry = {.count = 1, .reusable = false}}, SHIFT(217),
  [249] = {.entry = {.count = 1, .reusable = true}}, SHIFT(158),
  [251] = {.entry = {.count = 1, .reusable = true}}, SHIFT(246),
  [253] = {.entry = {.count = 1, .reusable = true}}, SHIFT(251),
  [255] = {.entry = {.count = 1, .reusable = true}}, SHIFT(57),
  [257] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6),
  [259] = {.entry = {.count = 1, .reusable = true}}, SHIFT(17),
  [261] = {.entry = {.count = 1, .reusable = false}}, SHIFT(263),
  [263] = {.entry = {.count = 1, .reusable = true}}, SHIFT(420),
  [265] = {.entry = {.count = 1, .reusable = false}}, SHIFT(229),
  [267] = {.entry = {.count = 1, .reusable = false}}, SHIFT(430),
  [269] = {.entry = {.count = 1, .reusable = true}}, SHIFT(373),
  [271] = {.entry = {.count = 1, .reusable = true}}, SHIFT(374),
  [273] = {.entry = {.count = 1, .reusable = false}}, SHIFT(209),
  [275] = {.entry = {.count = 1, .reusable = false}}, SHIFT(216),
  [277] = {.entry = {.count = 1, .reusable = true}}, SHIFT(190),
  [279] = {.entry = {.count = 1, .reusable = false}}, SHIFT(113),
  [281] = {.entry = {.count = 1, .reusable = false}}, SHIFT(115),
  [283] = {.entry = {.count = 1, .reusable = true}}, SHIFT(15),
  [285] = {.entry = {.count = 1, .reusable = true}}, SHIFT(422),
  [287] = {.entry = {.count = 1, .reusable = true}}, SHIFT(333),
  [289] = {.entry = {.count = 1, .reusable = true}}, SHIFT(356),
  [291] = {.entry = {.count = 1, .reusable = true}}, SHIFT(381),
  [293] = {.entry = {.count = 1, .reusable = false}}, SHIFT(284),
  [295] = {.entry = {.count = 1, .reusable = false}}, SHIFT(283),
  [297] = {.entry = {.count = 1, .reusable = true}}, SHIFT(185),
  [299] = {.entry = {.count = 1, .reusable = false}}, SHIFT(278),
  [301] = {.entry = {.count = 1, .reusable = false}}, SHIFT(333),
  [303] = {.entry = {.count = 1, .reusable = true}}, SHIFT(165),
  [305] = {.entry = {.count = 1, .reusable = true}}, SHIFT(240),
  [307] = {.entry = {.count = 1, .reusable = true}}, SHIFT(154),
  [309] = {.entry = {.count = 1, .reusable = true}}, SHIFT(323),
  [311] = {.entry = {.count = 1, .reusable = false}}, SHIFT(323),
  [313] = {.entry = {.count = 1, .reusable = true}}, SHIFT(255),
  [315] = {.entry = {.count = 1, .reusable = true}}, SHIFT(272),
  [317] = {.entry = {.count = 1, .reusable = true}}, SHIFT(212),
  [319] = {.entry = {.count = 1, .reusable = true}}, SHIFT(267),
  [321] = {.entry = {.count = 1, .reusable = true}}, SHIFT(159),
  [323] = {.entry = {.count = 1, .reusable = true}}, SHIFT(261),
  [325] = {.entry = {.count = 1, .reusable = true}}, SHIFT(253),
  [327] = {.entry = {.count = 1, .reusable = true}}, SHIFT(132),
  [329] = {.entry = {.count = 1, .reusable = true}}, SHIFT(245),
  [331] = {.entry = {.count = 1, .reusable = true}}, SHIFT(207),
  [333] = {.entry = {.count = 1, .reusable = true}}, SHIFT(266),
  [335] = {.entry = {.count = 1, .reusable = true}}, SHIFT(166),
  [337] = {.entry = {.count = 1, .reusable = true}}, SHIFT(171),
  [339] = {.entry = {.count = 1, .reusable = true}}, SHIFT(86),
  [341] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_last_declaration, 3, 0, 14),
  [343] = {.entry = {.count = 1, .reusable = true}}, SHIFT(175),
  [345] = {.entry = {.count = 1, .reusable = true}}, SHIFT(423),
  [347] = {.entry = {.count = 1, .reusable = false}}, SHIFT(175),
  [349] = {.entry = {.count = 1, .reusable = true}}, SHIFT(208),
  [351] = {.entry = {.count = 1, .reusable = true}}, SHIFT(362),
  [353] = {.entry = {.count = 1, .reusable = true}}, SHIFT(363),
  [355] = {.entry = {.count = 1, .reusable = true}}, SHIFT(188),
  [357] = {.entry = {.count = 1, .reusable = false}}, SHIFT(150),
  [359] = {.entry = {.count = 1, .reusable = false}}, SHIFT(214),
  [361] = {.entry = {.count = 1, .reusable = true}}, SHIFT(252),
  [363] = {.entry = {.count = 1, .reusable = true}}, SHIFT(141),
  [365] = {.entry = {.count = 1, .reusable = true}}, SHIFT(137),
  [367] = {.entry = {.count = 1, .reusable = true}}, SHIFT(142),
  [369] = {.entry = {.count = 1, .reusable = true}}, SHIFT(161),
  [371] = {.entry = {.count = 1, .reusable = true}}, SHIFT(130),
  [373] = {.entry = {.count = 1, .reusable = true}}, SHIFT(163),
  [375] = {.entry = {.count = 1, .reusable = true}}, SHIFT(60),
  [377] = {.entry = {.count = 1, .reusable = true}}, SHIFT(230),
  [379] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__selector, 1, 0, 1),
  [381] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__value, 1, 0, 3),
  [383] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__value, 1, 0, 3),
  [385] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__selector, 1, 0, 1),
  [387] = {.entry = {.count = 1, .reusable = true}}, SHIFT(122),
  [389] = {.entry = {.count = 1, .reusable = true}}, SHIFT(225),
  [391] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_scope_statement, 2, 0, 0),
  [393] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_scope_statement, 2, 0, 0),
  [395] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_import_statement, 3, 0, 0),
  [397] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_import_statement, 3, 0, 0),
  [399] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_media_statement, 3, 0, 0),
  [401] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_media_statement, 3, 0, 0),
  [403] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_charset_statement, 3, 0, 0),
  [405] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_charset_statement, 3, 0, 0),
  [407] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_keyframes_statement, 3, 0, 8),
  [409] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_keyframes_statement, 3, 0, 8),
  [411] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_supports_statement, 3, 0, 0),
  [413] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_supports_statement, 3, 0, 0),
  [415] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_block, 2, 0, 0),
  [417] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_block, 2, 0, 0),
  [419] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_at_rule, 2, 0, 0),
  [421] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_at_rule, 2, 0, 0),
  [423] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_rule_set, 2, 0, 0),
  [425] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_rule_set, 2, 0, 0),
  [427] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_at_rule, 3, 0, 0),
  [429] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_at_rule, 3, 0, 0),
  [431] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_class_name_repeat1, 2, 0, 0),
  [433] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_class_name_repeat1, 2, 0, 0),
  [435] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_name_repeat1, 2, 0, 0), SHIFT_REPEAT(50),
  [438] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__selector, 1, 0, 0),
  [440] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__value, 1, 0, 0),
  [442] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__selector, 1, 0, 0),
  [444] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__value, 1, 0, 0),
  [446] = {.entry = {.count = 1, .reusable = true}}, SHIFT(164),
  [448] = {.entry = {.count = 1, .reusable = true}}, SHIFT(277),
  [450] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_import_statement, 4, 0, 0),
  [452] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_import_statement, 4, 0, 0),
  [454] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_media_statement, 4, 0, 0),
  [456] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_media_statement, 4, 0, 0),
  [458] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_namespace_statement, 4, 0, 13),
  [460] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_namespace_statement, 4, 0, 13),
  [462] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_keyframe_block_list, 2, 0, 0),
  [464] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_keyframe_block_list, 2, 0, 0),
  [466] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_block, 3, 0, 0),
  [468] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_block, 3, 0, 0),
  [470] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_name, 1, 0, 0),
  [472] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_name, 1, 0, 0),
  [474] = {.entry = {.count = 1, .reusable = true}}, SHIFT(97),
  [476] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_declaration, 4, 0, 14),
  [478] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_declaration, 4, 0, 14),
  [480] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_at_rule, 4, 0, 0),
  [482] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_at_rule, 4, 0, 0),
  [484] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_import_statement, 5, 0, 0),
  [486] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_import_statement, 5, 0, 0),
  [488] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_keyframe_block_list, 3, 0, 0),
  [490] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_keyframe_block_list, 3, 0, 0),
  [492] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_postcss_statement, 3, 0, 0),
  [494] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_postcss_statement, 3, 0, 0),
  [496] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_block, 4, 0, 0),
  [498] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_block, 4, 0, 0),
  [500] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_scope_statement, 5, 0, 0),
  [502] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_scope_statement, 5, 0, 0),
  [504] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_declaration, 5, 0, 14),
  [506] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_declaration, 5, 0, 14),
  [508] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_declaration, 6, 0, 14),
  [510] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_declaration, 6, 0, 14),
  [512] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_scope_statement, 9, 0, 0),
  [514] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_scope_statement, 9, 0, 0),
  [516] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_namespace_statement, 3, 0, 0),
  [518] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_namespace_statement, 3, 0, 0),
  [520] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_string_value, 2, 0, 0),
  [522] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_string_value, 2, 0, 0),
  [524] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_string_value, 3, 0, 0),
  [526] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_string_value, 3, 0, 0),
  [528] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_name, 2, 0, 0),
  [530] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_name, 2, 0, 0),
  [532] = {.entry = {.count = 1, .reusable = true}}, SHIFT(50),
  [534] = {.entry = {.count = 1, .reusable = true}}, SHIFT(301),
  [536] = {.entry = {.count = 1, .reusable = true}}, SHIFT(200),
  [538] = {.entry = {.count = 1, .reusable = true}}, SHIFT(234),
  [540] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(171),
  [543] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_declaration_repeat1, 2, 0, 0),
  [545] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(187),
  [548] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(423),
  [551] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(214),
  [554] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(362),
  [557] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(363),
  [560] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(173),
  [563] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(174),
  [566] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(188),
  [569] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(150),
  [572] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(214),
  [575] = {.entry = {.count = 1, .reusable = true}}, SHIFT(13),
  [577] = {.entry = {.count = 1, .reusable = true}}, SHIFT(157),
  [579] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_arguments_repeat1, 2, 0, 0),
  [581] = {.entry = {.count = 1, .reusable = true}}, SHIFT(92),
  [583] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_last_declaration, 4, 0, 14),
  [585] = {.entry = {.count = 1, .reusable = true}}, SHIFT(211),
  [587] = {.entry = {.count = 1, .reusable = true}}, SHIFT(126),
  [589] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_pseudo_class_arguments_repeat1, 2, 0, 0),
  [591] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_pseudo_class_arguments_repeat1, 2, 0, 0), SHIFT_REPEAT(187),
  [594] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_pseudo_class_arguments_repeat1, 2, 0, 0), SHIFT_REPEAT(423),
  [597] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_pseudo_class_arguments_repeat1, 2, 0, 0), SHIFT_REPEAT(217),
  [600] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_pseudo_class_arguments_repeat1, 2, 0, 0), SHIFT_REPEAT(362),
  [603] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_pseudo_class_arguments_repeat1, 2, 0, 0), SHIFT_REPEAT(363),
  [606] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_pseudo_class_arguments_repeat1, 2, 0, 0), SHIFT_REPEAT(173),
  [609] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_pseudo_class_arguments_repeat1, 2, 0, 0), SHIFT_REPEAT(174),
  [612] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_pseudo_class_arguments_repeat1, 2, 0, 0), SHIFT_REPEAT(188),
  [615] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_pseudo_class_arguments_repeat1, 2, 0, 0), SHIFT_REPEAT(150),
  [618] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_pseudo_class_arguments_repeat1, 2, 0, 0), SHIFT_REPEAT(217),
  [621] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_pseudo_class_arguments_repeat2, 2, 0, 0),
  [623] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pseudo_class_selector, 3, 0, 10),
  [625] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_pseudo_class_selector, 3, 0, 10),
  [627] = {.entry = {.count = 1, .reusable = true}}, SHIFT(16),
  [629] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pseudo_class_selector, 2, 0, 0),
  [631] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_pseudo_class_selector, 2, 0, 0),
  [633] = {.entry = {.count = 1, .reusable = true}}, SHIFT(12),
  [635] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_postcss_statement_repeat1, 2, 0, 0),
  [637] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_postcss_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(189),
  [640] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_postcss_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(420),
  [643] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_postcss_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(229),
  [646] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_postcss_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(373),
  [649] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_postcss_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(374),
  [652] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_postcss_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(209),
  [655] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_postcss_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(216),
  [658] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_postcss_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(190),
  [661] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_postcss_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(202),
  [664] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_postcss_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(111),
  [667] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pseudo_class_selector, 2, 0, 2),
  [669] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_pseudo_class_selector, 2, 0, 2),
  [671] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__query, 1, 0, 4),
  [673] = {.entry = {.count = 1, .reusable = true}}, SHIFT(123),
  [675] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__query, 1, 0, 4),
  [677] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pseudo_class_selector, 3, 0, 0),
  [679] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_pseudo_class_selector, 3, 0, 0),
  [681] = {.entry = {.count = 1, .reusable = true}}, SHIFT(64),
  [683] = {.entry = {.count = 1, .reusable = true}}, SHIFT(189),
  [685] = {.entry = {.count = 1, .reusable = false}}, SHIFT(202),
  [687] = {.entry = {.count = 1, .reusable = false}}, SHIFT(111),
  [689] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pseudo_element_selector, 3, 0, 11),
  [691] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_pseudo_element_selector, 3, 0, 11),
  [693] = {.entry = {.count = 1, .reusable = true}}, SHIFT(11),
  [695] = {.entry = {.count = 1, .reusable = true}}, SHIFT(67),
  [697] = {.entry = {.count = 1, .reusable = true}}, SHIFT(228),
  [699] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pseudo_element_selector, 2, 0, 5),
  [701] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_pseudo_element_selector, 2, 0, 5),
  [703] = {.entry = {.count = 1, .reusable = true}}, SHIFT(226),
  [705] = {.entry = {.count = 1, .reusable = true}}, SHIFT(315),
  [707] = {.entry = {.count = 1, .reusable = true}}, SHIFT(293),
  [709] = {.entry = {.count = 1, .reusable = true}}, SHIFT(197),
  [711] = {.entry = {.count = 1, .reusable = true}}, SHIFT(227),
  [713] = {.entry = {.count = 1, .reusable = true}}, SHIFT(274),
  [715] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pseudo_element_arguments, 3, 0, 0),
  [717] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_pseudo_element_arguments, 3, 0, 0),
  [719] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute_selector, 3, 0, 9),
  [721] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_attribute_selector, 3, 0, 9),
  [723] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_namespace_selector, 2, 0, 0),
  [725] = {.entry = {.count = 1, .reusable = true}}, SHIFT(399),
  [727] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_namespace_selector, 2, 0, 0),
  [729] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_descendant_selector, 3, 0, 0),
  [731] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_descendant_selector, 3, 0, 0),
  [733] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute_selector, 5, 0, 9),
  [735] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_attribute_selector, 5, 0, 9),
  [737] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_universal_selector, 1, 0, 0),
  [739] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_universal_selector, 1, 0, 0),
  [741] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_id_selector, 2, 0, 6),
  [743] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_id_selector, 2, 0, 6),
  [745] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pseudo_class_with_selector_arguments, 4, 0, 0),
  [747] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_pseudo_class_with_selector_arguments, 4, 0, 0),
  [749] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_adjacent_sibling_selector, 3, 0, 0),
  [751] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_adjacent_sibling_selector, 3, 0, 0),
  [753] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_selector, 3, 0, 0),
  [755] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_selector, 3, 0, 0),
  [757] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pseudo_element_arguments, 2, 0, 0),
  [759] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_pseudo_element_arguments, 2, 0, 0),
  [761] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pseudo_class_arguments, 4, 0, 0),
  [763] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_pseudo_class_arguments, 4, 0, 0),
  [765] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sibling_selector, 3, 0, 0),
  [767] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_sibling_selector, 3, 0, 0),
  [769] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_namespace_selector, 3, 0, 0),
  [771] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_namespace_selector, 3, 0, 0),
  [773] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_id_selector, 3, 0, 12),
  [775] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_id_selector, 3, 0, 12),
  [777] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pseudo_class_selector, 4, 0, 10),
  [779] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_pseudo_class_selector, 4, 0, 10),
  [781] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pseudo_element_selector, 3, 0, 5),
  [783] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_pseudo_element_selector, 3, 0, 5),
  [785] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pseudo_element_selector, 4, 0, 11),
  [787] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_pseudo_element_selector, 4, 0, 11),
  [789] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute_selector, 4, 0, 15),
  [791] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_attribute_selector, 4, 0, 15),
  [793] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pseudo_class_with_selector_arguments, 3, 0, 0),
  [795] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_pseudo_class_with_selector_arguments, 3, 0, 0),
  [797] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pseudo_class_nth_child_arguments, 3, 0, 0),
  [799] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_pseudo_class_nth_child_arguments, 3, 0, 0),
  [801] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_selector, 2, 0, 0),
  [803] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_selector, 2, 0, 0),
  [805] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pseudo_class_selector, 3, 0, 2),
  [807] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_pseudo_class_selector, 3, 0, 2),
  [809] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__nth_child_pseudo_class_selector, 2, 0, 0),
  [811] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__nth_child_pseudo_class_selector, 2, 0, 0),
  [813] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pseudo_element_arguments, 4, 0, 0),
  [815] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_pseudo_element_arguments, 4, 0, 0),
  [817] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pseudo_class_arguments, 3, 0, 0),
  [819] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_pseudo_class_arguments, 3, 0, 0),
  [821] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pseudo_class_arguments, 2, 0, 0),
  [823] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_pseudo_class_arguments, 2, 0, 0),
  [825] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sibling_selector, 2, 0, 0),
  [827] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_sibling_selector, 2, 0, 0),
  [829] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute_selector, 6, 0, 15),
  [831] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_attribute_selector, 6, 0, 15),
  [833] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_adjacent_sibling_selector, 2, 0, 0),
  [835] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_adjacent_sibling_selector, 2, 0, 0),
  [837] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pseudo_class_nth_child_arguments, 5, 0, 0),
  [839] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_pseudo_class_nth_child_arguments, 5, 0, 0),
  [841] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_child_selector, 2, 0, 0),
  [843] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_child_selector, 2, 0, 0),
  [845] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pseudo_class_with_selector_arguments, 2, 0, 0),
  [847] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_pseudo_class_with_selector_arguments, 2, 0, 0),
  [849] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_child_selector, 3, 0, 0),
  [851] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_child_selector, 3, 0, 0),
  [853] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pseudo_class_selector, 4, 0, 0),
  [855] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_pseudo_class_selector, 4, 0, 0),
  [857] = {.entry = {.count = 1, .reusable = true}}, SHIFT(184),
  [859] = {.entry = {.count = 1, .reusable = true}}, SHIFT(326),
  [861] = {.entry = {.count = 1, .reusable = false}}, SHIFT(275),
  [863] = {.entry = {.count = 1, .reusable = false}}, SHIFT(326),
  [865] = {.entry = {.count = 1, .reusable = true}}, SHIFT(331),
  [867] = {.entry = {.count = 1, .reusable = false}}, SHIFT(331),
  [869] = {.entry = {.count = 1, .reusable = true}}, SHIFT(37),
  [871] = {.entry = {.count = 1, .reusable = false}}, SHIFT(37),
  [873] = {.entry = {.count = 1, .reusable = true}}, SHIFT(210),
  [875] = {.entry = {.count = 1, .reusable = false}}, SHIFT(210),
  [877] = {.entry = {.count = 1, .reusable = true}}, SHIFT(320),
  [879] = {.entry = {.count = 1, .reusable = false}}, SHIFT(320),
  [881] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_integer_value, 1, 0, 0),
  [883] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_integer_value, 1, 0, 0),
  [885] = {.entry = {.count = 1, .reusable = false}}, SHIFT(204),
  [887] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_float_value, 1, 0, 0),
  [889] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_float_value, 1, 0, 0),
  [891] = {.entry = {.count = 1, .reusable = false}}, SHIFT(206),
  [893] = {.entry = {.count = 1, .reusable = true}}, SHIFT(198),
  [895] = {.entry = {.count = 1, .reusable = false}}, SHIFT(198),
  [897] = {.entry = {.count = 1, .reusable = true}}, SHIFT(231),
  [899] = {.entry = {.count = 1, .reusable = false}}, SHIFT(231),
  [901] = {.entry = {.count = 1, .reusable = true}}, SHIFT(335),
  [903] = {.entry = {.count = 1, .reusable = false}}, SHIFT(335),
  [905] = {.entry = {.count = 1, .reusable = true}}, SHIFT(318),
  [907] = {.entry = {.count = 1, .reusable = false}}, SHIFT(318),
  [909] = {.entry = {.count = 1, .reusable = true}}, SHIFT(172),
  [911] = {.entry = {.count = 1, .reusable = true}}, SHIFT(438),
  [913] = {.entry = {.count = 1, .reusable = true}}, SHIFT(281),
  [915] = {.entry = {.count = 1, .reusable = true}}, SHIFT(342),
  [917] = {.entry = {.count = 1, .reusable = true}}, SHIFT(343),
  [919] = {.entry = {.count = 1, .reusable = false}}, SHIFT(259),
  [921] = {.entry = {.count = 1, .reusable = false}}, SHIFT(260),
  [923] = {.entry = {.count = 1, .reusable = true}}, SHIFT(180),
  [925] = {.entry = {.count = 1, .reusable = false}}, SHIFT(249),
  [927] = {.entry = {.count = 1, .reusable = false}}, SHIFT(281),
  [929] = {.entry = {.count = 1, .reusable = true}}, SHIFT(303),
  [931] = {.entry = {.count = 1, .reusable = false}}, SHIFT(303),
  [933] = {.entry = {.count = 1, .reusable = true}}, SHIFT(294),
  [935] = {.entry = {.count = 1, .reusable = false}}, SHIFT(294),
  [937] = {.entry = {.count = 1, .reusable = true}}, SHIFT(221),
  [939] = {.entry = {.count = 1, .reusable = false}}, SHIFT(221),
  [941] = {.entry = {.count = 1, .reusable = true}}, SHIFT(321),
  [943] = {.entry = {.count = 1, .reusable = false}}, SHIFT(321),
  [945] = {.entry = {.count = 1, .reusable = true}}, SHIFT(309),
  [947] = {.entry = {.count = 1, .reusable = false}}, SHIFT(309),
  [949] = {.entry = {.count = 1, .reusable = true}}, SHIFT(39),
  [951] = {.entry = {.count = 1, .reusable = false}}, SHIFT(39),
  [953] = {.entry = {.count = 1, .reusable = true}}, SHIFT(336),
  [955] = {.entry = {.count = 1, .reusable = false}}, SHIFT(336),
  [957] = {.entry = {.count = 1, .reusable = true}}, SHIFT(306),
  [959] = {.entry = {.count = 1, .reusable = false}}, SHIFT(306),
  [961] = {.entry = {.count = 1, .reusable = true}}, SHIFT(305),
  [963] = {.entry = {.count = 1, .reusable = false}}, SHIFT(305),
  [965] = {.entry = {.count = 1, .reusable = true}}, SHIFT(233),
  [967] = {.entry = {.count = 1, .reusable = false}}, SHIFT(233),
  [969] = {.entry = {.count = 1, .reusable = true}}, SHIFT(29),
  [971] = {.entry = {.count = 1, .reusable = false}}, SHIFT(29),
  [973] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_call_expression, 2, 0, 7),
  [975] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_call_expression, 2, 0, 7),
  [977] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_color_value, 2, 0, 0),
  [979] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_color_value, 2, 0, 0),
  [981] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parenthesized_value, 3, 0, 0),
  [983] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parenthesized_value, 3, 0, 0),
  [985] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_grid_value, 3, 0, 0),
  [987] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_grid_value, 3, 0, 0),
  [989] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_arguments, 2, 0, 0),
  [991] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_arguments, 2, 0, 0),
  [993] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_binary_expression, 3, 0, 0),
  [995] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_binary_expression, 3, 0, 0),
  [997] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_grid_value, 4, 0, 0),
  [999] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_grid_value, 4, 0, 0),
  [1001] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_arguments, 3, 0, 0),
  [1003] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_arguments, 3, 0, 0),
  [1005] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_arguments, 4, 0, 0),
  [1007] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_arguments, 4, 0, 0),
  [1009] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_integer_value, 2, 0, 0),
  [1011] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_integer_value, 2, 0, 0),
  [1013] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_float_value, 2, 0, 0),
  [1015] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_float_value, 2, 0, 0),
  [1017] = {.entry = {.count = 1, .reusable = true}}, SHIFT(168),
  [1019] = {.entry = {.count = 1, .reusable = true}}, SHIFT(148),
  [1021] = {.entry = {.count = 1, .reusable = false}}, SHIFT(218),
  [1023] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_declaration_repeat1, 2, 0, 0),
  [1025] = {.entry = {.count = 1, .reusable = true}}, SHIFT(93),
  [1027] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_last_declaration, 5, 0, 14),
  [1029] = {.entry = {.count = 1, .reusable = true}}, SHIFT(177),
  [1031] = {.entry = {.count = 1, .reusable = true}}, SHIFT(127),
  [1033] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_declaration_repeat1, 1, 0, 0),
  [1035] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_declaration_repeat1, 1, 0, 0),
  [1037] = {.entry = {.count = 1, .reusable = false}}, SHIFT(223),
  [1039] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_pseudo_class_arguments_repeat1, 1, 0, 0),
  [1041] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_pseudo_class_arguments_repeat1, 1, 0, 0),
  [1043] = {.entry = {.count = 1, .reusable = true}}, SHIFT(73),
  [1045] = {.entry = {.count = 1, .reusable = true}}, SHIFT(262),
  [1047] = {.entry = {.count = 1, .reusable = true}}, SHIFT(179),
  [1049] = {.entry = {.count = 1, .reusable = false}}, SHIFT(179),
  [1051] = {.entry = {.count = 1, .reusable = false}}, SHIFT(312),
  [1053] = {.entry = {.count = 1, .reusable = true}}, SHIFT(68),
  [1055] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_postcss_statement_repeat1, 1, 0, 0),
  [1057] = {.entry = {.count = 1, .reusable = true}}, SHIFT(176),
  [1059] = {.entry = {.count = 1, .reusable = false}}, SHIFT(176),
  [1061] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_postcss_statement_repeat1, 1, 0, 0),
  [1063] = {.entry = {.count = 1, .reusable = true}}, SHIFT(41),
  [1065] = {.entry = {.count = 1, .reusable = true}}, SHIFT(47),
  [1067] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2),
  [1069] = {.entry = {.count = 1, .reusable = false}}, SHIFT(170),
  [1071] = {.entry = {.count = 1, .reusable = true}}, SHIFT(25),
  [1073] = {.entry = {.count = 1, .reusable = true}}, SHIFT(149),
  [1075] = {.entry = {.count = 1, .reusable = true}}, SHIFT(429),
  [1077] = {.entry = {.count = 1, .reusable = true}}, SHIFT(418),
  [1079] = {.entry = {.count = 1, .reusable = true}}, SHIFT(26),
  [1081] = {.entry = {.count = 1, .reusable = true}}, SHIFT(28),
  [1083] = {.entry = {.count = 1, .reusable = true}}, SHIFT(31),
  [1085] = {.entry = {.count = 1, .reusable = true}}, SHIFT(32),
  [1087] = {.entry = {.count = 1, .reusable = true}}, SHIFT(33),
  [1089] = {.entry = {.count = 1, .reusable = true}}, SHIFT(24),
  [1091] = {.entry = {.count = 1, .reusable = true}}, SHIFT(247),
  [1093] = {.entry = {.count = 1, .reusable = false}}, SHIFT(192),
  [1095] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_selectors, 1, 0, 0),
  [1097] = {.entry = {.count = 1, .reusable = false}}, SHIFT(186),
  [1099] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_selectors_repeat1, 2, 0, 0),
  [1101] = {.entry = {.count = 1, .reusable = false}}, SHIFT(412),
  [1103] = {.entry = {.count = 1, .reusable = false}}, SHIFT(109),
  [1105] = {.entry = {.count = 1, .reusable = false}}, SHIFT(407),
  [1107] = {.entry = {.count = 1, .reusable = true}}, SHIFT(59),
  [1109] = {.entry = {.count = 1, .reusable = false}}, SHIFT(59),
  [1111] = {.entry = {.count = 1, .reusable = true}}, SHIFT(124),
  [1113] = {.entry = {.count = 1, .reusable = false}}, SHIFT(406),
  [1115] = {.entry = {.count = 1, .reusable = false}}, SHIFT(112),
  [1117] = {.entry = {.count = 1, .reusable = true}}, SHIFT(405),
  [1119] = {.entry = {.count = 1, .reusable = true}}, SHIFT(404),
  [1121] = {.entry = {.count = 1, .reusable = true}}, SHIFT(316),
  [1123] = {.entry = {.count = 1, .reusable = false}}, SHIFT(248),
  [1125] = {.entry = {.count = 1, .reusable = false}}, SHIFT(279),
  [1127] = {.entry = {.count = 1, .reusable = true}}, SHIFT(389),
  [1129] = {.entry = {.count = 1, .reusable = false}}, SHIFT(345),
  [1131] = {.entry = {.count = 1, .reusable = true}}, SHIFT(162),
  [1133] = {.entry = {.count = 1, .reusable = true}}, SHIFT(400),
  [1135] = {.entry = {.count = 1, .reusable = true}}, SHIFT(121),
  [1137] = {.entry = {.count = 1, .reusable = true}}, SHIFT(129),
  [1139] = {.entry = {.count = 1, .reusable = true}}, SHIFT(302),
  [1141] = {.entry = {.count = 1, .reusable = true}}, SHIFT(285),
  [1143] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_keyframe_block_list_repeat1, 2, 0, 0),
  [1145] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_keyframe_block_list_repeat1, 2, 0, 0), SHIFT_REPEAT(341),
  [1148] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_keyframe_block_list_repeat1, 2, 0, 0), SHIFT_REPEAT(408),
  [1151] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_keyframe_block_list_repeat1, 2, 0, 0), SHIFT_REPEAT(388),
  [1154] = {.entry = {.count = 1, .reusable = true}}, SHIFT(56),
  [1156] = {.entry = {.count = 1, .reusable = true}}, SHIFT(341),
  [1158] = {.entry = {.count = 1, .reusable = true}}, SHIFT(408),
  [1160] = {.entry = {.count = 1, .reusable = true}}, SHIFT(388),
  [1162] = {.entry = {.count = 1, .reusable = true}}, SHIFT(63),
  [1164] = {.entry = {.count = 1, .reusable = true}}, SHIFT(84),
  [1166] = {.entry = {.count = 1, .reusable = true}}, SHIFT(89),
  [1168] = {.entry = {.count = 1, .reusable = true}}, SHIFT(178),
  [1170] = {.entry = {.count = 1, .reusable = true}}, SHIFT(181),
  [1172] = {.entry = {.count = 1, .reusable = true}}, SHIFT(273),
  [1174] = {.entry = {.count = 1, .reusable = false}}, SHIFT(181),
  [1176] = {.entry = {.count = 1, .reusable = true}}, SHIFT(413),
  [1178] = {.entry = {.count = 1, .reusable = true}}, SHIFT(222),
  [1180] = {.entry = {.count = 1, .reusable = true}}, SHIFT(196),
  [1182] = {.entry = {.count = 1, .reusable = true}}, SHIFT(256),
  [1184] = {.entry = {.count = 1, .reusable = true}}, SHIFT(80),
  [1186] = {.entry = {.count = 1, .reusable = true}}, SHIFT(257),
  [1188] = {.entry = {.count = 1, .reusable = true}}, SHIFT(292),
  [1190] = {.entry = {.count = 1, .reusable = true}}, SHIFT(49),
  [1192] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_binary_query, 3, 0, 0),
  [1194] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_unary_query, 2, 0, 0),
  [1196] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_feature_query, 5, 0, 16),
  [1198] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_selector_query, 4, 0, 0),
  [1200] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_grid_value_repeat1, 2, 0, 0),
  [1202] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parenthesized_query, 3, 0, 0),
  [1204] = {.entry = {.count = 1, .reusable = true}}, SHIFT(268),
  [1206] = {.entry = {.count = 1, .reusable = true}}, SHIFT(75),
  [1208] = {.entry = {.count = 1, .reusable = true}}, SHIFT(409),
  [1210] = {.entry = {.count = 1, .reusable = false}}, SHIFT(397),
  [1212] = {.entry = {.count = 1, .reusable = true}}, SHIFT(224),
  [1214] = {.entry = {.count = 1, .reusable = true}}, SHIFT(417),
  [1216] = {.entry = {.count = 1, .reusable = true}}, SHIFT(307),
  [1218] = {.entry = {.count = 1, .reusable = true}}, SHIFT(160),
  [1220] = {.entry = {.count = 1, .reusable = true}}, SHIFT(81),
  [1222] = {.entry = {.count = 1, .reusable = true}}, SHIFT(304),
  [1224] = {.entry = {.count = 1, .reusable = true}}, SHIFT(61),
  [1226] = {.entry = {.count = 1, .reusable = true}}, SHIFT(53),
  [1228] = {.entry = {.count = 1, .reusable = true}}, SHIFT(43),
  [1230] = {.entry = {.count = 1, .reusable = true}}, SHIFT(87),
  [1232] = {.entry = {.count = 1, .reusable = true}}, SHIFT(291),
  [1234] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_import_statement_repeat1, 2, 0, 0),
  [1236] = {.entry = {.count = 1, .reusable = true}}, SHIFT(133),
  [1238] = {.entry = {.count = 1, .reusable = true}}, SHIFT(195),
  [1240] = {.entry = {.count = 1, .reusable = false}}, SHIFT(95),
  [1242] = {.entry = {.count = 1, .reusable = false}}, SHIFT(372),
  [1244] = {.entry = {.count = 1, .reusable = true}}, SHIFT(269),
  [1246] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_keyframe_block, 2, 0, 0),
  [1248] = {.entry = {.count = 1, .reusable = true}}, SHIFT(324),
  [1250] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4),
  [1252] = {.entry = {.count = 1, .reusable = false}}, SHIFT(280),
  [1254] = {.entry = {.count = 1, .reusable = false}}, SHIFT(348),
  [1256] = {.entry = {.count = 1, .reusable = false}}, SHIFT(349),
  [1258] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_import_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(256),
  [1261] = {.entry = {.count = 1, .reusable = false}}, SHIFT(128),
  [1263] = {.entry = {.count = 1, .reusable = false}}, SHIFT(377),
  [1265] = {.entry = {.count = 1, .reusable = false}}, SHIFT(271),
  [1267] = {.entry = {.count = 1, .reusable = false}}, SHIFT(378),
  [1269] = {.entry = {.count = 1, .reusable = false}}, SHIFT(131),
  [1271] = {.entry = {.count = 1, .reusable = false}}, SHIFT(350),
  [1273] = {.entry = {.count = 1, .reusable = true}}, SHIFT(287),
  [1275] = {.entry = {.count = 1, .reusable = false}}, SHIFT(289),
  [1277] = {.entry = {.count = 1, .reusable = false}}, SHIFT(358),
  [1279] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_keyframe_block, 3, 0, 0),
  [1281] = {.entry = {.count = 1, .reusable = false}}, SHIFT(290),
  [1283] = {.entry = {.count = 1, .reusable = true}}, SHIFT(201),
  [1285] = {.entry = {.count = 1, .reusable = false}}, SHIFT(203),
  [1287] = {.entry = {.count = 1, .reusable = false}}, SHIFT(364),
  [1289] = {.entry = {.count = 1, .reusable = false}}, SHIFT(365),
  [1291] = {.entry = {.count = 1, .reusable = false}}, SHIFT(205),
  [1293] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_selectors_repeat1, 2, 0, 0), SHIFT_REPEAT(25),
  [1296] = {.entry = {.count = 1, .reusable = true}}, SHIFT(236),
  [1298] = {.entry = {.count = 1, .reusable = false}}, SHIFT(371),
  [1300] = {.entry = {.count = 1, .reusable = false}}, SHIFT(96),
  [1302] = {.entry = {.count = 1, .reusable = false}}, SHIFT(219),
  [1304] = {.entry = {.count = 1, .reusable = false}}, SHIFT(375),
  [1306] = {.entry = {.count = 1, .reusable = false}}, SHIFT(376),
  [1308] = {.entry = {.count = 1, .reusable = false}}, SHIFT(220),
  [1310] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_string_value_repeat1, 2, 0, 0),
  [1312] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_string_value_repeat1, 2, 0, 0), SHIFT_REPEAT(377),
  [1315] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_string_value_repeat2, 2, 0, 0),
  [1317] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_string_value_repeat2, 2, 0, 0), SHIFT_REPEAT(378),
  [1320] = {.entry = {.count = 1, .reusable = false}}, SHIFT(347),
  [1322] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_arguments_repeat1, 2, 0, 0), SHIFT_REPEAT(164),
  [1325] = {.entry = {.count = 1, .reusable = false}}, SHIFT(359),
  [1327] = {.entry = {.count = 1, .reusable = true}}, SHIFT(319),
  [1329] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_grid_value_repeat1, 2, 0, 0), SHIFT_REPEAT(178),
  [1332] = {.entry = {.count = 1, .reusable = true}}, SHIFT(156),
  [1334] = {.entry = {.count = 1, .reusable = true}}, SHIFT(20),
  [1336] = {.entry = {.count = 1, .reusable = true}}, SHIFT(62),
  [1338] = {.entry = {.count = 1, .reusable = true}}, SHIFT(199),
  [1340] = {.entry = {.count = 1, .reusable = true}}, SHIFT(248),
  [1342] = {.entry = {.count = 1, .reusable = true}}, SHIFT(437),
  [1344] = {.entry = {.count = 1, .reusable = true}}, SHIFT(232),
  [1346] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_keyframe_block_repeat1, 2, 0, 0), SHIFT_REPEAT(324),
  [1349] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_keyframe_block_repeat1, 2, 0, 0),
  [1351] = {.entry = {.count = 1, .reusable = true}}, SHIFT(22),
  [1353] = {.entry = {.count = 1, .reusable = true}}, SHIFT(276),
  [1355] = {.entry = {.count = 1, .reusable = true}}, SHIFT(300),
  [1357] = {.entry = {.count = 1, .reusable = true}}, SHIFT(88),
  [1359] = {.entry = {.count = 1, .reusable = false}}, SHIFT(285),
  [1361] = {.entry = {.count = 1, .reusable = true}}, SHIFT(136),
  [1363] = {.entry = {.count = 1, .reusable = true}}, SHIFT(441),
  [1365] = {.entry = {.count = 1, .reusable = true}}, SHIFT(140),
  [1367] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_pseudo_class_arguments_repeat2, 2, 0, 0), SHIFT_REPEAT(13),
  [1370] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_selectors, 2, 0, 0),
  [1372] = {.entry = {.count = 1, .reusable = true}}, SHIFT(322),
  [1374] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_to, 1, 0, 0),
  [1376] = {.entry = {.count = 1, .reusable = true}}, SHIFT(151),
  [1378] = {.entry = {.count = 1, .reusable = true}}, SHIFT(27),
  [1380] = {.entry = {.count = 1, .reusable = true}}, SHIFT(288),
  [1382] = {.entry = {.count = 1, .reusable = true}}, SHIFT(296),
  [1384] = {.entry = {.count = 1, .reusable = false}}, SHIFT(194),
  [1386] = {.entry = {.count = 1, .reusable = false}}, SHIFT(135),
  [1388] = {.entry = {.count = 1, .reusable = true}}, SHIFT(143),
  [1390] = {.entry = {.count = 1, .reusable = true}}, SHIFT(65),
  [1392] = {.entry = {.count = 1, .reusable = true}}, SHIFT(237),
  [1394] = {.entry = {.count = 1, .reusable = true}}, SHIFT(299),
  [1396] = {.entry = {.count = 1, .reusable = true}}, SHIFT(194),
  [1398] = {.entry = {.count = 1, .reusable = true}}, SHIFT(414),
  [1400] = {.entry = {.count = 1, .reusable = true}}, SHIFT(90),
  [1402] = {.entry = {.count = 1, .reusable = true}}, SHIFT(83),
  [1404] = {.entry = {.count = 1, .reusable = true}}, SHIFT(76),
  [1406] = {.entry = {.count = 1, .reusable = true}}, SHIFT(116),
  [1408] = {.entry = {.count = 1, .reusable = true}}, SHIFT(18),
  [1410] = {.entry = {.count = 1, .reusable = true}}, SHIFT(411),
  [1412] = {.entry = {.count = 1, .reusable = true}}, SHIFT(135),
  [1414] = {.entry = {.count = 1, .reusable = true}}, SHIFT(101),
  [1416] = {.entry = {.count = 1, .reusable = true}}, SHIFT(367),
  [1418] = {.entry = {.count = 1, .reusable = true}}, SHIFT(55),
  [1420] = {.entry = {.count = 1, .reusable = true}}, SHIFT(30),
  [1422] = {.entry = {.count = 1, .reusable = true}}, SHIFT(282),
  [1424] = {.entry = {.count = 1, .reusable = true}}, SHIFT(118),
  [1426] = {.entry = {.count = 1, .reusable = true}},  ACCEPT_INPUT(),
  [1428] = {.entry = {.count = 1, .reusable = true}}, SHIFT(23),
};

enum ts_external_scanner_symbol_identifiers {
  ts_external_token__descendant_operator = 0,
  ts_external_token__pseudo_class_selector_colon = 1,
  ts_external_token___error_recovery = 2,
};

static const TSSymbol ts_external_scanner_symbol_map[EXTERNAL_TOKEN_COUNT] = {
  [ts_external_token__descendant_operator] = sym__descendant_operator,
  [ts_external_token__pseudo_class_selector_colon] = sym__pseudo_class_selector_colon,
  [ts_external_token___error_recovery] = sym___error_recovery,
};

static const bool ts_external_scanner_states[4][EXTERNAL_TOKEN_COUNT] = {
  [1] = {
    [ts_external_token__descendant_operator] = true,
    [ts_external_token__pseudo_class_selector_colon] = true,
    [ts_external_token___error_recovery] = true,
  },
  [2] = {
    [ts_external_token__pseudo_class_selector_colon] = true,
  },
  [3] = {
    [ts_external_token__descendant_operator] = true,
    [ts_external_token__pseudo_class_selector_colon] = true,
  },
};

#ifdef __cplusplus
extern "C" {
#endif
void *tree_sitter_css_external_scanner_create(void);
void tree_sitter_css_external_scanner_destroy(void *);
bool tree_sitter_css_external_scanner_scan(void *, TSLexer *, const bool *);
unsigned tree_sitter_css_external_scanner_serialize(void *, char *);
void tree_sitter_css_external_scanner_deserialize(void *, const char *, unsigned);

#ifdef TREE_SITTER_HIDE_SYMBOLS
#define TS_PUBLIC
#elif defined(_WIN32)
#define TS_PUBLIC __declspec(dllexport)
#else
#define TS_PUBLIC __attribute__((visibility("default")))
#endif

TS_PUBLIC const TSLanguage *tree_sitter_css(void) {
  static const TSLanguage language = {
    .abi_version = LANGUAGE_VERSION,
    .symbol_count = SYMBOL_COUNT,
    .alias_count = ALIAS_COUNT,
    .token_count = TOKEN_COUNT,
    .external_token_count = EXTERNAL_TOKEN_COUNT,
    .state_count = STATE_COUNT,
    .large_state_count = LARGE_STATE_COUNT,
    .production_id_count = PRODUCTION_ID_COUNT,
    .supertype_count = SUPERTYPE_COUNT,
    .field_count = FIELD_COUNT,
    .max_alias_sequence_length = MAX_ALIAS_SEQUENCE_LENGTH,
    .parse_table = &ts_parse_table[0][0],
    .small_parse_table = ts_small_parse_table,
    .small_parse_table_map = ts_small_parse_table_map,
    .parse_actions = ts_parse_actions,
    .symbol_names = ts_symbol_names,
    .symbol_metadata = ts_symbol_metadata,
    .public_symbol_map = ts_symbol_map,
    .alias_map = ts_non_terminal_alias_map,
    .alias_sequences = &ts_alias_sequences[0][0],
    .lex_modes = (const void*)ts_lex_modes,
    .lex_fn = ts_lex,
    .external_scanner = {
      &ts_external_scanner_states[0][0],
      ts_external_scanner_symbol_map,
      tree_sitter_css_external_scanner_create,
      tree_sitter_css_external_scanner_destroy,
      tree_sitter_css_external_scanner_scan,
      tree_sitter_css_external_scanner_serialize,
      tree_sitter_css_external_scanner_deserialize,
    },
    .primary_state_ids = ts_primary_state_ids,
    .name = "css",
    .max_reserved_word_set_size = 0,
    .metadata = {
      .major_version = 0,
      .minor_version = 25,
      .patch_version = 0,
    },
  };
  return &language;
}
#ifdef __cplusplus
}
#endif
